/* tbas.c 
 * Copyleft (C) 2009-2014 Antonio Maschio 
 * @ <ing dot antonio dot maschio at gmail dot com>
 * Update: 09 September 2017
 * 
 * This is tbas, a BASIC interpreter written in c99 for the UNIX world.
 * You must have gcc 4.X to compile (I suppose, though even 3.X may suffice).
 * 
 * This program is free software: you can redistribute it and/or modify it 
 * under the terms of the GNU General Public License as published by the Free 
 * Software Foundation, either version 3 of the License, or (at your option) 
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for 
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with 
 * this program. If not, see <http:/www.gnu.org/licenses/>.
 *
 * It's GPL! Enjoy!
 */
#include "tbas.h"	// catch-all header file


/******************************************************************************
 * The following routines are related to the APPEND and PREPEND statements
 ******************************************************************************/


/* closePrepend()
 * close procedures opened by the PREPEND function, that is closing current
 * input file (the prepended text), copies on it all the lines of the original
 * file (the appended text) and removes the temporary file, reopening the
 * channel on the original file.
 * Called by FAPPEND(), FCLOSE(), END(), FSCRATCH() and its companions.
 * Also, resets the isPrepend flag.
 * ch: the channel on which to act the changes
 * System function for the PREPEND function */
void closePrepend(int ch) {
	FILE *f;
	char stamp[COMPSTR];

	/* assure the file is open */
	if (!channel[ch].isopen) perror_(__LINE__,80,l);
	/* assures PREPEND is active */
	if (!channel[ch].isPrepend) return;

	/* queue text to current */
	f=fopen(channel[ch].ioname,"r");
	if (!f) perror_(__LINE__,151,l,channel[ch].ioname); 
	while (fgets(stamp, MAXSTRLEN, f)) 
		fputs(stamp, channel[ch].stream);
	fclose(f);

	/* reassign and reset */
	fclose(channel[ch].stream);
	remove(channel[ch].ioname);
	rename(channel[ch].queuename,channel[ch].ioname);
	channel[ch].queuename[0]='\0';
	channel[ch].isPrepend=FALSE;
	channel[ch].stream=fopen(channel[ch].ioname,"a+");
	if (!channel[ch].stream) perror_(__LINE__,151,l,channel[ch].ioname);
}

/* FAPPEND()
 * implementation of the APPEND statement for sequential files
 * the # (sign number) is not required
 * set file pointer to the end of file, so that subsequent writing operations 
 * are performed by adding text to the original file, not overwriting it */
void FAPPEND(void) {
	int ch;
	
	/* argument required */
	if (!*p) perror_(__LINE__,125,l);

	do {
		/* sequential file */
		if (*p=='#') p++;
		ch=(int)S();
		/* check if channel is valid */
		if (ch<1 || ch>STREAMS) perror_(__LINE__,133,l);
		/* assures PREPEND is closed */
		if (channel[ch].isPrepend) closePrepend(ch);
		/* check if file exists */
		if (!channel[ch].isopen) perror_(__LINE__,80,l);
		/* check if file is random, which results in an error*/
		if (channel[ch].type!=SEQ) perror_(__LINE__,101,l);
		/* check is file is read-only */
		if (channel[ch].isReadonly)
			perror_(__LINE__,155,l,channel[ch].ioname);
		/* set file in write mode and move seek to end */
		channel[ch].wayout=VWRITE;
		fseek(channel[ch].stream,0,SEEK_END);
		if (*p==',' || *p==';') p++;
	} while (*p);
}


/* FPREPEND()
 * perform the PREPEND statement, which is the contrary of the APPEND
 * statement, that is set writing text to the beginning of the file, until
 * a new state change
 * The PREPEND statement may be given once only at a time. */
void FPREPEND(void) {
	int ch;
	char stamp[COMPSTR];

	/* argument required */
	if (!*p) perror_(__LINE__,125,l);

	do {
		/* sequential file */
		if (*p=='#') p++;
		ch=(int)S();
		/* check if channel is valid */
		if (ch<1 || ch>STREAMS) perror_(__LINE__,133,l);
		/* assures PREPEND is not already active on this channel
		 * or print a warning and proceed with next attribution */
		if (channel[ch].isPrepend) {
			perror_(__LINE__,180,l,ch);
			if (*p==',' || *p==';') p++;
			continue;
		}
		/* check if file exists */
		if (!channel[ch].isopen) perror_(__LINE__,80,l);
		/* check if file is random, which results in an error*/
		if (channel[ch].type!=SEQ) perror_(__LINE__,101,l);
		/* check is file is read-only */
		if (channel[ch].isReadonly)
			perror_(__LINE__,155,l,channel[ch].ioname);
		/* set file in prepend mode */
		channel[ch].isPrepend=TRUE;
		/* build queue file name */
		sprintw(stamp,COMPSTR,"%0d_%0d%0d%0d",ch,\
				getHour(),getMin(),getSec());
		strncpy_(channel[ch].queuename,TEMPROUTE,COMPSTR);
		strncat_(channel[ch].queuename,stamp,COMPSTR);
		/* close current channels */
		fclose(channel[ch].stream);
		channel[ch].stream=fopen(channel[ch].queuename,"w+");
		if (!channel[ch].stream) perror_(__LINE__,142,l);
		/* set file in write mode */
		channel[ch].wayout=VWRITE;
		if (*p==',' || *p==';') p++;
	} while (*p);
}
	
	
/******************************************************************************
 * The following routines are related to the LIBRARY/LIBRARIES statements
 ******************************************************************************/


/* includeLibs()
 * add to the existing running program all libs from an external program, 
 * identified by its file name
 * unitname: contains the external program name.
 * Note: in the external unit, only SUBs and DEFs are loaded in the caller 
 * memory, and subsequently preParsed; all comments and root statements of the 
 * included file are discarded. 
 * System function for the OPEN AS LIBRARY/LIBRARY functions */
void includeLibs(char *unitname) {
	int oldEndCore=endCore;

	/* load lines one by one starting from position after endCore 
	 * Note: new endCore value is set into LOAD() */
	LOAD(m,unitname,endCore,ONLY_SUBS);

	/* preParse new lines (to add SUBs and DEFS) 
	 * Note: endCore was updated into LOAD() */
	preParse(m,oldEndCore,endCore);
}


/* OPENASLIBRARY()
 * perform the OPENASLIBRARY/LIBRARY statements with the following
 * syntax:
 *
 * [OPEN AS] LIBRARY "lib1", "lib2", ...
 *
 * where "lib1", "lib2"... are BASIC programs separated by comma or semicolon */
void OPENASLIBRARY(void) {
	char unitname[COMPSTR], *pb;
	int c=0;

	while (TRUE) {
		/* argument required */
		if (!*p) perror_(__LINE__,125,l);
		strncpy_(unitname,SS(),COMPSTR);
		setupName(unitname,&c,&c);
		pb=p;
		includeLibs(unitname);
		p=pb;
		if (*p==',' || *p==';') p++;
		else break;
	}
}


/******************************************************************************
 * The following routines define the CLEAR, COMMON, ERASE and REDIM statements
 ******************************************************************************/

/* TCLEAR() 
 * perform the CLEAR statement, to clear the screen. 
 * NOte: the 'clear' statement for the UNIX terminal depends on the terminfo 
 * or termcap database, and searches into the environment for the terminal 
 * type, in order to deduce how to clear the screen, but this is done 
 * implicitly, and thus tbas does not take in account any other action. */
void TCLEAR(void) {
#ifndef NOUNIX
	system("clear");
#else
	system("cls");
#endif
}
	

/* TCOMMON() 
 * perform the COMMON statement, to pass variables to a chained program,
 * according to the following syntax:
 * 
 * COMMON A7,name$,C22,Days(),Extract$
 * to pass only variables in the list
 * 
 * COMMON ALL
 * to pass all variables;
 *
 * COMMON NONE
 * to pass no variables;
 *
 * Note: open-close parentheses () indicate arrays. */
void TCOMMON(void) {
	int i=0, done=FALSE, cn;

	/* argument required */
	if (!*p) perror_(__LINE__,125,l);

	/* ALL case, equivalent to OPTION RESET OFF */
	if (!strncasecmp_(p,"ALL",3)) {
		p+=3;
		isCommon=FALSE;
		isToReset=FALSE;
		return;
	}
	/* NONE case, equivalent to OPTION RESET ON */
	else if ((cn=recog(2,p,"NONE","CLEAR"))) {
		p+=cn;
		/* As a safety measure, the isCommon flag is reset */
		for(i=0; i<LIMIT*2;i++) vn[i].isCommon=FALSE;
		isCommon=FALSE;
		isToReset=TRUE;
		return;
	}
	/* list case */
	else while (*p) {
		char name[COMPSTR], *nn=name;
		isCommon=TRUE;
		/* advance to next, discarding separators */
		if (*p==',' || *p==';') p++;
		if (isbeginnamechar(p)){
			while (isnamechar(p)) *nn++=*p++;
			*nn='\0';
		}
		else perror_(__LINE__,35,l);
		i=1;
		done=FALSE;
		while (i<=decCount) {
			/* variable name found */
			if (!strncmp(name,vn[i].name,COMPSTR)) {
				if (*p=='(' && *(p+1)==')') {
					/* if () is found: array*/
					if(!vn[i].isArr) perror_(__LINE__,34,l);
					p+=2;
				}
				/* if () is not found: no array*/
				else if (vn[i].isArr) perror_(__LINE__,34,l); 
				vn[i].isCommon=done=TRUE;
				break;
			}
			/* erase the Common property in all other cases 
			 * Note: I check against not-TRUE because isCommon
			 * may be uninitialized...*/
			else if (vn[i].isCommon!=TRUE) vn[i].isCommon=FALSE;
			i++;
			
		}
		/* if variable was not found, issue an error*/
		if (!done) perror_(__LINE__,170,l);
		isToReset=FALSE;
	}
}


/* ERASE()
 * perform the ERASE statement, to erase arrays and variables from memory, 
 * with the following syntax:
 *
 * ERASE <var>, <var>, .
 * 
 * Its general action is to erase references for variables and arrays, 
 * free arrays memory and set strings and numbers to null. */
void ERASE(void) {
	int arr=0, type=0, i=decCount;
	char name[COMPSTR], *n=name;

	/* argument required */
	if (!*p) perror_(__LINE__,125,l);

	/* get DEC name */
	while (isnamechar(p)) *n++=*p++;
	if (*p=='%') *n++=*p++;
	else if (*p=='$') *n++=*p++;
	*n='\0';

	/* search DEC item */
	while (i>0) {	// search all variables
		if (vn[i].name!=NIL && !strcmp(vn[i].name,name)) break;
		i--;
	}

	/* if not found is a wrong ERASE */
	if (i<=0) perror_(__LINE__,42,l);

	/* erase DECLARED */
	type=vn[i].rettype;
	vn[i].rettype=0;
	arr=vn[i].isArr;
	vn[i].isArr=0;
	vn[i].isInt=0;
	vn[i].isCommon=FALSE;

	/* debug features */
	if (isDebug) {
		printf(INVCOLOR);
		printf("Erased %s [%d] %s\n",arr?"array":vn[i].isConst?"const":"variable", i,\
				vn[i].name!=NIL?vn[i].name:"obj.");
		printf(RESETCOLOR);
	}

	/* task is done for DECLARE variables; reset memory */
	if (!arr) {
		if (type==NUM) P[i]=0.0;
		else if (type==STR) PS[i][0]='\0';
	}
	/* free array memory and reset array pointers */
	else if (arr && type==NUM) {
		if (dim[i].elem!=NIL) {
			free(dim[i].elem);
			dim[i].elem=NIL;
		}
		dim[i].type=dim[i].brow=dim[i].bcol=0;
		dim[i].row=dim[i].col=0;
	}
	else if (arr && type==STR){
		if (dimS[i].elem!=NIL) {
			free(dimS[i].elem);
			dimS[i].elem=NIL;
		}
		dimS[i].type=dimS[i].brow=dimS[i].bcol=0;
		dimS[i].row=dimS[i].col=0;
	}

	/* get rid of it in the vn array by putting it last and discarding it */
	swapDEC(i,decCount);
	decCount--;

	if (!strncasecmp_(p,"()",2)) p+=2;

	/* repeat for other variables */
	if (*p==',') {
		p++;
		ERASE();
	}
}


/* REDIM()
 * perform the REDIM statement, to redimension existing arrays, with the
 * following syntax:
 *
 * REDIM <arr>(newM,newN),<arr>(newM,newN),...
 * 
 * REDIM redefines memory for existing arrays with conservation of memory. 
 * Note: if the new row and col values determine a space which can be contained 
 * in the original array space (brow and bcol), a simple redefinition of row 
 * and col is done. If the new row and col determine a space which is larger 
 * that the existing space, a new space is created, the old space is copied in 
 * the new space, the array is completely redefined (brow, bcol, row and col 
 * are all set to new values), and it appears as if it was created as such from
 * the start; in this last case only, the original space is freed. */
void REDIM(void) {
	int cs, type=NUM, var=0;
	int M=0,N=0,T=1, total=0, newtotal=0;
	
	/* argument required */
	if (!*p) perror_(__LINE__,125,l);

	/* get variable id */
	if ((cs=findDEC(p)) && _arr) var=cs, p+=_len, type=_rettype;
	else perror_(__LINE__,42,l);

	/* get new dimensions M and N (required!) */
	if (*p=='(') {
		p++;
		M=(int)S();
		if (*p==',') {
			p++;
			N=(int)S();
			T=2;
		}
	}
	else perror_(__LINE__,35,l);
	
	/* check if the required space can be contained in the existing space, 
	 * and if not create a new space, both for vectors and matrices, both 
	 * for the numerical and the strings worlds */
	/* numerical arrays */
	if (type==NUM) {
		if (dim[var].type==1) {
			newtotal=sizeof(double)*((M+1)*2+1);
			total=sizeof(double)*((dim[var].row+1)*2+1);
		}
		else if (dim[var].type==2) {
			newtotal=sizeof(double)*((M+1)*(N+1)+1);
			total=sizeof(double)*((dim[var].brow+1)*(dim[var].bcol+1)+1);
		}
		/* space is not sufficient */
		if (newtotal>total) {
			double *elemcopy=xmalloc((size_t)newtotal);
			if (!elemcopy) perror_(__LINE__,190,l);
			memset(elemcopy, 0, newtotal);
			if (dim[var].elem!=NIL) {
				memcpy(elemcopy,dim[var].elem,total);
				free(dim[var].elem);
				dim[var].elem=NIL;
			}
			dim[var].elem=xmalloc((size_t)newtotal);
			if (!dim[var].elem) perror_(__LINE__,190,l);
			memcpy(dim[var].elem,elemcopy,total);	
			dim[var].brow=M;
			dim[var].bcol=N;
			free(elemcopy);
		}
		/* re-attribute values */
		dim[var].row=M;
		dim[var].col=N;
		dim[var].type=T;
	}
	/* string arrays */
	else {
		if (dimS[var].type==1) {
			newtotal=sizeof(char)*((M+1)*COMPSTR+1);
			total=sizeof(char)*((dimS[var].brow+1)*COMPSTR+1);
		}
		else if (dimS[var].type==2) {
			newtotal=sizeof(char)*((M+1)*(N+1)*COMPSTR+1);
			total=sizeof(char)*((dimS[var].brow+1)*(dimS[var].bcol+1)*COMPSTR+1);
		}
		/* space is not sufficient */
		if (newtotal>total) {
			char *elemcopy=xmalloc((size_t)newtotal);
			if (!elemcopy) perror_(__LINE__,190,l);
			memset(elemcopy, 0, newtotal);
			if (dimS[var].elem!=NIL) {
				memcpy(elemcopy,dimS[var].elem,total);
				free(dimS[var].elem);
				dimS[var].elem=NIL;
			}
			dimS[var].elem=xmalloc((size_t)newtotal);
			if (!dimS[var].elem) perror_(__LINE__,190,l);
			memcpy(dimS[var].elem,elemcopy,total);	
			dimS[var].brow=M;
			dimS[var].bcol=N;
			free(elemcopy);
		}
		/* re-attribute values */
		dimS[var].row=M;
		dimS[var].col=N;
		dimS[var].type=T;
	}

	/* repeat for other variables */
	if (*p==',') {
		p++;
		REDIM();
	}
}

/******************************************************************************
 * The following routines define the QUEUE, ROUTE and LPRINT statements
 ******************************************************************************/

/* openRoute()
 * setup the routing channel for printing in a file and then send it to
 * the printer (with ROUTE or at the end of program) 
 * routefile is the file name to be open and setup in the routing channel
 * System function for the QUEUE and LPRINT statements */
void openRoute(void) {
	channel[ROUTECHANNEL].stream=fopen(routefile,"w+");
	if (!channel[ROUTECHANNEL].stream) perror_(__LINE__,153,l);
	strncpy_(channel[ROUTECHANNEL].ioname,routefile,COMPSTR);
	channel[ROUTECHANNEL].type=SEQ;
	channel[ROUTECHANNEL].mode=NUMBER;
	channel[ROUTECHANNEL].fcounter=0;
	channel[ROUTECHANNEL].wayout=VWRITE;
	channel[ROUTECHANNEL].currseek=1;
	channel[ROUTECHANNEL].reclen=NUMRECLEN;
	channel[ROUTECHANNEL].randomaccessed=FALSE;
	channel[ROUTECHANNEL].quote=NOQUOTE;
	channel[ROUTECHANNEL].margin=MARGINDEF;
	channel[ROUTECHANNEL].page=0;
	channel[ROUTECHANNEL].pagerow=1;
	channel[ROUTECHANNEL].height=PAGEHEIGHT;
	channel[ROUTECHANNEL].pagenumber=FALSE;
	channel[ROUTECHANNEL].writer=NOUSE;
	channel[ROUTECHANNEL].reader=NOUSE;
	channel[ROUTECHANNEL].fprintn=1;
}


/* QUEUE 
 * perform the QUEUE statement, that queues all console output to a temporary
 * file rather than to screen from now on. Syntax:
 *
 * QUEUE
 * QUEUE "filename"
 * 
 * If followed by nothing, QUEUE uses a standard temporary file name to store 
 * the printing, and this file will be deleted at the end.
 * If followed by a string, that 'string' is taken as the temporary file name; 
 * in this case the file is retained at the end of program. */
void QUEUE(void) {
	char stamp[COMPSTR];
	sprintw(stamp,COMPSTR,"_%0d%0d%0d",getHour(),getMin(),getSec());
	/* name given */
	if (*p) {
		strncpy_(routefile,SS(),COMPSTR);
		fileStays=TRUE;
	}
	/* tempname */
	else {
		strncpy_(routefile,TEMPROUTE,COMPSTR);
		strncat_(routefile,stamp,COMPSTR);
		fileStays=FALSE;
	}
	openRoute();
	isQueued=TRUE;
}


/* LPRINT()
 * perform the LPRINT/LINEPRINT statement, that temporary redirects the
 * output to a file, which will be sent to printer when ROUTE is executed
 * or at the end of execution by END(). Syntax is the very same of PRINT
 * It is the very same statement of PRINT, which is called accordingly */
void LPRINT(void) {
	char stamp[COMPSTR];
	sprintw(stamp,COMPSTR,"%0d%0d%0d",getHour(),getMin(),getSec());
	/* this is the first LPRINT statement */
	if (!isLPrintEnabled) {
		strncpy_(routefile,TEMPROUTE,COMPSTR);
		strncat_(routefile,stamp,COMPSTR);
		openRoute();
		fileStays=FALSE;
		isLPrintEnabled=TRUE;
	}
	isLPrint=TRUE;
	PRINT(TRUE);
}


/* ROUTE()
 * perform the ROUTE statement, that sends the route file content to the 
 * default printer; this command cannot be executed before the QUEUE statement.
 * Notes: if fileStays is TRUE, do not remove the route file; if otherwise 
 * fileStays is FALSE, save the file to a secure position and update the
 * queueFileCount counter.
 * At the end, the QUEUE feature is disabled, and standard output is restored.
 * Note: for the sake of completeness, in END(), if for any reason, isQueue
 * is still TRUE at exit, the ROUTE() function is called again. */
void ROUTE(void) {
	char command[COMPSTR];
	if (!(isQueued+isLPrintEnabled)) perror_(__LINE__,154,l);
	if (channel[ROUTECHANNEL].stream)
		fclose(channel[ROUTECHANNEL].stream);
	/* if ROUTE is open, send the file to printer */
	if (isRoute) {
		sprintw(command,COMPSTR,"%s %s",ROUTESTRING,routefile);
		system(command);
	}
	/* or else, copy file to secure user file (in case of a temporary
	 * file created by the O.S.) */
	else if (!fileStays) {
		char com[COMPSTR];
		sprintw(com,COMPSTR,"cp -f %s %s%d.txt",routefile,TEMPROUTE,queueFileCount);
		system(com);
		queueFileCount++;
	}
	isQueued=FALSE;
	isLPrint=isLPrintEnabled=FALSE;
	if (!fileStays) remove(routefile);
	routefile[0]=channel[ROUTECHANNEL].ioname[0]='\0';
}


/******************************************************************************
 * The following routines define the GET#, PUT# and WAIT statements
 ******************************************************************************/

/* DOGET() 
 * perform the GET# statement, according to the following syntax:
 *
 * GET #N, M
 * GET #N, A$
 * 
 * If followed by a numerical variable, it will be instantiated with the ASCII 
 * value read in the range 0-255.
 * If followed by a string variable, this will be instantiated with the 
 * one-character string containing the ASCII character read.
 * Warning: characters in the range 128-255 are printed following the terminal 
 * coding, and this may vary on each computer. */
void DOGET(void) {
	int val=0, ch=0, cn, isNum=TRUE;

	/* argument required */
	if (!*p) perror_(__LINE__,125,l);

	/* sharp symbol is optional*/
	if (*p=='#') p++;

	/* get channel */
	if (isanynumber(p)) {
		ch=(int)S();
		/* checks if channel is valid */
		if (ch<1 || ch>STREAMS) perror_(__LINE__,133,l);
		/* checks if file is open and RANDOM and in READ mode*/
		if (!channel[ch].isopen) perror_(__LINE__,80,l);
		if (channel[ch].type!=SEQ) perror_(__LINE__,101,l);
		if (channel[ch].wayout!=VREAD) perror_(__LINE__,72,l);
		if (feof(channel[ch].stream)) 
			perror_(__LINE__,58,l,channel[ch].ioname);
	}
	else perror_(__LINE__,65,l,makechar(p),"a valid channel id");

	/* discard comma or semicolon */
	if (*p==',' || *p==';') p++;

	/* determine output data */
	if (isanystring(p)) {
		if ((cn=findDEC(p)) && _rettype==STR && !_arr) p+=_len;
		else perror_(__LINE__,42,l);
		isNum=FALSE;
	}
	else {
		if ((cn=findDEC(p)) && _rettype==NUM && !_arr) p+=_len;
		else perror_(__LINE__,42,l);
	}	
	/* get value and put into proper variable */
	val=fgetc(channel[ch].stream);
	if (isNum) P[cn]=val;
	else PS[cn][0]=val, PS[cn][1]='\0';
}


/* DOPUT() 
 * perform the PUT# statement, putting to file the ASCII value of the variable 
 * following, according to the following syntax:
 *
 * PUT #N, A$
 * PUT #N, M
 * 
 * If followed by a string variable, convert the first character in the
 * string to its ASCII value and puts it to file;
 * otherwise, put the numerical value as ASCII code to file */
void DOPUT(void) {
	int ch=0, val=0;
	char str[COMPSTR];

	/* argument required */
	if (!*p) perror_(__LINE__,125,l);

	/* sharp symbol is optional*/
	if (*p=='#') p++;

	/* get channel */
	if (isanynumber(p)) {
		ch=(int)S();
		/* checks if channel is valid */
		if (ch<1 || ch>STREAMS) perror_(__LINE__,133,l);
		/* checks if file is open and RANDOM and in READ mode*/
		if (!channel[ch].isopen) perror_(__LINE__,80,l);
		if (channel[ch].type!=SEQ) perror_(__LINE__,101,l);
		if (channel[ch].wayout!=VWRITE) perror_(__LINE__,74,l);
	}
	else perror_(__LINE__,65,l,makechar(p),"a validi channel id");

	/* discard comma or semicolon */
	if (*p==',' || *p==';') p++;

	/* determine output data */
	if (isanystring(p)) {
		strncpy_(str,SS(),COMPSTR);
		val=str[0];
	}
	else val=(int)S();

	/* put value onto stream */
	fputc(val,channel[ch].stream);
}


/* DOWAIT()
 * perform the WAIT statement, according to the following syntax:
 *
 * WAIT
 * WAIT N
 * 
 * If the argument is present, wait for that amount of seconds, otherwise
 * wait for 3 seconds. 
 * Note: the current implementation doesn't check for interrupts or flag,
 * while waiting, because it does not return control to the system */
/* FUTURE WORK TODO: make it more fluid, make it read some state and make 
 * it pass through parse indefinitely until something happens (maybe through 
 * a variable state that makes parse() NOT update variable 'l' and thus 
 * re-run this statement) */
void DOWAIT(void) {
	struct timeval wait_time;  
	time_t wait_now, wait_until;
	int howmuch=0;

	/* get user seconds; if nothing is specified, wait for 3 seconds */
	if (*p) howmuch=(int)S();
	else howmuch=3;
	/* check value */
	if (howmuch<0) howmuch*=-1;
	if (howmuch>60) howmuch=60;
	/* get current instant */
	gettimeofday(&wait_time, NULL);
	wait_now=((double)wait_time.tv_sec);
	/* wait for the desired time */
	do {
		gettimeofday(&wait_time, NULL);
		wait_until=((double)wait_time.tv_sec);
	} while (wait_until<wait_now+howmuch);
}


/******************************************************************************
 * The following routines define the ON ERROR/NO DATA/ON ATTENTION statements 
 ******************************************************************************/

/* ONERROR()
 * implementation of the ON ERROR statement; in can be used in these ways:
 *
 * ON ERROR [GOTO|THEN] <label>
 * - This form traps all errors, redirecting to line with label <label>
 * 
 * ON ERROR <err> [GOTO|THEN] <label>
 * - This form traps error with code <err>, redirecting to line with label 
 *   <label>
 * 
 * ON ERROR
 * - This form disables the ON ERROR features
 * 
 * If ON ERROR is used more than once, only the most recent will be used.
 * ON ERROR can be used with 'WITH ERROR IN' and 'WITH ERROR USE' structures */
void ONERROR(void) {
	int val=0, trap=0;
	/* if ON ERROR is followed by nothing, disable the ON ERROR feature */
	if (!*p) {
		isOnErrorPos=isOnErrorTrap=errLine=errValue=0;
		return;
	}
	/* parse and discard GO TO / THEN */
	if (!strncasecmp_(p,"GOTO",4)||(!strncasecmp_(p,"THEN",4))) p+=4;
	/* there's an any-name label */
	if (isbeginnamechar(p)) val=getLabelAddr(p, &p);
	/* there's a number: is it a code number or an address? */
	else {
		char *pp=p;
		val=strtol(pp,&pp,10);
		/* there's something after the number: it is the code */
		if (*pp) {
			trap=val;
			/* parse and discard GO TO / THEN */
			if (!strncasecmp_(pp,"GOTO",4)||\
					(!strncasecmp_(pp,"THEN",4))) 
				pp+=4;
			val=getLabelAddr(pp,&pp);
			p=pp;
		}
		/* there's nothing after the number: re-get it as label */
		else val=getLabelAddr(p,&p);
	}
	/* enable the ON ERROR feature */
	if (val>=0) {
		isOnErrorPos=val;
		isOnErrorTrap=trap;
	}
	/* or else print error */
	else syserr_(__LINE__,61,l);
}


/* ONATTENTION()
 * implementation of the ON ATTENTION statement, to intercept the user
 * interrupt and redirect program flow; in can be used in these ways:
 *
 * ON ATTENTION [GOTO|THEN] <label>
 * - This form traps interrupts redirecting to line with label <label>
 * 
 * ON ATTENTION
 * - This form disables the ON ATTENTION features
 * 
 * Note: ON ATTENTION makes no use of perror_(), and thus is does not activate
 * errValue and errLine; rather, it sets and uses arrValue and arrLine, for
 * it proper functioning; ON ATTENTION can be used with all the error-catching 
 * strategies, since it only cares to intercept the user interrupts */
void ONATTENTION(void) {
	int val=0;
	/* if ON ATTENTION is followed by nothing, disable the feature */
	if (!*p) {
		isOnAttentionPos=arrLine=arrValue=0;
		/* Restore trapping interrupts */
		signal(SIGHUP, sigint);
		signal(SIGINT, sigint);
		return;
	}
	/* parse and discard GO TO / THEN */
	if (!strncasecmp_(p,"GOTO",4)||(!strncasecmp_(p,"THEN",4))) p+=4;
	/* get destination label */
	val=getLabelAddr(p,&p);
	/* enable the ON ATTENTION feature */
	if (val>=0) {
		/* Restore trapping interrupts */
		signal(SIGHUP, sigtrap);
		signal(SIGINT, sigtrap);
		isOnAttentionPos=val;
	}
	/* or else print error */
	else syserr_(__LINE__,61,l);
}


/* JUMP()
 * implementation of the JUMP statement, to be used only within an ON ERROR, ON
 * ATTENTION or NO DATA structure, and provided the error that caused the jump 
 * to the handling section appeared. Syntax:
 *
 * JUMP <pl>
 * JUMP(<pl>)
 *
 * The jump is done to the BASIC file physical line number <pl>, not the label;
 * it was conceived to be used in conjunction with NXL() and ESM()/ERL(), 
 * thus the argument of JUMP must be a number */
void JUMP(void) {
	int val=0;
	
	/* argument required */
	if (!*p) syserr_(__LINE__,125,l);

	/* cannot use JUMP if ON ERROR/NO DATA features are both disabled 
	 * i.e. no error condition was raised */
	if (!(isOnErrorPos + isNoDataPos + isOnAttentionPos)) {
		errValue=errLine=arrValue=arrLine=0;
		syserr_(__LINE__,21,l);
	}
	/* get destination label */
	if (*p) val=(int)S(); // no getLabelAddr() here!
	/* check if destination is legal and JUMP!*/
        if (val<1 || val > endCore) syserr_(__LINE__,61,l,val);
	else l=val-1;
}


/* RESUME()
 * implementation of the RESUME statement, to address execution to the line 
 * following the line that caused error; RESUME cannot be used out of an 
 * ON ERROR/ON ATTENTION feature and, in any case, not if no error/attention 
 * condition is active; this guarantees that since the ON ERROR/ON ATTENTION 
 * handler is not a confined code, as WHEN ERROR handlers are, the eventual 
 * execution of such lines in normal condition (that is with no error occurred
 * and anyway out of an ON ERROR segment) an error is printed, avoiding further
 * execution of improper code. */
void RESUME(void) {
	/* cannot use RESUME outside of a ON ERROR/ON ATTENTION structure */
	if (!(errValue+arrValue)) syserr_(__LINE__,13,l);
	if (!(isOnErrorPos+isOnAttentionPos)) syserr_(__LINE__,12,l);
	/* re-execute the failing line resetting error conditions */
	if (errValue) {
		l=errLine; // go to line following the error line!
		errLine=errValue=0;
	}
	else if (arrValue) {
		l=arrLine; // go to line following the interrupt line!
		arrLine=arrValue=0;
	}
}


/* NODATA()
 * implementation of the NO DATA statement, to catch the error condition of the
 * OUT OF DATA message (error 110); it can be used in these ways:
 *
 * NO DATA [GOTO|THEN] <label>
 * - This form traps error 110 redirecting to line with label <label>
 * 
 * NO DATA
 * -  This form disables the NO DATA features
 * 
 * Note: If NO DATA is used more than once, only the most recent will be used.
 * NO DATA can be used with 'WITH ERROR IN' and 'WITH ERROR USE' structures
 * Note: this command may be simulated by ON ERROR 110 GOTO <label>, but in
 * perror_() NO DATA has precedence over ON ERROR and is evaluated earlier. */
void NODATA(void) {
	int val=0;
	/* if NO DATA is followed by nothing, disable the NO DATA feature */
	if (!*p) {
		isNoDataPos=isOnErrorTrap=errLine=errValue=0;
		return;
	}
	/* parse and discard GO TO / THEN */
	if (!strncasecmp_(p,"GOTO",4)||(!strncasecmp_(p,"THEN",4))) p+=4;
	/* get destination label */
	val=getLabelAddr(p,&p);
	/* enable the NO DATA feature */
	if (val>=0) {
		isNoDataPos=val;
		isOnErrorTrap=110; // preset code
	}
	/* or else print error */
	else syserr_(__LINE__,61,l);
}


/******************************************************************************
 * The following routines define the WHEN ERROR IN/USE/RETRY and the WHEN
 * ERROR USE <handler>/HANDLER/END HANDLER statements.
 * These statements come from the OpenVMS and HP traditions, that implemented 
 * such statements in their BASIC language for their Eighties mainframe; 
 * they in turn owe the design of WHEN ERROR to the Dartmouth tradition evolved 
 * in the TRUE BASIC compiler devised by John Kemeny.
 ******************************************************************************/

/* WHENERROR()
 * activate the WHEN ERROR IN/WHEN ERROR USE <handler> features, that intercept
 * the error (in perror_() and get to go to the handler (USE section for WHEN
 * ERROR IN and the handler <handler> address for WHEN ERROR USE), handler
 * where the error is treated by user code. Syntax:
 *
 * WHEN ERROR IN 
 * WHEN ERROR USE <handler>
 *
 * The first form isolates the fragments of code between WHEN and USE, and if
 * an error arises, the USE section is called, to perform the resolution
 * statements. The second form set <handler> (which must exist) as the
 * resolution executor. */
void WHENERROR(void) {
	int i=0;
	char hname[COMPSTR], *h=hname;
	
	/* check if WHEN ERROR is inside the sub */
	if (isDefRun || isSubRun) syserr_(__LINE__,22,l);
	
	/* perform WHEN ERROR IN */
	if (!strncasecmp_(p,"IN",2)) {
		isWhenErrorInPos=l; 	// WHEN address
		isUsePos=whenAddr[l];	// USE address
		return;
	}

	/* if after ERROR there is 'USE', perform WHEN ERROR USE <handler>;
	 * otherwise it's rubbish text */
	if (!strncasecmp_(p,"USE",3)) p+=3;
	else syserr_(__LINE__,19,l);

	/* get WHEN ERROR USE handler name */
	if (isbeginnamechar(p)) {
		while (isnamechar(p)) *h++=*p++;
		*h='\0';
	}
	else syserr_(__LINE__,11,l);

	/* check if WHEN ERROR USE is used into a SUB, which is not possible */
	if (subTOS) {
		if (sn[subTOS].start<l && sn[subTOS].end>l)
			syserr_(__LINE__,23,l);
	}

	/* search handler in the handler database created by preParse
	 * and, if found, set public handler number */
	while (i<=handlerTOS) {
		if (handler[i].hname && !strcmp(handler[i].hname,hname)) break;
		i++;
	}

	/* if handler is not found print error */
	if (!i || i>handlerTOS) syserr_(__LINE__,20,l,hname);
	/* set parameters */
	else { 
		isHandlerNum=i;
		isWhenErrorUsePos=l;		// WHEN pos
		endWhenUse=whenAddr[l];		// END WHEN pos
	}
}


/* USE()
 * activate the USE handler section of the WHEN ERROR IN feature; 
 * in this handler, errors found are regularly printed and errValue and errLine 
 * are available as error codes, because they were instantiated by perror_() */
void USE(void) {
	/* cannot use USE outside of a WHEN ERROR IN structure */
	if (!isWhenErrorInPos) syserr_(__LINE__,18,l);
	/* no error occurred: go to END WHEN immediately */
	if (!errValue) l=whenAddr[l]-1;
	/* error: execute the USE section and store the END WHEN address
	 * for eventual future usage (the EXIT HANDLER statement) */
	else endWhenIn=whenAddr[l], isUseActive=TRUE;
}


/* EXITWHEN()
 * causes the immediate exit of the USE handler section of the WHEN ERROR IN
 * feature, jumping to END WHEN unconditionally */
void EXITWHEN(void) {
	/* cannot use EXITWHEN outside of a Use section */
	if (!isUseActive) syserr_(__LINE__,24,l);
	/* go to END WHEN immediately */
	l=endWhenIn-1;
}


/* RETRY()
 * address execution to the line that caused error, hopefully after the user
 * has treated the error in the USE section; 
 * RETRY cannot be used outside of a USE section (the only one with error 
 * codes enabled and not null) */
void RETRY(void) {
	/* cannot use RETRY outside of a USE handler/WHEN ERROR IN structure 
	 * or out of a HANDLER handler/WHEN ERROR USE structure*/
	if (!(isUseActive + isHandlerActive)) syserr_(__LINE__,17,l);
	/* re-execute the failing line resetting error conditions */
	l=errLine-1;
	errLine=errValue=0;
	isUseActive=isHandlerActive=0;
}


/* DOCONTINUE()
 * perform the CONTINUE statement (which works without argument) to resume
 * execution from line following the mistaken line; it works in a similar
 * way of RETRY, but the mistaken line is overridden.
 * Note: in case arguments are used, pass execution to EXIT, which controls
 * CONTINUE in the cycles statements */
void DOCONTINUE(void) {
	/* cannot use CONTINUE outside of a USE handler/WHEN ERROR IN structure 
	 * or out of a HANDLER handler/WHEN ERROR USE structure*/
	if (!*p && !(isUseActive + isHandlerActive)) syserr_(__LINE__,16,l);
	else if (!*p) {
		/* go to line following the error condition */
		l=errLine;
		errLine=errValue=0;
		isUseActive=isHandlerActive=0;
	}
	else EXIT(1);
}


/* HANDLER()
 * start a handler section of a WHEN ERROR USE structure; syntax:
 *
 * HANDLER <handler>
 *
 * where <handler> is any valid name; the resolution code is terminated by
 * END HANDLER; this function checks also if it's the active handler that calls 
 * the handler or if the normal flow is getting here; in such a case, jump 
 * safely to END HANDLER. */
void HANDLER(void) {
	/* check if HANDLER is inside the sub */
	if (isDefRun || isSubRun) syserr_(__LINE__,23,l);
	
	/* no error occurred or this is not the right handler: 
	 * go to END HANDLER immediately */
	if (!errValue || handler[isHandlerNum].pos!=l) { 
		l=handlerAddr[l]-1;
	}
	/* otherwise don't do nothing and enter the handler */
	else isHandlerActive=TRUE;
}


/* ENDHANDLER()
 * perform the END HANDLER statement to terminate a handler section of a 
 * WHEN ERROR USE structure; if it was a handler that called the handler 
 * (and not a normal program flow) jump back to the END WHEN line of the 
 * calling WHEN ERROR USE structure (that is the one responsible for the 
 * enabling of the current handler reference, the public handler number). */
void ENDHANDLER(void) {
	/* go to END WHEN in case handler was active */
	if (errValue) l=endWhenUse-1, isHandlerActive=FALSE;
	/* otherwise don't do nothing and let program flow after END HANDLER */
}


/* EXITHANDLER()
 * perform the EXIT HANDLER statement: in a WHEN ERROR structure, jump to the 
 * END WHEN, to exit the WHEN section; variables and flags are all reset 
 * accordingly */
void EXITHANDLER(void) {
	/* EXIT HANDLER as handler terminator in a WHEN ERROR USE structure */
	if (isWhenErrorUsePos) l=endWhenUse-1;
	/* EXIT HANDLER in a WHEN ERROR IN structure */
	else if (isWhenErrorInPos) l=endWhenIn-1, isHandlerActive=FALSE;
	else syserr_(__LINE__,15,l); // warning
}


/* ENDWHEN()
 * perform the END WHEN statement, to terminate a WHEN ERROR structure, 
 * resetting parameters and letting prosecuting the program execution 
 * starting from subsequent line */
void ENDWHEN(void) {
	char *pp;
	/* cannot terminate a WHEN ERROR structure outside of it */
	if (!(isWhenErrorInPos+isWhenErrorUsePos)) {
		errValue=errLine=0;
		syserr_(__LINE__,14,l);
	}
	/* reset data of the WHEN ERROR IN features */
	if (checkCommand(m[whenAddr[l]],"WHEN ERROR IN",&pp)) {
		isWhenErrorInPos=isUsePos=endWhenIn=0;
		isUseActive=FALSE;
	}
	/* reset data of the WHEN ERROR USE features */
	if (checkCommand(m[whenAddr[l]],"WHEN ERROR USE",&pp)) {
		isWhenErrorUsePos=isHandlerNum=endWhenUse=0;
		isHandlerActive=FALSE;
	}
	/* reset the error codes */
	errLine=errValue=0;
}


/* CAUSEERROR()
 * raise an error condition; it can be used with this syntax:
 *
 * CAUSE ERROR
 * - raise error condition 0 (a default error condition not related to BASIC)
 * 
 * CAUSE ERROR <n>
 * - raise error condition <n> 
 * 
 * If <n> is not a valid error number code, it is resized accordingly 
 * (no error message, here, it would be misinterpreted!) */
void CAUSEERROR(void) {
	int val=0;
	if (!isdigit(*p)) val=0;
	else val=strtol(p,&p,10);
	if (val < 0) val=0;
	if (val > ERRLIM) val=ERRLIM;
	perror_(__LINE__,val,ERRORCODE); // it MUST use perror_()!
}


/******************************************************************************
 * The following routines define the OPTION and COMMAND statements; The COMMAND 
 * statement is not a typical BASIC statement. The OPTION statement was 
 * instead implemented in various form in many BASIC dialects since Seventies.
 ******************************************************************************/


/* COMMAND()
 * read following string (in any format) and send it to system, with the
 * following syntax:
 *
 * COMMAND command
 * COMMAND(command)
 * 
 * Note: The argument may also be enclosed in parentheses.
 * Note: this is a potentially dangerous command, since no control is
 * made upon the string. To be used it with care. */
void COMMAND(void) {
	char sys[COMPSTR];
	int par=FALSE;
	/* argument required */
	if (!*p) perror_(__LINE__,125,l);

	if (*p=='(') par=TRUE, p++;
	if (!isanystring(p)) perror_(__LINE__,161,l);
	strncpy_(sys,SS(),COMPSTR);
	if (*p==')') p++;
	else if (par) {
		printf("p=|%s|\n",p);
		perror_(__LINE__,37,l);
	}
	system(sys);
}


/* OPTION()
 * evaluates some extra-BASIC option, those influencing the interpreter and
 * not the language;
 * This option is built to maximize the compatibility effect, showing no errors
 * in case of unrecognized option or unresolved/unfitting arguments
 *
 * OPTION				default value
 * --------------------------------------------------
 * OPTION ANGLE DEGREES|RADIANS         RADIANS
 * OPTION BASE 0|1                      0
 * OPTION CAPS ON|OFF*                  OFF 
 * OPTION CASE ON|OFF*                  OFF
 * OPTION COMPARISON RELATIVE|ABSOLUTE  RELATIVE
 * OPTION DEBUG ON|OFF*                 OFF
 * OPTION DIFFERENCE 0n|OFF            6|OFF
 * OPTION ECHO ON|OFF*                  ON (for UNIX only)
 * OPTION EXPLICIT ON|OFF*              OFF
 * OPTION FORMAT AMERICAN|EUROPEAN      AMERICAN
 * OPTION HEADER                        OFF (caught into preparse)
 * OPTION PRECISION -1|016|OFF         OFF/-1
 * OPTION PROMPT ON|OFF*                ON
 * OPTION RAWPRINT ON|OFF*              OFF
 * OPTION RESET ON|OFF*                 ON
 * OPTION ROUTING ON|OFF*               ON
 * OPTION SPACING ON|OFF*               OFF
 * OPTION TAB 0n|OFF                   ZONEBASE|OFF
 * OPTION TIMING                        OFF
 * OPTION VINTAGE                       OFF  (caught into preparse)
 * OPTION ZERO 0n|OFF*                 0|OFF
 * Note: OPTION with (*) can be enabled with the sole identifier (e.g. OPTION 
 * SPACING ON can be written also as OPTION SPACING, with obvious purposes)
 * Note: since the header should be written before any output, OPTION HEADER
 * and OPTION VINTAGE are caught into preParse() 
 * Note: OPTION is iterable */
void OPTION(void) {
	/* OPTION ANGLE RADIANS|DEGREES 
	 * This option sets the default angle measure in radians or degrees.
	 * Default is RADIANS */
	if (!strncasecmp_(p,"ANGLE",5)) {
		p+=5;
		if (!strncasecmp_(p,"RADIANS",7)) isRadians=TRUE, p+=7;
		else if (!strncasecmp_(p,"DEGREES",7)) isRadians=FALSE, p+=7;
	}
	/* Corollaries to the previous statement: OPTION RADIANS and
	 * OPTION DEGREES are also valid as a bonus.*/
	else if (!strncasecmp_(p,"RADIANS",7)) isRadians=TRUE, p+=7;
	else if (!strncasecmp_(p,"DEGREES",7)) isRadians=FALSE, p+=7;
	/* OPTION BASE 0|1 
	 * This option sets the array index to 0 if OPTION BASE 0 is used, 
	 * to 1 if OPTION BASE 1 is used; in any case, matrices are not 
	 * affected, since they work always with index 1 (no change with 
	 * respect to decb).
	 * Default is 0 */
	else if (!strncasecmp_(p,"BASE",4)) {
		int u;
		p+=4;
		u=(int)S();
		if (!u) BASE=0;
		else BASE=1;
	}
	/* OPTION CAPS ON|OFF 
	 * This option sets all output to capital letters
	 * Neither the inner memory state of string nor the PRINT USING 
	 * statement nor the INPUT and MAT INPUT statements are affected by 
	 * this option (strings are input as-is); the effect is only related
	 * to the output, to what appears on the screen.
	 * Default is OFF */
	else if (!strncasecmp_(p,"CAPS",4)) {
		p+=4;
		if (!*p || *p==',') isCap=TRUE;
		else if (!strncasecmp_(p,"ON",2)) isCap=TRUE, p+=2;
		else if (!strncasecmp_(p,"OFF",3)) isCap=FALSE, p+=3;
	}
	/* OPTION CASE ON|OFF 
	 * This option sets case sensitiveness in case of string comparison.
	 * Default is OFF */
	else if (!strncasecmp_(p,"CASE",4)) {
		p+=4;
		if (!*p || *p==',') isCaseSensitive=TRUE;
		else if (!strncasecmp_(p,"ON",2)) isCaseSensitive=TRUE, p+=2;
		else if (!strncasecmp_(p,"OFF",3)) isCaseSensitive=FALSE, p+=3;
	}
	/* OPTION COMPARISON RELATIVE|ABSOLUTE 
	 * This option sets relative or absolute string comparison (relative
	 * considers trimming of spaces, while absolute considers strings
	 * as they are, trailing spaces included).
	 * Default is RELATIVE */
	else if (!strncasecmp_(p,"COMPARISON",10)) {
		p+=10;
		if (!strncasecmp_(p,"RELATIVE",8)) isComparison=RELATIVE;
		else if (!strncasecmp_(p,"ABSOLUTE",8)) isComparison=ABSOLUTE;
		p+=8;
	}
	/* OPTION DEBUG ON|OFF 
	 * This option enable/disable debug 
	 * Default is OFF */
	else if (!strncasecmp_(p,"DEBUG",5)) {
		p+=5;
		if (!*p || *p==',') isDebug=TRUE;
		else if (!strncasecmp_(p,"ON",2)) isDebug=TRUE, p+=2;
		else if (!strncasecmp_(p,"OFF",3)) isDebug=FALSE, p+=3;
	}
	/* OPTION DIFFERENCE <n>|OFF
	 * This option sets the quasi-equal feature (|= operator for numbers), 
	 * which evaluates TRUE the equivalence between two numbers whose
	 * difference is in absolute value lesser than 10^(-<n>)
	 * if OFF is used, the default value is set.
	 * <n> may be any numerical formula result, whose positive integer 
	 * part is used; a zero value cause the setting of the default value.
	 * Default is 1.0E-6)|OFF */
	else if (!strncasecmp_(p,"DIFFERENCE",10)) {
		int val;
		p+=10;
		if (!strncasecmp_(p,"OFF",3)) ZEROEQUAL=1.0E-6, p+=3;
		else {
			val=abs((int)S());
			if (!val) ZEROEQUAL=1.0E-6;
			else ZEROEQUAL=pow(10,-val);
		}
	}
#ifndef NOUNIX
	/* OPTION ECHO ON|OFF 
	 * This option enable/disable echoing during input
	 * Default is ON 
	 * Note: for UNIX only */
	else if (!strncasecmp_(p,"ECHO",4)) {
		p+=4;
		termiosHasOccurred=TRUE;
		if (!strncasecmp_(p,"OFF",2)) {
			modes.c_lflag &= ~ECHO;
			tcsetattr(0, 0, &modes);
			isEcho=FALSE;
			p+=3;
		}
		else if (!*p || !strncasecmp_(p,"ON",2)) {
			tcsetattr(0, 0, &savemodes);
			isEcho=TRUE;
			p+=2;
		}
	}
	else if (!strncasecmp_(p,"NOECHO",6)) {
			modes.c_lflag &= ~ECHO;
			tcsetattr(0, 0, &modes);
			isEcho=FALSE;
			p+=6;
	}
#endif
	/* OPTION EXPLICIT ON|OFF 
	 * This option enable/disable explicit declaration of variables
	 * Default is OFF */
	else if (!strncasecmp_(p,"EXPLICIT",8)) {
		p+=8;
		if (!*p || *p==',') isExplicit=TRUE;
		else if (!strncasecmp_(p,"OFF",3)) isExplicit=FALSE, p+=3;
		else if (!strncasecmp_(p,"ON",2)) isExplicit=TRUE, p+=2;
	}
	/* OPTION FORMAT AMERICAN|EUROPEAN 
	 * This option sets American format (dot for decimal separator,
	 * comma for triple unities - e.g. 123,456.789 for 123456.789) or
	 * European format (comma for decimal separator, tick for triple 
	 * unities - e.g. 123'456,789).
	 * Default is AMERICAN */
	else if (!strncasecmp_(p,"FORMAT",6)) {
		p+=6;
		if (!strncasecmp_(p,"AMERICAN",8)) 
			isAmericanFormat=TRUE;
		else if (!*p || !strncasecmp_(p,"EUROPEAN",8)) 
			isAmericanFormat=FALSE;
		p+=8;
	}
	/* OPTION HEADER is intercepted into preParse() */
	else if (!strncasecmp_(p,"HEADER",6)) p+=6;
	/* OPTION PRECISION <n> 
	 * This option sets numerical output precision for floats;
	 * if OFF is used, the default output is set ("automatic" display");
	 * <n> may be any numerical formula result, whose integer part is used;
	 * if PRECISION = -1 the default output is set ("automatic" display); 
	 * if PRECISION is in the range 0-16 then printNUM() is intercepted
	 * and print is made through fprintf() with the desired precision.
	 * Since emitNUM() is used as the actual printing routine, the numeric 
	 * output in length and counter updating is automatic. 
	 * Default is -1|OFF */
	else if (!strncasecmp_(p,"PRECISION",9)) {
		p+=9;
		if (!strncasecmp_(p,"OFF",3)) PRECISION=-1, p+=3;
		else SETDIGITS(); // it advances p
	}
	/* OPTION PROMPT ON|OFF 
	 * This option sets the displaying of the "?" prompt with INPUT and 
	 * MAT INPUT statements.
	 * Default is ON */
	else if (!strncasecmp_(p,"PROMPT",6)) {
		p+=6;
		if (!*p || *p==',') isPrompt=TRUE;
		else if (!strncasecmp_(p,"ON",2)) isPrompt=TRUE, p+=2;
		else if (!strncasecmp_(p,"OFF",3)) isPrompt=FALSE, p+=3;
	}
	/* OPTION RAWPRINT ON|OFF 
	 * This option sets the raw print style, in which the fcounter is not
	 * considered and thus the printing is as-is.
	 * Default is OFF */
	else if (!strncasecmp_(p,"RAWPRINT",8)) {
		p+=8;
		if (!*p || *p==',') isRawPrint=TRUE;
		else if (!strncasecmp_(p,"ON",2)) isRawPrint=TRUE, p+=2;
		else if (!strncasecmp_(p,"OFF",3)) {
			isRawPrint=FALSE, p+=3;
			/* with OFF, reset parameters for normal output */
			channel[CONSOLE].fcounter=0;
			channel[CONSOLE].pagerow=1;
			channel[CONSOLE].page++;
		}
	}
	/* OPTION RESET ON|OFF 
	 * This option controls the resetting of memory between two 
	 * consecutive CHAIN calls.
	 * Default is RESET */
	else if (!strncasecmp_(p,"RESET",5)) {
		p+=5;
		if (!*p || *p==',') isToReset=TRUE;
		else if (!strncasecmp_(p,"ON",2)) isToReset=TRUE, p+=2;
		else if (!strncasecmp_(p,"OFF",3)) isToReset=FALSE, p+=3;
	} 
	/* OPTION ROUTING ON|OFF 
	 * This option sets the sending to printer of the QUEUE temporary file
	 * Default is ON */
	else if (!strncasecmp_(p,"ROUTING",7)) {
		p+=7;
		if (!*p || *p==',') isRoute=TRUE;
		else if (!strncasecmp_(p,"ON",2)) isRoute=TRUE, p+=2;
		else if (!strncasecmp_(p,"OFF",3)) isRoute=FALSE, p+=3;
	}
	/* OPTION SPACING ON|OFF 
	 * This option controls the printing of compact string matrices,
	 * setting the printing of a single space to separate the output.
	 * Default is OFF */
	else if (!strncasecmp_(p,"SPACING",7)) {
		p+=7;
		if (!*p || *p==',') isSpacing=TRUE;
		else if (!strncasecmp_(p,"ON",2)) isSpacing=TRUE, p+=2;
		else if (!strncasecmp_(p,"OFF",3)) isSpacing=FALSE, p+=3;
	} 
	/* OPTION TAB <n>|OFF
	 * This option sets the tab value
	 * if OFF is used, the standard value is set.
	 * <n> may be any numerical formula result, whose positive integer 
	 * part is used; a zero value sets standard value.
	 * Default is 0|OFF */
	else if (!strncasecmp_(p,"TAB",3)) {
		int val;
		p+=3;
		if (!strncasecmp_(p,"OFF",3)) ZONE=ZONEBASE, p+=3;
		else {
			val=(int)S();
			if (!val) ZONE=ZONEBASE;
			else if (val>=MARGINMIN && val<=MARGINMAX) ZONE=val;
		}
	}
	/* OPTION TIMING ON|OFF 
	 * This option controls the printing of the execution time in seconds
	 * of the running program (or the total of the CHAINed programs)
	 * Default is OFF */
	else if (!strncasecmp_(p,"TIMING",6)) {
		p+=6;
		isTiming=TRUE;
	} 
	/* OPTION VINTAGE is intercepted into preParse() */
	else if (!strncasecmp_(p,"VINTAGE",7)) p+=7;
	/* OPTION ZERO <n>|OFF
	 * This option sets the zero feature, which prints as zero all numbers
	 * that as absolute value are lesser than 10^(-<n>), and enables the
	 * printing of a signed 'inf' in case the number is beyond INF or MINFF
	 * if OFF is used, the feature is disabled.
	 * if OPTION ZERO is used without parameters, use 1E-6 as default limit
	 * <n> may be any numerical formula result, whose positive integer 
	 * part is used; a zero value disables the feature.
	 * Default is 0|OFF */
	else if (!strncasecmp_(p,"ZERO",4)) {
		int val;
		p+=4;
		if (!*p) ZEROLIM=1E-6,isZeroLim=TRUE;
		else if (!strncasecmp_(p,"OFF",3)) isZeroLim=FALSE, p+=3;
		else {
			val=-abs((int)S());
			if (!val) isZeroLim=FALSE;
			else ZEROLIM=pow(10,val), isZeroLim=TRUE;
		}
	}
	/* remove every other instance of OPTION written in the list
	 * (substitute a comma to the 'N' in order to let the program go on */
	else if (!strncasecmp_(p,"OPTION",6)) p+=5, *p=',';
	/* print a warning advising the OPTION is not recognized */
	else perror_(__LINE__,79,l);

	/* iterate */
	if (*p==',') {
		p++;
		OPTION();
	}
}


/* SETDIGITS()
 * set PRECISION to the number following with the following syntax:
 *
 * SETDIGITS <n>
 * SETDIGITS(<n>)
 *
 * where n is any integer value in the range 0PRECLIM (by default=16)
 * number is restrained to the range 0-PRECLIM */
void SETDIGITS(void) {
	int tt=(int)S();
	if (tt<0) tt=-1;
	else if (tt>PRECLIM) tt=PRECLIM;
	PRECISION=tt;
}


/******************************************************************************
 * The following routines define the behavior of all the statements involving 
 * the USING feature
 *****************************************************************************/


/* FWRITEUSINGSHARP()
 * perform the 'PRINT/WRITE USING', the 'PRINT/WRITE USING #N' and the 
 * 'PRINT/WRITE #N, USING' statements.
 * The algorithm is: get the image and call printFormatted() upon it.
 * flag: if it's TRUE, perform the PRINT statement, otherwise WRITE#, printing
 * a line number in front of new lines 
 * isChan: if TRUE the form is PRINT USING #1, and comes from exec(): in this
 * case the number after the statement is not the image line but the channel
 * and tbas must be informed...
 * ch: a channel id in case command comes from FWRITESHARP()
 * p either points to the character after USING# in the first form or to the 
 * USING string in the second for*/
void FWRITEUSINGSHARP(int flag, int isChan, int ch) {
	char image[COMPSTR];
	int jj;

	/* clean buffer */
	for (jj=0;jj<COMPSTR;jj++) image[jj]='\0';	
	
	/* statement in the form PRINT/WRITE USING #1,<spec>,list
	 * this form comes exec(): recover the # character */
	if (isChan) p--;

	/* statement in the form PRINT/WRITE #1[,]USING<spec>,list
	 * this form comes from FWRITESHARP() */
	if (!strncasecmp_(p,"USING",5)) p+=5;
	/* statement in the form PRINT/WRITE USING #1,<spec>,list
	 * this form comes from parse() */
	else if (*p=='#') {
		p++;
		ch=(int)S();
		if (!channel[ch].isopen) perror_(__LINE__,80,l);
		if (channel[ch].type!=SEQ) perror_(__LINE__,101,l);
		if (ch<1 || ch>9) perror_(__LINE__,133,l);
		/* checks comma/colon, which must be present */
		if (','!=*p && ':'!=*p) 
			perror_(__LINE__,65,l,makechar(p),": OR ,");
		else p++;
	}
	/* else ch is assigned */
	
	/* string image */
	if (isanystring(p)) strncpy_(image,SS(),COMPSTR);
	/* line image */
	else if (isdigit(*p)) {
		int line=strtol(p,&p,10);
		int v=getAddr(line);
		char *mp, *im=image;
		if (v<0) perror_(__LINE__,61,l);
		if (m[v]==NIL) perror_(__LINE__,55,l);
		mp=m[v];
		while (*mp && isblank(*mp)) mp++;
		if (*mp!=':' && strncasecmp_(mp,"IMAGE",5)) 
			perror_(__LINE__,105,l);
		if (*mp==':') mp++; // skip ':'
		else if (!strncasecmp_(mp,"IMAGE",5)) {
			mp+=5;
			/* skip the space after IMAGE */
			if (isblank(*mp)) mp++;
		}
		while (*mp && *mp!='\n') *im++=*mp++;
		*im='\0';
		/* if there are no variables, but there's an image line,
		 * print it bare */
		if (!*p && image[0]) {
			fprintf(channel[ch].stream,"%s\n",image);
			return;
		}
	}
	else perror_(__LINE__,48,l);

	/* checks comma */
	if (*p!=',' && *p!=';') perror_(__LINE__,65,l,makechar(p),",");
	else p++;
	if (image[0]) printFormatted(image, p, !flag, 0.0);
	else perror_(__LINE__,49,l);
	/* checks if WRITE# and PRINT# are mixed for the same channel
	 * in sequential files, which is an error */
	if (channel[ch].writer==NOUSE && ch>0) {
		if (flag) channel[ch].writer=VPRINT;
		else channel[ch].writer=VWRITE;
	}
	else if (channel[ch].writer==VWRITE && flag) perror_(__LINE__,103,l);
	else if (channel[ch].writer==VPRINT && !flag) perror_(__LINE__,103,l);	
	fprintf(channel[ch].stream,"%s",_FORMAT);
}


/******************************************************************************
 * The following routines define the behavior of all the FILES statements
 ******************************************************************************/
/* For the files types, consult:
 * Ref. LBMAA-A-D 10.1, 10-1, "Types of Data Files"
 * Ref. LBMAA-A-D 10.1.1, 10-1, "Sequential Access Files" 
 * Ref. LBMAA-A-D 10.1.2, 10-2, "Random Access Files" */


/* FUPDATE()
 * perform the UPDATE statement, which ensures data flushing on current file 
 * in the argument channel. Syntax:
 *
 * UPDATE [#]<n>
 * UPDATE [:]<n>
 *
 * where <n> is any channel number id. the # or : symbols are optional. This 
 * function does not check what kind of file is in the argument (sequential 
 * or random); if this file is not open, no errors are reported. */
void FUPDATE(void) {
	int ch=0;

	/* argument required */
	if (!*p) perror_(__LINE__,125,l);

	/* UPDATE is valid for streams, not for the terminal */
	while (*p) {
		/* discard file type id */
		if (*p==':' || *p=='#') p++;
		ch=(int)S(); 
		if (ch < 1 || ch > STREAMS) perror_(__LINE__,133,l);
		/* assures PREPEND is closed */
		if (channel[ch].isPrepend) closePrepend(ch);
		/* flush channel data */
		if (channel[ch].isopen) fflush(channel[ch].stream);
		if (*p==',' || *p==';') p++;
	}
}


/* FOPEN()
 * implementation of the OPEN statement in a more modern fashion than 
 * FILE/FILES.
 * Syntax:
 *
 * OPEN "xxx" FOR [NEW] <id>[,] AS [RANDOM] [FILE] [:|#]<ch>[,] TYPE NUM|STR<n>
 * OPEN "xxx" AS LIBRARY
 * 
 * Notes:
 * - if READONLY is specified after OPEN, the read-only INPUT is assumed
 * - "xxx" may be any file name; 
 * - <id> may be INPUT or OUTPUT
 * - if the word NEW is found and <id> is OUTPUT, then the file is erased if 
 *   found. NEW is ignored in case it is used with INPUT (there's no meaning in 
 *   opening a file for input and erase it before starting reading!)
 * - if READONLY is found and <id> is not specified or it is INPUT, the 
 *   read-only INPUT is assumed
 * - <ch> must be any valid channel 1-9 not yet open; the channel is SEQ by
 *   default and may be preceded by # (dummy); if <ch> is preceded by ":",
 *   the RANDOM type is set
 * - if RANDOM is found after AS, the type is set as RANDOM
 * - if READONLY is found after AS, the read-only (in INPUT) is assumed
 * - if FILE is found before the channel, it is ignored
 * - if <ch> is preceded by # (optional) RANDOM must not be used
 * - if <ch> is preceded by : (optional) RANDOM is optional
 * - if TYPE is used, the type is set as RANDOM (NUM or STR); 
 * - if type is not used and RANDOM is set elsewhere, NUM is assumed (numeric 
 *   random type); 
 * - if TYPE STR <n> is used, STR is assumed (string random type); <n> is the 
 *   record length for random string type; if <n> is not set, STRRECLEN is used
 * - if no FOR <id> is found, INPUT is assumed. Thus, file must exist.
 * - if no AS <ch> is found (e.g. OPEN "name" FOR OUTPUT), channel is assigned
 *   according to the first available channel (but the user in unaware of this)
 * - if no FOR and no AS are used (e.g. OPEN "name"), INPUT and the first
 *   available channel are assumed. Thus, file must exist.
 * - if READONLY is not used, any INPUT will be treated as "r+" (read-mode with
 *   possibility to turn the mode to write-mode in append-mode)
 * - unlike FILE, OPEN treats one file at each invocation.
 * - if LIBRARY is found after AS, all other segments are ignored, and OPEN
 *   behaves like OPEN AS LIBRARY */
void FOPEN(void) {
	int currChan=0;		// channel #
	int id=FORINPUT;	// FORINPUT | FOROUTPUT
	int type=SEQ;		// SEQ by default
	int reclen=NUMRECLEN;	// NUM rec len by default
	int mode=NUMBER;	// NUMBER by default
	int isNew=FALSE;	// NEW holder TRUE | FALSE
	int wayout=VREAD;	// SEQ wayout type
	int fchar,flen;		// name-type detection aids
	int isReadonly=FALSE;	// read-only flag
	char op[COMPSCR], *q=op;	// holds file name

	/* there must be at least one argument */
	if (!*p) perror_(__LINE__,65,l,FFLFVT,"A FILE NAME");

	/* look for READONLY flag (OPEN READONLY ... case) */
	if (!strncasecmp_(p,"READONLY",8)) isReadonly=TRUE, p+=8;
	
	/* find first available free channel */
	for (currChan=1; currChan<=STREAMS; currChan++)
		if (!channel[currChan].isopen) break;

	/* get file name (copy input text into op);
	 * Note: unquoted constants are not permitted with OPEN */
	if (isanystring(p)) {
		if (*p=='\"') {
			p++;
			while (*p && *p!='\"') *q++=*p++;
			if (*p=='\"') p++;
			*q='\0';
		}
		else strncpy_(op,SS(),COMPSCR);
	}
	else perror_(__LINE__,88,l);
	
	/* determine file characteristics: first attribution by name */
	setupName(op,&fchar,&flen);
	if (flen<0) perror_(__LINE__,65,l,makechar(op),"; OR ,");
	if (fchar=='%') type=RANDOM, mode=NUMBER, reclen=NUMRECLEN;
	else if (fchar=='$') {
		type = RANDOM, mode=STRING;
		if (flen>0) reclen=flen;
		else reclen=NOUSE;
		if ((reclen<1 || reclen>STRRECMAX) && reclen!=NOUSE) 
			perror_(__LINE__,56,l);
	}

	if (*p==',' || *p==';') p++;

	/* FOR <id> specifier INPUT OR OUTPUT */
	if (!strncasecmp_(p,"FOR",3)) {
		p+=3;
		/* intercept NEW (ignored with INPUT) */
		if (!strncasecmp_(p,"NEW",3)) {
			p+=3;
			isNew=TRUE;
		}
		/* look for READONLY flag (OPEN  ... FOR READONLY INPUT case) */
		else if (!strncasecmp_(p,"READONLY",8)) {
			p+=8;
			isReadonly=TRUE;
			id=FORINPUT;
		}
		/* search for INPUT */
		if (!strncasecmp_(p,"INPUT",5)) {
			p+=5;
			if (doesExist(op)==DOWRITE) 
				perror_(__LINE__,158,l,basename(op));
			id=FORINPUT;
		}
		else if (!strncasecmp_(p,"OUTPUT",6)) {
			p+=6;
			if (doesExist(op)==DOREAD) 
				perror_(__LINE__,157,l,basename(op));
			if (isReadonly) perror_(__LINE__,156,l,basename(op));
			id=FOROUTPUT;
		}
		else perror_(__LINE__,150,l);
	}
	
	if (*p==',' || *p==';') p++;

	/* AS <ch> specifier: second attribution by RANDOM specifier 
	 * and by channel id % or #*/
	if (!strncasecmp_(p,"AS",2)) {
		int this=0;
		p+=2;
		/* load library if required */
		if (!strncasecmp_(p,"LIBRARY",7)) {
			p+=7;
			includeLibs(op);
			return;
		}
		/* set random if required */
		else if (!strncasecmp_(p,"RANDOM",6)) 
			p+=6, type=RANDOM, mode=NUMBER, reclen=NUMRECLEN;
		/* look for READONLY flag (OPEN  ...  AS READONLY ... case) */
		else if (!strncasecmp_(p,"READONLY",8)) {
			p+=8;
			isReadonly=TRUE;
			/* cannot open as read-only a write file */
			if (id==FOROUTPUT) perror_(__LINE__,156,l,op); 
		}
		/* discard FILE if found */
		if (!strncasecmp_(p,"FILE",4)) p+=4;
		/* discard the # if found */
		if (*p=='#' || *p==':') this=*p++;
		/* make checks */
		if (this==':') type=RANDOM, mode=NUMBER, reclen=NUMRECLEN;
		else if (this=='#') type=SEQ, mode=STRING, reclen=STRRECLEN;
		currChan=(int)S();
		if (currChan>STREAMS || currChan<1) perror_(__LINE__,133,l);
	}
	
	if (*p==',' || *p==';') p++;

	/* TYPE specifier: third attribution by direct specifier */
	if (!strncasecmp_(p,"TYPE",4)) {
		p+=4;
		if (!strncasecmp_(p,"NUM",3)) 
			p+=3, type=RANDOM, mode=NUMBER, reclen=NUMRECLEN;
		else if (!strncasecmp_(p,"STR",3)) {
			p+=3;
			if (*p) flen=(int)S();
			else flen=STRRECLEN;
			type=RANDOM, mode=STRING, reclen=flen;
			if (reclen<1 || reclen>MAXSTRLEN) 
				perror_(__LINE__,56,l);
		}
	}

	/* check if channel is already opened by another command, and
	 * check if file has neither read nor write permissions*/
	if (channel[currChan].isopen) perror_(__LINE__,83,l,currChan);
	if (!doesExist(op)) perror_(__LINE__,152,l,basename(op)); 

	/* erase name if already set */
	if (channel[currChan].ioname[0]) channel[currChan].ioname[0]='\0'; 

	/* name cannot be opened more than once */
	if (checkNames(currChan,op)) perror_(__LINE__,82,l,op);

	/* set INPUT channel values */ 
	if (id==FORINPUT) {
		if (isReadonly) channel[currChan].stream=fopen(op,"r");
		else channel[currChan].stream=fopen(op,"r+");
		if (!channel[currChan].stream) perror_(__LINE__,151,l,op);
		else wayout=VREAD;
	}
	/* set OUTPUT channel values */
	else if (id==FOROUTPUT) {
		/* test safe open */
		channel[currChan].stream=fopen(op,"r+");
		/* if file does not exist, or if isNew is required 
		 * create it empty */
		if (!channel[currChan].stream || isNew) {
			if (channel[currChan].stream)
				fclose(channel[currChan].stream);
			channel[currChan].stream=fopen(op,"w+");
		}
		/* if file exist and isNew is not required,
		 * create it for append */
		else channel[currChan].stream=fopen(op,"a+");
		wayout=VWRITE;
	}
	/* check result: if file couldn't anyway be open, raise an error */
	if (!channel[currChan].stream) perror_(__LINE__,151,l,op);
	/* otherwise set it as open */
	else channel[currChan].isopen=TRUE;

	/* setup file data */
	strncpy_(channel[currChan].ioname,op,COMPSTR);
	channel[currChan].type=type;
	channel[currChan].mode=mode;
	channel[currChan].fcounter=0;
	channel[currChan].wayout=wayout; // only for SEQ
	channel[currChan].currseek=1;
	channel[currChan].reclen=reclen;
	channel[currChan].randomaccessed=FALSE;
	channel[currChan].quote=NOQUOTE;
	channel[currChan].margin=MARGINDEF;
	channel[currChan].page=0;
	channel[currChan].pagerow=1;
	channel[currChan].height=PAGEHEIGHT;
	channel[currChan].pagenumber=FALSE;
	channel[currChan].writer=NOUSE;
	channel[currChan].reader=NOUSE;
	channel[currChan].fprintn=1;
	channel[currChan].isReadonly=isReadonly;
	channel[currChan].isPrepend=FALSE;
}


/* FFILE()
 * implementation of the FILE/FILES statements. 
 * isFiles: flag: if TRUE, execute FILES (which does not accept variables and 
 * is executed before program start by preParse); if FALSE, execute FILE (which
 * accepts variables and is executed during runtime).
 * l: contains current line number for perror_() (needed for FILES, because 
 * in preParse() execution has not begun yet);
 * p holds address of character after FILE/FILES */
void FFILE(int isFiles, int l) {
	int currChan=0;		// channel #
	int type, mode, filelen, reclen;
	/* the default mode for FILES/FILE opened file is read mode */
	int wayout=VREAD;
	int fchar,flen;
	char op[COMPSCR];
	char fop[COMPFIL];

	/* there must be at least one argument */
	if (!*p) {
		if (isFiles) perror_(__LINE__,88,l) ;
		else perror_(__LINE__,65,l,FFLFVT,"# OR :");
	}

	/* perform the FILES statement */
	if (isFiles) do {
		char *o;
		op[0]='\0';o=op;
		reclen=STRRECLEN; mode=STRING; type=SEQ;
		while (channel[currChan].isopen) currChan++; 
		if (currChan>STREAMS) perror_(__LINE__,59,l);
		if (*p==',') {
			p++;
			continue;
		}

		/* variables are not admitted with FILES because FILES
		 * is executed into preParse(), when any variable is still
		 * uninstantiated, and thus not known */
		if (isstrvar(p)) perror_(__LINE__,88,l);
		/* get file name, enclosed into double quotes */
		if (*p=='\"') {
			p++;
			while (*p && *p!='\"') *o++=*p++;
			if (*p=='\"') p++; // skip closing quotes
		}
		/* name as constant */
		else while (*p && *p != ',' && *p !=';') *o++=*p++;
		*o='\0';


		/* discard separators , and ; on the input line
		 * (for next iteration) */
		if (*p==',' || *p==';') p++;

		/* determine file characteristics */
		setupName(op,&fchar,&flen);
		if (flen<0) perror_(__LINE__,65,l,makechar(op),"; OR ,");
		if (fchar=='%') type=RANDOM, mode=NUMBER, reclen=NUMRECLEN;
		else if (fchar=='$') {
			type = RANDOM, mode=STRING;
			if (flen>0) reclen=flen;
			else reclen=NOUSE;
			if ((reclen<1 || reclen>STRRECMAX) && reclen!=NOUSE)
				perror_(__LINE__,56,l);
		}
		
		/* set channel values */
		/* open file, checks before if file has already been
		 * opened into another channel stream */
		if (checkNames(currChan,op)) perror_(__LINE__,82,l,op);

		/* set channel values */
		channel[currChan].stream=fopen(op,"r+");
		channel[currChan].isopen=TRUE;
		/* in case file does not exist, create it empty 
		 * Note: I couldn't find references about this, but I take it 
		 * for granted */
		if (!channel[currChan].stream) {
			channel[currChan].stream=fopen(op,"w");
			fclose(channel[currChan].stream);
			channel[currChan].stream=fopen(op,"r+");
			if (reclen==NOUSE) reclen=STRRECLEN;
			wayout=VREAD;
		}
		/* if file exists and type is RANDOM and mode is STRING
		 * get first item to set filelen to file length */
		else {
			if (mode==STRING && type==RANDOM) {
				char op1[COMPSTR];
				fread(op1,sizeof(char),MAXSTRLEN,
						channel[currChan].stream);
				filelen=(int)strtod(op1+1,NULL);
				if (filelen>STRING) filelen-=STRING;
				else perror_(__LINE__,101,l);
				if (filelen>MAXSTRLEN || filelen<1) 
					perror_(__LINE__,56,l);
				if (reclen==NOUSE) reclen=filelen;
				else if (reclen!=filelen) perror_(__LINE__,84,l);
				/* file existed not as random file: reset it */
				if (reclen<0) {
					fclose(channel[currChan].stream);
					channel[currChan].stream=fopen(op,"w");
					fclose(channel[currChan].stream);
					reclen=STRRECLEN;
				}
				else if (reclen!=filelen) 
					perror_(__LINE__,84,l);
				/* reset file position */
				fseek(channel[currChan].stream,0,SEEK_SET);
			}
		}

		/* print identifier in case of random files */
		if (type==RANDOM) {
			int i;
			for (i=0; i<COMPFIL; i++) fop[i]='\0';
			if (mode==STRING)
				sprintw(fop,COMPFIL,"%d",reclen+STRING);
			else if (mode==NUMBER)
				sprintw(fop,COMPFIL,"%d",reclen+NUMBER);
			fseek(channel[currChan].stream,0,SEEK_SET);
			fwrite(fop,sizeof(char),
					FIRSTLEN,channel[currChan].stream);
		}

		/* setup file */
		strncpy_(channel[currChan].ioname,op,COMPSTR);
		channel[currChan].type=type;
		channel[currChan].mode=mode;
		channel[currChan].fcounter=0;
		channel[currChan].wayout=wayout; // (only for SEQ)
		channel[currChan].currseek=1;
		channel[currChan].reclen=reclen;
		channel[currChan].randomaccessed=FALSE;
		channel[currChan].quote=NOQUOTE;
		channel[currChan].margin=MARGINDEF;
		channel[currChan].page=0;
		channel[currChan].pagerow=1;
		channel[currChan].height=PAGEHEIGHT;
		channel[currChan].pagenumber=FALSE;
		channel[currChan].writer=NOUSE;
		channel[currChan].reader=NOUSE;
		channel[currChan].fprintn=1;
		channel[currChan].isReadonly=FALSE;
		channel[currChan].isPrepend=FALSE;
	} while (*p); 

	/* perform the FILE command */
	else do {
		reclen=STRRECLEN;
		mode=STRING;
		type=SEQ;

		/* get type and channel */
		if (*p==':') type=RANDOM, p++;
		else if (*p=='#') p++;
		else if (!isdigit(*p)) perror_(__LINE__,133,l); // tested
		currChan=strtol(p,&p,10);
		if (currChan>STREAMS || currChan<1) perror_(__LINE__,133,l);
		if (*p==',' || *p==':') p++;
		else perror_(__LINE__,65,l,makechar(p),": OR ,");
		
		/* check if channel is already opened by another command */
		if (channel[currChan].isopen) perror_(__LINE__,83,l,currChan);
		
		/* get file name (copy input text into op);
		 * Note: unquoted constants are not permitted with FILE */
		if (isanystring(p)) strncpy_(op,SS(),COMPSCR);
		else perror_(__LINE__,37,l); // tested

		/* discard separators , and ; on the input line */
		if (*p==',' || *p==';') p++;

		/* determine file characteristics */
		setupName(op,&fchar,&flen);
		if (flen<0) perror_(__LINE__,65,l,makechar(op),"; OR ,");
		if (fchar=='%') type=RANDOM, mode=NUMBER, reclen=NUMRECLEN;
		else if (fchar=='$') {
			type = RANDOM, mode=STRING;
			if (flen) reclen=flen;
			else reclen=NOUSE;
			if ((reclen<1 || reclen>STRRECMAX) && reclen!=NOUSE)
				perror_(__LINE__,56,l);
		}
		/* the FILE: statement must be followed by a specifier for 
		 * the random file type (this behavior has been tested on the
		 * LCM), and this specifier must reside into the double quotes 
		 * or in the string variable */
		else if (type==RANDOM) perror_(__LINE__,101,l);

		/* if FILE finds open the required channel, it disconnects it 
		 * and the relative file, and reopens the channel connecting it
		 * to the new file; old file type is immaterial */
		if (channel[currChan].stream!=NIL ||\
			       	*channel[currChan].ioname) {
			fflush(channel[currChan].stream);
			fclose(channel[currChan].stream);
		}

		/* name cannot be opened more than once;
		 * Note: I couldn't find news about this but I take it for 
		 * granted */
		if (checkNames(currChan,op)) perror_(__LINE__,82,l,op);

		/* set channel values */ 
		channel[currChan].stream=fopen(op,"r+");
		channel[currChan].isopen=TRUE;
		/* in case file does not exist, create it empty 
		 * Note: I couldn't find references about this, but I take it 
		 * for granted */
		if (!channel[currChan].stream) {
			channel[currChan].stream=fopen(op,"w");
			fclose(channel[currChan].stream);
			channel[currChan].stream=fopen(op,"r+");
			if (reclen==NOUSE) reclen=STRRECLEN;
			wayout=VREAD;
		}
		/* if file exists and type is RANDOM and mode is STRING
		 * get first item to set file length */
		else {
			if (mode==STRING && type==RANDOM) {
				char op1[COMPSTR];
				fread(op1,sizeof(char),MAXSTRLEN,
						channel[currChan].stream);
				filelen=(int)strtod(op1+1,NULL);
				if (filelen>STRING) filelen-=STRING;
				else perror_(__LINE__,101,l);
				if (filelen>MAXSTRLEN || filelen<1) 
					perror_(__LINE__,56,l);
				if (reclen==NOUSE) reclen=filelen;
				/* file existed not as random file: reset it */
				if (reclen<0) {
					fclose(channel[currChan].stream);
					channel[currChan].stream=fopen(op,"w");
					fclose(channel[currChan].stream);
					reclen=STRRECLEN;
				}
				else if (reclen!=filelen) 
					perror_(__LINE__,84,l);
				/* reset file position */
				fseek(channel[currChan].stream,0,SEEK_SET);
			}
		}

		/* print identifier in case of random files */
		if (type==RANDOM) {
			int i;
			for (i=0; i<COMPFIL; i++) fop[i]='\0';
			if (mode==STRING)
				sprintw(fop,COMPFIL,"%d",reclen+STRING);
			else if (mode==NUMBER)
				sprintw(fop,COMPFIL,"%d",reclen+NUMBER);
			fseek(channel[currChan].stream,0,SEEK_SET);
			fwrite(fop,sizeof(char),
					FIRSTLEN,channel[currChan].stream);
		}
		/* setup file */
		strncpy_(channel[currChan].ioname,op,COMPSTR);
		channel[currChan].type=type;
		channel[currChan].mode=mode;
		channel[currChan].fcounter=0;
		channel[currChan].wayout=wayout; // only for SEQ
		channel[currChan].currseek=1;
		channel[currChan].reclen=reclen;
		channel[currChan].randomaccessed=FALSE;
		channel[currChan].quote=NOQUOTE;
		channel[currChan].margin=MARGINDEF;
		channel[currChan].page=0;
		channel[currChan].pagerow=1;
		channel[currChan].height=PAGEHEIGHT;
		channel[currChan].pagenumber=FALSE;
		channel[currChan].writer=NOUSE;
		channel[currChan].reader=NOUSE;
		channel[currChan].fprintn=1;
		channel[currChan].isReadonly=FALSE;
		channel[currChan].isPrepend=FALSE;
	} while (*p); 
}


/* FMARGIN()
 * implementation of the MARGIN statement */
void FMARGIN(void) {
	int ch, msize;

	/* argument required */
	if (!*p) perror_(__LINE__,125,l);

	/* MARGIN ALL statement */
	if (!strcasecmp(p,"ALL")) {
		/* get margin size */
		p+=3; // 'ALL' specifier must be followed by direct formula
		if (*p==',' || *p==':') p++;
		msize=(int)S();
		if (msize<1 || msize>MARGINMAX) perror_(__LINE__,98,l);
		for (ch=1;ch<=STREAMS;ch++)
			if (channel[ch].isopen && channel[ch].type==SEQ)
				channel[ch].margin=msize;
	}
	/* files or terminal MARGIN statement */
	else do {
		if (*p==':') perror_(__LINE__,101,l);
		if (*p=='#') { 
			p++; 
			ch=(int)S();
			if (ch<1 || ch>STREAMS) perror_(__LINE__,133,l);
		}
		else ch=CONSOLE; // terminal
		if (!channel[ch].isopen) perror_(__LINE__,80,l);
		if (channel[ch].type!=SEQ) perror_(__LINE__,100,l);
		if (*p==':' || *p==',') p++;
		msize=(int)S();
		if (msize<1 || msize>MARGINMAX) perror_(__LINE__,98,l);
		channel[ch].margin=msize;
		if (*p==',' || *p==';') p++;
	} while (*p);
}


/* FNOMARGIN()
 * implementation of the NOMARGIN [ALL] statement */
void FNOMARGIN(void) {
	int ch;

	/* NOMARGIN ALL statement for files only */
	if (!strcasecmp(p,"ALL")) {
		p+=3;
		if (*p) perror_(__LINE__,65,l,makechar(p),TERMINAT);
		for (ch=1;ch<=STREAMS;ch++)
			if (channel[ch].isopen && channel[ch].type==SEQ)
				channel[ch].margin=MARGINDEF;
	}
	/* NOMARGIN alone for terminal */
	else if (!*p) channel[CONSOLE].margin=MARGINDEF;
	/* NOMARGIN for single channels */
	else while (*p) {
		if (*p==':') perror_(__LINE__,101,l);
		/* get channel */
		if (*p==',' || *p==';') ch=CONSOLE; // implicit
		else if (*p=='#') { 
			p++; 
			ch=(int)S();
			if (ch<1 || ch>STREAMS) perror_(__LINE__,133,l); 
		}
		else ch=(int)S();
		if (!channel[ch].isopen) perror_(__LINE__,80,l);
		if (channel[ch].type!=SEQ) perror_(__LINE__,100,l);
		channel[ch].margin=MARGINDEF;
		if (*p==',' || *p==';') p++;
	}
}


/* FPAGE()
 * implementation of the PAGE [ALL] statement */
void FPAGE(void) {
	int ch, psize;

	/* argument required */
	if (!*p) perror_(__LINE__,125,l);

	/* PAGE ALL statement */
	if (!strncasecmp_(p,"ALL",3)) {
		/* get page size */
		p+=3; // 'ALL' specifier must be followed by direct formula
		if (*p==',' || *p==':') p++;
		psize=(int)S();
		if (psize<1 || psize> MAXPAGEHEIGHT) perror_(__LINE__,116,l);
		for (ch=1;ch<=STREAMS;ch++)
			if (channel[ch].isopen && channel[ch].type==SEQ)
				channel[ch].height=psize;
	}
	/* files or terminal statement */
	else do {
		if (*p==':') perror_(__LINE__,101,l);
		if (*p=='#') { 
			p++; 
			ch=(int)S(); 
			if (ch<1 || ch>STREAMS) perror_(__LINE__,133,l); 
		}
		else ch=CONSOLE; // terminal
		if (!channel[ch].isopen) perror_(__LINE__,80,l);
		if (channel[ch].type!=SEQ) perror_(__LINE__,100,l);
		if (*p==':' || *p==',') p++;
		psize=(int)S();
		if (psize<1 || psize> MAXPAGEHEIGHT) perror_(__LINE__,116,l);
		channel[ch].height=psize;
		if (*p==',' || *p==';') p++;
	} while (*p);
}


/* FNOPAGE()
 * implementation of the NOPAGE [ALL] statement */
void FNOPAGE(void) {
	int ch;

	/* NOPAGE ALL statement (for files) */
	if (!strcasecmp(p,"ALL")) {
		p+=3;
		if (*p) perror_(__LINE__,65,l,makechar(p),TERMINAT);
		for (ch=1;ch<=STREAMS;ch++)
			if (channel[ch].isopen && channel[ch].type==SEQ)
				channel[ch].height=0;
	}
	/* NOPAGE alone for terminal */
	else if (!*p) channel[CONSOLE].height=0;
	/* NOPAGE for single channels */
	else while (*p) {
		if (*p==':') perror_(__LINE__,101,l);
		/* get channel */
		if (*p==',' || *p==';') ch=CONSOLE; // implicit
		else if (*p=='#') { 
			p++; 
			ch=(int)S(); 
			if (ch<1 || ch>STREAMS) perror_(__LINE__,133,l); 
		}
		else ch=(int)S();
		if (!channel[ch].isopen) perror_(__LINE__,80,l);
		if (channel[ch].type!=SEQ) perror_(__LINE__,100,l);
		channel[ch].height=0;
		if (*p==',' || *p==';') p++;
	}
}


/* FPAGENUM()
 * implementation of the PAGE NUM [ALL]/NOPAGE NUM [ALL] statements 
 * state: a flag to set/reset pagenum mode; if TRUE, set PAGENUM mode, if 
 * FALSE, set NOPAGENUM mode*/
void FPAGENUM(int state) {
	int ch;

	/* files: PAGENUM ALL/NOPAGENUM ALL statements */
	if (!strncasecmp_(p,"ALL",3)) {
		p+=3; 
		if (*p) perror_(__LINE__,65,l,makechar(p),TERMINAT);
		for (ch=1;ch<=STREAMS;ch++) {
			if (channel[ch].isopen && channel[ch].type==SEQ) {
				channel[ch].page=1;
				channel[ch].pagerow=1;			
				channel[ch].pagenumber=state;
			}
		}
	}
	/* PAGENUM/NOPAGENUM alone for terminal */
	else if (!*p) {
		channel[CONSOLE].page=1;
		channel[CONSOLE].pagerow=1;			
		channel[CONSOLE].pagenumber=state;
	}
	/* PAGENUM/NOPAGENUM for single channels */
	else while (*p) {
		if (*p==':') perror_(__LINE__,101,l);
		/* get channel */
		if (*p==',' || *p==';') ch=CONSOLE; // implicit
		else if (*p=='#') { 
			p++; 
			ch=(int)S(); 
			if (ch<1 || ch>STREAMS) perror_(__LINE__,133,l); 
		}
		else ch=(int)S();
		if (!channel[ch].isopen) perror_(__LINE__,80,l);
		if (channel[ch].type!=SEQ) perror_(__LINE__,100,l);
		channel[ch].page=1;
		channel[ch].pagerow=1;
		channel[ch].pagenumber=state;
		if (*p==',' || *p==';') p++;
	} 
}


/* FILLPAGE()
 * statement to execute the page filling on terminal/file; it executes the 
 * <PA> (page advance) as a statement and not as a PRINT specifier */
void FFILLPAGE(void) {
	int ch=0;

	/* Note: the FILLPAGE ALL has little sense, and is avoided. */
	/* FILLPAGE for terminal */
	if (!*p) {
		fillPage(CONSOLE);	
		channel[CONSOLE].fprintn=0;
		channel[CONSOLE].fcounter=0;
	}
	/* FILLPAGE for single channels */
	else while (*p) {
		if (*p==':') perror_(__LINE__,101,l);
		/* get channel */
		if (*p==',' || *p==';') ch=CONSOLE; // implicit
		else if (*p=='#') {
			p++;
			ch=(int)S();
			if (ch<1 || ch>STREAMS) perror_(__LINE__,133,l);
		}
		else if (!*p) ch=CONSOLE; // terminal
		if (!channel[ch].isopen) perror_(__LINE__,80,l);
		if (channel[ch].type!=SEQ) perror_(__LINE__,100,l);
		fillPage(ch);
		channel[ch].fprintn=0;
		channel[ch].fcounter=0;
		if (*p==',' || *p==';') p++;
	}
}


/* FQUOTE()
 * implementation of the QUOTE/NOQUOTE [ALL] statement
 * quotemode: a flag to set/reset quote mode; if TRUE, set QUOTE mode, if 
 * FALSE, set NOQUOTE mode */
void FQUOTE(int quotemode) {
	int ch;

	/* QUOTE ALL/NOQUOTE ALL statement for files only */
	if (!strcasecmp(p,"ALL")) {
		p+=3;
		if (*p) perror_(__LINE__,65,l,makechar(p),TERMINAT);
		for (ch=1;ch<=STREAMS;ch++)
			if (channel[ch].isopen && channel[ch].type==SEQ)
				channel[ch].quote=quotemode;
	}
	/* QUOTE/NOQUOTE alone for terminal */
	else if (!*p) channel[CONSOLE].quote=quotemode;
	/* QUOTE #N/NOQUOTE #N for single channels */
	else while (*p) {
		if (*p==':') perror_(__LINE__,101,l);
		/* get channel */
		if (*p==',' || *p==';') ch=CONSOLE; // implicit
		else if (*p=='#') { 
			p++; 
			ch=(int)S();
			if (ch<1 || ch>STREAMS) perror_(__LINE__,133,l); 
		}
		else ch=(int)S();
		if (!channel[ch].isopen) perror_(__LINE__,80,l);
		if (channel[ch].type!=SEQ) perror_(__LINE__,100,l);
		channel[ch].quote=quotemode;
		if (*p==',' || *p==';') p++;
	}
}


/* PRINT()
 * Wrapper to the PRINT family statements to select the correct path (designed
 * for PRINT/PRINT#/WRITE#) 
 * flag: if TRUE, perform PRINT/PRINT#, otherwise perform WRITE#
 * p points to the character after PRINT or WRITE */
void PRINT(int flag) {
	/* PRINT#/WRITE */
	if (*p=='#') FWRITESHARP(flag);
	/* PRINT:/WRITE: */
	else if (*p==':') FWRITECOLON();
	/* PRINT */
	else FWRITESHARP(TRUE);
}


/* FWRITESHARP()
 * implementation of the PRINT/PRINT#/WRITE# statement (sequential files only) 
 * flag: if TRUE, perform PRINT/PRINT#, otherwise perform WRITE# */
void FWRITESHARP(int flag) {
	int ch, marg;
	int pagenum=0;

	/* choose destination */
	/* print on screen/terminal or on route file */
	if (*p!='#') {
		if (!(isQueued+isLPrint)) ch=CONSOLE;
		else ch=ROUTECHANNEL;
	}
	/* print on file: get file channel */
	else {
		p++;
		ch=(int)S();
		if (ch<1 || ch>STREAMS) perror_(__LINE__,133,l);
		if (*p==','||*p==':') p++; // optional comma/colon
		if (!channel[ch].isopen) perror_(__LINE__,80,l);
	
		/* if p points to USING, it's a PRINT/WRITE #1[,]USING<spec>
		 * form of the command - delegate FWRITEUSINGSHARP() */
		if (!strncasecmp_(p,"USING",5)) {
			FWRITEUSINGSHARP(flag,FALSE,ch);
			return;
		}
	}
	
	/* set default print CR state */
	channel[ch].fprintn=1;

	/* if file is new, set writer mode */
	if (channel[ch].writer==NOUSE && ch>0) {
		if (flag) channel[ch].writer=VPRINT;
		else channel[ch].writer=VWRITE;
	}
	/* checks if WRITE# and PRINT# are mixed for the same channel
	 * in sequential files, which is an error */
	else {
		if (flag && channel[ch].type==SEQ) {
			if (channel[ch].writer==VWRITE) 
				perror_(__LINE__,103,l);
		}
		if (!flag && channel[ch].type==SEQ) {
			if (channel[ch].writer==VPRINT) 
				perror_(__LINE__,103,l);
		}
	}

	/* checks if this seq file is in READ mode, which is an error */
	if (channel[ch].wayout==VREAD) perror_(__LINE__,73,l);

	/* get current margin for subsequent evaluations */
	marg=channel[ch].margin;

	/* check if PAGE row counter is already beyond limit
	 * Note: a channel[ch].page=0 means that PAGE option is off
	 * Note: pagenum reserves two lines for the page footer */ 
	if (channel[ch].pagerow>channel[ch].height-pagenum) {
		fillPage(ch);
		channel[ch].fprintn=0;
		channel[ch].fcounter=0;
		channel[ch].pagerow=1;
	}
	/* WRITE#/PRINT# without argument */
	if (!*p) channel[ch].fprintn=1;
	/* WRITE#/PRINT# with argument */
	else while (*p) {
		/* WRITE# begins a line with a 5-digit number, 
		 * starting from 1000, followed by a tab */
		if (!flag) {
			if (!channel[ch].fcounter) {
				/* for WRITE, the MARGIN must be at least 7 or 
				 * line number and the tab can't be printed */
				if (marg<=MARGINMIN) 
					perror_(__LINE__,99,l);
				/* line numbers must be <= DATLIM */
			       	if (writeSeqStart>DATLIM) 
					perror_(__LINE__,71,l);
				fprintf(channel[ch].stream,"%05d\t",\
						writeSeqStart);
				writeSeqStart+=writeSeqStep;
				channel[ch].fcounter=6;
			}
		}

		/* found an open bracket */
		if (*p=='(') {
			char *j=p;
			while (*j=='(') j++;
			if (isanystring(j)) {
				p++;
				FWRITESHARP(flag);
				if (*p==')') p++;
				else perror_(__LINE__,37,l);
				continue;
			}
		}
		/* print a string (explicit or into a variable or into a string 
		 * expression); in QUOTE mode, string is enclosed into double 
		 * quotes and preceded by a blank */
		/* print a string (explicit or into a variable or into a string 
		 * expression); in QUOTE mode, string is enclosed into double 
		 * quotes and preceded by a blank */
		if (isanystring(p)) {
			int i;
			char res[COMPSTR]; res[0]='\0';
			strncpy_(res,SS(),COMPSTR);
			for (i=0;i<COMPSTR;i++) pad[i]='\0';
			/* quote mode for channel */
			if (channel[ch].quote==QUOTE) {
				/* string in quote mode cannot contain quote
				 * character (ASCII 34) */
				if (strchr(res,34)) perror_(__LINE__,86,l);
				strncpy_(pad," ",COMPSTR);
				/* quotes only if a blank/comma/tab is found */
				if (isblankcommatab(res) || !strlen(res)) {
					strncat_(pad,"\"",COMPSTR);
					strncat_(pad,res,COMPSTR); //sp+dq
					strncat_(pad,"\"",COMPSTR);
				}
				else strncat_(pad,res,MAXSTRLEN-1); // space
				/* check length: if string is too long, it
				 * would be printed broken, and this would
				 * screw the QUOTE mode feature up */
				if ((int)strlen(pad)>marg) 
					perror_(__LINE__,111,l);
				/* if string is too long, it would be printed
				 * broken (part here and part on next line)
				 * so put it entirely to next line */
				if ((int)strlen(pad)+channel[ch].fcounter>marg)
					FCR(ch);
				printContained(ch,pad,marg);
				if (channel[ch].fcounter>marg && !isRawPrint) {
					FCR(ch);
					channel[ch].fprintn=0;
				}
				else channel[ch].fprintn=1;
			}
			/* normal noquote mode for channel */
			else {
				strncpy_(pad,res,COMPSTR);
				if ((int)strlen(pad)+channel[ch].fcounter>marg)
					if (!isRawPrint) FCR(ch);
				printContained(ch,pad,marg);
				channel[ch].fprintn=1;
			}
			/* if after a comma/semicolon there's a TAB, the
			 * comma/semicolon function must not be performed */
			if (!strncasecmp_(p,";TAB(",5)||\
					!strncasecmp_(p,",TAB(",5)) p++;
			/* p is updated into SS() */
			/* only the + operator is admitted */
			if (*p && *p!='+') {
				if ((isro(p))) perror_(__LINE__,35,l);
			}
		}
		/* adjust comma position (go to next ZONE) 
		 * Note: WRITE starts counting ZONEs after the five numbers
		 * plus the TAB character, so we must subtract this value */
		else if (*p==',') {
			int k=flag?0:6; // WRITE line-number space
			if (!channel[ch].fcounter) channel[ch].fcounter+=k;
			if (!channel[ch].fcounter) FAC(ch);
			while ((channel[ch].fcounter)%ZONE) FAC(ch);
			/* in case of multiple commas, force advancing */
			if (*(p+1)==',') FAC(ch);
			p++, channel[ch].fprintn=0;
		}
		/* adjust semicolon position (actually doing nothing) */
		else if (*p==';') {
			/* in case of sequent comma, force advancing */
			if (*(p+1)==',') FAC(ch);
			p++, channel[ch].fprintn=0;
		}
		/* <PA> delimiter for PRINT/PRINT#/WRITE# */
		else if (is_pa(p)) {
			fillPage(ch);
			p+=4;
			channel[ch].fprintn=0;
			channel[ch].fcounter=0;
		}
		/* TAB() function */
		else if (!strncasecmp_(p,"TAB(",4)) {
			int i,tab;
			p+=4;
			channel[ch].fprintn=0;
			tab=(int)S();
			tab=(abs(tab)%marg);
			if (')'==*p) p++; // skip ')'
			else perror_(__LINE__,37,l);
			if (tab > channel[ch].fcounter)
				for (i=0; i<tab-channel[ch].fcounter; i++) 
					fputc(' ',channel[ch].stream);
			channel[ch].fcounter=tab;
		}
		/* case of closed calculation */
		else if (*p==')') {
			return;
		}
		/* print a numerical expression;
		 * regardless of QUOTE mode, a space is printed after the 
		 * number to prevent contiguous characters and a space is
		 * printed before the number if the number is negative
		 * (the positive number takes the empty sign space with it) */ 
		else {
			double res;
			char out[COMPSTR], *o=out, pad[COMPSTR];
			*o='\0';
			res=S();
			/* catch special 'inf' string in case of OPTION ZERO */
			if (res>=INFF && isZeroLim) {
				strncpy_(pad," inf",COMPSTR);
			}
			else if (res<=MINFF && isZeroLim) {
				strncpy_(pad,"-inf",COMPSTR);
			}
			/* normal processing */
			else {
				printNUM(ch,res,1,out);
				if (res<0 && channel[ch].quote==QUOTE) 
					sprintw(pad,COMPSTR," %s",out);
				else strncpy_(pad,out,COMPSTR);	
			}
			
			/* print number */
			if ((int)strlen(pad)>marg && !isRawPrint) 
				perror_(__LINE__,111,l);
			else {
				fprintf(channel[ch].stream,"%s",pad);
				channel[ch].fcounter+=(int)strlen(pad);
			}
			channel[ch].fprintn=1;
			
			/* if after a comma/semicolon there's a TAB, the
			 * comma/semicolon function must not be performed */
			if (!strncasecmp_(p,";TAB(",5)||\
					!strncasecmp_(p,",TAB(",5)) p++;
			/* p is updated into S() */
		}
	}
	
	/* close LPRINT */
	isLPrint=FALSE;

	/* update the cursor */
	if (channel[ch].fprintn) FCR(ch);

}


/* FWRITECOLON()
 * implementation of the WRITE:/PRINT: statement (random files only) 
 * Note: WRITE: and PRINT: statements are completely equivalent, 
 * and may be used on the same channel consecutively and/or alternatively */
void FWRITECOLON(void) {
	int ch, gap;
	char op[COMPFIL];

	/* argument required */
	if (!*p) perror_(__LINE__,125,l);

	/* get channel and make checks */
	p++;
	ch=(int)S();
	if (ch<1 || ch>STREAMS) perror_(__LINE__,133,l);
	if (!channel[ch].isopen) perror_(__LINE__,80,l);
	if (channel[ch].type!=RANDOM) perror_(__LINE__,101,l);
	if (*p==':' || *p==',') p++;
	else perror_(__LINE__,38,l);

	/* calculate gap in case of file already accessed*/
	if (channel[ch].randomaccessed) {
		gap=ftell(channel[ch].stream)/channel[ch].reclen;
		if (channel[ch].currseek>gap) fillGap(ch,gap);
	}
	/* case of file not yet accessed 
	 * print random identifier and record length */
	else if (channel[ch].mode==STRING) {
		channel[ch].randomaccessed=TRUE;
		sprintw(op,COMPFIL,"%d",channel[ch].reclen+STRING);
		fseek(channel[ch].stream,0,SEEK_SET);
		fwrite(op,sizeof(char),
				FIRSTLEN,channel[ch].stream);
		channel[ch].currseek=1;
	}
	/* case of file not yet accessed  
	 * print random identifier and record length */
	else if (channel[ch].mode==NUMBER) {
		channel[ch].randomaccessed=TRUE;
		sprintw(op,COMPFIL,"%d",channel[ch].reclen+NUMBER);
		fseek(channel[ch].stream,0,SEEK_SET);
		fwrite(op,sizeof(char),
				FIRSTLEN,channel[ch].stream);
		channel[ch].currseek=1;
	}
	
	/* get elements to be written */
	do {
		/* string variable or literal */
		if (channel[ch].mode==STRING) {
			/* checks type */
			if (channel[ch].mode!=STRING) perror_(__LINE__,45,l);
			/* get element and print it on file */
			strncpy_(op,SS(),COMPFIL);
			if ((int)strlen(op)>channel[ch].reclen) 
				perror_(__LINE__,113,l);
			fwrite(op,sizeof(char),channel[ch].reclen,
				channel[ch].stream);
			channel[ch].currseek++;
		}
		/* it's a numeric formula */
		else {
			double val;
			/* checks type */
			if (channel[ch].mode!=NUMBER) perror_(__LINE__,45,l);

			/* get element and print it on file */
			val=S();
			if (val>INFF) val=INFF;
			else if (val<MINFF) val=MINFF;
			printNUM(ch,val,TRUE,op);
			fwrite(op,sizeof(char),channel[ch].reclen,
				channel[ch].stream);
			channel[ch].currseek++;
		}
		if (*p==',' || *p==';') p++;
		else if (*p) perror_(__LINE__,38,l);

	} while (*p);
}


/* FREADSHARP()
 * implementation of the READ#/INPUT# statements
 * Note: READ# and INPUT# differ: READ expects a line number for each data file 
 * line, and this number is discarded; if no number is found, an error is 
 * issued; INPUT#, does not expect a line number, and if found, treats it as a 
 * datum. These commands may not be mixed in the same channel.
 * flag: if TRUE, perform INPUT#, otherwise perform READ# 
 * mode: if TRUE, read a whole line (LINPUT#/LINEINPUT and LREAD#/LINEREAD#
 * mode), else read a datum in the standard way (INPUT# and READ# mode) */
void FREADSHARP(int flag, int mode) {
	char line[FILELINE+1], *u;
	int c, ch, var, j,isArr=FALSE,type=NUM;
	static int startofline=TRUE;

	/* argument required */
	if (!*p) perror_(__LINE__,125,l);

	/* get channel and make checks */
	ch=(int)S();
	if (ch<1 || ch>STREAMS) perror_(__LINE__,133,l);
	if (*p==',' || *p==':') p++;
	else perror_(__LINE__,38,l);
	if (!channel[ch].isopen) perror_(__LINE__,69,l);
	if (channel[ch].type!=SEQ) perror_(__LINE__,101,l);

	/* checks if sequential file is in WRITE mode, which is an error */
	if (channel[ch].wayout==VWRITE) perror_(__LINE__,70,l);
	
	/* checks if end of file */
	if (testEOF(ch)) perror_(__LINE__,25,l);
		
	/* checks if READ# and INPUT # are mixed for the same
	 * channel, which results in an error */
	if (channel[ch].reader==NOUSE) {
		if (flag) channel[ch].reader=VINPUT;
		else channel[ch].reader=VREAD;
		fseek(channel[ch].stream,0,SEEK_SET);
	}
	else {
		if (flag && channel[ch].reader==VREAD) 
			perror_(__LINE__,102,l);
		if (!flag && channel[ch].reader==VINPUT) 
			perror_(__LINE__,102,l);
	}
	for (j=0; j<=FILELINE;j++) line[j]='\0';
	
	/* read first character */
	c=fgetc(channel[ch].stream);

	/* repeat for all variables */
	while (*p) {
		int cs;
		double val=0.0;

		/* restore parameters */
		isArr=FALSE,type=NUM;
		
		/* is it an existing variable? */
		if ((cs=findDEC(p))) {
			var=cs, p+=_len;
			type=vn[cs].rettype;
			isArr=vn[cs].isArr;
		}
		else perror_(__LINE__,42,l);
		
		/* get next available character */
		u=line;
		while (isblank(c) || c=='\n') c=fgetc(channel[ch].stream);

		/* READ# reads and discards starting line number */
		if (!flag && startofline) {
			if (isdigit(c)) {
				while (isdigit(c)) {
					c=fgetc(channel[ch].stream);
				}
			}
			else perror_(__LINE__,97,-1,l);
			startofline=FALSE;
			while (isblank(c)) c=fgetc(channel[ch].stream);
		}
		
		/* read a string enclosed into double quotes */
		if (c=='\"' && !mode) {
			c=fgetc(channel[ch].stream);
			while (c!=EOF && c!='\n' && c!='\"' &&\
					(int)(u-line)<FILELINE) {
				*u++=c;
				c=fgetc(channel[ch].stream);
			}
			if (c=='\"') c=fgetc(channel[ch].stream);
		        if (c==',') c=fgetc(channel[ch].stream);	
		}
		/* read a normal string or number-string
		 * (the comma separates strings as well as spaces */
		else if (!mode) {
			while (!isblank(c) && c!=',' && c!='\n' && c!=EOF &&\
					(int)(u-line)<FILELINE) {
				*u++=c; // store even one starting space
				c=fgetc(channel[ch].stream);
			}
		        if (c==',') c=fgetc(channel[ch].stream);	
		}
		/* read a line until EOL */
		else {
			while (c && c!=EOF && c!='\n'&& (int)(u-line)<FILELINE){
				*u++=c; // store even one starting space
				c=fgetc(channel[ch].stream);
			} 
		}
		*u='\0'; // close line!

		/* consume spaces on the file */
		while (isblank(c)) c=fgetc(channel[ch].stream); 
		
		/* Reset counter if end of line is reached */
		if (c=='\n'|| !c) {
			startofline=TRUE;
		}
		/* treat exhausted lines 
		 * (last character read was an end-of-line) */
		else if (c==EOF) break;

		/* treat lines beginning with spaces */
		u=line;
		while (*u && (isblank(*u) || *u==',')) u++, startofline=FALSE;

		/* get numeric value */
		if (type==NUM && isnm(u)) {
			char *pp=p;
			p=u;
			val=S();
			u=p;
			p=pp;
		}

		/* numeric variable */
		if (!isArr && type==NUM){
			if (vn[var].isInt) P[var]=(int)val;
			else P[var]=val;
		}
		/* string variable) */
		else if (!isArr && type==STR) strncpy_(PS[var],u,COMPSTR);
		/* numeric or string array */
		else if (isArr) {
			int M=0, N=0, T=1;
			if (*p=='(') p++;
			else perror_(__LINE__,37,l);
			M=(int)S();
			if (*p==',') {
				p++;
				N=(int)S();
				T=2;
			}
			p++;
			if (vn[var].isInt) val=(int)val;
			if (type==NUM) setArrayVal(var,T,M,N,val);
			else setArrayStr(var,T,M,N,u);
		}
		if (*p==',') p++;
	}

	/* discard all remaining spaces in the line */
	while (c==' ' || c=='\t' || c=='\n') 
		c=fgetc(channel[ch].stream);

	/* reset file position */
	u=line;
	while (*u) u++;
	fseek(channel[ch].stream,-(int)strlen(line)+(int)(u-line)-1,SEEK_CUR);
}


/* FREADCOLON()
 * implementation of the READ:/INPUT: statements
 * Note: The READ: and INPUT: statements are completely equivalent, and may be 
 * used in the same channel consecutively without raising errors  */
void FREADCOLON(void) {
 	char op[FILELINE+1];
	int i,ch;

	/* argument required */
	if (!*p) perror_(__LINE__,125,l);

	/* get channel and make checks */
	ch=(int)S();
	if (ch<1 || ch>STREAMS) perror_(__LINE__,133,l);
	if (*p==',' || *p==':') p++;
	else perror_(__LINE__,38,l);
	if (!channel[ch].isopen) perror_(__LINE__,80,l);
	if (channel[ch].type!=RANDOM) perror_(__LINE__,101,l);

	/* checks if end of file */
	if (testEOF(ch)) perror_(__LINE__,25,l);
		
	/* in case this command is executed on a not accessed file,
	 * read data file line 0 */
	if (!channel[ch].randomaccessed) {
		int rl;
		int ft=ftell(channel[ch].stream);
		channel[ch].randomaccessed=TRUE;
		fseek(channel[ch].stream,0,SEEK_SET);
		fread(op,sizeof(char),FIRSTLEN,channel[ch].stream);
		op[0]='0'; // get rid of  identifier
		rl=(int)strtod(op,NULL);
		if (channel[ch].mode==STRING) {
			rl-=STRING;
			if (rl>MAXSTRLEN||rl<1) perror_(__LINE__,56,l);
		}
		else if (channel[ch].mode==NUMBER) rl-=NUMBER;
		if (rl!=channel[ch].reclen) perror_(__LINE__,84,l);
		fseek(channel[ch].stream,ft,SEEK_SET); // restore position
		channel[ch].currseek=1;
	}

	/* repeat for all variables */
	while (*p) {
		int cs;

		/* clear buffer and read data from file */
		for (i=0;i<COMPFIL;i++) op[i]='\0';
		fread(op,sizeof(char),channel[ch].reclen,channel[ch].stream);

		/* detect vars */
		if ((cs=findDEC(p))) {
			p+=_len;
			/* numeric vars */
			if (vn[cs].rettype==NUM && !vn[cs].isArr)
				P[cs]=strtod(op,NULL);
			/* string variables */
			if (vn[cs].rettype==STR && !vn[cs].isArr) 
				strncpy_(PS[cs],op,COMPSTR);
			/* string and numeric arrays */
			else if (vn[cs].isArr) {
				int M=0, N=0,T=1;
				if (*p=='(') p++;
				else perror_(__LINE__,37,l);
				M=(int)S();
				if (*p==',') {
					p++;
					N=(int)S();
					T=2;
				}
				if (*p==')') p++;
				else perror_(__LINE__,37,l);
				if (vn[cs].rettype==STR) 
					setArrayStr(cs,T,M,N,op);
				else setArrayVal(cs,T,M,N,strtod(op,NULL));
			}
		}
		channel[ch].currseek++;
		if (*p==','||*p==';') p++;
	}
}


/* FRESTORE()
 * implementation of the RESTORE statement for files;
 * omitting the # (sign number) or the : (colon) is and error */
void FRESTORE(void) {
	int ch;
	
	/* argument required */
	if (!*p) perror_(__LINE__,125,l);

	do {
		/* sequential file */
		if (*p=='#') {
			if (*p=='#') p++;
			ch=(int)S();
			/* check if channel is valid */
			if (ch<1 || ch>STREAMS) perror_(__LINE__,133,l);
			/* check if file exists */
			if (!channel[ch].isopen) perror_(__LINE__,81,l);
			/* assure PREPEND is closed */
			if (channel[ch].isPrepend) closePrepend(ch);
			/* check if file is random, which results in an error*/
			if (channel[ch].type!=SEQ) perror_(__LINE__,101,l);
			/* set file in read mode and move seek to start */
			channel[ch].wayout=VREAD;
			fseek(channel[ch].stream,0,SEEK_SET);
		}
		/* random file */
		else if (*p==':') {
			p++;
			ch=(int)S();
			/* check if channel is valid */
			if (ch<1 || ch>STREAMS) perror_(__LINE__,133,l);
			/* check if file exists */
			if (!channel[ch].isopen) perror_(__LINE__,81,l);
			/* check if file is seq, which results in an error*/
			if (channel[ch].type!=RANDOM) perror_(__LINE__,101,l);
			/* set to first item */
			channel[ch].currseek=1;
			channel[ch].randomaccessed=FALSE;
			fseek(channel[ch].stream,channel[ch].reclen,SEEK_SET);
		}
		else perror_(__LINE__,38,l);
		if (*p==',' || *p==';') p++;
	} while (*p);
}


/* FCLOSE()
 * implementation of the CLOSE statement for files; 
 * Note: the # (sign number) or the : (colon) are optional */
void FCLOSE(void) {
	int ch;
	
	/* argument required */
	if (!*p) perror_(__LINE__,125,l);

	do {
		/* get channel */
		if (*p=='#' || *p==':') p++;
		if (isanynumber(p)) ch=(int)S();
		else perror_(__LINE__,38,l);
		/* check if channel is valid */
		if (ch<1 || ch>STREAMS) perror_(__LINE__,133,l);
		/* check if file is not open 
		 * Note: I call perror_ on isopen and not the call
		 * on stream, because this last may not be portable 
		 * (e.g. BSD systems and Linux) */
		if (!channel[ch].isopen) perror_(__LINE__,149,l,ch);
		/* assure PREPEND is closed */
		if (channel[ch].isPrepend) closePrepend(ch);
		/* close file and reset flags 
		 * (the check is left for getting rid of any doubt) */
		if (channel[ch].stream) fclose(channel[ch].stream);
		channel[ch].isopen=FALSE;
		strncpy_(channel[ch].ioname,"",COMPSTR);
		channel[ch].type=NOUSE;
		channel[ch].mode=NOUSE;
		channel[ch].fcounter=0;
		channel[ch].wayout=NOUSE;
		channel[ch].currseek=1;
		channel[ch].reclen=0;
		channel[ch].randomaccessed=FALSE;
		channel[ch].quote=NOQUOTE;
		channel[ch].margin=MARGINDEF;
		channel[ch].page=0;
		channel[ch].pagerow=1;
		channel[ch].height=PAGEHEIGHT;
		channel[ch].pagenumber=FALSE;
		channel[ch].writer=NOUSE;
		channel[ch].reader=NOUSE;
		channel[ch].fprintn=1;
		
		if (*p==',' || *p==';') p++;
	} while (*p);
}


/* FSCRATCH() 
 * implementation of the SCRATCH statement;
 * omitting the # (sign number) or the : (colon) is interpreted as a SCRATCH on 
 * a sequential file */
void FSCRATCH(void) {
	int ch;
	
	/* argument required */
	if (!*p) perror_(__LINE__,125,l);

	do {
		/* sequential file */
		if (*p=='#' || isdigit(*p)) {
			if (*p=='#') p++;
			ch=(int)S();
			/* check if channel is valid */
			if (ch<1 || ch>STREAMS) perror_(__LINE__,133,l);
			/* assures PREPENDED is closed */
			if (channel[ch].isPrepend) closePrepend(ch);
			/* check if file exists */
			if (!channel[ch].isopen) perror_(__LINE__,80,l);
			/* check if file is random, which results in an error*/
			if (channel[ch].type!=SEQ) perror_(__LINE__,101,l);
			/* check is file is read-only */
			if (channel[ch].isReadonly)
				perror_(__LINE__,155,l,channel[ch].ioname);
			/* erase and reset to write mode */
			if (channel[ch].stream) {
				fclose(channel[ch].stream);
				channel[ch].stream=
					fopen(channel[ch].ioname,"w+");
				channel[ch].wayout=VWRITE;
			}
		}
		/* random file */
		else if (*p==':') {
			p++;
			ch=(int)S();
			/* check if channel is valid */
			if (ch<1 || ch>STREAMS) perror_(__LINE__,133,l);
			/* check if file exists */
			if (!channel[ch].isopen) perror_(__LINE__,80,l);
			/* check if file is seq, which results in an error*/
			if (channel[ch].type!=RANDOM) perror_(__LINE__,101,l);
			/* erase and reset to first item */
			if (channel[ch].stream) {
				fclose(channel[ch].stream);
				channel[ch].stream=
					fopen(channel[ch].ioname,"w+");
				channel[ch].currseek=1;
				channel[ch].randomaccessed=FALSE;
			}
		}
		else perror_(__LINE__,38,l);
		if (*p==',' || *p==';') p++;
	} while (*p);
}


/* FSET() 
 * implementation of the SET statement */
void FSET(void) {
	int ch=0, setpos=0, eof=0, reclen=0;

	/* there must be an argument */
	if (!*p) perror_(__LINE__,125,l);
	
	do {
		/* colon can be omitted */
		if (*p==':') p++;
		
		/* get and check channel */
		ch=(int)S();
		if (ch<1 || ch>STREAMS) perror_(__LINE__,133,l);
		if (!channel[ch].isopen) perror_(__LINE__,80,l);
		if (channel[ch].type!=RANDOM) perror_(__LINE__,101,l);

		 /* get and check position */
		if (*p!=':' && *p!=',') perror_(__LINE__,38,l);
		p++;
		setpos=(int)S();
		reclen=channel[ch].reclen;
		if (setpos<=0) perror_(__LINE__,120,l);
		/* fill positions in case SET value is past EOF */
		eof=getEOF(ch);
		if (setpos>eof) {
			fseek(channel[ch].stream,0,SEEK_END);
			fwrite("\0",sizeof(char),(setpos-eof-1)*reclen,
					channel[ch].stream);
		}
		/* set position */
		fseek(channel[ch].stream,FIRSTLEN+(setpos-1)*reclen,SEEK_SET);
		channel[ch].currseek=setpos;
		if (*p==',' || *p==';') p++;
	} while (*p);		
}


/* CHAIN()
 * implementation of the CHAIN command;
 * enable a program to load and parse an external program.
 * Note: the CHAIN statement is the last to be executed, so anything following
 * it is discarded, both in listing and in the program flow. In practice, the 
 * caller program transfers the control to the CHAINed file.
 * The CHAINed file doesn't inherit the caller's variables
 * The starting line of the CHAINed program may be specified after a comma */
void CHAIN(void) {
	char fname[COMPSCR], *tf=fname;
	FILE* fd;
	int z=0, ph=TRUE;
	
	/* there must be an argument */
	if (!*p) perror_(__LINE__,125,l);
	
	/* Note: the CHAIN command may be issued into a FOR, or with a pending
	 * GOSUB subroutine, or within a multiline DEF function; so there's no
	 * need to prevent the call to CHAIN in those cases */

	/* get file name 
	 * name may be given as CHAIN FILE1 or CHAIN "FILE1" or CHAIN A$; 
	 * FILE1 without quotes must of course not contain quotes, spaces,
	 * commas; if it contains any of such peculiar characters, it must be 
	 * written enclosed by double quotes */
	while (isblank(*p)) p++;
	/* get name from string (case a)*/
	if (!isstrvar(p)) 
		{	
			/* string enclosed by "" */
			if (*p=='\"') {
				p++;
				while (*p && *p!='\"') *tf++=*p++;
				*tf='\0';
				p++;
			}
			/* plain name until comma or end of string */
			else {
				while (*p && *p!=',' && *p!='\0') *tf++=*p++;
				*tf='\0';
			}
		}
	/* get name from variable (case b) */
	else strncpy_(fname,SS(),COMPSCR);
	while (isblank(*p)) p++;

	/* get starting line, if given as argument; check if into bounds */
	if (*p){
		if (*p!=',') perror_(__LINE__,38,l);
		p++; // skip comma
		startingLine=(int)S(); // label value
		if (startingLine<0 || startingLine>DATLIM) 
			perror_(__LINE__,94,l);
		ph=FALSE;// still label
	}
	/* if no argument was found, set default starting line */
	else startingLine=1; // physical address

	/* checks file existence */
	setupName(fname,&z,&z);
	if (!(fd=fopen(fname,"r"))) perror_(__LINE__,1,-1,fname);
	else fclose(fd);
	
	/* set CHAIN flag, load file name */
	isChain=TRUE;
	strncpy_(filename,fname,COMPSCR);
	OLD();

	/* convert label address to physical address */
	if (!ph) {
		startingLine=getAddr(startingLine);
		if (startingLine<0) perror_(__LINE__,107,l);
	}

	/* not existing line in the CHAINed file: search for next */
	while (!m[startingLine] && startingLine<=DATLIM) startingLine++;
	/* irregular line (line inside a multiline DEF FN, for instance) */
	if (fm[startingLine]) perror_(__LINE__,89,l);
	/* all's good: execute the CHAINed program 
	 * Note: RUN erases all memory and calls both preParse() and parse() */
	RUN();
}


/******************************************************************************
 * The following routines define the behavior of the MAT statements, which were
 * included as a CardBasic option since 1964 version II of Dartmouth BASIC, and
 * as standard since 1966 version III.
 ******************************************************************************/

/* MATDOT()
 * The DOT statement (this is a math-function rather than a matrix-function)
 * return dot product between two unidimensional vectors, yielding a real
 * argument vectors are taken from stream
 * Vectors may be vertical or horizontal in either side */
double MATDOT(void) {
	int var1=0,var2=0;
	double res=0.0;
	int cs, dim1=0, dim2=0;

	/* there must be an argument */
	if (!*p) perror_(__LINE__,35,l);
	
	/* get first matrix codes */
	forceArray=TRUE;
	if ((cs=findDEC(p)) && _rettype==NUM && _arr) var1=cs,p+=_len;
	else perror_(__LINE__,42,l);
	if (*p++!=',') perror_(__LINE__,35,l);
	/* get second matrix codes */
	if ((cs=findDEC(p)) && _rettype==NUM && _arr) var2=cs,p+=_len;
	else perror_(__LINE__,42,l);
	forceArray=FALSE;

	/* check types (they must exist and not be matrices) */
	if (dim[var1].row==0 && dim[var1].col==0) perror_(__LINE__,51,l);
	if (dim[var2].row==0 && dim[var2].col==0) perror_(__LINE__,51,l);
	if (dim[var1].row>0 && dim[var1].col>0) perror_(__LINE__,51,l); 
	if (dim[var2].row>0 && dim[var2].col>0) perror_(__LINE__,51,l); 
	
	/* get dimensions */
	if (dim[var1].row!=0 && dim[var1].col==0)      dim1=dim[var1].row;
	else if (dim[var1].row==0 && dim[var1].col!=0) dim1=dim[var1].col;
	else perror_(__LINE__,51,l);
	if (dim[var2].row!=0 && dim[var2].col==0)      dim2=dim[var2].row;
	else if (dim[var2].row==0 && dim[var2].col!=0) dim2=dim[var2].col;
	else perror_(__LINE__,51,l);
	
	/* check dimensions validity */
	if (dim1!=dim2) perror_(__LINE__,117,l);

	/* perform dot product */
	else {
		double vec1[dim1+2], vec2[dim1+2]; // dim1=dim2!!
		int i;
		
		/* copy var1 from either type */
		if (dim[var1].type==1) {
			for (i=0;i<=dim1;i++)
				vec1[i]=getArrayVal(var1,1,i,0);
		}	
		else {
			if (dim[var1].row!=0 && dim[var1].col==0)
				for (i=0;i<=dim1;i++) 
					vec1[i]=getArrayVal(var1,2,i,0);
			else if (dim[var1].row==0 && dim[var1].col!=0) 
				for (i=0;i<=dim1;i++) 
					vec1[i]=getArrayVal(var1,2,0,i);
		}
		/* copy var 2 from either type */
		if (dim[var2].type==1) {
			for (i=0;i<=dim1;i++)
				vec2[i]=getArrayVal(var2,1,i,0);
		}
		else {
			if (dim[var2].row!=0 && dim[var2].col==0)
				for (i=0;i<=dim2;i++) 
					vec2[i]=getArrayVal(var2,2,i,0);
			else if (dim[var2].row==0 && dim[var2].col!=0) 
				for (i=0;i<=dim2;i++) 
					vec2[i]=getArrayVal(var2,2,0,i);
		}

		/* calculate dot product from copies */
		for (i=0;i<=dim1;i++) res+=vec1[i]*vec2[i];
	}
	return res;
}


/* MATbyK(result-matrix)
 * multiply a matrix by an expression, both taken from stream; read multiplier
 * and argument matrix from stream, multiply to the result matrix
 * varr: the result matrix code, checked into MATtype() and passed here */
void MATbyK(int varr){
	double exp=S();
	int var1=0, i=0,j=0,m=0,n=0,t=0, cs=0;

	/* there must be an argument */
	if (!*p) perror_(__LINE__,125,l);
	
	/* result matrix existence was checked into MATtype() */

	/* check syntax */
	if (*p++!= ')') perror_(__LINE__,35,l);
	if (*p++!= '*') perror_(__LINE__,35,l);
	
	/* get argument matrix code */
	forceArray=TRUE;
	cs=findDEC(p);
	if (cs && _rettype==NUM && _arr) var1=cs,p+=_len;
	else perror_(__LINE__,42,l);
	forceArray=FALSE;

	/* check dimensions coherence and redimension result matrix */
	if ((dim[varr].brow+1)*(dim[varr].bcol+1)<\
			(dim[var1].row+1)*(dim[var1].col+1)) 
		perror_(__LINE__,76,l);
	else {
		m=dim[varr].row=dim[var1].row;
		n=dim[varr].col=dim[var1].col;
		t=dim[varr].type;
	}

	/* perform multiplication by expression */
	/* Legal procedures for special cases:
	 * MAT A=(K)*A */
	i=1; j=(t==1)?0:1;
	while (i<=m) {
		setArrayVal(varr,t,i,j,exp*getArrayVal(var1,t,i,j));
		j++;
		if (j>n) {i++; j=(t==1)?0:1;}
	}
}


/* MATTRN(result-matrix)
 * The MAT TRN statement for transposing a numerical matrix; read argument 
 * matrix from stream, transpose it to the result matrix 
 * Note: there's not much sense in transposing a string matrix, isn't it?
 * varr: the result matrix, checked into MATtype() and passed here */
void MATTRN(int varr){
	int var1,i,j,t,t1,y,cs;

	/* argument required */
	if (!*p) perror_(__LINE__,125,l);

	/* result matrix existence was checked into MATtype */

	/* get argument matrix code */
	forceArray=TRUE;
	if (*p++=='(') {
		cs=findDEC(p);
		if (cs && _rettype==NUM && _arr) var1=cs, p+=_len;
		else perror_(__LINE__,42,l);
	}
	else perror_(__LINE__,35,l);
	forceArray=FALSE;

	/* check dimensions coherence */
	if ((dim[varr].brow+1)*(dim[varr].bcol+1)<\
			(dim[var1].row+1)*(dim[var1].col+1))
		perror_(__LINE__,76,l);

	/* Legal procedures for special cases:
	 * MAT C=TRN(C) 
	 * calculate values for result matrix and copy argument matrix to
	 * temporary matrix for enabling the legal case MAT C=TRN(C) 
	 * TEMPVAR = var1 */
	t=dim[varr].type;
	t1=dim[var1].type;
	dimArray(TEMPVAR,t1,dim[var1].row,dim[var1].col);
	for(i=0;i<=dim[var1].row;i++) for(j=0;j<=dim[var1].col;j++)
		setArrayVal(TEMPVAR,t1,i,j,getArrayVal(var1,t1,i,j));

	/* transpose argument vector (previously copied over temporary) 
	 * to result horizontal vector */
	if (dim[var1].type==1) {
		dim[varr].type=2;
		dim[varr].col=dim[var1].row;
		dim[varr].row=0; // ??
		i=1;
		while (i<=dim[varr].col) {
			setArrayVal(varr,2,0,i,getArrayVal(TEMPVAR,1,i,0));
			i++;
		}
	}
	/* transpose argument horizontal vector (previously copied over 
	 * temporary) to result vector */
	else if (dim[var1].type==2 && dim[var1].row==0 && dim[var1].col) {
		dim[varr].type=1;
		dim[varr].row=dim[var1].col;
		dim[varr].col=0; // ??
		i=1;
		while (i<=dim[varr].row) {
			setArrayVal(varr,1,i,0,getArrayVal(TEMPVAR,t1,0,i));
			i++;
		}
	}
	/* transpose argument matrix (previously copied over temporary) 
	 * to result matrix */
	else {
		/* redimension result matrix */
		y=dim[var1].row;
		dim[varr].row=dim[var1].col;
		dim[varr].col=y;
		i=1;
		while (i<=dim[varr].col) {
			j=1;
			while (j<=dim[varr].row) {
			   setArrayVal(varr,t,j,i,getArrayVal(TEMPVAR,t1,i,j));
				j++;
			}
			i++;
		}
	}
}


/* MATINV(result-matrix)
 * The MAT INV statement for inverting a matrix
 * Read parameters from the stream, setup parameters and pass them to 
 * matinvert(), the **real** matrix inverter. 
 * varr: the result matrix, checked into MATtype() and passed here */
void MATINV(int varr){
	int var1=0, cs=0;

	/* argument required */
	if (!*p) perror_(__LINE__,125,l);

	/* existence was checked into MATtype() */
	
	/* get argument matrix code */
	forceArray=TRUE;
	if (*p++=='(') {
		cs=findDEC(p);
		if (cs && _rettype==NUM && _arr) var1=cs,p+=_len;
		else perror_(__LINE__,42,l);
		if (*p!=')') perror_(__LINE__,37,l);
		else p++;
	}
	else perror_(__LINE__,42,l);
	forceArray=FALSE;

	/* ELSE case */
	invElseCase=FALSE; // reset each time
	if (!strncasecmp_(p,"ELSE",4)) p+=4, invElseCase=TRUE;

	/* check matrix coherence */
	if (dim[var1].type!=2) perror_(__LINE__,76,l);

	/* redimension result matrix if its dimensions are greater than
	 * matrix to invert
	 * Note: in the expression MAT C=INV(A) C may be a vector; it is turned 
	 * to square, provided it has sufficient space for the NxN elements;
	 * Note: MAT A=INV(A) is considered legal (no control) */
	if ((dim[varr].brow+1)*(dim[varr].bcol+1)>=\
			(dim[var1].row+1)*(dim[var1].col+1)) {
		dim[varr].type=2; // in case of a transformed vector
		dim[varr].row = dim[var1].row;
		dim[varr].col = dim[var1].col;
	}
	else perror_(__LINE__,76,l);
	
	/* delegate inversion to matinvert */
	matinvert(varr,var1);

	/* check the ELSE CASE */
	if (!detValue && invElseCase) {
		p+=4;
		GOTO();
	}
}


/* matinvert(result-matrix, argument-matrix)
 * Invert matrix 'argument' into matrix 'result' and calculate the determinant 
 * using the Gauss factorization scheme. Variables:
 * - varr: the unknown result global inverted matrix (NxN) in the BASIC code
 * - var1: the global knowns matrix to be inverted (NxN) in the BASIC code, it
 *    is used to form ma
 * ma is the local composed matrix (2NxN) (to the left the knowns matrix, to 
 *    the right the identity matrix); this matrix is invisible to BASIC
 * x is the local calculated inverted matrix (NxN), invisible to BASIC and 
 *    later copied back to varr. 
 * detValue is the global determinant value, visible to BASIC through DET.
 * Copyleft Antonio Maschio - March 2008 */
void matinvert(int varr, int var1) {
	int n = (dim[var1].row), d=2*n+1; 	
	int i=0, j=0, k=0;
	double x[n+1][n+1],ma[n+1][d+1];
	double mul=0.0,piv=0.0;

	/* set determinant base multiplier */
	detValue=1.0;
	
	/* create composed matrix AI, with the same rows number of A,
	 * and twice the columns number of A; in the second half lies I */
	for (i=1;i<=n;i++) for(j=1;j<=n;j++) {
		ma[i][j]=getArrayVal(var1,2,i,j);
		/* setting I to all zeroes */
		ma[i][j+n+1]=0.0;
	}
	/* set diagonal of I to all ones */
	for (i=1;i<=n;i++) ma[i][i+n+1]=1.0;
	
	/* erase destination matrix */
	for (i=1; i<=n; i++) for (j=1; j<=n; j++) x[i][j]=0.0;
	
	/* Gauss factorization over AI with partial pivoting */
	for (k=1;k<n;k++) {
		piv=ma[k][k];
		if (!piv) {
			double maxx=0;
			int r=k;
			/* k is the row where pivot is zero
			 * now find the row where pivot is max */
			for (i=k+1;i<=n;i++) 
				if (maxx<fabs(ma[i][k])){ 
					maxx=ma[i][k]; 
					r=i; 
				}
			/* matrix is singular if r hasn't changed */
			if (r==k) { 
				if (!invElseCase) perror_(__LINE__,121,l); 
				detValue=0.0; 
				return; 
			}
			/* exchange rows r and k */
			for (i=1;i<=d;i++) {
				maxx=ma[r][i];
				ma[r][i]=ma[k][i];
				ma[k][i]=maxx;
			}
			piv=ma[k][k];
			/* matrix is singular if pivot is still null */
			if (!piv) { 
				if (!invElseCase) perror_(__LINE__,121,l); 
				detValue=0.0; 
				return; 
			}
			/* update determinant sign */
			detValue*=-1; 
		}
		/* update the rest of the rows */
		for (i=k+1;i<=n;i++) {
			mul=ma[i][k]/piv;
			for (j=k;j<=d;j++) ma[i][j]=ma[i][j]-mul*ma[k][j];
		} 
		
	}

	/* calculate determinant */
	for (i=1;i<=n;i++) detValue*=ma[i][i];
	
	/* check singularity; this should not be the case, since singularity
	 * should emerge from the calculation, but if, after all, determinant
	 * is null, print a message, to warn the user without stopping
	 * the program in any case.
	 * Actually, in some cases (notably, for the inverse of test matrix 
	 * {1 2 3}{4,5,6}{7,8,9}) this algorithm returns determinant zero
	 * and therefore the matrix is singular (not invertible); for this
	 * I take comfort using some online matrix inversion algorithms (mainly
	 * matrix.reshish.com, www.quickmath.com, matrixcalc.org/en/), which
	 * don't return any inverted matrix because the determinant is zero */
	if (detValue==0.0) {
		if (!invElseCase) perror_(__LINE__,121,l);
		return;
	}

	/* calculate the inverse of A through back-substitution */
	for (i=1;i<=n;i++) {
		x[n][i]=ma[n][i+n+1]/ma[n][n];
		for (j=n-1;j>=0;j--) {
			piv=0;
			for(k=j+1;k<=n;k++) piv+=ma[j][k]*x[k][i];
			x[j][i]=(ma[j][i+n+1]-piv)/ma[j][j];
		}
	}

	/* set varr to the resulting inverted matrix */
	for (i=0; i<=n; i++) for (j=0; j<=n; j++) 
		setArrayVal(varr,2,i,j,x[i][j]);
}


/* MATIDN(result-matrix)
 * The MAT IDN statement for setting a square array to identity matrix;
 * var: the code of the matrix to be IDNed (there is no argument matrix) */
void MATIDN(int var){
	int m=0, n=0, t=0;
	int i=0, j=0;

	/* set array values */
	m=dim[var].row;
	n=dim[var].col;
	t=dim[var].type;
	if (m!=n || t!=2) perror_(__LINE__,122,l);

	/* resizing values (only square matrices) */
	if (*p=='(') {
		p++;
		m=(int)S();
		if (*p==',') {
			p++;
			n=(int)S();
		}
		else perror_(__LINE__,122,l);
		if (m!=n) perror_(__LINE__,122,l); 
		if ((m+1)*(n+1)<=(dim[var].brow+1)*(dim[var].bcol+1)) {
			dim[var].row=m;
			dim[var].col=n;
			dim[var].type=t;
		}
		else perror_(__LINE__,176,l);
		if (*p!=')') perror_(__LINE__,37,l);
		else p++;
	}

	/* result matrix existence was checked into MATtype() */

	/* set values for IDN */
	for (i=0;i<=m;i++) for (j=0;j<=n;j++) setArrayVal(var,2,i,j,0.0);
	for (i=0;i<=m;i++) setArrayVal(var,2,i,i,1.0);
}


/* MATZER(result-matrix, option-value)
 * The MAT ZER and MAT CON statements for setting (or resetting) an array to 
 * all zeroes (MAT ZER) or given value (MAT CON), depending on option value 
 * 'op' (= 0 for ZER, 1 for CON) */
void MATZER(int var, double op){
	int t=0, m=0, n=0;
	int i=0, j=0;

	/* set array values */
	t=dim[var].type;
	m=dim[var].row;
	n=dim[var].col;

	/* resizing values */
	if (*p=='(') {
		p++;
		m=(int)S();
		n=0;
		t=1;
		if (*p==',') {
			p++;
			n=(int)S();
			t=2;
		}
		if ((m+1)*(n+1)<=(dim[var].brow+1)*(dim[var].bcol+1)) {
			dim[var].row=m;
			dim[var].col=n;
			dim[var].type=t;
		}
		else perror_(__LINE__,176,l);
		if (*p!=')') perror_(__LINE__,37,l);
		else p++;
	}

	/* result matrix existence was checked into MATtype() */
	
	/* set values for ZER or CON */
	if (t==1) for (i=0;i<m;i++) setArrayVal(var,t,i+1,0,op);
	else for (i=0;i<m;i++) for (j=0;j<n;j++) 
		setArrayVal(var,t,i+1,j+1,op);
}


/* MATNUL(result-matrix)
 * The MAT NUL$ statement for setting a string matrix to all void strings */
void MATNUL(int var){
	int t=0, m=0, n=0;
	int i=0, j=0;

	/* set array values */
	t=dimS[var].type;
	m=dimS[var].row;
	n=dimS[var].col;

	/* resizing values */
	if (*p=='(') {
		p++;
		m=(int)S();
		n=0;
		t=1;
		if (*p==',') {
			p++;
			n=(int)S();
			t=2;
		}
		if ((m+1)*(n+1)<=(dimS[var].brow+1)*(dimS[var].bcol+1)) {
			dimS[var].row=m;
			dimS[var].col=n;
			dimS[var].type=t;
		}
		else perror_(__LINE__,176,l);
		if (*p!=')') perror_(__LINE__,37,l);
		else p++;
	}

	/* result matrix existence was checked into MATtype() */
	
	/* set values for NUL$ */
	if (t==1) for (i=0;i<m;i++) setArrayStr(var,t,i+1,0, NULLSTR);
	else for (i=0;i<m;i++) for (j=0;j<n;j++) 
		setArrayStr(var,t,i+1,j+1, NULLSTR);
}


/* MATsum(result-matrix,matrix1,matrix2,option-value)
 * Add two matrices or subtract the second from the first.
 * varr: the result matrix code
 * var1: first matrix addendum
 * var2: second matrix addendum
 * op: if -1, do subtraction, if 1 do addition */
void MATsum(int varr, int var1, int var2, int op){
	int i=0,j=0;
	int M=0, N=0;
	int T=dim[varr].type;
	int T1=dim[var1].type;
	int T2=dim[var2].type;
	
	/* existences of result matrix and of argument matrices var1 and var2 
	 * have already been checked into MATtype() */

	/* check matching of dimensions 
	 * Note: one can only sum vectors with vectors, matrices with
	 * matrices, and not use unidimensional vectors or
	 * degenerated matrices of one row or of one column */
	if (dim[var1].row!=dim[var2].row ||\
			dim[var1].col!=dim[var2].col)
		perror_(__LINE__,76,l);
	if (T1!=T2) perror_(__LINE__,76,l);

	/* Legal procedures for special cases:
	 * MAT A=A+B, A=A-B 
	 * MAT A=B+A, A=B-A (which should succeed for the same reasons)
	 * MAT A=A+A, A=A-A */

	/* redimension result matrix */
	if ((dim[var1].brow+1)*(dim[var1].bcol+1)>\
			(dim[varr].brow+1)*(dim[varr].bcol+1))
		perror_(__LINE__,76,l);
	else {
		M=dim[varr].row=dim[var1].row;
		N=dim[varr].col=dim[var1].col;
	}

	/* create dummy matrix dimensionally equal to varr */
	dimArray(TEMPVAR,T,M,N);
	
	/* perform sum and copy data back to varr;
	 * TEMPVAR serves as a buffer for not overwriting any values */
	i=1; j=(T==1)?0:1;
	while (i<=M) {
		setArrayVal(TEMPVAR,T,i,j,\
				getArrayVal(var1,T1,i,j)+\
				op*getArrayVal(var2,T2,i,j));
		setArrayVal(varr,T,i,j,getArrayVal(TEMPVAR,T,i,j));
		j++;
		if (j>N) {i++; j=(T==1)?0:1;}
	}

}


/* MATmul(result-matrix,matrix1,matrix2)
 * The cross-product between two matrices or vectors 
 * varr: the result matrix code
 * var1: first matrix 
 * var2: second matrix */
void MATmul(int varr, int var1, int var2){
	int i,j,k;
	int M=dim[varr].row;
	int N=dim[var1].col;
	int P=dim[var2].col;
	int T=dim[varr].type;
	int T1=dim[var1].type;
	int T2=dim[var2].type;

	/* existences of result matrix and of argument matrices var1 and var2 
	 * have already been checked into MATtype() */

	/* instantiate temporary matrix with result matrix dimensions */
	dimArray(TEMPVAR,T,M,dim[varr].col);

	/* check dimensions validity */
	if ((dim[varr].brow+1)*(dim[varr].bcol+1)<\
			(dim[var1].row+1)*(dim[var2].col+1))
		perror_(__LINE__,76,l);

	/* redimension result matrix (and temporary) */
	M=dim[varr].row=dim[TEMPVAR].row=dim[var1].row;
	dim[varr].col=dim[TEMPVAR].col=dim[var2].col;
	
	 /* Legal procedures for special cases:
	 * MAT B=A*A 
	 * MAT A=A*B, A=B*A and A=A*A */

	/* perform cross multiplication and copy data back to varr 
	 * TEMPVAR serves as a buffer for not overwriting any values */
	if (T==2) {
		for(i=1;i<=M;i++) {
			for(j=1;j<=P;j++) {
				double res=0.0;
				for(k=1;k<=N;k++)
					res+=getArrayVal(var1,T1,i,k)*\
						getArrayVal(var2,T2,k,j);
				setArrayVal(TEMPVAR,T,i,j,res);
			}
		}
		for (i=1;i<=M;i++)  for (j=1;j<=P;j++)
			setArrayVal(varr,T,i,j,getArrayVal(TEMPVAR,T,i,j));
	}
	else if (T==1) {
		for(i=1;i<=M;i++) {
			double res=0.0;
			for(k=1;k<=N;k++)
				res+=getArrayVal(var1,T1,i,k)*\
					getArrayVal(var2,T2,k,0);
			setArrayVal(TEMPVAR,T,i,0,res);
		}
		for (i=1;i<=M;i++)
			setArrayVal(varr,T,i,0,getArrayVal(TEMPVAR,T,i,0));
	}
}



/* MATtype()
 * this function is invoked by parse when a MAT statement is met (other than 
 * MAT READ, MAT PRINT or MAT INPUT, which follow their way), to choose the 
 * correct command. 
 * Controls are done on result matrix (instantiating it if not existing), and
 * in case of addition, subtraction and multiplication between matrices, 
 * control is done on argument matrices; peculiar functions like IDN, CON,
 * ZER, INV and TRN check argument matrix on their own. */
void MATtype(void) {
	int varr=0; 
	int var1=0;
	int var2=0;
	int op=0;
	int mode=NUM, cs;

	/* get result matrix code */
	forceArray=TRUE;
	cs=findDEC(p);
	if (cs && _rettype==NUM && _arr) varr=cs, p+=_len;
	else if (cs && _rettype==STR && _arr) varr=cs, p+=_len, mode=STR;
	else perror_(__LINE__,42,l);
	forceArray=FALSE;

	if (*p++!='=') perror_(__LINE__,35,l);
	
	/* string mat operations */
	if (mode==STR) {
		/* perform the erasing of string matrix (delegate MATNUL) */
		if (!strncasecmp_(p,"NULL$",5)) {
			p+=5; 
			MATNUL(varr);
			return;
		}
		else if (!strncasecmp_(p,"NUL$",4)) {
			p+=4; 
			MATNUL(varr);
			return;
		}
		/* perform the matrix copying (delegate matStrCopy) */
		else {
			/* get first matrix operand code */
			forceArray=TRUE;
			if ((cs=findDEC(p)) && _rettype==STR && _arr) 
				var1=cs,p+=_len;
			else perror_(__LINE__,42,l);
			forceArray=FALSE;
			if (!*p) matStrCopy(varr, var1);
			return;
		}
	}
	/* numeric mat functions */
	else if (!strncasecmp_(p,"ZERO",4)) p+=4, MATZER(varr,0.0);
	else if (!strncasecmp_(p,"ZER",3)) p+=3, MATZER(varr,0.0);
	else if (!strncasecmp_(p,"UNITY",5)) p+=5, MATZER(varr,1.0);
	else if (!strncasecmp_(p,"CON",3)) p+=3, MATZER(varr,1.0);
	else if (!strncasecmp_(p,"TRANSPOSE",9)) p+=9, MATTRN(varr);
	else if (!strncasecmp_(p,"TRN",3)) p+=3, MATTRN(varr);
	else if (!strncasecmp_(p,"IDENTITY",8)) p+=8, MATIDN(varr);
	else if (!strncasecmp_(p,"IDN",3)) p+=3, MATIDN(varr);
	else if (!strncasecmp_(p,"INVERT",6)) p+=6, MATINV(varr);
	else if (!strncasecmp_(p,"INV",3)) p+=3, MATINV(varr);
	else if (*p=='(') p++, MATbyK(varr);
	/* numeric mat operations */
	else {
		/* get first matrix operand code */
		forceArray=TRUE;
		if ((cs=findDEC(p)) && _rettype==NUM && _arr)
			var1=cs,p+=_len;
		else perror_(__LINE__,42,l);
		forceArray=FALSE;
		
		/* perform the simple copy (no operators) */
		if (!*p) matCopy(varr, var1);
		/* perform the matrix operations */
		else {
			/* get op */
			if (*p!='+' && *p!='-' && *p!='*' && *p!='\0') 
				perror_(__LINE__,35,l);
			else op=*p++;

			/* get second matrix operand code */
			forceArray=TRUE;
			cs=findDEC(p);
			if (op!='\0') {
				if (cs && _rettype==NUM && _arr) 
					var2=cs, p+=_len;
				else perror_(__LINE__,42,l);
			}
			forceArray=FALSE;
			
			/* perform numerical operations */
			     if (op=='-') MATsum(varr, var1, var2, -1);
			else if (op=='+') MATsum(varr, var1, var2, 1);
			else if (op=='*') MATmul(varr, var1, var2);
			else perror_(__LINE__,35,l);
		}
	}
}


/* matCopy(result-matrix, argument-matrix)
 * copy src argument matrix into dst result matrix;
 * dst: destination matrix
 * src: source matrix
 * Automatic redimension is performed in case target matrix can hold argument 
 * matrix dimensions */
void matCopy(int dst, int src){
	int i,j;
	/* source data */
	int ms=dim[src].row;
	int ns=dim[src].col;
	int ts=dim[src].type;
	/* destination data */
	int MD=dim[dst].row;
	int ND=dim[dst].col;
	int TD=dim[dst].type;

	/* existence of argument matrix var1 has already 
	 * been checked into MATtype() */

	/* check immediate compatibility */
	if (dst==src) return;
	if (ts!=TD || ts==0 || TD==0) perror_(__LINE__,76,l);
	
	/* redimension */
	if (ms!=MD || ns!= ND) {
		if ((MD+1)*(ND+1)>=(dim[src].brow+1)*(dim[src].bcol+1)) {
			MD=dim[dst].row=dim[src].row;
			ND=dim[dst].col=dim[src].col;
		}
		else perror_(__LINE__,76,l);
	}
	
	/* check compatibility */
	if (ms!=MD || ns!=ND) perror_(__LINE__,76,l);

	/* copy (copy column 0 and row 0 too!) */
	if (ts==2) for (i=0;i<=ms;i++) for (j=0;j<=ns;j++)
		setArrayVal(dst,TD,i,j,getArrayVal(src,ts,i,j));
	else if (ts==1) for (i=0;i<=ms;i++)
		setArrayVal(dst,TD,i,0,getArrayVal(src,ts,i,0));
}


/* matStrCopy(result-matrix, argument-matrix)
 * copy src argument matrix into dst result matrix;
 * dst: destination matrix
 * src: source matrix
 * Automatic redimension is performed in case target matrix can hold argument 
 * matrix dimensions */
void matStrCopy(int dst, int src){
	int i,j;
	/* source data */
	int ms=dimS[src].row;
	int ns=dimS[src].col;
	int ts=dimS[src].type;
	/* destination data */
	int MD=dimS[dst].row;
	int ND=dimS[dst].col;
	int TD=dimS[dst].type;

	/* existence of argument matrix var1 has already 
	 * been checked into MATtype() */

	/* check immediate compatibility */
	if (dst==src) return;
	if (ts!=TD || ts==0 || TD==0) perror_(__LINE__,76,l);
	
	/* redimension */
	if (ms!=MD || ns!= ND) {
		if ((MD+1)*(ND+1)>=(dimS[src].brow+1)*(dimS[src].bcol+1)) {
			MD=dimS[dst].row=dimS[src].row;
			ND=dimS[dst].col=dimS[src].col;
		}
		else perror_(__LINE__,76,l);
	}
	
	/* check compatibility */
	if (ms!=MD || ns!=ND) perror_(__LINE__,76,l);

	/* copy (copy column 0 and row 0 too!) */
	if (ts==2) for (i=0;i<=ms;i++) for (j=0;j<=ns;j++)
		setArrayStr(dst,TD,i,j,getArrayStr(src,ts,i,j));
	else if (ts==1) for (i=0;i<=ms;i++)
		setArrayStr(dst,TD,i,0,getArrayStr(src,ts,i,0));
}


/* MATFINPUT()
 * The MAT INPUT# wrapper
 * Multiple-matrices printing enabled */
void MATFINPUT(int isWrite) {
	int ch=0;
	
	/* argument required */
	if (!*p) perror_(__LINE__,125,l);

	/* print on file: get file channel */
	ch=(int)S();
	if (ch<1 || ch>STREAMS) perror_(__LINE__,133,l);
	if (*p==',' || *p==':') p++; // optional comma/colon
	if (!channel[ch].isopen) perror_(__LINE__,80,l);

	/* delegate MATINPUT */
	MATINPUT(ch,isWrite);
}


/* MATINPUT()
 * The MAT INPUT statement
 * mode 1 is for numerical matrices, mode 2 for string matrices.
 * Multiple-matrices input enabled, each with its own question point, and
 * the ability to break the input and require a continuation adding & after 
 * the last input number */ 
void MATINPUT(int ch, int isWrite) {
	char line[COMPSCR], *u=line, *pp=p;
	int cs=0, mode=0, M=0, N=0, T=0, var=0;

	numValue=0; 
	/* argument required */
	if (!*p) perror_(__LINE__,125,l);

	/* get user input */
	if (!ch) {
		if (isPrompt) fprintf(stdout,INPUTS);
		p=pp; u=line; *u='\0';
		strncpy_(line,getString(),COMPSCR);
		if (!isEcho) fprintf(stdout,"\n");
		if ((int)strlen(line)>MAXSTRLEN) perror_(__LINE__,124,l);
	}
	/* get file input */
	else {
		fgets(line,COMPSCR,channel[ch].stream);
		if ((int)strlen(line)>0) {
			if (line[(int)strlen(line)-1]=='\n')
				line[(int)strlen(line)-1]='\0';
		}
		numValue++;
		u=line;
		/* read line number and discard it */
		if (isWrite) {
			while (isblank(*u)) u++;
			if (isdigit(*u)) while (isdigit(*u)) u++;
			else perror_(__LINE__,97,-1,l);
		}
	}
	
	/* get array type */
	while (isblank(*u)) u++;
	forceArray=TRUE;
	cs=findDEC(p);
	if (cs && _rettype==STR && _arr) var=cs, mode=2, p+=_len;
	else if (cs && _rettype==NUM && _arr) var=cs, p+=_len, mode=1;
	else perror_(__LINE__,42,l);
	forceArray=FALSE;

	/* setup for horizontal arrays or normal matrices */
	if (mode==1) {
		T=dim[var].type; 
		if (!dim[var].row && dim[var].col) M=0, N=1;
		else M=1, N=1;
	}
	else {
		T=dimS[var].type;
		if (!dimS[var].row && dimS[var].col) M=0, N=1;
		else M=1, N=1;
	}

	/* INPUT a numerical vector */
	if (mode==1 && T==1 && *u) do {
		char *pp=p;
		double res;
		if (*u==',') u++;
		while (isblank(*u)) u++;
		/* asking for more in case of ampersand */
		if (!ch) {
			if (*u=='&') {
				if (isPrompt) fprintf(stdout,INPUTS);
				u=line;
				*u='\0';
				strncpy_(line,getString(),COMPSCR);
				if (!isEcho) fprintf(stdout,"\n");
			}
		}
		/* asking for more in case of file reading */
		else if (!*u) {
			if (fgets(line,COMPSCR,channel[ch].stream)) {
				if ((int)strlen(line)>0) {
					if (line[(int)strlen(line)-1]=='\n')
						line[(int)strlen(line)-1]='\0';
				}
				u=line;
				/* read line number and discard it */
				if (isWrite) {
					while (isblank(*u)) u++;
					if (isdigit(*u)) 
						while (isdigit(*u)) u++;
					else perror_(__LINE__,97,-1,l);
				}
			}
			else break;
		}
		while (isblank(*u)) u++;
		if (!*u) break;
		p=u; res=S();
		setArrayVal(var,T,M,0,res);
		u=p; p=pp;
		M++;
		if (M>dim[var].row) break;
		numValue++;
		while (isblank(*u)) u++;
		if (*u==',') u++;
	} while (TRUE);
	/* INPUT from keyboard a numerical matrix or horizontal vector 
	 * Note: the MAT INPUT for a matrix works in this way: first item is 
	 * attributed to element at pos (1,1), the remaining are lined in 
	 * memory consecutively, excluding item at pos (0,X) or (X,0)
	 * Note: M and N go through the whole matrix/vector 
	 * Note: M is set to 0 in case of horizontal vector */
	else if (mode==1 && T==2 && *u) do {
		char *pp=p;
		double res;
		if (*u==',') u++;
		while (isblank(*u)) u++;
		/* asking for more in case of ampersand */
		if (!ch) {
			if (*u=='&') {
				fprintf(stdout," ?");
				u=line;
				*u='\0';
				strncpy_(line,getString(),COMPSCR);
				if (!isEcho) fprintf(stdout,"\n");
			}
		}
		/* asking for more in case of file reading */
		else if (!*u) {
			if (fgets(line,COMPSCR,channel[ch].stream)) {
				if ((int)strlen(line)>0) {
					if (line[(int)strlen(line)-1]=='\n')
						line[(int)strlen(line)-1]='\0';
				}
				u=line;
				/* read line number and discard it */
				if (isWrite) {
					while (isblank(*u)) u++;
					if (isdigit(*u)) 
						while (isdigit(*u)) u++;
					else perror_(__LINE__,97,-1,l);
				}
			}
			else break;
		}
		while (isblank(*u)) u++;
		if (!*u) break;
		p=u; res=S();
		setArrayVal(var,T,M,N,res);
		u=p; p=pp;
		N++;
		if (N>dim[var].col) {
			M++;
			N=1;
			if (M>dim[var].row) break;
		}
		numValue++;
		while (isblank(*u)) u++;
		if (*u==',') u++;
	} while (TRUE);
	/* INPUT from keyboard a string vector */
	else if (*u && T==1) do {
		char op[COMPSTR], *P1;
		P1=op, *P1='\0';
		if (*u==',') u++;
		while (isblank(*u)) u++;
		/* asking for more in case of ampersand */
		if (!ch) {
			if (*u=='&') {
				fprintf(stdout," ?");
				u=line;
				*u='\0';
				strncpy_(line,getString(),COMPSCR);
				if (!isEcho) fprintf(stdout,"\n");
			}
		}
		/* asking for more in case of file reading */
		else if (!*u) {
			if (fgets(line,COMPSCR,channel[ch].stream)) {
				if ((int)strlen(line)>0) {
					if (line[(int)strlen(line)-1]=='\n')
						line[(int)strlen(line)-1]='\0';
				}
				u=line;
				/* read line number and discard it */
				if (isWrite) {
					while (isblank(*u)) u++;
					if (isdigit(*u)) 
						while (isdigit(*u)) u++;
					else perror_(__LINE__,97,-1,l);
				}
			}
			else break;
		}
		while (isblank(*u)) u++;
		if (!*u) break;
		/* grab next input string until comma or blank */
		if (*u!='\"') {
			while (*u && *u!=',' && *u!='&' &&!isblank(*u)) 
				*P1++=*u++;
			*P1='\0';
		}
		/* grab next input string until closing quotes */
		else {
			u++;
			while (*u && *u!='\"' && *u!='&') *P1++=*u++;
			*P1='\0';
			if (*u=='\"') u++;
			if (*u=='&') *--u=',';
			while (isblank(*u)) u++;
		}
		setArrayStr(var,T,M,0,op);
		M++;
		if (M>dimS[var].row) break;
		numValue++;
		while (isblank(*u)) u++;
	} while (TRUE);
	/* INPUT from keyboard a string matrix 
	 * Note: the MAT INPUT for a matrix works in this way: first item is 
	 * attributed to element at pos (1,1), the remaining are lined in 
	 * memory consecutively, including item at pos (X,1)
	 * Note: M and N go through the whole matrix/vector */
	else if (*u && T==2) do {
		char op[COMPSTR], *P1;
		P1=op, *P1='\0';
		if (*u==',') u++;
		while (isblank(*u)) u++;
		/* asking for more in case of ampersand */
		if (!ch) {
			if (*u=='&') {
				if (isPrompt) fprintf(stdout,INPUTS);
				u=line;
				*u='\0';
				strncpy_(line,getString(),COMPSCR);
				if (!isEcho) fprintf(stdout,"\n");
			}
		}
		/* asking for more in case of file reading */
		else if (!*u) {
			if (fgets(line,COMPSCR,channel[ch].stream)) {
				if ((int)strlen(line)>0) {
					if (line[(int)strlen(line)-1]=='\n')
						line[(int)strlen(line)-1]='\0';
				}
				u=line;
				/* read line number and discard it */
				if (isWrite) {
					while (isblank(*u)) u++;
					if (isdigit(*u)) 
						while (isdigit(*u)) u++;
					else perror_(__LINE__,97,-1,l);
				}
			}
			else break;
		}
		while (isblank(*u)) u++;
		if (!*u) break;
		/* grab next input string until comma or blank */
		if (*u!='\"') {
			while (*u && *u!=',' && *u!='&' && !isblank(*u)) 
				*P1++=*u++;
			*P1='\0';
		}
		/* grab next input string until closing quotes */
		else {
			u++;
			while (*u && *u!='\"' && *u!='&') *P1++=*u++;
			*P1='\0';
			if (*u=='\"') u++;
			if (*u=='&') *--u=',';
			while (isblank(*u)) u++;
		}
		setArrayStr(var,T,M,N,op);
		N++;
		if (N>dimS[var].col) {
			M++;
			N=1;
			if (M>dimS[var].row) break;
		}
		numValue++;
		while (isblank(*u)) u++;
	} while (TRUE);
	/* NUM returns the total number of items INPUT so far, even 
	 * those not used; so count remaining items (future usage) */
	if (*u && mode==1) {
		numValue++;
		while (*u) {
			if (*u==',') numValue++;
			u++;
		}
	}

	/* redo in case of comma */
	if (*p==',') p++;
	if (*p) MATINPUT(ch, isWrite);
	
	/* reset printing position */
	channel[0].fprintn=channel[0].fcounter=0;
}


/* MATREAD()
 * The MAT READ statement-
 * Multiple-matrices reading enabled */
void MATREAD(void){
	int cs, var=0, m=0, n=0, mode=0; // mode=0 => number, 1 => string
	
	/* argument required */
	if (!*p) perror_(__LINE__,125,l);
	
	/* start loop on all MAT READ arrays */	
	do {
		mode=0;
		/* get variable code */
		forceArray=TRUE;
		cs=findDEC(p);
		if (cs && _rettype==NUM && _arr) 
			var=cs, p+=_len;
		else if (cs && _rettype==STR && _arr) 
			var=cs, p+=_len, mode=1;
		else perror_(__LINE__,65,l,makechar(p+1),TERMINAT);
		forceArray=FALSE;
		
		/* READ numerical array */
		if (!mode && _rettype==NUM) {
			if (dataLimit==0) perror_(__LINE__,50,l);
			/* use implicit data */
			else if (*p=='\0' || *p==',') {
				m=dim[var].row;
				n=dim[var].col;
			}
			else perror_(__LINE__,34,l);
	
			/* assign DATA values */
			matReadAssign(var,dim[var].type,m,n);	
	
		} 
		/* READ string array */
		else if (_rettype==STR) {
			if (dataLimitS==0) perror_(__LINE__,50,l);
			/* use implicit data */
			else if (*p=='\0' || *p==',') {
				m=dimS[var].row;
				n=dimS[var].col;
			}
			else perror_(__LINE__,34,l);

			/* assign DATA values */
			matStrReadAssign(var,dimS[var].type,m,n);	

		}
		else perror_(__LINE__,34,l);
	} while (*p++==',');
}


/* WFCR()
 * print a carriage return doing the MAT WRITE line
 * prepending a five digits number followed by a tab */
void WFCR(int ch, int isWrite) {
	FCR(ch);
	if (isWrite && ch) { 
		fprintf(channel[ch].stream,"%05d\t",writeSeqStart);
		writeSeqStart+=writeSeqStep;
	}
}


/* MATFPRINT()
 * The MAT PRINT# wrapper
 * Multiple-matrices printing enabled */
void MATFPRINT(int isWrite) {
	int ch=0;
	
	/* argument required */
	if (!*p) perror_(__LINE__,125,l);

	/* print on file: get file channel */
	ch=(int)S();
	if (ch<1 || ch>STREAMS) perror_(__LINE__,133,l);
	if (*p==',' || *p==':') p++; // optional comma/colon
	if (!channel[ch].isopen) perror_(__LINE__,80,l);

	/* delegate MATPRINT */
	MATPRINT(ch, isWrite);
}


/* MATPRINT()
 * The MAT PRINT statement 
 * Multiple-matrices printing enabled */
void MATPRINT(int ch, int isWrite) {
	int var=0, m=0, n=0, t=0;	// matrix data
	int mode=0, ma;			// mode=TRUE=string array, ma=margin
	int cs,i,j;			// counters
	char outstr[COMPSTR];
	static int spacing=0;		// 1=comma, 2=semicolon, 0=vertical

	/* argument required */
	if (!*p) perror_(__LINE__,125,l);

	/* resolve irregular format */
	if (!isbeginnamechar(p)) perror_(__LINE__,35,l);

	/* get variable codes */
	forceArray=TRUE;
	cs=findDEC(p);
	if (cs && _arr &&_rettype==NUM) var=cs, p+=_len;
	else if (cs && _arr && _rettype==STR) var=cs,mode=1,p+=_len;
	else perror_(__LINE__,42,l);
	forceArray=FALSE;
	if (*p==',') spacing=1,p++;
	else if (*p==';') spacing=2, p++;

	/* print starting WRITE line */
	if (isWrite && ch) { 
		fprintf(channel[ch].stream,"%05d\t",writeSeqStart);
		writeSeqStart+=writeSeqStep;
	}

	/* PRINT a string array */
	if (mode) {
		/* store string matrix/vector basic data */
		t=dimS[var].type;
		m=dimS[var].row;
		n=dimS[var].col; 
		ma=channel[ch].margin;

		/* print a string vector */
		if (t==1) {
			i=1;
			do {
				strncpy_(outstr,getArrayStr(var,t,i,0),COMPSTR);
				if (outstr[0]) printContained(ch,outstr,ma);
				else printContained(ch,"|           |",ma);
				/* add a separator space */
				if (isSpacing) printContained(ch," ",ma);
				if (spacing==1) { 
					while (channel[ch].fcounter%ZONE) 
						FAC(ch);
				}
				else if (!spacing) WFCR(ch,isWrite);
				if (channel[ch].fcounter>ma) WFCR(ch,isWrite);
				i++;
			} while (i<=m);
			if (spacing>0) FCR(ch);
			FCR(ch);
		}
		/* print a horizontal string vector */
		else if (!m && n) {
			i=1;
			do {
				strncpy_(outstr,getArrayStr(var,t,0,i),COMPSTR);
				if (outstr[0]) printContained(ch,outstr,ma);
				else printContained(ch,"|           |",ma);
				/* add a separator space */
				if (isSpacing) printContained(ch," ",ma);
				if (spacing<2) { 
					while (channel[ch].fcounter%ZONE) 
						FAC(ch);
				}
				if (channel[ch].fcounter>ma) WFCR(ch,isWrite);
				i++;
			} while (i<=n);
			FCR(ch);
			FCR(ch);
		}
		/* print a string matrix */
		else {
			i=1;
			do {
				j=1;
				do {
				   strncpy_(outstr,getArrayStr(var,t,i,j),COMPSTR);
				   if (outstr[0]) 
				      printContained(ch,outstr,ma);
				   else 
				      printContained(ch,"|           |",ma);
				   /* add a separator space */
				   if (isSpacing) printContained(ch," ",ma);
				   if (channel[ch].fcounter>ma)WFCR(ch,isWrite);
				   if (spacing!=2) { 
				      while(channel[ch].fcounter%ZONE) FAC(ch);
				   }
				   if (channel[ch].fcounter>ma)WFCR(ch,isWrite);
				   j++;
				} while (j<=n);
				i++;
				if (i<=m) WFCR(ch,isWrite);
			} while (i<=m);
			FCR(ch);
			FCR(ch);
		}
	}

	/* PRINT a numerical array/matrix */
	else {
		/* get numeric matrix/vector elements */
		t=dim[var].type;
		m=dim[var].row;
		n=dim[var].col;
		ma=channel[ch].margin;
		
		/* print a numeric vector */
		if (t==1) {
			i=1;
			do {
				channel[ch].fcounter=\
					printNUM(ch,getArrayVal(var,t,i,0),0,NIL);
				if (spacing==1) {
				      while (channel[ch].fcounter%ZONE) FAC(ch);
				}
				else if (!spacing) WFCR(ch,isWrite);
				if (channel[ch].fcounter>ma) WFCR(ch,isWrite);
				i++;
			} while (i<=m);
			if (spacing>0) FCR(ch);
			FCR(ch);
		}
		/* print a numeric horizontal vector */
		else if (!m && n) {
			i=1;
			do {
				channel[ch].fcounter=\
					printNUM(ch,getArrayVal(var,t,0,i),0,NIL);
				if (spacing<2) {
					while (channel[ch].fcounter%ZONE) 
						FAC(ch);
				}
				if (channel[ch].fcounter>ma) 
					WFCR(ch,isWrite);
				i++;
			} while (i<=n);
			FCR(ch);
			FCR(ch);
		}
		/* print a numeric matrix */
		else {
			i=1;
			do {
				j=1;
				do {
					channel[ch].fcounter=printNUM(ch,getArrayVal(var,t,i,j),0,NIL);
					if (channel[ch].fcounter>ma)
						WFCR(ch,isWrite);
					if (spacing!=2) { 
					     while(channel[ch].fcounter%ZONE) 
						FAC(ch);
					}
					if (channel[ch].fcounter>ma)
						WFCR(ch,isWrite);
					j++;
				} while (j<=n);
				i++;
				if (i<=m) WFCR(ch,isWrite);
			} while (i<=m);
			FCR(ch);
			FCR(ch);
		}
	}
	
	/* if there's something else to print, recall MATPRINT again 
	 * or close operations resetting data */
	if (*p==',' || *p==';') p++;
	if (*p) MATPRINT(ch,isWrite);
	else spacing=0;
}


/****************************************************************************** 
 * The following routines define the behavior of the statements involved with 
 * the CHANGE statement;
 * the CHANGE statement was included since 1967 version IV (not including the 
 * add-on variation CHANGE..TO..BIT, which appeared in 1969 version V). 
 ******************************************************************************/


/* CHANGE() 
 * The CHANGE statement
 * convert the explicit string (a variable or a constant enclosed into 
 * parentheses) to the series of ASCII character it's composed of, and store it 
 * into the vector specified after TO;
 * convert also the vector which holds the string data to the string variable
 * specified after TO. 
 * The vector need not to be dimensioned if the string is shorter than 10 
 * characters (pos. 0 holds the length, pos. 1 through 10 hold the string), but
 * if it's longer it must be declared, otherwise an error will be raised 
 * Note: the CHANGE statement accepts any kind of strings as argument, even any
 * DECLAREd string, but not string arrays elements */
void CHANGE(void) {
	char op[COMPSTR], *P1=op;
	int var=0, varS=0, len, cs;

	/* argument required */
	if (!*p) perror_(__LINE__,125,l);

	/* check general syntax */
	if (check(p,"TO")); else perror_(__LINE__,37,l);

	forceArray=TRUE;
	/* in case first argument is a numerical variable (which stands for 
	 * the array name) perform CHANGE vector TO string */
	if ((cs=findDEC(p)) && _arr && _rettype==NUM) {
		int c,val;

		/* get var name */
		forceArray=FALSE;
		var=cs, p+=_len;

		/* check vector existence and dimension 
		 * (only unidimensional vectors are valid) */
		if (dim[var].type==2) perror_(__LINE__,62,l);
		
		/* check syntax */
		if (strncasecmp_(p," O",2)) perror_(__LINE__,38,l);
		p+=2;

		/* check if a string variable is used */
		cs=findDEC(p);
		if (cs) {
			if (vn[cs].isArr) perror_(__LINE__,93,l);
			if (vn[cs].rettype!=STR) perror_(__LINE__,106,l);
			varS=cs;
		}
		
		/* fill string if len was correct
		 * (in general, error 90 depends on a file corruption) */
		len=(int)getArrayVal(var,1,0,0);
		if (len>=MAXSTRLEN) len=MAXSTRLEN;
		else if (len<=0) perror_(__LINE__,90,l);
		for (c=0; c<len; c++) {
			val=getArrayVal(var,1,c+1,0);
			/* Note; illegal character is a negative val, here */
			if (val<0) perror_(__LINE__,87,l);
			PS[varS][+c]=val;
		}
		PS[varS][len]='\0';
	}
	
	/* otherwise perform CHANGE string TO vector */
	else {
		int i=0;
		forceArray=FALSE;

		/* get string variable code */
		if ((cs=findDEC(p)) && !_arr && _rettype==STR) var=cs;
		else if (isanystring(p));
		else perror_(__LINE__,93,l);
		
		/* store string 
		 * Note: tbas is able of reading here string arrays elements 
		 * too, but for uniformity with the CHANGE vector TO string 
		 * feature, only string variables are allowed here. */
		*P1='\0';
		strncpy_(op,SS(),COMPSTR);
		len=(int)strlen(op);
		/* check syntax */
		if (strncasecmp_(p," O",2)) perror_(__LINE__,38,l);
		p+=2;
		
		/* check if a string variable is used */
		forceArray=TRUE;
		cs=findDEC(p);
		if (cs) {
			if (vn[cs].rettype!=NUM) perror_(__LINE__,106,l);
			if (!vn[cs].isArr && dim[cs].type!=2) 
				perror_(__LINE__,62,l);
		}
		forceArray=FALSE;

		/* get vector code */
		if (cs) var=cs, p+=_len;
		else perror_(__LINE__,62,l);

		/* check vector existence and dimension 
		 * (only unidimensional vectors are valid) */
		if (dim[var].type==2) perror_(__LINE__,76,l);
		if (len>dim[var].row) perror_(__LINE__,106,l);
	
		/* do the string conversion, storing in 0 string length */
		setArrayVal(var,1,i++,0,len);
		while (*P1) setArrayVal(var,1,i++,0,*P1++);
	}
}


/******************************************************************************
 * The following routines define the remaining standard BASIC statements;
 * most were present in BASIC since 1964 (apart from PRINT, which has been
 * reunited with PRINT#/WRITE#) but many (above all loops and cycles) have been 
 * introduced after the quick BASIC language defined a sort of standard for 
 * the BASIC language after 1985, when it was introduced for DOS.
 ******************************************************************************/

/* EXITSUB()
 * the EXIT SUB statement to exit prematurely from a subroutine 
 * Note: if the subroutine is a function, and the return value is not
 * yet been calculated, a warning is printed */
void EXITSUB(void) {
	if (!isSubRun) perror_(__LINE__,221,l);
	if (!isSubCalc) perror_(__LINE__,220,l); // warning
	l=sn[isSubVar].end;
	
}


/* RETURNTOSUB() 
 * the RETURN statement for SUB 
 * Not to be confused with the RETURN statement for the GOSUB/RETURN feature 
 * Note: the RETURN statement into SUB does not work as the return command 
 * in C or other similar languages; whereas in those language 'return' exit
 * immediately from the function, RETURN here is a mere return-value-storage;
 * any time it is met into SUB, the return value is stored, and will be
 * returned when out of SUB (for function, not procedures). To specifically
 * exit after a return, use it in conjunction with EXIT SUB */
void RETURNTOSUB(void) {
	if (sn[isSubVar].type==FUNC && sn[isSubVar].rettype==NUM)
		sn[isSubVar].numres=S();
	else if (sn[isSubVar].type==FUNC && sn[isSubVar].rettype==STR) {
		if (sn[isSubVar].strres!=NIL && sn[isSubVar].strres[0]) {
			free(sn[isSubVar].strres);
			sn[isSubVar].strres=NIL;
		}
		sn[isSubVar].strres=xmalloc((size_t)COMPSTR);
		if (!sn[isSubVar].strres) perror_(__LINE__,195,l);
		else strncpy_(sn[isSubVar].strres,SS(),COMPSTR);
	}
	else perror_(__LINE__,216,l);
	isSubCalc=TRUE;
}


/* EXPORT()
 * the EXPORT statement 
 * marks the variable in the list as exportable; at the end of the SUB
 * they won't be overwritten.
 * Note: if a global variable is set as to be EXPORTED, nothing happens 
 * Note: if an undeclared variable is set as to be EXPORTED, it is created 
 * as new*/
void EXPORT(void) {
	int cs;

	/* EXPORT cannot be used in main or into a DEF FN */
	if (isDefRun || !isSubRun) perror_(__LINE__,201,l);

	/* argument required */
	if (!*p) perror_(__LINE__,125,l);

	/* export all in one turn 
	 * Note: if ALL is followed by something else (a letter, a number, the 
	 * comma), it is not interpreted as the ALL  token, but is left to the 
	 * rest of the algorithm to find if it's an any-name variable beginning 
	 * with ALL */
	if (!strncasecmp_(p,"ALL",3) && !*(p+3)) {
		saveSubCount[XCount]=decCount;
		p+=3;
		return;
	}

	/* check variable list and set mask */
	while (*p) {
		cs=findDEC(p);
		/* main-DECLARE variables 
		 * Note: these variables need not any rearranging */
		if (cs && cs<=saveSubCount[XCount] && !vn[cs].isArr){
			p+=_len;
		}
		/* sub-DECLARE variables 
		 * Note: in order to EXPORT these variables, they need to
		 * be rearranged: the chosen variable, if not already set
		 * as the first variable after saveSubCount, is swapped
		 * with the variable in that place; subsequently, the
		 * saveSubCount index is incremented to include the new
		 * swapped variable in the caller world */
		else if (cs && cs>saveSubCount[XCount] && !vn[cs].isArr){
			int dd=saveSubCount[XCount]+1;
			/* swap vn structure and values */
			if (cs>dd) swapDEC(cs,dd); // avoid cs=dd
			saveSubCount[XCount]++;
			p+=_len;
		}
		/* main DECLARE arrays 
		 * Note: these variables need not any rearranging */
		else if (cs && cs<=saveSubCount[XCount] && vn[cs].isArr) {
			p+=_len;
			if (!strncmp(p,"()",2)) p+=2;
		}
		/* sub-DECLARE arrays 
		 * Note: see sub-DECLARE variables process */
		else if (cs && cs>saveSubCount[XCount] && vn[cs].isArr) {
			int dd=saveSubCount[XCount]+1;
			/* swap vn structure and values */
			if (cs>dd) swapDEC(cs,dd); // avoid cs=dd
			p+=_len;
			saveSubCount[XCount]++;
			/* the () token is optional */
			if (!strncmp(p,"()",2)) p+=2;
		}

		/* check syntax: if there's something else not recognized,
		 * it's a global variable or a not existing name*/
		if (*p && *p!=',' && *p!=';') perror_(__LINE__,37,l);
		else if (*p) p++;
	}
}


/* INPUT()
 * The INPUT/LINPUT statements
 * if mode is 1 perform INPUT, if 0 perform LINPUT/LINEINPUT (get line up to 
 * EOL) */
void INPUT(int mode) {
	char line[SCREENLINE+1], *u=line, *pb=p;
	int done=1, var=0, isArr=FALSE, type=NUM;

	/* argument required */
	if (!*p) perror_(__LINE__,125,l);

	/* detect a possible LINPUT: statement */
	if (*p==':') perror_(__LINE__,92,l);

	/* input welcome string in the form INPUT "prompt";<varlist>
	 * only quoted strings are accepted */
	if (*p=='\"') {
		p++; // skip first "
		while (*p!='\"') fprintf(stdout,"%c", *p++);
		p++; // skip last "
		if (*p==',' || *p==';') p++;
		pb=p; // updating
	}

redo:
	/* get user input and store it into u
	 * the "?" character is printed only in case of true isPrompt*/
	*u='\0'; u=line;
	if (isPrompt) fprintf(stdout,INPUTS);
	strncpy_(line,getString(),COMPSCR);
	if ((int)strlen(line)>MAXSTRLEN) perror_(__LINE__,124,l);
	if (!isEcho) fprintf(stdout,"\n");

	/* repeat for all variables */
	do {
		int cs;
		char op[COMPSTR], *P1=op;
		double val=0.0;
		
		/* restore parameters */
		isArr=FALSE,type=NUM;

		/* there's another input string: recap */
		if (*p=='\"') {
			INPUT(mode);
			return;
		}
		/* or is it an existing variable? */
		else if ((cs=findDEC(p))) {
			var=cs, p+=_len;
			type=vn[cs].rettype;
			if (*p=='(') isArr=TRUE;
			else isArr=FALSE;
		}
		else perror_(__LINE__,42,l); 

		/* get numerical value
		 * Note: in the following case, for numerical variables
		 * and numerical arrays, I use S() and not strtol, so any 
		 * formula may be used with INPUT. Numbers may also exceed 
		 * the 8 digits limit and be in any format */
		if (type==NUM) {
			char *pp=p;
			p=u;
			val=S();
			u=p;
			p=pp;
		}
		/* get string value */
		else {
			P1=op; *P1='\0';
			/* INPUT mode */
			if (mode) {
				/* INPUT: grab string until comma */
				if (*u!='\"') {
					while (*u && *u!=',') *P1++=*u++;
					*P1='\0';
				}
				/* INPUT: grab string until closing quotes */
				else {
					u++;
					while (*u && *u!='\"') *P1++=*u++;
					*P1='\0';
					if (*u=='\"') u++;
				}
			}
			/* LINPUT mode */
			else {
				/* LINPUT: grab string until EOL */
				while (*u && *u!='\n') *P1++=*u++;
				*P1='\0';
			}
		}

		/* numeric variable */
		if (!isArr && type==NUM){
			if (vn[var].isInt) P[var]=(int)val;
			else P[var]=val;
		}
		/* string variable */
		else if (!isArr && type==STR) strncpy_(PS[var],op,COMPSTR);
		/* numeric or string array */
		else if (isArr) {
			int M=0, N=0, T=1;
			if (*p=='(') p++;
			else perror_(__LINE__,37,l);
			M=(int)S();
			if (*p==',') {
				p++;
				N=(int)S();
				T=2;
			}
			p++;
			if (vn[var].isInt) val=(int)val;
			if (type==NUM) setArrayVal(var,T,M,N,val);
			else setArrayStr(var,T,M,N,op);
		}
		else done=0;
		if (*p==',') p++;
		if (!*p) break;

		while (*u && *u!=',') u++;
		if (*u==',') u++;
		while (isblank(*u)) u++;
		/* if user input is over, but there are still variables
		 * to be filled, get another user input */
		if (*p && *p!='\"' && !*u) {
			int i;
			u=line;
			for(i=0;i<SCREENLINE;i++) line[i]='\0';
			if (isPrompt) fprintf(stdout,INPUTS);
			strncpy_(line,getString(),COMPSCR);
		}
	} while (*p && done);

	/* redo in case of any mistaken input, resetting parameters */
	if (!done) {
		perror_(__LINE__,91,l);
		/* reset values for repeating */
		p=pb;
		u=line;
		done=1;
		goto redo;
	}

	/* reset printing position */
	channel[CONSOLE].fprintn=channel[CONSOLE].fcounter=0;
}


/* RANDOMIZE()
 * perform the RANDOMIZE statement.
 * The algorithm (depending from current time) is the following:
 *   - divide current 'usec' measure with the sum of the value obtained summing 
 *     months (12), days (31), hours (23), minutes (59) and seconds (59), which 
 *     always gives 184;
 *   - divide this value (always in the range [1,184]) with the logarithm of 
 *     276.00 which is 184.00x1.5, obtaining a value between 0.177923252 and 
 *     32.73787838, a range with medium value 16.45790082, the starting default
 *     value of the getRandom() function 
 * Used in conjunction with the getRandom() function. */
void RANDOMIZE(void) {
	int b;
	gettimeofday(&tv, NULL);
       	b = tv.tv_usec%184;
	rand2Seed=b/LOG276;
}


/* ONGOTO()
 * The ON..GOTO/ON..THEN/ON..GOSUB statements */
void ONGOTO(void) {
	int addr[ONLIM], ac=0;
	int val=0, isOtherwise=FALSE, tempad=0;

	/* argument required */
	if (!*p) perror_(__LINE__,125,l);

	if (!check(p,"GOTO")) {
		if (!check(p,"GOSUB")) {
			if (check(p,"THEN"));
			else perror_(__LINE__,35,l);
		}
	}

	for (val=0; val<ONLIM; val++) addr[val]=0;
	
	/* take the integer control value */
	val=S();
	/* check for a valid value */
	if (isDefRun) {
		if (val<=fn[isDefVar].start || val>fn[isDefVar].end+1)
			perror_(__LINE__,188,l,isDefVar,val);
	}
	else if (isSubRun) {
		if (val<=sn[isSubVar].start || val>sn[isSubVar].end)
			perror_(__LINE__,189,l,\
				sn[isSubVar].name!=NIL?sn[isSubVar].name:"-",\
				val);
	}

	/* if formula is followed by a comma, skip it */
	if (*p==',') p++;

	/* perform ON GOTO */
	if (!strncasecmp_(p," OTO",4)||(!strncasecmp_(p," HEN",4))){
		int v=0;
		p+=4;
		/* at least an address must be given */
		if (!*p) perror_(__LINE__,35,l);
		addr[ac++]=getLabelAddr(p,&p); // first address
		while (*p==',') {
			p++;
			if (!strncasecmp_(p,"OTHERWISE",9)) {
				p+=9,isOtherwise=TRUE;
				addr[ac++]=getLabelAddr(p,&p);
				break;
			}
			addr[ac++]=getLabelAddr(p,&p);
		}
		/* go to OTHERWISE */
		if (val>ac || val<1) {
			/* ON..GOTO overflow causes a halt if no OTHERWISE
			 * was specified */
			if (!isOtherwise) perror_(__LINE__,109,l) ;
			else {
				v=addr[ac-1];
				if (norow(v)) perror_(__LINE__,61,l);
				l=v-1;
			}
		}
		/* goto to nth address */
		else {
			v=addr[val-1];
			if (norow(v)) perror_(__LINE__,61,l);
			l=v-1;
		}
	}
	/* perform ON GOSUB */
	else if (!strncasecmp_(p," OSUB",5)) {
		p+=5;
		/* at least an address must be given */
		if (!*p) perror_(__LINE__,35,l);
		addr[ac++]=getLabelAddr(p,&p);
		while (*p++==',') {
			if (!strncasecmp_(p,"OTHERWISE",9)) 
				p+=9,isOtherwise=TRUE;
			addr[ac++]=getLabelAddr(p,&p);
		}

		/* ON..GOSUB overflow  */
		if (val>ac || val<1) {
			if (!isOtherwise) perror_(__LINE__,109,l) ;
			else tempad=addr[ac-1];
		}
		else tempad=addr[val-1];
		gosubStack[++gosubTOS]=l;
		l=tempad-1;
	}
	else perror_(__LINE__,35,l);
}


/* IFTHEN()
 * The IF..THEN/IF..GOTO statements, and treat the IF-THEN part of the
 * IF-THEN/ELSE/ELSEIF/ENDIF clause
 * Note: THEN and GOTO are perfectly interchangeable */
void IFTHEN(void) {
	int flag=0;

	/* IF is not followed by THEN/GOTO */
	if (!check(p,"THEN")) {
		if (check(p,"GOTO"));
		else perror_(__LINE__,35,l);
	}
	
	/* get IF..THEN condition status from a string relation or 
	 * from a numerical expression */
	flag=(int)S(); // formula-by-value enabled
	
	/* if there's a comma, skip it */
	if (*p==',') p++;

	/* perform THEN which can execute addresses and statements */
	if (!strncasecmp_(p," HEN",4)) {
		p+=4;
		/* if IF..THEN is followed by a number, jump to address */
		if (isdigit(*p)) { 
			if (flag) GOTO(); 
		}
		/* treat multiline instructions (THEN/ELSE) */
		else if (!*p) {
			if (flag) ifStack[++ifTOS]=CASEEND;
			else {
				ifStack[++ifTOS]=CASEVALUE;
				l=ifAddr[l]-1; // jump to next ELSE/END
			}
		}
		/* treat single instruction or error */
		else if (flag) {
			strncpy_(parseSpace,p,COMPSTR);
			parseLine(parseSpace,l);
		}
	}
	/* perform GOTO, which can only jump to addresses */
	else if (!strncasecmp_(p," OTO",4)) {
		p+=4;
		if (flag) GOTO();
	}
	/* otherwise it's an incorrect IF */
	else perror_(__LINE__,36,l);
}


/* ELSEIF() 
 * The ELSE IF execution 
 * if in the stack there's an ENDTOKEN, relative statement have been already 
 * executed, so jump to END IF;
 * if otherwise in the stack there's an CONTINUE, call IFTHEN to proceed */ 
void ELSEIF(void) {
	if (!ifTOS) perror_(__LINE__,206,l);
	/* jump to END IF in case of true condition (THEN clause) */
	if (ifStack[ifTOS]==CASEEND) l=ifAddr[l]-1;
	else {
		ifTOS--;
		IFTHEN();
	}
}


/* ELSE()
 * The ELSE execution 
 * if in the stack there's an ENDTOKEN, relative statement have been already 
 * executed, so jump to END IF;
 * if otherwise in the stack there's a CONTINUE, do nothing */
void ELSE(void) {
	if (!ifTOS) perror_(__LINE__,206,l);
	/* jump to END IF in case of true condition (THEN clause) */
	if (ifStack[ifTOS]==CASEEND) l=ifAddr[l]-1;
}


/* ENDIF()
 * the ENDIF execution
 * its task is merely to clear the stack position of the current IF-ENDIF 
 * level clause, leaving the underlying levels untouched */
void ENDIF(void) {
	if (!ifTOS) perror_(__LINE__,207,l);
	/* pop (clear) */
	ifTOS--;
}


/* GOTO()
 * The GOTO Statement
 * called also by IF..THEN when followed by and address */
void GOTO(void){
        int num;
        
	/* argument required */
	if (!*p) perror_(__LINE__,125,l);

        /* check if destination is legal (the number may be a formula) 
	 * Note: the control to a self call is not done, so the control is 
	 * delegated to the BASIC programmer */
	num=getLabelAddr(p,&p);
	if (norow(num)) perror_(__LINE__,61,l);
	if (isDefRun) {
		if ((num<fn[isDefVar].start)||(num>(fn[isDefVar].end+1)))
			perror_(__LINE__,39,l);
	}
	else if (fm[num]) perror_(__LINE__,39,l);
	
	/* Note: a jump from inside a SUB to outside a SUB is not illegal,
	 * but the programmer must know what is being done (and eventually
	 * make a jump back to restore pointers) */

	/* jump! */
	l=num-1;
}


/* GOSUB()
 * The GOSUB Statement */
void GOSUB(void){
        int num;

	/* argument required */
	if (!*p) perror_(__LINE__,125,l);

	/* check if GOSUB is within a DEF FN */
	if (isDefRun || isSubRun) perror_(__LINE__,30,l);

	/* check if destination is legal
	 * The number may be a formula */
	num=getLabelAddr(p,&p);
	if (norow(num)) perror_(__LINE__,61,l);
	if (fm[num]) perror_(__LINE__,39,l);

	/* check if GOSUB calls itself */
	if (isARing(num)) perror_(__LINE__,126,l);

	/* perform GOSUB and check GOSUB stack depth */
        gosubStack[++gosubTOS]=l;
	if (gosubTOS >= LIMIT) perror_(__LINE__,196,l); 
	l=num-1;
}


/* RETURN()
 * The RETURN Statement */ 
void RETURN(void){
	/* RETURN without GOSUB? */
        if (gosubTOS<=0) perror_(__LINE__,118,l);
	/* check if RETURN is within a DEF FN */
	if (isDefRun || isSubRun) perror_(__LINE__,54,l);
        /* restore pointers */
        l=gosubStack[gosubTOS--];
}


/* WHILE()
 * The WHILE Statement */
void WHILE(void) {
        int nextstep=0, cond;

	/* argument required */
	if (!*p) perror_(__LINE__,125,l);

	/* make checks and calculate condition value */
	if (whilecount >= FORLIM) perror_(__LINE__,53,l);
	if (isanystring(p)) cond=strRel();
	else cond=(int)S();
	if (*p) perror_(__LINE__,35,l);
	
	/* the search for closing END WHILE matching the variable 
	 * was done in preParse() and stored into whileAddr[] */
	nextstep=whileAddr[l];

	/* If 'nextstep' is still null, no closing END WHILE was found,
	 * so end program emitting a warning */
	if (!nextstep) perror_(__LINE__,28,l);

        /* update stack values */
	whileStack[++whileTOS]=nextstep;    // TOS-1 = ENDWHILE line
	whileStack[++whileTOS]=l;           // TOS   = WHILE instruction
	whilecount++;

	/* set DEFFN counter */
	if (isDefRun) isDefWhile++;
	
	/* check if WHILE must be performed (only if condition is true) */
	if (!cond) {
		whileTOS-=2;
		whilecount--;
		/* jump beyond END WHILE */
		l=nextstep;
	}
}


/* EXIT()
 * The EXIT statement 
 * Select the following keywords (DO/FOR/WHILE) and unload relative
 * decision structures and cycles 
 * if mode = 0 perform EXIT, if = 1 perform CONTINUE */
void EXIT(int mode) {
	int i=0, rl=0; 	// return line for end address

	/* argument required */
	if (!*p) perror_(__LINE__,125,l);

	/* first pass: find return address line rl
	 * Note: return line is the physical last line of the specified
	 * object of statement
	 * EXIT: l will be put equal to rl, and parse() will increment then 
	 * l to rl+1 and thus jumping beyond the last specified cycle
	 * CONTINUE: l will be put equal to rl-mode, and parse() will increment
	 * l to rl+1-mode = rl+1-1 = rl itself, thus continuing the cycle */
	i=l+1;
	while (*p) {
		if (!strncasecmp_(p,"FOR",3)) {
			p+=3;
			while (!forAddr[i] && i<=endCore) i++;
			if (i>endCore) perror_(__LINE__,199,l);
			else rl=i;
		}
		else if (!strncasecmp_(p,"WHILE",5)) {
			p+=5;
			while (!whileAddr[i] && i<=endCore) i++;
			if (i>endCore) perror_(__LINE__,198,l);
			else rl=i;
		}
		else if (!strncasecmp_(p,"DO",2)) {
			p+=2;
			while (!doAddr[i] && i<=endCore) i++;
			if (i>endCore) perror_(__LINE__,197,l);
			else rl=i;
		}
		else perror_(__LINE__,202,l);
		if (*p==',' || *p==';') p++;
		i++;
	}

	/* check return address */
	if (rl<1 || rl>endCore) perror_(__LINE__,203,l);

	/* second pass: scan source from l=current line to rl=return line to 
	 * do a real unload of all IF, SELECT, WHILE, DO, FOR in the range */
	for (i=l+1;i<=rl-mode;i++) {
		if (ifAddr[i] && cAddr[i]<l) ifTOS--;
		else if (selectAddr[i] && cAddr[i]<l) 
			selectTOS-=3, selectcount--;
		else if (whileAddr[i] && cAddr[i]<l) 
			whileTOS-=2, whilecount--;
		else if (doAddr[i] && cAddr[i]<l) doTOS-=2, docount--;
		else if (forAddr[i] && cAddr[i]<l) forTOS-=6, forcount--;
	}

	/* set program count to prosecution line */
	l=rl-mode;
}


/* BREAK()
 * The BREAK statement 
 * Exits from the most external loop and unload relative
 * decision structures and cycles */
void BREAK(void) {
	int i=0, rl=0; 	// return line for end address

	/* first pass: find return address line rl
	 * Note: return line is the physical last line of the specified
	 * object of statement */
	i=l+1;
	while (i<=endCore) {
		if (cAddr[i]) {
			if (forAddr[i] || whileAddr[i] || doAddr[i]) {
				rl=i;
				break;
			}
		}
		i++;
	}

	/* break used out of any cycle */
	if (i>endCore) perror_(__LINE__,204,l);

	/* check return address */
	if (rl<1 || rl>endCore) perror_(__LINE__,204,l);

	/* second pass: scan source from l=current line to rl=return line to 
	 * do a real unload of all IF, SELECT, WHILE, DO, FOR in the range */
	for (i=l+1;i<=rl;i++) {
		if (ifAddr[i] && cAddr[i]<l) ifTOS--;
		else if (selectAddr[i] && cAddr[i]<l) 
			selectTOS-=3, selectcount--;
		else if (whileAddr[i] && cAddr[i]<l) 
			whileTOS-=2, whilecount--;
		else if (doAddr[i] && cAddr[i]<l) doTOS-=2, docount--;
		else if (forAddr[i] && cAddr[i]<l) forTOS-=6, forcount--;
	}

	/* set program count to prosecution line */
	l=rl;
}


/* ENDWHILE()
 * The END WHILE statement
 * Its only scope is to jump back to WHILE, because if condition is false
 * it's WHILE() that jumps over END WHILE, putting the cycle to an end*/
void ENDWHILE(void) {
	if (!whilecount) perror_(__LINE__,209,l);
	/* jump unconditionally back to WHILE */
	l=whileStack[whileTOS]-1;
	/* unload WHILE data, which will be reloaded by the call to WHILE */
	whileTOS -= 2;
	whilecount--;
}


/* DOWHILE()
 * The DOWHILE Statement */
void DOWHILE(void) {
        int nextstep=0, cond;

	/* argument required */
	if (!*p) perror_(__LINE__,125,l);

	/* make checks and calculate condition value */
	if (docount >= FORLIM) perror_(__LINE__,53,l);
	if (isanystring(p)) cond=strRel();
	else cond=(int)S();
	if (*p) perror_(__LINE__,35,l);
	
	/* the search for closing LOOP 
	 * was done in preParse() and stored into doAddr[] */
	nextstep=doAddr[l];

	/* If 'nextstep' is still null, no closing END WHILE was found,
	 * so end program emitting a warning */
	if (!nextstep) perror_(__LINE__,28,l);

        /* update stack values */
	doStack[++doTOS]=nextstep;    // TOS-1 = ENDWHILE line
	doStack[++doTOS]=l;           // TOS   = WHILE instruction
	docount++;

	/* set DEFFN counter */
	if (isDefRun) isDefDo++;
	
	/* check if WHILE must be performed (only if condition is true) */
	if (!cond) {
		doTOS-=2;
		docount--;
		/* jump beyond LOOP */
		l=nextstep;
	}
}


/* DOUNTIL()
 * The DOUNTIL Statement */
void DOUNTIL(void) {
        int nextstep=0, cond;

	/* argument required */
	if (!*p) perror_(__LINE__,125,l);

	/* make checks and calculate condition value */
	if (docount >= FORLIM) perror_(__LINE__,53,l);
	if (isanystring(p)) cond=strRel();
	else cond=(int)S();
	if (*p) perror_(__LINE__,35,l);
	
	/* the search for closing LOOP 
	 * was done in preParse() and stored into doAddr[] */
	nextstep=doAddr[l];

	/* If 'nextstep' is still null, no closing END WHILE was found,
	 * so end program emitting a warning */
	if (!nextstep) perror_(__LINE__,28,l);

        /* update stack values */
	doStack[++doTOS]=nextstep;    // TOS-1 = ENDWHILE line
	doStack[++doTOS]=l;           // TOS   = WHILE instruction
	docount++;

	/* set DEFFN counter */
	if (isDefRun) isDefDo++;
	
	/* check if WHILE must be performed (only if condition is false) */
	if (cond) {
		doTOS-=2;
		docount--;
		/* jump beyond LOOP */
		l=nextstep;
	}
}


/* LOOP()
 * The LOOP statement for the DO WHILE|UNTIL .. LOOP
 * Its scope is merely to jump back to DO WHILE, because if condition is true
 * it's DOWHILE() that jumps over LOOP, putting the cycle to an end*/
void LOOP(void) {
	if (!docount) perror_(__LINE__,219,l);
	/* jump unconditionally back to WHILE */
	l=doStack[doTOS]-1;
	/* unload WHILE data, which will be reloaded by the call to WHILE */
	doTOS -= 2;
	docount--;
}


/* LOOPWHILE()
 * The LOOPWHILE Statement for the DO..LOOP WHILE */
void LOOPWHILE(void) {
        int prevstep=0, cond;

	/* argument required */
	if (!*p) perror_(__LINE__,125,l);

	/* make checks and calculate condition value */
	if (docount >= FORLIM) perror_(__LINE__,53,l);
	if (isanystring(p)) cond=strRel();
	else cond=(int)S();
	if (*p) perror_(__LINE__,35,l);
	
	/* the search for opening DO 
	 * was done in preParse() and stored into doAddr[] */
	prevstep=doAddr[l];

	/* If 'prevtep' is still null, no closing END WHILE was found,
	 * so end program emitting a warning */
	if (!prevstep) perror_(__LINE__,28,l);

        /* update stack values */
	docount++;

	/* set DEFFN counter */
	if (isDefRun) isDefDo++;

	/* unload stack data */
	doTOS -= 2;
	docount--;
	
	/* check if LOOP must be performed (only if condition is true) */
	if (cond) l=prevstep;
}


/* LOOPUNTIL()
 * The LOOPUNTIL Statement for the DO..LOOP UNTIL */
void LOOPUNTIL(void) {
        int prevstep=0, cond;

	/* argument required */
	if (!*p) perror_(__LINE__,125,l);

	/* make checks and calculate condition value */
	if (docount >= FORLIM) perror_(__LINE__,53,l);
	if (isanystring(p)) cond=strRel();
	else cond=(int)S();
	if (*p) perror_(__LINE__,35,l);
	
	/* the search for opening DO 
	 * was done in preParse() and stored into doAddr[] */
	prevstep=doAddr[l];

	/* If 'prevtep' is still null, no closing END WHILE was found,
	 * so end program emitting a warning */
	if (!prevstep) perror_(__LINE__,28,l);

        /* update stack values */
	docount++;

	/* set DEFFN counter */
	if (isDefRun) isDefDo++;
	
	/* unload stack data */
	doTOS -= 2;
	docount--;
	
	/* check if LOOP must be performed (only if condition is false) */
	if (!cond) l=prevstep;
}


/* SELECT()
 * the SELECT statement */
void SELECT(void) {
	int type=0;
	char sf[COMPSTR], sres[COMPSTR], *f=sf, *pb;
	double nres;

	/* update count */
	selectcount++;

	/* case of string formula */
	if (isanystring(p)) {
		int len=0;
		type=STR;
		while (*p) *f++=*p++;
		*f='\0';
		pb=p;
		p=sf;
		strncpy_(sres,SS(),COMPSTR);
		p=pb;
		if (strformula[selectcount]!=NIL) {
			free(strformula[selectcount]);
			strformula[selectcount]=NIL;
		}
		len=((int)strlen(sres)+1)*sizeof(char);
		strformula[selectcount]=xmalloc((size_t)len);
		if (strformula[selectcount]==NIL) perror_(__LINE__,175,l);
		strncpy_(strformula[selectcount],sres,len);
	}
	/* case of numeric formula */
	else if (*p) {
		type=NUM;
		nres=S();
		numformula[selectcount]=nres;
	}
	/* case of TRUE clauses (mute select) */
	else isMuteSelect[selectcount]=TRUE;

	/* set up select stack */
	selectStack[++selectTOS]=selectcount;   // TOS -2 count
	selectStack[++selectTOS]=type;		// TOS -1 type
	selectStack[++selectTOS]=CASEVALUE;	// TOS    operation
}


/* CASE() 
 * the CASE statement of the SELECT-END SELECT decision structure */
void CASE(void) {
	double nval;
	char sval[COMPSTR];
	int count,type,op, flag;

	/* check is CASE is possible */
	if (selectTOS<3) perror_(__LINE__,224,l);

	count=selectStack[selectTOS-2];
	type=selectStack[selectTOS-1];
	op=selectStack[selectTOS];

	/* end operation, because another previous CASE has matched */
	if (op==CASEEND) {
		l=selectAddr[l]-1;
		return;
	}

	/* if = is used, skip it */
	if (*p=='=') p++;

	/* calculate flag for mute select */
	if (isMuteSelect[count]) {
		if (isanystring(p)) flag=strRel();
		else flag=S();
	}
	/* calculate range numeric value */
	else if (type==NUM && check(p,"TO")) {
		double lowrange,highrange;
		lowrange=S();
		if (!strncasecmp_(p," O",2)) p+=2;
		highrange=S();
		if (numformula[count]>=lowrange &&\
				numformula[count]<=highrange) 
			flag=TRUE;
		else flag=FALSE;
	}
	/* calculate range numeric value */
	else if (type==STR && check(p,"TO")) {
		char lowrange[COMPSTR],highrange[COMPSTR];
		strncpy_(lowrange,SS(),COMPSTR);
		if (!strncasecmp_(p," O",2)) p+=2;
		strncpy_(highrange,SS(),COMPSTR);
		if (strcmp(lowrange,strformula[count])<=0 &&\
				strcmp(strformula[count],highrange)<=0) 
			flag=TRUE;
		else flag=FALSE;
	}
	/* calculate string value */
	else if (type==STR) {
		strncpy_(sval,SS(),COMPSTR);
		if (!strcmp(sval,strformula[count])) flag=TRUE;
		else flag=FALSE;
	}
	/* calculate numeric value */
	else if (type==NUM) {
		nval=S();
		if (nval==numformula[count]) flag=TRUE;
		else flag=FALSE;
	}

	/* values do match! Let program flow */
	if (flag) selectStack[selectTOS]=CASEEND;
	/* values don't match: look for next CASE/ENDSELECT*/
	else {
		int end=selectAddr[l], lb=l+1;
		char *ll;
		while (lb<=end) {
			ll=m[lb];
			if (!strncasecmp_(ll,"CASE",4)) break;
			else if (checkCommand(ll,"END SELECT",&ll)) break;
			lb++;
		}
		/* jump to next position */
		l=lb-1;
	}
}


/* CASEDEFAULT() 
 * the CASE DEFAULT statement of the SELECT-END SELECT decision structure */
void CASEDEFAULT(void) {
	/* check if CASE DEFAULT is possible */
	if (selectTOS<3) perror_(__LINE__,224,l);

	/* end operation, because another previous CASE has matched */
	if (selectStack[selectTOS]==CASEEND) l=selectAddr[l]-1;
	
	/* Ensure that this is the last CASE DEFAULT used */
	selectStack[selectTOS]=CASEEND;
}


/* ENDSELECT()
 * the END SELECT statement of the SELECT-END SELECT decision structure */
void ENDSELECT(void) {
	if (selectTOS<3) perror_(__LINE__,222,l);
	isMuteSelect[selectStack[selectTOS-2]]=FALSE;
	selectTOS-=3;
	selectcount--;
}


/* FOR()
 * The FOR..TO..STEP/FOR..TO..BY Statements.
 * STEP and BY are interchangeable */
void FOR(void){
        int var=0, nextstep=0, minus=1, cs=0;
        double start=0.0, end=0.0, step=0.0, varvalue=0.0;

	/* argument required */
	if (!*p) perror_(__LINE__,125,l);

	/* make checks and grab variable name */
	if (forcount >= FORLIM) perror_(__LINE__,53,l);
	if (!isalpha(*p)) perror_(__LINE__,65,l,makechar(p),"a letter or underscore");
	else if ((cs=findDEC(p)) && !vn[cs].isArr && vn[cs].rettype==NUM) {
		var=cs, p+=_len;
	}
	else perror_(__LINE__,37,l);
        if (*p!='=') perror_(__LINE__,65,l,makechar(p),"=");
	else p++;
	
	/* save values of FOR variable index */
	varvalue = P[var];
	if (*p=='-') minus=-1, p++;
	if (check(p,"TO")) start=minus*S();
	else perror_(__LINE__,35,l);
	if (*p != ' ' && *(p+1) !='O') perror_(__LINE__,35,l);
	
	/* case of a user step using BY */
	if (check(p,"BY")) {
		p+=2, minus=1;
		if (*p=='-') minus=-1, p++;
		end=minus*S();
		p+=2;
		if (!*p) perror_(__LINE__,37,l);
		step=S();
	}	
	/* case of a user step using STEP */
	else if (check(p,"STEP")) {
		p+=2, minus=1;
		if (*p=='-') minus=-1, p++;
		end=minus*S();
		p+=4;
		if (!*p) perror_(__LINE__,37,l);
		step=S();
	}	
	/* case of an implicit step = 1 */
	else {
		p+=2, minus=1;
		if (*p=='-') minus=-1, p++;
		end=minus*S();
		step=1.0;
	}

	/* set starting value for index variable */
	P[var]=start;

	/* the search for closing NEXT matching the variable 
	 * was done in preParse() and stored into forAddr[] */
	nextstep=forAddr[l];

	/* If 'nextstep' is still null, no closing NEXT was found,
	 * so end program emitting a warning */
	if (!nextstep) {
		perror_(__LINE__,29,l);
	}
	
	/* if step is meaningless and thus not executed, go beyond FOR..NEXT, 
	 * restoring index variable value, and return */
	if ((end-start>0 && step<0) || (end-start<0 && step>0) || step==0) {
		l=nextstep-1;
		P[var]=varvalue;
		return;
	}

        /* update stack values */
	forStack[++forTOS]=nextstep;	// TOS-5 = NEXT program line
	forStack[++forTOS]=l;           // TOS-4 = FOR instruction
        forStack[++forTOS]=var;		// TOS-3 = var index
        forStack[++forTOS]=end;         // TOS-2 = end value
        forStack[++forTOS]=step;        // TOS-1 = step value
        forStack[++forTOS]=start;       // TOS   = start/current value
	forcount++;

	/* set DEFFN counter */
	if (isDefRun) isDefFor++;
}


/* NEXT()
 * The NEXT... Statement
 * Note: see the note in FOR().
 * Note: this version handles multiple addressing, like NEXT J,I, as well as
 * a NEXT alone; in this case, the last variable in the FOR cycle is used as 
 * the current NEXT variable. */
void NEXT(void) {
        int var=0, cs=0;
        double step=0.0, end=0.0;

nextvar:

	/*check if a NEXT is possible */
	if (forTOS<6) perror_(__LINE__,47,l);

	/* get index of first variable in line */
        if ((cs=findDEC(p)) && !vn[cs].isArr && vn[cs].rettype==NUM) {
		var=cs, p+=_len;
	}
	else if (!*p) var=forStack[forTOS-3];
	else perror_(__LINE__,35,l);

	/* update cycle values */
	step = forStack[forTOS-1];
	end  = forStack[forTOS-2];

	/* check index; if different from what stored 
	 * with the previous FOR instruction => NEXT error */
	if (var != (int)forStack[forTOS-3]) perror_(__LINE__,47,l);

        /* Note: if beyond the limit, put cycle to an end; after that, if there 
	 * are other indices in the list continue with the most outer cycle;
	 * Note: the variable to check for is the variable itself (this is 
	 * actually meaningless for normal cycles, but when the index is 
	 * altered inside the cycle this is important).
	 * Note: tbas cannot be absolutely precise with limits and indices
	 * with fractional parts because they may not be correctly stored
	 * in the C double format; it runs (yes!) into a more modern computer, 
	 * but with its limitations too! So do expect that some mislead cycle
	 * may happen. */
        if (((step > 0) && (P[var]+step > end)) || \
		((step < 0) && (P[var]+step < end))) {
                        if (forTOS>5) forTOS -= 6;
			forcount--;
			if (isDefRun) isDefFor--;
			/* restart in case of multiple NEXT */
			if (*p==',') {
				p++;
				goto nextvar;
			}
	}
	/* else reiterate; update counter variable (current +/- step*/
        else {
        	forStack[forTOS] += step; 
	        P[var]+=step;
	        l=(int)forStack[forTOS-4];
	}
	/* reset flag for the FOR..NEXT cycle inside a DEF FN 
	 * in case no cycle is left - this should be never needed, but... */
	if (!forcount && isDefFor) isDefFor=FALSE;
}
	

/* READ()
 * The READ statement 
 * this function checks the next available item in the READ line, and stores in
 * it the first available correspondent item (number or string)
 * Note: READ and DATA work on two dedicated arrays, one for numbers, one for 
 * strings, whose activation flags are dataLimit and dataLimitS and whose 
 * pointers are dc and dcs; these arrays are built in the DATA section of 
 * preParse() (there is no DATA command, actually) */
void READ(void) {
	int var=0, isArr=FALSE, type=NUM;
	
	/* there must be at least one DATA to read! */
	if (!dataLimit && !dataLimitS) perror_(__LINE__,50,l);
	
	/* argument required */
	if (!*p) perror_(__LINE__,125,l);

	/* start READ */
	do {
		int cs;
		char op[COMPSTR];
		double val=0.0;
		isArr=FALSE,type=NUM;

		/* is it an existing variable? */
		if ((cs=findDEC(p))) {
			var=cs, p+=_len;
			type=vn[cs].rettype;
			if (*p=='(') isArr=TRUE;
			else isArr=FALSE;
		}
		else perror_(__LINE__,42,l);

		/* check that there's still something to read */
		if (type==STR) {
			if (dcs>dataLimitS) perror_(__LINE__,110,l);
		}
		else if (type==NUM) {
			if (dc>dataLimit) perror_(__LINE__,110,l);
		}

		/* get numerical value */
		if (type==NUM) val=D[dc++];
		/* get string value */
		else strncpy_(op,DS[dcs++],COMPSTR);

		/* numeric variable */
		if (!isArr && type==NUM){
			if (vn[var].isInt) P[var]=(int)val;
			else P[var]=val;
		}
		/* string variable */
		else if (!isArr && type==STR) strncpy_(PS[var],op,COMPSTR);
		/* numeric or string array */
		else if (isArr) {
			int M=0, N=0, T=1;
			if (*p=='(') p++;
			else perror_(__LINE__,37,l);
			M=(int)S();
			if (*p==',') {
				p++;
				N=(int)S();
				T=2;
			}
			p++;
			if (vn[var].isInt) val=(int)val;
			if (type==NUM) setArrayVal(var,T,M,N,val);
			else setArrayStr(var,T,M,N,op);
		}
		if (*p==',') p++;
	} while (*p);
}


/* RESTORE()
 * the RESTORE statement. 
 * Note: a RESTORE followed by a number restore the pointers to the first 
 * label value not greater than this number 
 * If followed by * restores only numbers; if by $ only strings 
 * Note: if the specific * or $ symbol is used, the line number must exist, 
 * because we are requiring a precise entity; but if the simple RESTORE N is 
 * used, no error is printed because the string and number DATA are not 
 * necessarily lined. */
void RESTORE(void){
	int ok=TRUE;
	/* restore to a specific line label */
	if (isdigit(*p)) {
		int val=strtol(p,&p,10);
		/* restore number index; for both blocks the algorithm is:
		 * ~- search for first item in line with label 'val' 
		 * ~- if you find it, let the pointer there, but decrease it
		 *    until you find the first item in the line
		 * ~- otherwise raise an error */
		if (*p=='*') {
			p++;
			if (dataLimit) {
				dc=dataLimit;
				while (dc>=1) {
					if(ln[DN[dc]]>0 && ln[DN[dc]]<=val) 
						break;
					dc--;
				}
				while (DN[dc]==DN[dc-1]) dc--;
				if (dc<=0 || ln[DN[dc]]!=val) 
					perror_(__LINE__,114,l,"numeric",val);
			}
		}
		/* restore string index */
		else if (*p=='$') {
			p++;
			if (dataLimitS) {
				dcs=dataLimitS;
				while (dcs>=1) {
					if(ln[DSN[dcs]]>0 && ln[DSN[dcs]]<=val) 
						break;
					dcs--;
				}
				while (DSN[dcs]==DSN[dcs-1]) dcs--;
				if (dcs<=0 || ln[DSN[dcs]]!=val) 
					perror_(__LINE__,114,l,"string",val);
			}
		}
		/* restore both number and string indices
		 * for both blocks the algorithm is:
		 * ~- search for first item in line with label 'val' 
		 * ~- if you find it, let the pointer there, but decrease it
		 *    until you find the first item in the line
		 * ~- if you don't find it, restore it to the value it had */
		else if (!*p) {
			if (dataLimit) {
				int dcb=dc;
				dc=dataLimit;
				while (dc>=1) {
					if(ln[DN[dc]]>0 && ln[DN[dc]]<=val) 
						break;
					dc--;
				}
				while (DN[dc]==DN[dc-1]) dc--;
				if (dc<=0 || ln[DN[dc]]<val) dc=dcb; 
			}
			if (dataLimitS) {
				int dcsb=dcs;
				dcs=dataLimitS;
				while (dcs>=1) {
					if(ln[DSN[dcs]]>0 && ln[DSN[dcs]]<=val) 
						break;
					dcs--;
				}
				while (DSN[dcs]==DSN[dcs-1]) dcs--;
				if (dcs<=0 || ln[DSN[dcs]]<val) dcs=dcsb;
			}
		}
	}
	/* restore number index */
	else if (*p=='*') {
		if (dataLimit) dc=1; else ok=FALSE;
	}
	/* restore string index */
	else if (*p=='$') {
		if (dataLimitS) dcs=1; else ok=FALSE;
	}
	/* restore both number and string indices*/
	else if (!*p) {
		if (dataLimit+dataLimitS) dc=dcs=1; else ok=FALSE;
	}
	else perror_(__LINE__,35,l);
	/* warning in case something went wrong */	
	if (!ok) perror_(__LINE__,50,l);
}


/* multiLET()
 * wrapper for LET(), in order to catch multiple assignments 
 * if a colon is found, put a NIL there and call LET(); 
 * then call recursively itself on the rest
 * if a colon is not found, call LET() */
int multiLET(void) {
	char *br=NIL; // break

	/* find next colon */
	if ((br=strcasestr_oq(p,":"))) {
		*br='\0';
		if (!LET()) return FALSE;
		if (br) p=br+1;
		multiLET();
	}
	else return LET();
	return TRUE;
}


/* LET()
 * The LET statement (even for implicit LET)
 * address simple assignment (e.g X=3) and multi-variables assignments
 * (e.g. X=Y=T=3) 
 * return TRUE if assignment can be accomplished, FALSE otherwise 
 * Note: a multi-variable assignment proceeds from right to left, that is
 * the first item is updated last; this is important in cases like
 * LET A=B(A)=expression, where B(A) is updated with the 'old' value of A,
 * while in LET B(A)=A=expression, B(A) is updated with the 'new' value of A. 
 * Note: the effective addressing routine is delegated to assign() 
 * Note: this algorithm doesn't consider the equal signs into strings (in 
 * which case they are 'characters') or into parenthesis (in which case they 
 * are logical operators). */
int LET(void) {
	char *sep, *sep2, *lp;
	int v=0;
	
	/* LET is implicit */
	if (!strncasecmp_(p,"LET",3)) p+=3;
	
	/* argument required */
	if (!*p) perror_(__LINE__,125,l);


	/* if assignment begins with LEFT$( or with RIGHT$( or MID$(
	 * let assign decide; it may be a left assignment */
	lp=p;
	if (!strncasecmp_(p,"LEFT$(",6)) return assign();
	else if (!strncasecmp_(p,"RIGHT$(",7)) return assign();
	else if (!strncasecmp_(p,"MID$(",5)) return assign();

	/* short circuit for statements which are surely not assignments */
	if (!strchr(p,'=')) return FALSE;

	/* search last assignment (signaled by '=') out of strings and 
	 * out of parenthesis (in which case '=' is a logic operator) */
	v=0;
	while (TRUE) {
		if ((int)strlen(p)>0) sep=p+(int)strlen(p)-1;
		else sep=p;
		do {
			/* skip explicit strings */
			if (*sep=='\"') {
				sep--; // skip opening quotes
				while (sep>lp && *sep!='\"') sep--;
				sep--; // skip closing quotes
			}
			/* skip calculations in parenthesis
			 * (since we start from the end, we first search
			 * for ')' and finally for '(' ) */
			else if (*sep==')') v++, sep--;
			else if (*sep=='(' && v) v--, sep--;
			else if (*sep=='=' && v) sep--;
			else if (*sep=='=' && !v) break; // found!
			else if (sep>lp && *sep!='=') sep--;
		} while (sep>lp);

		/* maybe no assignment was found */
		if (*sep!='=') return FALSE;

		sep2=sep-1;
		while (sep2>lp && *sep2!='=') sep2--;

		/* proceed with assignment:
		 * it's a multi-variables assignment... */
		if (*sep2=='=') {
			p=sep2+1;
			if (!assign()) return FALSE;
			*sep='\0';
			continue;
		}
		/* ...or else it's a single assignment */
		else {
			p=lp;
			return assign();
		}
	}
	return FALSE;
}


/****************************************************************************
 * The following routines define the DEFFN and the SUB functions, used in
 * preParse to store data respectively of the DEF FN functions (single and
 * multi lines) and the SUB functions/procedure
 ****************************************************************************/

/* showDEF()
 * DEBUG FUNCTION - UNDOCUMENTED */
void showDEF(int i) {
	int j, y=2;

	printf(INVCOLOR);	
	printf("DEF#%d:\tName = FN%c\n",i,i);
	printf("\ttype: %s\n",fn[i].type==SINGLE?"single line":"multiline");
	printf("\tVars: #%d",fn[i].vars[0]);
	for (j=0;j<fn[i].vars[0];j++) {
		printf("\n\t   ");
		while (fn[i].vars[y]>0)
		       printf("%c",fn[i].vars[y++]);
		y+=3;
	}
	printf("\n\tLines:\n");
	if (fn[i].type==SINGLE) printf("\t%d\t|%s|\n",fn[i].start,fn[i].def);
	else for (j=fn[i].start;j<=fn[i].end;j++)
		printf("\t%d\t|%s|\n",j,m[j]);
	printf("End of FN%c\n\n",i);
	printf(RESETCOLOR);	
}


/* DEF()
 * The DEF Statement for numerical functions; handles multiple arguments 
 * Note: this function is called in preParse() before RUN().
 * The label for FN must be in the A..Z range and not subscripted.
 * Note: this function executes a storing only; the actual command (the 
 * function interpretation and execution) is performed into getFunc() for
 * function FNX; in case of an inline DEF FN, the string is stored into an 
 * array called fn[].def; in case of a multi line DEF FN they are stored into 
 * an array of char pointers called fn[].lines[] */
void DEF(int l) {
	int len, currDEF=*(p+2), mode=isSingleDEF(p);
	int varcount;

	/* check if the function name has already been used */
	if (fn[currDEF].type==SINGLE || fn[currDEF].type==MULTI) 
		perror_(__LINE__,31,l);

	/* if there are arguments, they must be saved */
	varcount=0;
	if (!strncasecmp_(p,"FN",2) && isalpha(currDEF) && *(p+3)=='(') {
		int vn=0;

		/* point to the first variable */
		p+=4;
		/* store all argument-variables */
		while (*p && *p!=')') {
			if (isstrvar(p) || isnumvar(p)) {
				fn[currDEF].vars[++varcount]=DECADD;
				while (isnamechar(p))
					fn[currDEF].vars[++varcount]=*p++;
				fn[currDEF].vars[++varcount]=-1; // termin
				fn[currDEF].vars[++varcount]=DECADD+NUM;
				vn++;
			}
			else perror_(__LINE__,35,l);
	
			/* string functions not admitted in DEFFN structure*/
			if (*p=='$') 
				perror_(__LINE__,65,l,makechar(p), TERMINAT);

			/* check syntax */
			if (*p==',') p++;
		}
		if (*p!=')') perror_(__LINE__,35,l);
		else p++;
	
		/* assign variable count */
		fn[currDEF].vars[0]=vn;
	}
	else fn[currDEF].vars[0]=0; 


	/* single line DEF: store in fn[].def the following formula, caring of 
	 * stripblanking it, because all this is done in parse() */
	if (mode==SINGLE) {
		char dline[SCREENLINE+1];
		int qq=0;
		for (qq=0;qq<=SCREENLINE;qq++) dline[qq]='\0';
		if (*p!='=') perror_(__LINE__,35,l);
		len=((int)strlen(++p)+1)*sizeof(char);
		stripBlanks(p,dline);
		/* empty line is an error */
		if (!dline[0]) perror_(__LINE__,35,l);
		if (fn[currDEF].def!=NIL) {
			free(fn[currDEF].def);
			fn[currDEF].def=NIL;
		}
		fn[currDEF].def=xmalloc((size_t)len);
		if (!fn[currDEF].def) perror_(__LINE__,192,l,currDEF);
		strncpy_(fn[currDEF].def,dline,len);
		fn[currDEF].type=mode;
		fn[currDEF].start=l;
	}
	/* multiline DEF: store lines */
	else {
		char *q;
		int i;
		/* find next available line after DEFFN */
		l++; while (!m[l] && l<=DATLIM) l++;
		fn[currDEF].start=l;
		fm[l]=TRUE;
		/* find last line (where FNEND is located) */
		while (l<=endCore) {
			if (m[l]) {
				q=m[l];
				while (isblank(*q)) q++;
				if (!strcasecmp(q,"FNEND")) break;
			}
			l++;
		}
		if (l>endCore) perror_(__LINE__,52,-1,currDEF);
		else fn[currDEF].end=l-1;

		/* set type */
		fn[currDEF].type=mode;

		/* find lines and mask them*/
		for (i=fn[currDEF].start; i<=fn[currDEF].end; i++) {
			if (m[i]) fm[i]=TRUE;
		}
	
		/* update count */
		isDefCount++;
	}
	
	/* DEBUGGING show DEF contents */	
	if (isDebug) showDEF(currDEF);
}


/* showSUB()
 * DEBUG FUNCTION - UNDOCUMENTED */
void showSUB(int i) {
	int j,y=2;

	printf(INVCOLOR);	
	printf("Sub#%d:\tName = %s\n",i,sn[i].name[0]?sn[i].name:"obj.");
	printf("\ttype: %s\n",sn[i].type==FUNC?"FUNC":"PROC");
	if (sn[i].type==FUNC) 
		printf("\treturn type: %s\n",sn[i].rettype==NUM?"number":"string");
	printf("\tVars: #%d",sn[i].vars[0]);
	for (j=0;j<sn[i].vars[0];j++) {
		int t=0;
		printf("\n\t   ");
		while (sn[i].vars[y]>0)
		       printf("%c",sn[i].vars[y++]);
		y++; // skip -1
		t=sn[i].vars[y++]-DECADD;
		if (t==STR) printf(", string");
		else if (t==NUM) printf(", numerical");
		y++;
	}
	printf("\n\tLines:\n");
	for (j=sn[i].start;j<=sn[i].end;j++) {
		printf("\t%d\t|%s|\n",j,m[j]);
	}
	printf("End of SUB %s\n\n",sn[i].name[0]?sn[i].name:"obj.");
	printf(RESETCOLOR);	
}


/* SUB()
 * The SUB Statement for subroutines; handles multiple arguments 
 * Note: this function is called in preParse() before RUN().
 * Note: this function executes a storing only; the actual command (the 
 * function interpretation and execution) is performed by execSUB() called
 * into getFunc() and SS() for function names and in exec() for procedures */
void SUB(int l) {
	int len=l, rettype=NUM, type=PROC;
	char subname[COMPSTR], *sp=subname;
	char *q;
	int i,varcount, namelen=0;

	/* argument required */
	if (!*p) perror_(__LINE__,125,l);

	/* get sub name */
	if (isbeginnamechar(p)) {
		while (isnamechar(p)) *sp++=*p++;
	}
	else perror_(__LINE__,172,l);

	/* determine string return type */
	if (*p=='$') *sp++=*p++,rettype=STR;
	*sp='\0';

	/* setup sub */
	subCount++;
	sn[subCount].number=subCount;
	namelen=((int)strlen(subname)+1)*sizeof(char);
	if (sn[subCount].name!=NIL) {
		free(sn[subCount].name);
		sn[subCount].name=NIL;
	}
	sn[subCount].name=xmalloc((size_t)namelen);
	if (!sn[subCount].name) perror_(__LINE__,195,l);
	else strncpy_(sn[subCount].name,subname,namelen);

	/* if there are arguments, they must be saved */
	varcount=0;
	if (*p=='(') {
		int vn=0;
		/* point to the first variable */
		p++;
		/* store all argument-variables */
		while (*p && *p!=')') {
			if (isstrvar(p) || isnumvar(p)) {
				int ret=NUM;
				sn[subCount].vars[++varcount]=DECADD;
				while (isnamechar(p))
					sn[subCount].vars[++varcount]=*p++;
				if (*p=='%' && *(p+1)!='%') ret=NUMINT,
					sn[subCount].vars[++varcount]=*p++;
				else if (*p=='$') ret=STR,
					sn[subCount].vars[++varcount]=*p++;
				sn[subCount].vars[++varcount]=-1; // end
				sn[subCount].vars[++varcount]=DECADD+ret;
				vn++;
				if (*p=='(') perror_(__LINE__,129,l);
			}
			else perror_(__LINE__,35,l);

			/* check syntax */
			if (*p==',') p++;
		}
		if (*p!=')') perror_(__LINE__,35,l);
		else p++;
	
		/* assign variable count */
		sn[subCount].vars[0]=vn;
	}
	else sn[subCount].vars[0]=0; 

	/* find next available line after SUB */
	l++; while (!m[l] && l<DATLIM) l++;
	sn[subCount].start=l;
	fm[l]=TRUE;
	
	/* find last line (where END SUB is located)
	 * and check if RETURN is used */
	while (l<=endCore) {
		if (m[l]) {
			q=m[l];
			while (isblank(*q)) q++;
			if (strcasestr_oq(q,"RETURN")) type=FUNC;
			if (checkCommand(q,"END FUNCTION",&q)) 
				perror_(__LINE__,208,l);
			if (checkCommand(q,"END SUB",&q)) break;
		}
		l++;
	}
	if (l>endCore) perror_(__LINE__,210,len,subname);
	else sn[subCount].end=l-1;

	/* set type */
	sn[subCount].type=type;
	sn[subCount].rettype=rettype;

	/* find lines and mask them */
	for (i=sn[subCount].start; i<=sn[subCount].end; i++) {
		if (m[i]) fm[i]=TRUE;
	}
	/* DEBUGGING show SUB contents */	
	if (isDebug) showSUB(sn[subCount].number);
}


/****************************************************************************
 * The following routines define the DECLARE/DIM statements:
 * the DIM statement for numerical arrays with one-letter or one-letter-and-a-
 * number names was included since 1964 version II of the Dartmouth BASIC, 
 * and for strings since 1967 version IV
 ****************************************************************************/

/* showDEC()
 * DEBUG FUNCTION - UNDOCUMENTED 
 * flag= TRUE for DECLARE, false for implicit declaration into SUBs */
void showDEC(int decCount, int flag) {
	/* DEBUGGING show SUB contents */
	printf(INVCOLOR);
	if (flag) printf("Declared ");
	else printf("Declared implicit ");
	if (vn[decCount].isConst) printf("const ");
	else printf("var ");
	printf("%s [#%d], type %s",\
			vn[decCount].name!=NIL?vn[decCount].name:"obj.",\
			decCount,\
			vn[decCount].rettype==NUM?"numeric":"string");
	if (vn[decCount].isArr) {
		printf(" array");
		if (vn[decCount].rettype==NUM)
			printf(" (%dx%d)",1+dim[decCount].brow,1+dim[decCount].bcol);
		else
			printf(" (%dx%d)",1+dimS[decCount].brow,1+dimS[decCount].bcol);
	}
	printf("\n");
	printf(RESETCOLOR);
}	


/* DECLARE()
 * The DECLARE Statement for any-name variables 
 * Note: this function is run-time.
 * Note: this function executes a storing only; the actual command (the 
 * function interpretation and execution) is performed by findDEC(), called
 * into getFunc() and SS() for function names and in LET() for assignments, and
 * in all functions dealing with variables. */
void DECLARE(int isConst) {
	char decname[COMPSTR], *sp=decname;
	int var=0, type=NUM, namelen=0, isInt=FALSE;
	
	/* search for AS INTEGER, to stop parsing when due */
	check(p,"ASINTEGER");

	/* get var name */
	if (isbeginnamechar(p)) {
		while (isnamechar(p)) *sp++=*p++;
	}
	else perror_(__LINE__,170,l);

	/* determine string return type */
	if (*p=='%' && *(p+1)!='%') *sp++=*p++, isInt=TRUE;
	else if (*p=='%' && *(p+1)=='%' && *(p+2)=='%') *sp++=*p++, isInt=TRUE;
	else if (*p=='$') *sp++=*p++, type=STR;
	*sp='\0'; decCount++;
	if (decCount>2*LIMIT) perror_(__LINE__,223,l);
	var=decCount;
	vn[decCount].isArr=FALSE;
	vn[decCount].isInt=isInt;
	
	/* array figure without specified dimensions */
	if (!strncmp(p,"()",2) && !isConst) {
		if (type==STR) dimStrArray(var,2,10,10);
		else dimArray(var,2,10,10);
		vn[decCount].isArr=TRUE;
		p+=2;
	}
	/* array figure */
	else if (*p=='(' && !isConst) {
		int M=0,N=0,T=1;
		vn[decCount].isArr=TRUE;
		p++;
		/* get N (and N) */
		M=(int)S();
		if (M<0) perror_(__LINE__,76,l);
		if (*p==',') {
			p++;
			N=(int)S();
			if (N<0) perror_(__LINE__,76,l);
			T=2;
		}
		if (!M && N) T=2;
		else if (M && !N) T=1;
		
		/* check correspondence */
		if (type==STR) {
			/* variable should be still free, anyway... */
			if (dimS[var].type) perror_(__LINE__,63,l);
			if (dimS[var].type==1 && T==2) perror_(__LINE__,64,l);
			dimStrArray(var,T,M,N);
		}
		else {
			/* again, variable should be still free, anyway... */
			if (dim[var].type) perror_(__LINE__,63,l);
			if (dim[var].type==1 && T==2) perror_(__LINE__,64,l);
			dimArray(var,T,M,N);
		}
		/* check syntax */
		if (*p!=')') perror_(__LINE__,35,l);
		else p++;
	}
	/* user preset value for variables */
	else if (*p=='=') {
		p++;
		/* numeric variables */
		if (type==NUM) P[var]=S();
		/* string variables */
		else {
			char op[COMPSTR];
			strncpy_(op,SS(),COMPSTR);
			strncpy_(PS[var],op,COMPSTR);
		}
	}
	/* signal error in case of missing assignment to constants */
	else if (isConst) perror_(__LINE__,191,l);
	/* preset null values for uninstantiated variables */
	else  {
		/* numeric variables */
		if (type==NUM) P[var]=0.0;
		/* string variables */
		else if (type==STR) strncpy_(PS[var],"",COMPSTR);
	}

	/* integer */
	if (!strncasecmp_(p," SINTEGER",9)) {
		p+=9;
		if (type==STR) perror_(__LINE__,171,l);
		vn[decCount].isInt=TRUE;
		if (!vn[decCount].isArr) 
			P[var]=(int)P[var];
	}
	
	/* set constant flag */
	if (isConst) vn[decCount].isConst=TRUE;
	else vn[decCount].isConst=FALSE;
	
	/* set name/type */
	namelen=((int)strlen(decname)+1)*sizeof(char);
	if (vn[decCount].name!=NIL) {
		free(vn[decCount].name);
		vn[decCount].name=NIL;
	}
	vn[decCount].name=xmalloc((size_t)namelen);
	if (!vn[decCount].name) perror_(__LINE__,173,l);
	else strncpy_(vn[decCount].name,decname,namelen);
	vn[decCount].rettype=type;
	vn[decCount].isCommon=FALSE;

	/* debug features */
	if (isDebug) showDEC(decCount, TRUE);

	/* next assignment */
	if (*p==',' || *p==';') {
		p++;
		DECLARE(isConst);
	}
}


/******************************************
 *            BASIC INTERFACE             *
 ******************************************/


/* OLD() 
 * execute the OLD command, erasing memory and calling LOAD
 * Note: this statement is not available as a programming statement */
void OLD(void) {
	int i;

	/* erase core memory */
	for (i=0;i<MEMLIMIT;i++) {
		if (m[i]!=NIL && (int)strlen(m[i])>0) {
			free(m[i]);
			m[i]=NIL;
		}
	}

	/* LOAD */
	LOAD(m,filename,0,ALL_LINES);
}


/* LOAD() 
 * load file in memory, from position 'stt'
 * filestr: name of file to open
 * m: memory where program is stored
 * Note: this statement is is not available as a programming statement.
 * If one among '@', '&' or '#' is set as the first character of a line, 
 * the loading is stopped/suspended; this lets store documentation within the 
 * source file; this feature didn't belong to the DEC BASIC */
void LOAD(char **m, char *filestr, int stt, int what) {
	FILE *f;
	char B[SCREENLINE+10], *pB=B;
	int isSuspended=FALSE;
	int saveSubs=FALSE;
	int isAllBlanks=FALSE, linenum=stt;
	int isVoid=TRUE;

	/* open file */
	if (!(f=fopen(filestr,"r"))) {
		beginS=-1;
		perror_(__LINE__,1,-1,filestr);
	}

	/* load lines one by one */
	while(fgets(B,SCREENLINE+9,f) && linenum<MEMLIMIT-1) {
		linenum++; pB=B;
		while (isblank(*pB)) pB++;
		/* check if it's a void/NIL line (i.e. a line with
		 * no line number and no statements) */
		if (!*pB) continue;
		/* check if file is a normal or random file */
		if (*pB=='') perror_(__LINE__,97,-1);
		/* stop loading when '@' is met */
		if (*pB=='@') break;
		/* suspend loading between two '' */
		if (*pB=='&' && isSuspended) {
			isSuspended=FALSE;
			continue;
		}
		if (*pB=='&' && !isSuspended) {
			isSuspended=TRUE;
			continue;
		}
		if (isSuspended) continue;
		/* strip away '\n' left by fgets() */
		if ((int)strlen(B)>0) {
			if (B[(int)strlen(B)-1]=='\n') B[(int)strlen(B)-1]='\0';
			/* if '\n' was not found line is too long! */
			else perror_(__LINE__,10,linenum);
		}
		/* strip away ASCII 13 left by DOS files */
		if ((int)strlen(B)>0) {
			if (B[(int)strlen(B)-1]==13) B[(int)strlen(B)-1]='\0';
		}
		/* test for only blanks */
		if (isblank(*pB)) {
			char *jB=pB;
			while (isblank(*jB)) jB++;
			if (!jB) isAllBlanks=TRUE;
		}
		/* discard leading blanks */
		if (!isAllBlanks) while (isblank(*pB)) pB++;
		/* discard void lines and lines beginning with # */
		if ((!isAllBlanks && !*pB) || *pB=='#') continue;
		/* store line */
		else if (what==ONLY_SUBS) {
			/* set saving flag when SUB starts */
			if (!strncasecmp_(pB,"SUB",3)) saveSubs=TRUE;
			/* set saving flag with DEF */
			else if (!strncasecmp_(pB,"DEF",3)) {
				char *pC=pB+3;
				while (isblank(*pC)) pC++;
				/* set DEF flag with multiline DEF starts */
				if (isSingleDEF(pC)==MULTI) saveSubs=TRUE;
				/* save directly line with single line DEF*/
				else if (isSingleDEF(pC)==SINGLE) {
					storeSingleLineToCore(pB,linenum); 
					saveSubs=FALSE;
				}
				/* it's a malformed DEF */
				else perror_(__LINE__,37,l);
			}
			/* save only in case saving flag is TRUE */
			if (saveSubs) storeSingleLineToCore(pB,linenum);
			/* reset saving flag when SUB or multiline DEF end */
			if (checkCommand(pB,"END SUB",&pB)) saveSubs=FALSE;
			else if (!strncasecmp_(pB,"FNEND",5)) saveSubs=FALSE;
		}
		else storeSingleLineToCore(pB,linenum); 
		isVoid=FALSE;
	}

	/* check if file was entirely loaded */
	if (!feof(f) && linenum==MEMLIMIT-1) perror_(__LINE__,9,-2);

	/* check void files */
	if (isVoid) perror_(__LINE__,2,-1,filename);

	/* close file channel */
	if (f) fclose(f);

	endCore=linenum;
}


/* RUN()
 * the RUN command 
 * Note: this statement is executed at start automatically and is not 
 * available as a programming statement */
void RUN(void) {
	int i,j;

	opturn=0; // for safety

	/* COMMON passing */
	if (isCommon) {
		i=1;
		/* cycle through all variables */
		while (i<decCount) {
			/* in case of a not-COMMON variable, try exchanging
			 * it with the first COMMON onward */
			if (!vn[i].isCommon) {
				j=i+1;
				/* search next COMMON */
				while (!vn[j].isCommon && j<=decCount) j++;
				/* change only in the range of current vars */
				if (vn[j].isCommon && j<=decCount) {
					swapDEC(i,j);
					futureDecCount=i;
				}
			}
			/* in case of a COMMON variable, simply update
			 * future counter */
			else futureDecCount=i;
			i++;
		}
		/* rule last variable, in case it's COMMON (this means the
		 * last swap was not done) */
		if (vn[i].isCommon) futureDecCount=i;
		/* set DEF, SUB and DECLARE data */
		isDefRun=isDefVar=isDefCalc=isDefFor=isDefCount=0;
		isSubRun=isSubVar=isSubCalc=isSubFor=subCount=0;
		decCount=futureDecCount;
	}
	/* reset memory if required */
	else if (isToReset) {
		/* set numerical variables and relative flags*/
		for(i=0; i<LIMIT*2;i++) P[i]=0.0;
		/* set numerical arrays references */
		for(i=0; i<LIMIT*2;i++) if (dim[i].elem) *dim[i].elem=0.0;
		/* set string references */
		for(i=0; i<LIMIT*2;i++) PS[i][0]='\0';
		/* set string arrays references */
		for(i=0; i<LIMIT*2;i++) if (dim[i].elem) dimS[i].elem=NIL;
		/* reset COMMON flag */
		for(i=0; i<LIMIT*2;i++) vn[i].isCommon=FALSE;
		/* set DEF, SUB and DECLARE data */
		isDefRun=isDefVar=isDefCalc=isDefFor=isDefCount=0;
		isSubRun=isSubVar=isSubCalc=isSubFor=subCount=decCount=0;
		/* set DATA and MAT vars */
		dc=dcs=dataLimit=dataLimitS=0;
		detValue=0.0, numValue=0;
	}
	
	/* reset FOR/GOSUB references and FOR/NEXT address array 
	 * specific for each RUN() */
	gosubTOS=forTOS=forcount=0;

	/* reset channels info specific for each RUN() */
	for(i=1; i<=STREAMS; i++) channel[i].type=NOUSE;

	/* record starting time if not in a CHAINed program
	 * beginS is used also into the TIM function */
        if (!isChain) {
        	gettimeofday(&tv, NULL);
           	beginS = ((double)tv.tv_sec + tv.tv_usec / 1000000.0);
        }

	/* find first useful not empty line  */
	while (!m[startingLine] && startingLine<=DATLIM) startingLine++;
	
	/* check if start line is within a DEF FN */
	if (fm[startingLine]) perror_(__LINE__,39,l);
	/* preparsing operations */
	preParse(m,1,endCore);
	if (isNoExecute) return;
	isRunOn=TRUE;
	/* and finally, execute program if everything in preParse() is OK */
	parse(m,startingLine,endCore);
}



/* ABORT */
void ABORT(void) {
	/* ABORT alone behaves as STOP */
	if (!*p) END(EXIT_STOP);
	printf("%s\n",SS());
	END(EXIT_ABORT);
}


/* closeFiles()
 * close all PREPEND features and close all files in the queue */
int closeFiles(void) {
	int i=0, none=TRUE;

	for (i=1; i<=STREAMS;i++) {
		if (channel[i].isPrepend) closePrepend(i);
		if (channel[i].isopen) {
			if (channel[i].stream) 
				fclose(channel[i].stream);
			none=FALSE;
		}
	}
	return none;
}


/* END()
 * end program, accomplishing various tasks, such as closing opened channels,
 * print timing, fixing graphical stuff and closing regularly as C requires.
 * Note: as exit state, return the state passed to it as argument.
 * Note: as the returned states, I use:
 * EXIT_SUCCESS (which is 0) to indicate a proper successful execution
 * EXIT_FAILURE (which is 1) to indicate a stop after an error condition
 * EXIT_STOP (which is 2) to indicate an invoked STOP termination */
void END(int state) {
	int i, none=TRUE;

	/* TIMING OPERATIONS */
	/* record end time 
	 * Note: if file was not found, beginS remains negative */
	if (beginS<0) endS=0.0;
	else {
		gettimeofday(&tv, NULL);
	       	endS = ((double)tv.tv_sec + tv.tv_usec / 1000000.0);
		endS=endS-beginS;
	}

	/* FILE OPERATIONS */
	/* close and setup all files */
	none=closeFiles();

	/* free sub memory */
	for (i=1; i<=subCount;i++) {
		if (sn[i].strres) {
			free(sn[i].strres);
			sn[i].strres=NIL;
		}
	}
					
	/* DEBUGGING INFO */
	if (isDebug && !isNoExecute) {
		printf(INVCOLOR);
		printf("\ndebug: structures and files remnants:");
		if (!(ifTOS+selectcount+forcount+docount+whilecount))
			printf(" none.\n\n");
		else printf("\n\n");
		if (ifTOS) 
			printf("'If' remnants       : %d\n",ifTOS);
		if (selectcount)
			printf("'Select' remnants   : %d\n",selectcount);
		if (forcount)
			printf("'For-Next' remnants : %d\n",forcount);
		if (docount)
			printf("'Do-Loop' remnants  : %d\n",docount);
		if (whilecount)
			printf("'While' remnants    : %d\n",whilecount);
		if (!none) {
			printf("Streams left open   : ");
			for (i=1; i<=STREAMS;i++) {
				if (channel[i].isopen) {
					channel[i].isopen=FALSE;
					printf("%d, ",i);
				}
			}
			printf("\n");
			channel[CONSOLE].fcounter=0;
		}
		printf(RESETCOLOR);
	}

	/* ROUTING OPERATIONS */
	/* send routing file to printer in case ROUTE is not used */
	if (isQueued || isLPrintEnabled) ROUTE();

	/* GRAPHICAL OPERATIONS */
	/* CTRL-C was pressed: print to signal */
	if (isInterrupt) {
		fprintf(stdout, "\n\ntbas stopped after user interrupt at line %d",l);
		if (ln[l] && l>=0) fprintf(stdout," (labelled as %d)",ln[l]);
		fprintf(stdout,"\n\n");
	}
	/* SYSTEM READY string */
	else if (state==EXIT_SYSTEM) {
		fprintf(stdout, "\n\nREADY.\n\n");
	}
	/* add a Carriage Return in case of suspended PRINT */
	else if (channel[CONSOLE].fcounter) fprintf(stdout, "\n");

	/* in case the NOECHO option was specified, disable */
	if (!isEcho) {
		termiosHasOccurred=TRUE;
		tcsetattr(0, 0, &savemodes);
		modes.c_lflag &= ~ECHO;
	}
	/* in case of timing, add the timed calculation string */
	if (isTiming) {
		int hour=0,min=0;
		double secs=endS;
		/* precalculate intermediate time values */
		if (endS>60) {
			min=endS/60;
			endS=endS-min*60;
		}
		if (min>60) {
			hour=min/60;
			min=min-hour*60;
		}
		fprintf(stdout,"\n\n\n");
		/* depict time in condensed form */
		if (!hour && !min) 
			fprintf(stdout,"TIME:  %.2f SECS.\n\n",endS);
		else if (!hour)
			fprintf(stdout,"TIME:  %.2d:%.2d MINS   (%.2f SECS).\n\n",min,(int)(endS+0.5),secs);
		else 	
			fprintf(stdout,"TIME:  %.2d:%.2d:%.2d HOUR%c   (%.2f SECS).\n\n",hour,min,(int)(endS+0.5),hour>1?'S':'\0',secs);
	}
	
	/* QUITTING */
	/* flush all open streams, to ensure file data writing is completed */
	fflush(NULL);
	s_exit(state);
}


/* s_exit()
 * safe exit - avoid messing up of colours, and force the screen resetting
 * return 'state' as current state
 * system function */
void s_exit(int state) {
	/* check for closing brackets */
	if (termiosHasOccurred) tcsetattr(0, 0, &savemodes);
	if (isDebug) printf(RESETCOLOR);
	exit(state);
}


/******************************************************************************
 * The following functions define the whole parser engine:
 * SS() 	  : the strings expression parser
 * strRel()       : the strings relation evaluator (returns a truth value)
 * S() 		  : the numerical expression parser
 * getVal()	  : the single numerical value parser 
 * getFunc()	  : the numerical function parser
 * priority()	  : return numerical operator precedence
 * conceived April 2008 for the dib project
 * modified for vibes with strings insertion September-November 2009
 * updated and verified for tbas - February 2013, February-Summer 2014
 * Copyleft Antonio Maschio - 2008-2015 
 ******************************************************************************/

/* SS()
 * The PARSER for string expression; it's a general routine that contains
 * also evaluation of functions and concatenation of strings in series
 * if the + (plus) or & (ampersand) operators are used. 
 * The name SS() comes from S() and the second S means "strings".
 * The address of starting position into the flow stream is pointed by p 
 * The algorithm is as follows: 
 * until there is a string in any form, append it to mypad; return mypad. */
char* SS(void) {
	int i,ns,type=0,rettype=0,cn=0;
	double numres;
	char mypad[COMPSTR], strres[COMPSTR], *opT;

	/* create local pad and reset */
	for (i=0;i<COMPSTR;i++) mypad[i]='\0';
	MYPAD=mypad;
	p--;

	/* calculate string result
	 * accept concatenation with '+' and '&' */
	do {
		p++;
		/* open/close parenthesis */
		if (*p=='(') {
			p++;
			strncat(mypad,SS(),COMPSTR);
			if (*p==')') p++;
			else perror_(__LINE__,37,l);
		}
		/* explicit string */
		else if (*p=='\"') {
			char op[COMPSTR], *P1=op;
			p++;
			while (*p && *p!='\"') *P1++=*p++; 
			*P1='\0';
			if (*p=='\"') p++; // skip closing quotes
			if ((int)strlen(mypad)+(int)strlen(op)>MAXSTRLEN) 
				perror_(__LINE__,124,l);
			strncat_(mypad,op,COMPSTR);
		}
		/* SUB string functions */
		else if ((ns=findSUB(p,&opT,&type,&rettype))) {
			if (type==FUNC && rettype==STR)	{
				execSUB(ns,opT,&numres,strres);
				strncat_(mypad,strres,COMPSTR);
			}
			else perror_(__LINE__,218,l);
			if (*p==')') p++;
			else perror_(__LINE__,37,l);
		}
		/* INKEY$ (one-char input buffer) */
		else if (!strncasecmp_(p,"INKEY$",6)) {
			char cc[2];
			double extraS;
#ifndef NOUNIX
			struct termios thismodes, thissavemodes;
			int thistermiosHasOccurred=termiosHasOccurred;
			int secs=0;
			p+=6;
			/* get optional argument: timed inkey$ */
			if (*p=='(') {
				p++;
				secs=(int)S();
				if (secs<0) secs=-secs;
				if (secs>60) secs=60;
				if (*p==')') p++;
				else perror_(__LINE__,37,l);
			}

			/* delay time */
        		gettimeofday(&tv, NULL);
       			extraS = ((double)tv.tv_sec + tv.tv_usec / 1000000.0);

			/* set tcgetattr values backing up current state */
			tcgetattr(0, &thismodes);
			thissavemodes = thismodes;
			termiosHasOccurred=TRUE;
			thismodes.c_lflag &= ~ICANON;
			thismodes.c_lflag &= ~ECHO;
			/* normal INKEY$ */
			if (!secs) {
				tcsetattr(0, 0, &thismodes);
				cc[0] = (char)getchar();
				cc[1]='\0';
			}
			/* timed INKEY$ */
			else {
				thismodes.c_cc[VMIN]=0;
				thismodes.c_cc[VTIME]=secs*10;
				tcsetattr(0, 0, &thismodes);
				read(0, cc, 1);
				cc[1]='\0';
			}
			tcsetattr(0, 0, &thissavemodes);
			termiosHasOccurred=thistermiosHasOccurred;
#else			
			cc[0]=(char)getchar();
			cc[1]='\0';
#endif
			if ((int)strlen(mypad)+1>MAXSTRLEN) 
				perror_(__LINE__,124,l);
			strncat_(mypad,cc,COMPSTR);
			/* get timer again to exclude user delay from timing 
			 * and update beginS */
		        gettimeofday(&tv, NULL);
		       	extraS=((double)tv.tv_sec+tv.tv_usec/1000000.0)-extraS;
			beginS=beginS+extraS;
		}
		/* INPUT$ (& cut) string from keyboard */
		else if (!strncasecmp_(p,"INPUT$",6)) {
			char op[COMPSTR];
			op[0]='\0';
#ifndef NOUNIX			
			struct termios thismodes, thissavemodes;
			int thistermiosHasOccurred=termiosHasOccurred;
			int num=0;
			p+=6;
			/* no arguments in UNIX: use getString */
			if (*p!='(') strncpy_(op,getString(),COMPSTR);

			/* there's an argument in UNIX: string length */
			else {
				char *lp=op;
				int li=0;
				p++;
				num=(int)S();
				if (num<0) num*=-1;
				if (!num || num>MAXSTRLEN) num=MAXSTRLEN;
				if (*p!=')') perror_(__LINE__,37,l);
				else p++;
				/* set tcgetattr values 
				 * backing up current state */
				tcgetattr(0, &thismodes);
				thissavemodes = thismodes;
				termiosHasOccurred=TRUE;
				thismodes.c_lflag &= ~ICANON;
				thismodes.c_lflag &= ~ECHO;
				/* set up termios data, 
				 * included controlling features */
				tcsetattr(0, 0, &thismodes);
				while (li<num) {
					*lp=getchar();
					if (*lp==10 || *lp==13 || *lp=='\n') {
						*lp='\0';
						break;
					}
					else lp++;
					li++;
				}
				*lp='\0';
				tcsetattr(0, 0, &thissavemodes);
				termiosHasOccurred=thistermiosHasOccurred;
			}
#else
			/* Note: in Windows and in general, where termios.h
			 * is not available, the INPUT$ function simply returns
			 * the input string, and not argument is allowed */
			strncpy_(op,getString(),COMPSTR);
#endif
			if ((int)strlen(mypad)+(int)strlen(op)>MAXSTRLEN) 
				perror_(__LINE__,124,l);
			strncat_(mypad,op,COMPSTR);
		}
		/* FORMAT$() function (CDC tradition)
		 * ~return string with number formatted according to the
		 * laws of PRINT USING (through printFormatted().
		 * The intermediate result is stored into _FORMAT )*/
		else if (!strncasecmp_(p,"FORMAT$(",8)) {
			double val;
			char op[COMPSTR];
			p+=8;
			if (isanystring(p)) strncpy_(op,SS(),COMPSTR);
			else perror_(__LINE__,214,l);
			if (*p==',') p++;
			else perror_(__LINE__,37,l);
			if (isanynumber(p)) val=S();
			else perror_(__LINE__,215,l);
			printFormatted(op,NIL,FALSE,val);
			if ((int)strlen(mypad)+(int)strlen(_FORMAT)>MAXSTRLEN) 
				perror_(__LINE__,124,l);
			if (*p==')') p++;
			else perror_(__LINE__,37,l);
			strncat_(mypad,_FORMAT,COMPSTR);
		}
		/* TRIM$() function (CDC tradition)
		 * ~return a string with trimmed trailing spaces
		 * (ahead and behind) */
		else if (!strncasecmp_(p,"TRIM$(",6)) {
			char *OP;
			char op[COMPSTR];
			p+=6;
			if (isanystring(p)) strncpy_(op,SS(),COMPSTR);
			else perror_(__LINE__,214,l);
			if ((int)strlen(op)>0) OP=op+(int)strlen(op)-1;
			else OP=op;
			while ((int)(OP-op)>0 && isblank(*OP)) *OP--='\0';
			OP=op;
			while ((int)(OP-op)<MAXSTRLEN && isblank(*OP)) OP++;
			if ((int)strlen(mypad)+(int)strlen(OP)>MAXSTRLEN) 
					perror_(__LINE__,124,l);
			if (*p==')') p++;
			else perror_(__LINE__,37,l);
			strncat_(mypad,OP,COMPSTR);
		}
		/* ENV$() function
		 * ~return a string with the content of the environment variable
		 * passed as a string argument */
		else if (!strncasecmp_(p,"ENV$(",5)) {
			char env[COMPSTR], value[COMPSTR];
			p+=5;
			if (!isanystring(p)) perror_(__LINE__,218,l);
			strncpy_(env,SS(),COMPSTR);
			if (!getenv(env) || !env[0]) value[0]='\0';
			else strncpy_(value,getenv(env),COMPSTR);
			if ((int)strlen(mypad)+(int)strlen(value)>MAXSTRLEN) 
					perror_(__LINE__,124,l);
			if (*p==')') p++;
			else perror_(__LINE__,37,l);
			strncat_(mypad,value,COMPSTR);
		}
		/* DIR$ system variable
		 * ~return a string with the content of current working
		 * directory; the argument is dummy */
		else if (!strncasecmp_(p,"DIR$",4)) {
			char value[COMPSTR];
			p+=4;
			if (*p=='(') {
				p++;
				S();
				if (*p==')') p++;
			}
			strncpy_(value,getenv("PWD"),COMPSTR);
			if ((int)strlen(mypad)+(int)strlen(value)>MAXSTRLEN) 
					perror_(__LINE__,124,l);
			strncat_(mypad,value,COMPSTR);
		}
		/* ESM$()/ERR$() functions (CDC and HP tradition)
		 * ~return a string with error message of the code argument
		 * or current error message if argument is null 
		 * Note this function requires an argument! */
		else if ((cn=recog(2,p,"ESM$(","ERR$("))) {
			int o;
			char op[COMPSTR];
			p+=cn;
			if (!isanynumber(p)) perror_(__LINE__,218,l);
			o=(int)S();
			if (o<0) o=0;
			if (o>ERRLIM) o=ERRLIM;
			strncpy_(op,error[o?o:errValue],COMPSTR);
			if ((int)strlen(mypad)+(int)strlen(op)>MAXSTRLEN) 
					perror_(__LINE__,124,l);
			if (*p==')') p++;
			else perror_(__LINE__,37,l);
			strncat_(mypad,op,COMPSTR);
		}
		/* STRING$() function 
		 * ~return a string composed by n copies in the first argument
		 * of ASCII character or string in second argument */
		else if (!strncasecmp_(p,"STRING$(",8)) {
			int o;
			char op[COMPSTR];
			int c=0;
			p+=8;
			o=(int)S();
			if (*p!=',') perror_(__LINE__,37,l);
			else p++;
			if (o<0) o=1;
			if (isanystring(p)) strncpy_(op,SS(),COMPSTR);
			else {
				c=(char)S();
				sprintw(op,COMPSTR,"%c",c);
			}
			if ((int)strlen(mypad)+o*(int)strlen(op)>MAXSTRLEN) 
				perror_(__LINE__,124,l);
			if (*p==')') p++;
			else perror_(__LINE__,37,l);
			while (o>0) {
				strncat_(mypad,op,COMPSTR);
				o--;
			}
		}
		/* LPAD$() function (CDC tradition)
		 * ~return a string a$ in first argument left-padded with 
		 * spaces if shorter than the length n in second argument */
		else if (!strncasecmp_(p,"LPAD$(",6)) {
			int o=0, tl=0;
			char op[COMPSTR];
			p+=6;
			if (isanystring(p)) strncpy_(op,SS(),COMPSTR);
			else perror_(__LINE__,218,l);
			if (*p!=',') perror_(__LINE__,37,l);
			else p++;
			o=(int)S();
			if (o<0) perror_(__LINE__,37,l);
			tl=(int)strlen(op);
			o=o>tl?o:tl;
			if ((int)strlen(mypad)+o>MAXSTRLEN) 
				perror_(__LINE__,124,l);
			o-=tl;
			while (o-->0) strncat(mypad," ",COMPSTR);
			if (*p==')') p++;
			else perror_(__LINE__,37,l);
			strncat_(mypad,op,COMPSTR);
		}
		/* RPAD$() function (CDC tradition)
		 * ~return a string a$ in first argument right-padded with 
		 * spaces if shorter than the length n in second argument */
		else if (!strncasecmp_(p,"RPAD$(",6)) {
			int o=0, tl=0;
			char op[COMPSTR];
			p+=6;
			if (isanystring(p)) strncpy_(op,SS(),COMPSTR);
			else perror_(__LINE__,218,l);
			if (*p!=',') perror_(__LINE__,37,l);
			else p++;
			o=(int)S();
			if (o<0) perror_(__LINE__,37,l);
			tl=(int)strlen(op);
			o=o>tl?o:tl;
			if ((int)strlen(mypad)+o>MAXSTRLEN) 
				perror_(__LINE__,124,l);
			strncat_(mypad,op,COMPSTR);
			o-=tl;
			while (o-->0) strncat(mypad," ",COMPSTR);
			if (*p==')') p++;
			else perror_(__LINE__,37,l);
		}
		/* RPT$()/REPEAT$() function (CDC tradition)
		 * ~return a string composed by repeated copies of the string
		 * in the first argument by n copies in second argument */
		else if ((cn=recog(2,p,"RPT$(","REPEAT$("))) {
			int o;
			char op[COMPSTR];
			p+=cn;
			if (isanystring(p)) strncpy_(op,SS(),COMPSTR);
			else perror_(__LINE__,218,l);
			if (*p!=',') perror_(__LINE__,37,l);
			else p++;
			o=(int)S();
			if (o<0) perror_(__LINE__,37,l);
			if ((int)strlen(mypad)+o*(int)strlen(op)>MAXSTRLEN) 
				perror_(__LINE__,124,l);
			while (o>0) {
				strncat_(mypad,op,COMPSTR);
				o--;
			}
			if (*p==')') p++;
			else perror_(__LINE__,37,l);
		}
		/* WEEKDAY$()
		 * ~return current week day 
		 * Note: the algorithm for a casual date is simple and, for 
		 * dates between YEARLIMIT (currently set equal to 1583) and 
		 * the current year, it returns always correct 
		 * values (at least as far as I can test); the algorithm:
		 * ~- convert date to Gregorian day number
		 * ~- subtract 1 because we start with Sunday=0 not Sunday=1
		 * ~- subtract another 1
		 * ~- return mod 7: value is 0~6 (Sun~Sat) 
		 * ~Antonio Maschio - August 2014 */
		else if (!strncasecmp_(p,"WEEKDAY$(",9)) {
			double o;
			int cut=0;
			char mon[32];
			p+=9;
			o=S();
			if (*p==',') {
				p++;
				cut=(int)S();
			}
			if (cut<0) cut=0;
			if (cut>3) cut=3;
			if (o==0.0) {
				if (cut) 
				     strncpy_(mon,daylabel[getWDay()],cut+1);
				else 
				     strncpy_(mon,daylabel[getWDay()],32);
			}
			else {
				/* any date's weekday - some dates tested on
				 * http://www.timeanddate.com/date/weekday.html
				 * The leap year is implicit (02/29 converted
				 * 03/01 implicitly) */
				int w,ty,tm,td; // from decimal date
				ty=(int)o; // yy or yyyy
				tm=(int)((o-ty)*100); // mm
				td=(int)((o*100-tm-100*ty)*100+0.5); // day
				ty=setYear(ty);
				w=gregDayNumber(ty,tm,td)-2;
				if (cut) strncpy_(mon,daylabel[w%7],cut+1);
				else strncpy_(mon,daylabel[w%7],32);
			}
			if ((int)strlen(mypad)+(int)strlen(mon)>MAXSTRLEN) 
					perror_(__LINE__,124,l);
			if (*p==')') p++;
			else perror_(__LINE__,37,l);
			strncat_(mypad,mon,COMPSTR);
		}
		/* MONTH$() 
		 * ~if argument is 0, return current month
		 * ~if argument is in the format yy.mmdd or yyyy.mmdd, 
		 * extract month label */
		else if (!strncasecmp_(p,"MONTH$(",7)) {
			double o, cut=0;
			char mon[32];
			p+=7;
			o=S();
			if (*p==',') {
				p++;
				cut=(int)S();
			}
			if (cut<0) cut=0;
			if (cut>3) cut=3;
			/* get current values */
			if (o==0.0) {
				if (cut) 
				  strncpy_(mon,monthlabel[getMonth()],cut+1);
				else 
				  strncpy_(mon,monthlabel[getMonth()],32);
			}
			/* get decimal argument values */
			else {
				o-=(int)o;
				o=(int)(o*100+0.5);
				if (cut) 
				  strncpy_(mon,monthlabel[(int)(o)],cut+1);
				else 
				  strncpy_(mon,monthlabel[(int)(o)],32);
			}
			if ((int)strlen(mypad)+(int)strlen(mon)>MAXSTRLEN) 
				perror_(__LINE__,124,l);
			if (*p==')') p++;
			else perror_(__LINE__,37,l);
			strncat_(mypad,mon,COMPSTR);
		}
		/* CLK$ function (CDC tradition)
		 * ~return the hour in the format HH.MM.SS  */
		else if (!strncasecmp_(p,"CLK$",4)) {
			int h=getHour(),m=getMin(),s=getSec();
			char str[16];
			p+=4;
			sprintw(str,16," %02d.%02d.%02d",h,m,s);
			if ((int)strlen(mypad)+9>MAXSTRLEN) 
				perror_(__LINE__,124,l);
			strncat_(mypad,str,COMPSTR);	
		}
		/* CHR$ function 
		 * ~return a one-character string with the picture of the
		 * ASCII numeric value in argument */
		else if (!strncasecmp_(p,"CHR$(",5)) {
			char chop[2];
			p+=5;
			chop[0]=(unsigned char)S();
			chop[1]='\0';
			if ((int)strlen(mypad)+1>MAXSTRLEN)	
				perror_(__LINE__,124,l);
			if (*p==')') p++;
			else perror_(__LINE__,37,l);
			strncat_(mypad,chop,COMPSTR);
		}
		/* DATE$/DAT$ function (based on DAT$ of CDC tradition)
		 * ~return the date in the format YY/MM/DD (KRONOS/NOS feat.)*/
		else if ((cn=recog(2,p,"DATE$","DAT$"))) {
			int y=getYear(),m=getMonth(),d=getDay();
			char str[16];
			p+=cn;
			/* retain year's last two digits only */
			while (y>1000) y-=1000;
			while (y>100) y-=100;
			sprintw(str,16," %02d/%02d/%02d",y,m,d);
			if ((int)strlen(mypad)+9>MAXSTRLEN) 
				perror_(__LINE__,124,l);
			strncat_(mypad,str,COMPSTR);	
		}
		/* SPACE$/SPC$ function (SPC$ from Dartmouth tradition)
		 * ~return a string of as much spaces as the argument value */
		else if ((cn=recog(2,p,"SPACE$(","SPC$("))) {
			int sp;
			p+=cn;
			sp=(int)S();
			if (sp<0) sp=0;
		       	if (sp>MAXSTRLEN)sp=MAXSTRLEN; 
			if ((int)strlen(mypad)+sp>MAXSTRLEN)	
				perror_(__LINE__,124,l);
			if (*p==')') p++;
			else perror_(__LINE__,37,l);
			strncat_(mypad,spaces(sp),COMPSTR);
		}
		/* STR$/NUM$ function (NUM$ from HP BASIC tradition) 
		 * ~convert the value in argument to a string without
		 * trailing spaces ahead and behind */
		else if ((cn=recog(2,p,"STR$(","NUM$("))){
			double val;
			char num[64], *pnum=num;
			p+=cn;
			val=S();
			printNUM(0,val,TRUE,num);
			/* discard trailing spaces */
			if ((int)strlen(num)>0) pnum=num+(int)strlen(num)-1;
			else pnum=num;
			while ((int)(pnum-num)>0 && isblank(*pnum)) 
				*pnum--='\0';
			pnum=num;
			while ((int)(pnum-num)<64 && isblank(*pnum)) pnum++;
			/* check length 	*/
			if ((int)strlen(mypad)+(int)strlen(pnum)>MAXSTRLEN)
				perror_(__LINE__,124,l);
			if (*p==')') p++;			
			else perror_(__LINE__,37,l);
			strncat_(mypad,pnum,COMPSTR);
		}
		/* LEFT$ function 
		 * ~return the left substring of string in the first argument 
		 * from start according to the second argument length
		 * Note: second argument is optional, and if not specified 
		 * length is set to 1 */
		else if (!strncasecmp_(p,"LEFT$(",6)) {
			int llen;
			char op[COMPSTR], *P1=op;
			p+=6;
			strncpy_(op,SS(),COMPSTR);
			if (*p==',') {
				p++;
				llen=(int)S();
			}
			else if (*p==')') llen=1;
			else perror_(__LINE__,37,l);
			if (llen<0) llen=0; 
			if (llen>(int)strlen(op)+1) llen=(int)strlen(op)+1;
			*(P1+llen)='\0';
			if ((int)strlen(mypad)+(int)strlen(op)>MAXSTRLEN)
				perror_(__LINE__,124,l);
			if (*p==')') p++;
			else perror_(__LINE__,37,l);
			strncat_(mypad,op,COMPSTR);
		}
		/* MID$ function 
		 * ~return the substring of string in the first argument from 
		 * start to the second argument for the number of characters 
		 * contained in the third argument 
		 * This function differs from SEG$ in that the numerical
		 * arguments specify the start and length of selection, rather
		 * than the start and end of selection.
		 * Note: arguments may be 0 (return the NULL string) 
		 * Note: third argument is optional, and if not specified 
		 * length is set to 1 */
		else if (!strncasecmp_(p,"MID$(",5)) {
			int mstart=0,mlen=0;
			char op[COMPSTR], *P1=op;
			p+=5;
			strncpy_(op,SS(),COMPSTR);
			if (*p==',') p++;
			else perror_(__LINE__,37,l);
			mstart=(int)S();
			if (*p==',') {
				p++;
				mlen=(int)S();
			}
			else if (*p==')') mlen=1;
			else perror_(__LINE__,37,l);
			if (mstart<=0) mstart=1;
			if (mlen<0) mlen=0;
			if (mstart>(int)strlen(op)+1) mstart=(int)strlen(op)+1;
			if (mlen>(int)strlen(op)) mlen=(int)strlen(op);
			if ((int)strlen(mypad)+mlen+1<MAXSTRLEN) 
				*(P1+mstart-1+mlen)='\0';
			if ((int)strlen(mypad)+(int)strlen(P1+mstart-1)>MAXSTRLEN)
				perror_(__LINE__,124,l);
			if (*p==')') p++;
			else perror_(__LINE__,37,l);
			strncat_(mypad,P1+mstart-1,COMPSTR);
		}
		/* RIGHT$ function  
		 * ~return the right substring of the string in the first 
		 * argument from end according to the second argument length
		 * Note: second argument is optional, and if not specified 
		 * length is set to 1 */
		else if (!strncasecmp_(p,"RIGHT$(",7)) {
			int rlen=0;
			char op[COMPSTR], *P1=op;
			p+=7;
			strncpy_(op,SS(),COMPSTR);
			if (*p==',') {
				p++;
				rlen=(int)S();
			}
			else if (*p==')') rlen=1;
			else perror_(__LINE__,37,l);
			if (rlen<0) rlen=0;
			if (rlen>(int)strlen(op)) rlen=(int)strlen(op);
			rlen=(int)strlen(op)-rlen;
			if ((int)strlen(mypad)+rlen>MAXSTRLEN)
				perror_(__LINE__,124,l);
			if (*p==')') p++;
			else perror_(__LINE__,37,l);
			strncat_(mypad,P1+rlen,COMPSTR);
		}
		/* UPPER$ function 
		 * ~return the same string in argument turned to all upper
		 * characters in the alphabetical sections */
		else if ((cn=recog(2,p,"UPPER$(","UCASE$("))) {
			char op[COMPSTR], *P1=op;
			p+=cn;
			/* case of a char argument */
			if (*(p+1)==')') sprintw(op,COMPSTR,"%c",*p++);
			/* case of any string */
			else strncpy_(op,SS(),COMPSTR);
			/* string is too long */
			if ((int)strlen(mypad)+(int)strlen(op)>MAXSTRLEN)
				perror_(__LINE__,124,l);
			while (*P1) {
				*P1=toupper(*P1);
				P1++;
			}
			if (*p==')') p++;
			else perror_(__LINE__,37,l);
			strncat_(mypad,op,COMPSTR);
		}
		/* LOWER$ function 
		 * ~return the same string in argument turned to all lower
		 * characters in the alphabetical sections */
		else if ((cn=recog(2,p,"LOWER$(","LCASE$("))) {
			char op[COMPSTR], *P1=op;
			p+=cn;
			/* case of a char argument */
			if (*(p+1)==')') sprintw(op,COMPSTR,"%c",*p++);
			/* case of any string */
			else strncpy_(op,SS(),COMPSTR);
			/* string is too long */
			if ((int)strlen(mypad)+(int)strlen(op)>MAXSTRLEN)
				perror_(__LINE__,124,l);
			while (*P1) {
				*P1=tolower(*P1);
				P1++;
			}
			if (*p==')') p++;
			else perror_(__LINE__,37,l);
			strncat_(mypad,op,COMPSTR);
		}
		/* CAP$ function 
		 * ~return the same string in lower characters with the first
		 * letter capitalized */
		else if (!strncasecmp_(p,"CAP$(",5)) {
			char op[COMPSTR], *P1=op;
			p+=5;
			/* case of a char argument */
			if (*(p+1)==')') sprintw(op,COMPSTR,"%c",*p++);
			/* case of any string */
			else strncpy_(op,SS(),COMPSTR);
			/* string is too long */
			if ((int)strlen(mypad)+(int)strlen(op)>MAXSTRLEN)
				perror_(__LINE__,124,l);
			/* turn to lower */
			while (*P1) *P1=tolower(*P1), P1++;
			/* capitalize initials; rules:
			 * ~ capitalize first letter 
			 * ~ capitalize letter following a dash (e.g.
			 *   ROMAN-CATHOLIC is turned to Roman-Catholic 
			 * ~ capitalize the whole string (until space or
			 *   end-of-string) if % is met before the token */
			P1=op;
			*P1=toupper(*P1), P1++;
			while (*P1) {
				if (isblank(*P1)) {
					P1++;
					if (isalpha(*P1)) *P1=toupper(*P1);
				}
				else if (*P1=='-') {
					P1++;
					if (isalpha(*P1)) *P1=toupper(*P1);
				}
				P1++;
			}
			if (*p==')') p++;
			else perror_(__LINE__,37,l);
			strncat_(mypad,op,COMPSTR);
		}
		/* SEG$ function (Dartmouth tradition) 
		 * ~return the substring of the string in the first argument
		 * from the second argument to the third argument places 
		 * This function differs from MID$ in that the numerical
		 * arguments specify the start and end of selection, rather
		 * than the start and length of selection. 
		 * Note: If arguments are meaningless, inverted or identifying 
		 * a substring out of the length of string - e.g. 
		 * SEG$("free"10,14), the NULL string is returned */
		else if (!strncasecmp_(p,"SEG$(",5)) {
			char op[COMPSTR], *P1=op;
			int sstart,send;
			p+=5;
			strncpy_(op,SS(),COMPSTR);
			if (*p==',') p++;
			else perror_(__LINE__,37,l);
			sstart=(int)S();
			if (sstart<=0) sstart=1;
			if (*p==',') {
				p++;
				send=(int)S();
			}
			else send=strlen(op)+1;
		       	if (send<sstart) sstart=1, send=0;
			*(P1+send)='\0';
			if ((int)strlen(mypad)+(int)strlen(P1+sstart-1)>MAXSTRLEN)
				perror_(__LINE__,124,l);
			if (*p==')') p++;
			else perror_(__LINE__,37,l);
			strncat_(mypad,P1+sstart-1,COMPSTR);
		}
		/* HEX$ function 
		 * ~return the hex form string of the number in argument */
		else if (!strncasecmp_(p,"HEX$(",5)) {
			int val;
			char num[BITS+1];
			p+=5;
			val=(int)S();
			sprintw(num,BITS+1,"%X",val);
			/* check length 	*/
			if ((int)strlen(mypad)+(int)strlen(num)>MAXSTRLEN)
				perror_(__LINE__,124,l);
			if (*p==')') p++;			
			else perror_(__LINE__,37,l);
			strncat_(mypad,num,COMPSTR);
		}
		/* OCT$ function 
		 * ~return the octal form string of the number in argument */
		else if (!strncasecmp_(p,"OCT$(",5)) {
			int val;
			char num[BITS+1];
			p+=5;
			val=(int)S();
			sprintw(num,BITS+1,"%o",val);
			/* check length 	*/
			if ((int)strlen(mypad)+(int)strlen(num)>MAXSTRLEN)
				perror_(__LINE__,124,l);
			if (*p==')') p++;			
			else perror_(__LINE__,37,l);
			strncat_(mypad,num,COMPSTR);
		}
		/* BIN$ function 
		 * ~return the binary form string of the number in argument;
		 * if number is followed by 0, add all leading zeroes to the
		 * 32 bits number; if number is followed by 1 or by nothing,
		 * print the number in compact form; a negative number is
		 * anyway printed in full 32 bits. */
		else if (!strncasecmp_(p,"BIN$(",5)) {
			int val;
			int mode=1;	// no zeroes
			char num[BITS+1];
			p+=5;
			val=(int)S();
			if (*p==',') {
				p++;
				mode=(int)S();
				if (mode) mode=1; // always 1 for true
			}
			mode=fprintb(num,val,mode);
			if ((int)strlen(mypad)+mode>MAXSTRLEN)
				perror_(__LINE__,124,l);
			if (*p==')') p++;			
			else perror_(__LINE__,37,l);
			strncat_(mypad,num,COMPSTR);
		}
		/* REV$ function 
		 * ~reverse the string in argument */
		else if (!strncasecmp_(p,"REV$(",5)) {
			char op[COMPSTR],rop[COMPSTR];
			int i,h;
			p+=5;
			strncpy_(op,SS(),COMPSTR);
			i=h=strlen(op)-1;
			while (i>=0) {
				rop[h-i]=op[i];
				i--;
			}
			rop[h+1]='\0';
			if ((int)strlen(mypad)+(int)strlen(rop)>MAXSTRLEN)
				perror_(__LINE__,124,l);
			if (*p==')') p++;			
			else perror_(__LINE__,37,l);
			strncat_(mypad,rop,COMPSTR);
		}
		/* TIME$ function 
		 * ~return the standard time string in the form "HH:MM:SS" 
		 * Note: this has a dummy argument */
		else if (!strncasecmp_(p,"TIME$",5)) {
			char op[COMPSTR];
			p+=5;
			/* dummy argument */
			if (*p=='(') {
				p++;
				S();
				if (*p!=')') perror_(__LINE__,37,l);
				else p++;
			}
			/* build time string */
			sprintw(op,COMPSTR,"%02d:%02d:%02d",getHour(),getMin(),getSec());
			if ((int)strlen(mypad)+(int)strlen(op)>MAXSTRLEN)
				perror_(__LINE__,124,l);
			strncat_(mypad,op,COMPSTR);
		}
		/* USR$ function 
		 * ~return the log string (see tbas.h for LOGAS) 
		 * Note: this is a dummy function */
		else if (!strncasecmp_(p,"USR$",4)) {
			p+=4;
			/* dummy argument */
			if (*p=='(') {
				p++;
				S();
				if (*p!=')') perror_(__LINE__,37,l);
				else p++;
			}
			if ((int)strlen(mypad)+(int)strlen(LOGAS)>MAXSTRLEN)
				perror_(__LINE__,124,l);
			strncat_(mypad,LOGAS,COMPSTR);
		}
		/* numeric variables are illegal */
		else if ((ns=findDEC(p)) && !_arr && _rettype==NUM) {
			perror_(__LINE__,86,l);
		}
		/* string variables */
		else if ((ns=findDEC(p)) && !_arr) {
			int var=ns;
			p+=_len;
			if (PS[var][0]) {
				if ((int)strlen(mypad)+(int)strlen(PS[var])>MAXSTRLEN)
					perror_(__LINE__,124,l);
				if ((int)strlen(PS[var]))
					strncat_(mypad,PS[var],COMPSTR);
			}
		}
		/* string arrays */
		else if ((ns=findDEC(p)) && _arr) {
			int var=0,M=0,N=0,T=1;
			char op[COMPSTR];
			var=ns, p+=_len+1; // skip var, '$' and '('
			M=(int)S();
			if (*p==',') {
				p++;
				N=(int)S();
				if (dimS[var].type!=2) perror_(__LINE__,76,l);
				else T=2;
			}
			if (*p!=')') perror_(__LINE__,37,l);
			else p++;
			strncpy_(op,getArrayStr(var,T,M,N),COMPSTR);
			if ((int)strlen(mypad)+(int)strlen(op)>MAXSTRLEN)
				perror_(__LINE__,124,l);
			strncat_(mypad,op,COMPSTR); 
		}
		/* no possible evaluation */
		else {
			/* wrong string evaluation in string assignment */
			if (!numEval) perror_(__LINE__,86,l);
			/* wrong string relation in numeric assignment */
			else perror_(__LINE__,37,l); 
		}
		if (!*p || *p==',') break;
	} while (*p=='+' || *p=='&');
	strncpy_(sspad,mypad,COMPSTR);
	MYPAD=mypad;
	return MYPAD;
}


/* strRel()
 * evaluate a string expression, with operators <, >, =, # (for <> and ><),
 *  (for <= and =<) and ! (for >= and =>) and | (for ==); 
 * always return a truth value
 * Note: in relations, any kind of string variable (even string arrays)
 * may be used into the formula.
 * Note: since evaluation of string relations returns a truth value, which is
 * a number, it is useful only into IF..THEN statements and into S(); so they 
 * are recognized/used there, and not into SS() */
int strRel(void) {
	int rel=FALSE, operator;
	char op1[COMPSTR], op2[COMPSTR], *P1=op1, *P2=op2;

	/* read operands and operator */
	strncpy_(op1,SS(),COMPSTR);
	if (ispro(p)) operator='|', p+=2;	// ==
	else if (isdif(p)) operator='#', p+=2;	// <> or ><
	else if (isleq(p)) operator='', p+=2;	// <= or =<
	else if (isgeq(p)) operator='!', p+=2;	// >= or =>
	else operator=*p++;
	strncpy_(op2,SS(),COMPSTR);
	/* delete trailing spaces before and after strings */
	if (operator!='|' && isComparison==RELATIVE) {
		if ((int)strlen(op1)>0) {
			P1=op1+(int)strlen(op1)-1; 
			while ((int)(P1-op1)>0 && isblank(*P1)) *P1--='\0';
		}
		if ((int)strlen(op2)>0) {
			P2=op2+(int)strlen(op2)-1; 
			while ((int)(P2-op2)>0 && isblank(*P2)) *P2--='\0';
		}
		P1=op1; while ((int)(P1-op1)<COMPSTR && isblank(*P1)) P1++;
		P2=op2; while ((int)(P2-op2)<COMPSTR && isblank(*P2)) P2++;
	}

	/* apply operator */
	/* case of STRING and string as different strings */
	if (isCaseSensitive) switch(operator) {
		case '<': if (strcmp(P1,P2) < 0)  rel=TRUE; break;
		case '>': if (strcmp(P1,P2) > 0)  rel=TRUE; break;
		case '=': if (!strcmp(P1,P2))     rel=TRUE; break;
		case '|': if (!strcmp(P1,P2))     rel=TRUE; break;
		case '#': if (strcmp(P1,P2))      rel=TRUE; break;
		case '': if (strcmp(P1,P2) <= 0) rel=TRUE; break;
		case '!': if (strcmp(P1,P2) >= 0) rel=TRUE; break;
		default:  perror_(__LINE__,41,l);
	}
	/* case of STRING and string as same strings */
	else switch(operator) {
		case '<': if (strcasecmp(P1,P2) < 0)  rel=TRUE; break;
		case '>': if (strcasecmp(P1,P2) > 0)  rel=TRUE; break;
		case '=': if (!strcasecmp(P1,P2))     rel=TRUE; break;
		case '|': if (!strcasecmp(P1,P2))     rel=TRUE; break;
		case '#': if (strcasecmp(P1,P2))      rel=TRUE; break;
		case '': if (strcasecmp(P1,P2) <= 0) rel=TRUE; break;
		case '!': if (strcasecmp(P1,P2) >= 0) rel=TRUE; break;
		default:  perror_(__LINE__,41,l);
	}

	return rel;
}


/* priority()
 * return math priority for operator op, according to the following table:
 * 8: the minus and the NOT (caught in S()) are evaluated immediately
 * 7: (..) parentheses (caught in getVal) have the highest precedence
 * 6: power (^ and **), shift left/right and modulus
 * 5: multiplication, division
 * 4: addition, subtraction
 * 3: <, >, =, <=, >=, <> 
 * 2: &&, ||, !!, &, |, {, }
 * 1: IMP, EQV
 * 0: anything else has NIL priority and should cause error 
 * Note: string concatenation operators (in the syntax + and &) are evaluated 
 * directly into SS(); string relational operators (<, >, <=, >=, =, <>) are
 * evaluated into strRel(), which is invoked only into IFTHEN()*/
int priority(const int op) {
	if (op=='^'||op==']'||op=='['||op=='%') return 6;
	else if (op=='*'||op=='/'||op=='\\') return 5;
	else if (op=='+'||op=='-') return 4;
	else if (op=='<'||op=='>'||op=='='||op==''||op=='#'||op=='!') 
		return 3;
	else if (op==''||op==''||op==''||op=='&'||op=='|'||op=='{'||op=='}')
		return 2;
	else if (op==''||op=='') return 1;
	return 0;
}
#define PRIOR 6


/* S()
 * The PARSER for numerical expressions
 * The address of starting position into the flow stream is pointed by p 
 * The algorithm is as follows: 
 * an algebraic expression is turned into an RPN stack-based expression, 
 * with numbers and operators in order in their own stack; operations are 
 * performed from higher priorities down to lower and from left to right;  
 * priorities are given by the priority() function, which is part of this tool.
 * Note: the name S() is retained from Mr. Spinellis version. 
 * Note: the < and > operators (with their companion <= and >=) may not give 
 * correct results with big numbers, because of the ways of storing numbers; 
 * for instance, 1E29*10 is not stored in the same way of 1E30, because of 
 * rounding errors in calculating the multiplication (the error propagates 
 * rapidly when big numbers are calculated in series) */
double S(void) {
	double numStack[PSTACK];
	int opStack[PSTACK], prStack[PSTACK], numTOS=0, opTOS=0, ii,jj, level;
	char* pcopy=NIL;

	/* catch unary plus at start (simply discarded because redundant)
	 * Note: tbas parser is built in such a way that the unary sign 
	 * (+ or -) is tightly bound to the entity following it, so that 
	 * operations like -3+-2, +3+-2, -3-+2, +3-+2 are accepted */
	if (*p=='+') p++;
	/* catch unary minus at start */
	else if (*p=='-') {
		double res;
		p++; // skip '-'
		res=getVal();
		if (*p=='^' || istim(p)) {
			/* power following a unary minus has higher priority */
			double expon;
			int sign=1;
			if (*p=='^') p++;
			else p+=2;
			if (*p=='-') p++, sign=-1;
			else if (*p=='+') p++;
			expon=sign*getVal();
			res=-power(res,expon);
		}
		else res=-res;
		numStack[numTOS++]=res;
		opturn=1;
		/* short circuit: if number is followed by comma, semicolon
		 * or colon or closed parentheses or backslash or void, 
		 * return result immediately */
		if (!*p||*p==','||*p==';'||*p==':'||*p==')') 
		return res;
		/* short circuit #2: if the -a^x exponent is not followed
		 * by an operator, return result immediately */
		if (!isop(p)) return res;
	}

	pcopy=p;
	/* reverse algebraic expression into a stack-based structure */
	while (*p && !iscs(p) && *p!='\"' && *p!='\'' && !isblank(*p) \
			&& !(*p=='&'&&!*(p+1)) && *p!=':' && *p!='\?'){
		/* catch an operator */
		if ((isop(p)) && opturn) {
			     if (isand(p)) opStack[opTOS]='',p++;
			else if (is_or(p)) opStack[opTOS]='',p++; 
			else if (isqua(p)) opStack[opTOS]='',p++; 
			else if (isxor(p)) opStack[opTOS]='',p++; 
			else if (iseqv(p)) opStack[opTOS]='',p++; 
			else if (isimp(p)) opStack[opTOS]='',p++; 
			else if (isdif(p)) opStack[opTOS]='#',p++; 
			else if (isleq(p)) opStack[opTOS]='',p++; 
			else if (isgeq(p)) opStack[opTOS]='!',p++; 
			else if (issri(p)) opStack[opTOS]=']',p++; 
			else if (issle(p)) opStack[opTOS]='[',p++; 
			else if (istim(p)) opStack[opTOS]='^',p++; 
			else if (ismod(p)) opStack[opTOS]='%',p++; 
			else opStack[opTOS]=*p;
			prStack[opTOS]=priority(opStack[opTOS]);
			opTOS++;
			p++;
			opturn=0;
			/* operator followed by void is an illegal formula */
			if (!*p) perror_(__LINE__,37,l);
		}
		/* catch a closed parenthesis */
		else if (*p==')') break;
		/* catch the unary minus after an operator */
		else if (*p=='-' && !opturn) {
			p++;
			numStack[numTOS++]=-getVal();
			opturn=1;
		}
		/* catch the not as symbol after an operator */
		else if (*p=='~' && !opturn) {
			p++;
			numStack[numTOS++]=-(double)(!getVal());
			opturn=1;
		}
		/* catch the not as keyword after an operator */
		else if (!strncasecmp_(p,"NOT",3) && !opturn) {
			p+=3;
			numStack[numTOS++]=-(double)(!getVal());
			opturn=1;
		}
		/* catch the unary plus after an operator */
		else if (*p=='+' && !opturn) {
			p++;
			numStack[numTOS++]=getVal();
			opturn=1;
		}
		/* catch a number (literal or variable) */
		else {
			numStack[numTOS++]=getVal();
			opturn=1;
			if (*p && !isop(p) && opturn) break;
		}
		
		/* Since the string was evaluated, change pcopy to signal
		 * the work is done */
		pcopy=NIL;
	}

	/* No parsing has been done if the string is not void */
	if (pcopy==p) perror_(__LINE__,37,l);
	/* perform calculations from higher priorities downward
	 * Note: for tbas, any relation is also a math expression */
	for (ii=PRIOR;ii>=0;ii--) {
		/* calculations are performed from left to right */
		for (level=0;level<opTOS;level++) {
			if (prStack[level]==ii) {
				/* collapse stacks */
				switch(opStack[level]) {
// HERE BEGINS OPERATORS ANALYSIS
/* lesser than < */
case '<': numStack[level]=-(double)(numStack[level]<numStack[level+1]); break; 
/* greater than > */
case '>': numStack[level]=-(double)(numStack[level]>numStack[level+1]); break; 
/* not equal <> */
case '#': numStack[level]=-(double)(numStack[level]!=numStack[level+1]); break;
/* greater than or equal >= */
case '!': numStack[level]=-(double)(numStack[level]>=numStack[level+1]); break;
/* lesser than or equal <= */
case '': numStack[level]=-(double)(numStack[level]<=numStack[level+1]); break;
/* equal to */
case '=': numStack[level]=-(double)(numStack[level]==numStack[level+1]); break;
/* quasi-equal to */
case '': {
		  if (sign(numStack[level])!=sign(numStack[level+1]))
			numStack[level]=0;
		  else if (fabs((fabs(numStack[level]-fabs(numStack[level+1]))))<ZEROEQUAL)
			numStack[level]=-1;
		  else numStack[level]=0; 
	  }
	  break;
/* power ^ */
case '^': numStack[level]=power(numStack[level],numStack[level+1]); break;
/* subtraction - */
case '-': numStack[level]-=numStack[level+1]; break;
/* addition + */
case '+': numStack[level]+=numStack[level+1]; break;
/* division */
case '/': {
		double div=numStack[level+1];
		/* if divisor (div) is null, return INFF */
		if (fabs(div)<ZERO) {
			perror_(__LINE__,77,l);
			numStack[level]=numStack[level]<0?MINFF:INFF;
		}
		else numStack[level]/=div;
	  }
	  break;
/* integer division */
case '\\': {
		int div2=(int)numStack[level+1];
		int div1=(int)numStack[level];
		/* if divisor (div) is null, return INFF */
		if (abs(div2)<ZERO) {
			perror_(__LINE__,77,l);
			numStack[level]=numStack[level]<0?MINFF:INFF;
		}
		else numStack[level]=(double)((int)(div1/div2));  
	  }
	  break;
/* modulus operator */
case '%': {
		double o1=(int)numStack[level+1];
		double o=(int)numStack[level];
		/* if divisor (div) is null, return INFF */
		if (fabs(o1)<ZERO) {
			perror_(__LINE__,77,l);
			numStack[level]=numStack[level]<0?MINFF:INFF;
		}
		else numStack[level]=(double)(o-o1*(int)(o/o1));  
	  }
	  break;
/* multiplication * */
case '*': numStack[level]*=numStack[level+1]; break;
/* bitwise & */
case '&': numStack[level]=(int)numStack[level] & (int)numStack[level+1]; break;
/* bitwise | */
case '|': numStack[level]=(int)numStack[level] | (int)numStack[level+1]; break;
/* logical && AND */
case '': 
	numStack[level]=-(double)(!(!numStack[level])&&!(!numStack[level+1]));
	break;
/* logical || OR */
case '': 
	numStack[level]=-(double)(!(!numStack[level])||!(!numStack[level+1]));
	break;
/* logical !! XOR */
case '':
        if (!(!numStack[level]) != !(!numStack[level+1])) numStack[level]=-1;
	else numStack[level]=0;
	break;
/* logical -> IMP */
case '': 
        if (!(!numStack[level]) && !(!(!numStack[level+1]))) numStack[level]=0;
	else numStack[level]=-1;
	break;
/* logical ## EQV */
case '': 
        if (!(!numStack[level]) == !(!numStack[level+1])) numStack[level]=-1;
	else numStack[level]=0;
	break;
/* logical NOR */
case '{': 
	numStack[level]=-(double)!(!(!numStack[level])||!(!numStack[level+1]));
	break;
/* logical NAND */
case '}': 
	numStack[level]=-(double)!(!(!numStack[level])&&!(!numStack[level+1]));
	break;
/* ']' (shift right */
case ']': 
	numStack[level]=((int)numStack[level])>>((int)numStack[level+1]);
	break;
/* '[' (shift left */
case '[': 
	numStack[level]=((int)numStack[level])<<((int)numStack[level+1]);
	break;
// HERE ENDS OPERATORS ANALYSIS
				}
				/* shift stack positions */
				for (jj=level;jj<opTOS-1;jj++) {
					numStack[jj+1]=numStack[jj+2];
					opStack[jj]=opStack[jj+1];
					prStack[jj]=prStack[jj+1];
				}
				opTOS--;
				level--;
			}
		}
	}

	/* At this point, we have the final result into cell [0] */
	return numStack[0];
}


/* getVal()
 * get single value (number, variable, function or array element) 
 * the case of the unary minus for entities has been caught in S(); here is
 * treated the case of open parentheses
 * The double operators ++ and -- in the prefix notation are treated here
 * The address of starting position into the flow stream is pointed by p */
double getVal(void){
	double o=0;

	/* case of a real number */
	if (isnm(p)) o=strtod(p,&p);
	/* case of unary minus 
	 * (for general cases, e.g. before a parenthesis) */
	else if (*p=='-') {
		p++;
		o=-getVal();
	}
	/* case of ~ (tilde) - NOT operator
	 * (for general cases, e.g. before a parenthesis) */
	else if (*p=='~') {
		p++;
		o=-(double)(!getVal());
	}
	/* case of NOT
	 * (for general cases, e.g. before a parenthesis) */
	else if (!strncasecmp_(p,"NOT",3)) {
		p+=3;
		o=-(double)(!getVal());
	}
	/* case of open parenthesis */ 
	else if (*p=='(') { 
		p++;
		o=S();
		if (*p!=')') perror_(__LINE__,37,l);
		else p++; 
	}
	/* case of a string relation (which returns a numeric truth value) */
	else if (isanystring(p)) {
		numEval=TRUE;
		o=strRel();
		numEval=FALSE;
	}
	/* delegate function calculations to getFunc() */
	else if (isbeginnamechar(p)) { 
		o=getFunc();
		p++; 
	}
	/* unrecognized object */
	else perror_(__LINE__,37,l);
	
	return o;
}


/* getFunc()
 * functions calculator 
 * Note: I know trigonometry and hyperbolic functions since my High School days
 * and University years; in any case, for what about the organization of facts,
 * domain limits, graphs and alternative representations, I owe much  to the 
 * Wolfram|Alpha team guys (https://www.wolframalpha.com/), where a lot of info 
 * is freely available for anyone interested in math facts! 
 * Note: in the following:
 * - the symbol -> must be read as "belongs to",
 * - the symbol : means "such as" 
 * Note: the DEC BASIC had very powerful math functions, but they worked in
 * a very "strange" sense: if you tried to get SQR(-3) a warning appeared and
 * the SQR(3) was returned. This is not fair. Or, if you wanted to get LOG(-1)
 * a warning appeared, returning LOG(1).
 * This is not a good math behaviour, because I'd like, as a programmer, to
 * know if my algorithm has a fault and calculates wrong numbers, and the
 * warning, mixed to the output, may get unnoticed. 
 * On the other hand, if the function has an infinite for some domain values,
 * why stopping signaling the fact? In these cases, such limit may be output
 * as INFF, and this big big value is a substitute for the infinite value.
 * So I redesigned all math functions along these two principles:
 * 1. if the function goes to infinite for some values, INFF is returned, no 
 * error message is printed, with the convention that the sign of the INFF 
 * value depends on the sign of the function on the relative side (there may be 
 * cases where errors in representing the number cause shifts that fall in the 
 * wrong side, and INFF is printed with the opposite sign: a minor effect...)
 * 2. if the value of the function given as argument is out of dominion (for
 * instance ASIN(2) which is senseless), an error message is printed.
 * The previous rules define also the behavior of SQR and LOG, because negative
 * values fall in rule 2 (and an error message, not a warning, is printed), and
 * execution is stopped, while TAN(PI/2) falls in rule 1 (no error message is 
 * printed, INFF is returned) and execution is not stopped.*/
double getFunc(void) {
	double o=0.0, o1=0.0;
	int paren=TRUE, ns=0, type=0, rettype=0, cn=0;
	char strres[COMPSTR], *op;

	/* SUB FUNCTIONS - delegate execSUB() */
	if (!isValOn && (ns=findSUB(p,&op,&type,&rettype))) {
		if (type==FUNC && rettype==NUM)	{
			execSUB(ns,op,&o,strres);
		}
		else perror_(__LINE__,217,l);
		if (*p!=')') paren=FALSE;
	}
	/* FN (DEFFN structure) - delegate calcDEF() 
	 * Domain: depending on user function */
	else if (!isValOn && !strncasecmp_(p,"FN",2)) {
		p+=2;
		p=calcDEF();
		if (*p!=')') paren=FALSE;
		o=calcres;
	}
	
	/* MATH FUNCTIONS */ 

	/* ABS(x) absolute value function
	 * Domain: x -> R 
	 * Range:  y -> R : y >= 0 */
	else if (!strncasecmp_(p,"ABS(",4)) {
		p+=4;
		o=S();
		o=fabs(o);
	}
	/* ACOS(x) arc cosine function 
	 * Domain: x -> R : -1 <= x <= 1 
	 * Range:  y -> R : 0 <= y < pi (angular measure) */
	else if (!strncasecmp_(p,"ACOS(",5)) {
		p+=5;
		o=S();
		/* domain error */
		if (o<-1.0 || o>1.0) perror_(__LINE__,135,l);
		o=acos(o);
		if (!isRadians) o*=RATIO;
	}
	/* ACOSEC(x)/ACSC(x) arc cosecant function 
	 * Domain: x -> R - { ]-1,1[ }
	 * Range:  y -> R : -pi/2 < y < pi/2 (angular measure) */
	else if ((cn=recog(2,p,"ACOSEC(","ACSC("))) {
		p+=cn;
		o=S();
		/* domain error */
		if (o>-1 && o<1) perror_(__LINE__,137,l);
		o=asin(1.0/o);
		if (!isRadians) o*=RATIO;
	}
	/* ACOSECH(x)/ACSCH(x) hyperbolic arc cosecant function
	 * Domain: x -> R - { 0 }
	 * Range:  y -> R - { 0 }
	 * Note: case of x=0 is treated */
	else if ((cn=recog(2,p,"ACOSECH(","ACSCH("))) {
		p+=cn;
		o=S();
		if (fabs(o)<ZERO) o=(o<0.0?MINFF:INFF);
		else {
			o=1.0/o;
			o=log(o+sqrt(1.0+o*o));
		}
	}
	/* ACOSH(x) hyperbolic arc cosine function 
	 * Domain: x -> R : x >= 1
	 * Range:  y -> R : y >= 0 */
	else if (!strncasecmp_(p,"ACOSH(",6)) {
		p+=6;
		o=S();
		/* domain error */
		if (o<1.0) perror_(__LINE__,138,l);
		o=log(o+sqrt(o*o-1.0));
	}
	/* ACOT(x) arc cotangent function 
	 * Domain: x -> R 
	 * Range:  y -> R : 0 < y < pi (angular measure) */
	else if (!strncasecmp_(p,"ACOT(",5)) {
		p+=5;
		o=S();
		o=atan(1.0/o);
		if (o<0) o+=TBASPI;
		if (!isRadians) o*=RATIO;
	}
	/* ACOTH(x) hyperbolic arc cotangent function
	 * Domain: x -> R - { ]-1,1[ } 
	 * Range:  y -> R - { 0 } */
	else if (!strncasecmp_(p,"ACOTH(",6)) {
		p+=6;
		o=S();
		/* domain error */
		if (o>-1.0 && o<1.0) perror_(__LINE__,140,l);
		else if (o==-1.0) o=MINFF;
		else if (o==1.0) o=INFF;
		else o=0.5*log((o+1.0)/(o-1.0));
	}
	/* ASEC(x) arc secant function 
	 * Domain: x -> R - { ]-1,1[ }
	 * Range:  y -> R : 0 <= y < pi (angular measure) */
	else if (!strncasecmp_(p,"ASEC(",5)) {
		p+=5;
		o=S();
		/* domain error */
		if (o>-1.0 && o<1.0) perror_(__LINE__,136,l);
		o=acos(1.0/o);
		if (!isRadians) o*=RATIO;
	}
	/* ASECH(x) hyperbolic arc secant function 
	 * Domain: x -> R : 0 < x <= 1
	 * Range:  y -> R : y >= 0 
	 * Note: the cases x=0 and x=1 are treated */
	else if (!strncasecmp_(p,"ASECH(",6)) {
		p+=6;
		o=S();
		/* domain error */
		if (o<0.0 || o>1.0) perror_(__LINE__,141,l);
		else if (fabs(o)<ZERO) o=INFF; // o>0 here!
		else if (o==1.0) o=0.0;
		else {
			o=1.0/o;
			o=log(o+sqrt(o*o-1.0));
		}
	}
	/* ASIN(x) arc sin function 
	 * Domain: x -> R : -1 <= x <=1 
	 * Range:  y -> R : -pi/2 < y < pi/2 (angular measure) */
	else if (!strncasecmp_(p,"ASIN(",5)) {
		p+=5;
		o=S();
		/* domain error */
		if (o<-1.0 || o>1.0) perror_(__LINE__,134,l);
		o=asin(o);
		if (!isRadians) o*=RATIO;
	}
	/* ASINH(x) hyperbolic arc sin function 
	 * Domain: x -> R 
	 * Range:  y -> R */
	else if (!strncasecmp_(p,"ASINH(",6)) {
		p+=6;
		o=S();
		o=log(o+sqrt(1+o*o));
	}
	/* ATN(y[,x])|ATAN(y[,x]) arc tangent function (one or two arguments)
	 * From the manual page - man atan2:
	 * ~The atan2() function is used mostly to convert from rectangular 
	 * (x,y) to polar (r,theta) coordinates that must satisfy: 
	 * ~x = r*cos theta
	 * ~y = r*sin theta
	 * Note: In general, conversions to polar coordinates should be 
	 * computed like this:
	 * ~r := sqrt(x*x+y*y)
	 * ~theta := atan2(y,x)
	 * Domain:       x -> R
	 * Range: ATAN:  y -> R : -pi/2 < y < pi/2 (angular measure)  
	 * Range: ATAN2: y -> R : -pi < y < pi (angular measure) */
	else if ((cn=recog(2,p,"ATN(","ATAN("))) {
		p+=cn,o=S();
		/* atan2 case (sign of the two arguments decide the quadrant)*/
		if (*p==',') {
			p++;
			o1=S();
			o=atan2(o,o1);
		}
		/* atan case */
		else o=atan(o);
		if (!isRadians) o*=RATIO;
	}
	/* ATNH(x)|ATANH(x) hyperbolic arc tangent function 
	 * Domain: x -> R : -1 <= x <= 1
	 * Range:  y -> R 
	 * Note the cases x=1 and x=-1 are treated */
	else if ((cn=recog(2,p,"ATNH(","ATANH("))) {
		p+=cn;
		o=S();
		/* domain error */
		if (o<-1.0 || o>1.0) perror_(__LINE__,139,l);
		else if (o==-1.0) o=MINFF;
		else if (o==1.0) o=INFF;
		else o=0.5*log((1.0+o)/(1.0-o));
	}
	/* BIN(x$) binary conversion 
	 * ~return the decimal value of the binary number string
	 * in the argument */
	else if (!strncasecmp_(p,"BIN(",4)) {
		char op[COMPSTR], *bo;
		int n=0;
		p+=4;
		o=0.0;
		strncpy_(op,SS(),COMPSTR);
		bo=op+strlen(op)-1;
		while (bo>=op && (*bo=='1' || *bo=='0')) 
			o+=(*bo-48)*pow(2,n++), bo--;
		if (bo!=op) perror_(__LINE__,143,l); // arg not exhausted
	}
	/* CEIL(x)/CEILING(x) ceiling function 
	 * ~return the smallest following integer */
	else if ((cn=recog(2,p,"CEIL(","CEILING("))) {
		int k=0;
		p+=cn;
		o=S();
		if (o>0 && o!=(int)o) k=1;
		o=(int)o+k;  
	}
	/* COS(x) cosine function 
	 * Domain: x -> R (angular measure) 
	 * Range:  y -> R : -1 <= y <= 1 */
	else if (!strncasecmp_(p,"COS(",4)) {
		p+=4;
		o=S();
		if (!isRadians) o/=RATIO;
		o=cos(o);
	}
	/* COSEC(x)/CSC(x) cosecant function 
	 * Domain: x -> R - { pi/2+kpi } (angular measure) 
	 * Range:  y -> R - { ]-1,1[ } 
	 * Note: the case x=pi/2+kpi is treated after calculation of the sine 
	 * value, which must not be null */
	else if ((cn=recog(2,p,"COSEC(","CSC("))) {
		p+=cn;
		o=S();
		if (!isRadians) o/=RATIO;
		o=sin(o);
		/* Note: I check here against ZEROCALC=2E-13 and not ZERO 
		 * because the rendition of sin(pi) is never really zero, 
		 * since the argument is never really pi; so the absolute value 
		 * of sin(pi) is never lesser than ZERO, and the least value it
		 * can reach is around 2E-13 */
		if (fabs(o)<ZEROCALC) o=o<0.0?MINFF:INFF;
		else o=1.0/o;
	}
	/* COSECH(x)/CSCH(x) hyperbolic cosecant function
	 * Domain: x -> R - {0}
	 * Range:  y -> R - {0} 
	 * Note: the case x=0 is treated */
	else if ((cn=recog(2,p,"COSECH(","CSCH("))) {
		p+=cn;
		o=S();
	       	if (fabs(o)<ZERO) o=o<0.0?MINFF:INFF;
		else o=2.0/(exp(o)-exp(-o));
	}
	/* COSH(x) hyperbolic cosine function 
	 * Domain: x -> R
	 * Range:  y -> R : y >= 1 */
	else if (!strncasecmp_(p,"COSH(",5)) {
		p+=5;
		o=S();
		o=(exp(-o)+exp(o))/2.0;
	}
	/* COT(x) cotangent function 
	 * Domain: x -> R - { kpi } (angular measure) 
	 * Range:  y -> R 
	 * Note: the case x=kpi is treated */
	else if (!strncasecmp_(p,"COT(",4)) {
		p+=4;
		o=S();
		if (!isRadians) o/=RATIO;
		/* set value to the range 0-pi */
		while (o>TBASPI) o-=TBASPI;
		while (o<0) o+=TBASPI;
		if (fabs(o)<ZERO) o=INFF;
		else if (fabs(o-TBASPI)<ZERO) o=MINFF;
		else o=1.0/tan(o);
	}
	/* COTH(x) hyperbolic cotangent function
	 * Domain: x -> R - { 0 }
	 * Range:  y -> R - { ]-1,1[ } 
	 * Note: the case x=0 is treated */
	else if (!strncasecmp_(p,"COTH(",5)) {
		p+=5;
		o=S();
		if (fabs(o)<ZERO) o=o<0.0?MINFF:INFF;
		else o=(exp(o)+exp(-o))/(exp(o)-exp(-o));
	}
	/* CR (constant) return character of end-of-line \n */
	else if (!strncasecmp_(p,"CR",2) && !isalnum(*(p+2))) {
		p++;
		o=_CR;
		paren=FALSE;
	}
	/* DEG(x)/DEGREES(x) function - argument in radians */
	else if ((cn=recog(2,p,"DEG(","DEGREES("))) {
		p+=cn;
		o=S(); 
		o*=RATIO;
	}
	/* DET(x) function
	 * ~return the determinant of the square matrix in argument.
	 * Note: the matrix is internally inverted, but no attribution of
	 * the inversion is made to any variable; the process is done only
	 * to obtain a determinant. 
	 * Note: if matrix is singular, return 0.0 */
	else if (!strncasecmp_(p,"DET(",4)) {
		int ns=0;
		p+=4;
		detValue=0;
		forceArray=TRUE;
		ns=findDEC(p);
		if (!ns || _rettype!=NUM) perror_(__LINE__,128,l);
		if (!dim[ns].row || !dim[ns].col) perror_(__LINE__,128,l);
		if (dim[ns].row!=dim[ns].col) perror_(__LINE__,122,l);
		matinvert(0,ns); // 0 is the temp variable (not used by tbas)
		forceArray=FALSE;
		o=detValue;
		p+=_len;
	}
	/* DET variable 
	 * ~return the determinant of the last inverted matrix. If no matrix 
	 * was ever inverted, return 0.0  */
	else if (!strncasecmp_(p,"DET",3) && !isalnum(*(p+3))) {
		o=detValue;
		p+=2;
		paren=FALSE;
	}
	/* IDIV(x,y) integer division (Classic  Math) */
	else if (!strncasecmp_(p,"IDIV(",5)) {
		int A=0, B=0, div=0, mod=0;
		p+=5;
		A=(int)S(); // dividend
		if (*p!=',') perror_(__LINE__,35,l);
		else p++;
		B=(int)S(); // divisor
		modulo(A,B,&div,&mod,TRUE);
		o=(double)div;
	}
	/* DIV(x,y) integer division (Number Theory) */
	else if (!strncasecmp_(p,"DIV(",4)) {
		int A=0, B=0, div=0, mod=0;
		p+=4;
		A=(int)S(); // dividend
		if (*p!=',') perror_(__LINE__,35,l);
		else p++;
		B=(int)S(); // divisor
		modulo(A,B,&div,&mod,FALSE);
		o=(double)div;
	}
	/* DOT(A,B) dot product */
	else if (!strncasecmp_(p,"DOT(",4)) {
		p+=4;
		o=MATDOT();
	}
	/* EPS constant
	 * ~return the lowest number for this version, a number which is 
	 * at the lower edge of any interpretability */
	else if (!strncasecmp_(p,"EPS",3) && !isalnum(*(p+3))) {
		o=ZERO;
		p+=2;
		paren=FALSE;
	}
	/* ERFC(X) function
	 * ~return the complementary error function of X, computing the
	 * difference of the error function from 1.0; the error function is
	 * known as erf() in most languages, and its value may be obtained
	 * by 1-erfc().
	 * usage: internal testing (inner Pseudo-random Number Generator) */
	else if (!strncasecmp_(p,"ERFC(",5)) {
		p+=5,o=S(); // argument
		o=erfc(o);
		if (fabs(o)<ZERO) o=0.0;
	}
	/* EXP(x) exponential function (base e)
	 * Domain: x -> R 
	 * Range:  y -> R : y >= 0 */
	else if (!strncasecmp_(p,"EXP(",4)) {
		p+=4;
		o=S();
		o=exp(o);
	}
	/* EXPR(x$) function 
	 * ~return the value of the number or formula in the argument, 
	 * which must be a regular string */
	else if (!strncasecmp_(p,"EXPR(",5)) {
		char op[COMPSTR];
		p+=5;
		o=0.0;
		strncpy_(op,SS(),COMPSTR);
		/* Note: the only incorrect forms of the argument I can think 
		 * of is if it's null; if so, return 0.0 with a warning */
		if (!op[0]) perror_(__LINE__,131,l);
		/* in case it holds a number or a formula, turn string 
		 * to float using regular tbas rules */
		else {
			char *pp, pe[COMPSTR];
			stripBlanks(op,pe);
			pp=p;
			p=pe;
			o=S();
			if (*p) perror_(__LINE__,130,l);
			p=pp;
		}
	}
	/* FALSE constant
	 * ~return the false condition 0 */
	else if (!strncasecmp_(p,"FALSE",5) && !isalnum(*(p+5))) {
		o=FALSE;
		p+=4;
		paren=FALSE;
	}
	/* FLOOR(x) floor function 
	 * ~return the largest previous following integer */
	else if (!strncasecmp_(p,"FLOOR(",6)) {
		int k=0;
		p+=6;
		o=S();
		if (o<0.0 && (o-(int)o)!=0.0) k=1;
		o=(int)o-k;  
	}
	/* FRAC(x)/FP(x) factional part function 
	 * ~return fractional part of x */
	else if ((cn=recog(2,p,"FRAC(","FP("))) {
		p+=cn;
		o=S();
		o=o-(int)o;
	}
	/* GAMMA(x) function 
	 * ~return the gamma function value of x 
	 * Domain: x -> R
	 * Range:  y -> R */
	else if (!strncasecmp_(p,"GAMMA(",6)) {
		p+=6;
		o=S();
		o=tgamma(o);
	}
	/* HEX(x$) hexadecimal conversion 
	 * ~return the decimal value of the hex number string
	 * in the argument */
	else if (!strncasecmp_(p,"HEX(",4)) {
		char op[COMPSTR], *ho;
		p+=4;
		strncpy_(op,SS(),COMPSTR);
		o=strtol(op,&ho,16);
		if (*ho) perror_(__LINE__,143,l); 
	}
	/* TRUE constant
	 * ~return the true condition */
	else if (!strncasecmp_(p,"TRUE",4) && !isalnum(*(p+4))) {
		o=TRUE;
		p+=3;
		paren=FALSE;
	}
	/* INF constant
	 * ~return the larger number for this version, a number which is 
	 * at the higher edge of any interpretability */
	else if (!strncasecmp_(p,"INF",3) && !isalnum(*(p+3))) {
		o=INFF;
		p+=2;
		paren=FALSE;
	}
	/* INT(x)/IP(x) integer function
	 * ~return the integer part
	 * Domain: x -> R 
	 * Range:  y -> N */
	else if ((cn=recog(2,p,"INT(","IP("))) {
		p+=cn;
		o=(int)S(); 
	}
	/* INV(x) invert function
	 * ~return the ratio 1/x
	 * Domain: x -> R - {0}
	 * Range:  y -> R 
	 * Note: the case x=0 is treated */
	else if (!strncasecmp_(p,"INV(",4)) {
		p+=4;
		o=S();
		if (fabs(o)<ZERO) {
			perror_(__LINE__,77,l);
			o=o<0?MINFF:INFF;
		}
		else o=1.0/o;
	}
	/* LASTF constant
	 * ~return the last DEF calculation result */
	else if (!strncasecmp_(p,"LASTF",5) && !isalnum(*(p+5))) {
		o=calcres;
		p+=4;
		paren=FALSE;
	}
	/* LEN(x$)|LENGTH(x$) string length function 
	 * ~return length of the argument string */
	else if ((cn=recog(2,p,"LEN(","LENGTH("))) {
		char lop[COMPSTR];
		p+=cn;
		strncpy_(lop,SS(),COMPSTR);
		o=(double)strlen(lop);
	}
	/* LGAMMA(x) function 
	 * ~return the natural logarithm of the absolute value of the gamma 
	 * function of value x
	 * Domain: x -> R
	 * Range:  y -> R */
	else if (!strncasecmp_(p,"LGAMMA(",7)) {
		p+=7;
		o=S();
		o=lgamma(o);
	}
	/* LOGE(x)|LOG(x)|LN(x) natural logarithm function
	 * ~return the natural logarithm
	 * Domain: x -> R : x > 0 
	 * Range:  y -> R 
	 * Note: the case x=0 is treated */
	else if ((cn=recog(3,p,"LN(","LOG(","LOGE("))) {
		p+=cn;
		o=S();
		/* domain errors (negative and equal to zero) */
		if (o<0) perror_(__LINE__,95,l); 
		else if (o<ZERO) { // o>0 here!
			perror_(__LINE__,96,l);
			o=MINFF; 
		}
		else o=log(o);
	}
	/* LOG10(x)|CLOG(x)|LGT(x) 10-base logarithm function 
	 * ~return the logarithm in base 10
	 * Domain: x -> R : x > 0 
	 * Range:  y -> R
	 * Note: LGT comes from the CDC tradition 
	 * Note: the case x=0 is treated */
	else if ((cn=recog(3,p,"LGT(","LOG10(","CLOG("))) {
		p+=cn;
		o=S();
		/* domain errors (negative and equal to zero) */
		if (o<0) perror_(__LINE__,95,l); 
		else if (o<ZERO) { // o>0 here!
			perror_(__LINE__,96,l);
			o=MINFF; 
		}
		else o=log10(o);
	}
	/* MAX(x,y,z...) maximum value of a list */
	else if (!strncasecmp_(p,"MAX(",4)) {
		double o2;
		p+=4, o=0.0;
		while (*p!=')') {
			o2=S();
			if (o2>o) o=o2;
			if (*p==',') p++;
		}
	}
	/* VMAX(x) maximum value of a vector/matrix */
	else if (!strncasecmp_(p,"VMAX(",5)) {
		double o2;
	       	int cn=0;
		p+=5, o=0.0;
		forceArray=TRUE;
		cn=findDEC(p);
		forceArray=FALSE;
		if (!cn && vmaxLast) cn=vmaxLast;
		else vmaxLast=cn;
		/* invalid cases */
		if (cn && vn[cn].rettype==STR) perror_(__LINE__,215,l);
		else if (cn && !vn[cn].isArr) perror_(__LINE__,34,l);
		/* max of an array */
		else if (cn) {
			int i,j,m=dim[cn].row, n=dim[cn].col;
			p+=_len;
			for (i=0;i<=m;i++) {
				for (j=0;j<=n;j++) {
					o2=getArrayVal(cn,dim[cn].type,i,j);
					if (o2>o) o=o2;
				}
			}
		}
		else o=0;
	}
	/* MAYBE - random TRUE or FALSE */
	else if (!strncasecmp_(p,"MAYBE",5) && !isalnum(*(p+5))) {
		p+=4;
		o=getRandom();
		if (o<0.5) o=FALSE;
		else o=TRUE;
		paren=FALSE;
	}
	/* MIN(x,y,z...) minimum value of a list */
	else if (!strncasecmp_(p,"MIN(",4)) {
		double o2;
		p+=4, o=0.0;
		while (*p!=')') {
			o2=S();
			if (o2<o) o=o2;
			if (*p==',') p++;
		}
	}
	/* VMIN(x) minimum value of a vector/matrix */
	else if (!strncasecmp_(p,"VMIN(",5)) {
		double o2;
	       	int cn=0;
		p+=5, o=0.0;
		forceArray=TRUE;
		cn=findDEC(p);
		forceArray=FALSE;
		if (!cn && vminLast) cn=vminLast;
		else vminLast=cn;
		/* invalid cases */
		if (cn && vn[cn].rettype==STR) perror_(__LINE__,215,l);
		else if (cn && !vn[cn].isArr) perror_(__LINE__,34,l);
		/* min of an array */
		else if (cn){
			int i,j,m=dim[cn].row, n=dim[cn].col;
			p+=_len;
			for (i=0;i<=m;i++) {
				for (j=0;j<=n;j++) {
					o2=getArrayVal(cn,dim[cn].type,i,j);
					if (o2<o) o=o2;
				}
			}
		}
		else o=0;
	}
	/* MOD(x,y) integer modulus (Classic Math) */
	else if (!strncasecmp_(p,"MOD(",4)) {
		int A=0, B=0, div=0, mod=0;
		p+=4;
		A=(int)S(); // dividend
		if (*p!=',') perror_(__LINE__,43,l);
		else p++;
		B=(int)S(); // divisor
		modulo(A,B,&div,&mod,TRUE);
		o=mod;
	}
	/* REMAINDER(x,y) integer modulus (Number Theory) */
	else if (!strncasecmp_(p,"REMAINDER(",10)) {
		int A=0, B=0, div=0, mod=0;
		p+=10;
		A=(int)S(); // dividend
		if (*p!=',') perror_(__LINE__,43,l);
		else p++;
		B=(int)S(); // divisor
		modulo(A,B,&div,&mod,FALSE);
		o=mod;
	}
	/* NUM(x)
	 * ~MAT INPUT function with dummy argument */
	else if (!strncasecmp_(p,"NUM(",4)) { 
		p+=4;
		o=S(); // dummy
		o=numValue; 
	}
	/* NUM
	 * ~MAT INPUT variable */
	else if (!strncasecmp_(p,"NUM",3) && !isalnum(*(p+3))) {
		o=numValue;
		p+=2;
		paren=FALSE;
	}
	/* OCT(x$) octal conversion 
	 * ~return the decimal value of the octal number string
	 * in the argument */
	else if (!strncasecmp_(p,"OCT(",4)) {
		char op[COMPSTR], *ho;
		p+=4;
		strncpy_(op,SS(),COMPSTR);
		o=strtol(op,&ho,8);
		if (*ho) perror_(__LINE__,143,l); 
	}
	/* PI (greek PI) constant */
        else if (!strncasecmp_(p,"PI",2) && !isalnum(*(p+2))) {
		p++;
		o=TBASPI;
		if (!isRadians) o*=RATIO;
		paren=FALSE;
	}
	/* RAD(x)/RADIANS(x) radians function
	 * ~convert x to radians (argument in degrees) */
	else if ((cn=recog(2,p,"RAD(","RADIANS("))) {
		p+=cn;
		o=S(); 
		o/=RATIO;
	}
	/* RND(x) random function (dummy argument) 
	 * ~return a random number */
        else if (!strncasecmp_(p,"RND(",4)) {
	       p+=4;
	       o=S();
	       o=getRandom();
	}
	/* RND random variable 
	 * ~return a random number */
        else if (!strncasecmp_(p,"RND",3) && !isalnum(*(p+3))) {
		o=getRandom();
		p+=2;
		paren=FALSE;
	}
	/* ROUND(x[,y])|ROF(x[,y])|FIX(x[,y]) round function
	 * ~round x to integer (rounding last digit according to first decimal)
	 * or round x to yth decimal place, round last digit according to the
	 * digit in y+1 position
	 * Note: ROF comes from the CDC tradition */
        else if ((cn=recog(3,p,"ROF(","FIX(","ROUND("))) {
		double v;
		int r=0;
		p+=cn;
		v=S();
		/* no second argument: round to integer */
		if (*p==')') o=(int)(v+(v>0?0.5:-0.5));
		/* specified number of decimals to be rounded */
		else if (*p==',') {
			p++; // skip comma
			r=(int)S();
			o=(int)(v*pow(10.0,r)+(v>0?0.5:-0.5))/pow(10.0,r);
		}
		else perror_(__LINE__,37,l);
	}
	/* SEC(x) secant function
	 * Domain: x -> R - { kpi } (angular measure) 
	 * Range:  y -> R -{ ]-1,1[ } 
	 * Note: the case x=kpi is treated */
	else if (!strncasecmp_(p,"SEC(",4)) {
		p+=4;
		o=S();
		if (!isRadians) o/=RATIO;
		o=cos(o);
		/* Note: I check here against ZEROCALC=2E-13 and not ZERO 
		 * because the rendition of cos(pi/2) is never really zero, 
		 * since the argument is never really pi/2 */
		if (fabs(o)<ZEROCALC) o=o<0?MINFF:INFF;
		else o=1.0/o;
	}
	/* SECH(x) hyperbolic secant function 
	 * Domain: x -> R 
	 * Range:  y -> R : 0 <= y <= 1 */
	else if (!strncasecmp_(p,"SECH(",5)) {
		p+=5;
		o=S();
		o=2.0/(exp(o)+exp(-o));
	}
	/* SGN(x)/SIGN(x) sign function 
	 * Domain: x -> R
	 * Range   y -> N, : y = -1 or 0 or 1 */
	else if ((cn=recog(2,p,"SGN(","SIGN("))) {
		p+=cn;
		o=sign(S()); 
	}
	/* SHL(n,m) shift left function (2 parameters) */
	else if (!strncasecmp_(p,"SHL(",4)) {
		int a,b;
		p+=4;
		a=(int)S();
		if (*p!=',') perror_(__LINE__,43,l);
		else p++;
		b=(int)S();
		o=(double)(a<<b);
	}
	/* SHR(n,m) shift right function (2 parameters) */
	else if (!strncasecmp_(p,"SHR(",4)) {
		int a,b;
		p+=4;
		a=(int)S();
		if (*p!=',') perror_(__LINE__,43,l);
		else p++;
		b=(int)S();
		o=(double)(a>>b);
	}
	/* SIN(x) sin function 
	 * Domain: x -> R (angular measure) 
	 * Range:  y -> R : -1 <= y <= 1 */
	else if (!strncasecmp_(p,"SIN(",4)) {
		p+=4;
		o=S();
		if (!isRadians) o/=RATIO;
		o=sin(o);
	}
	/* SINH(x) hyperbolic sin function 
	 * Domain: x -> R 
	 * Range:  y -> R */
	else if (!strncasecmp_(p,"SINH(",5)) {
		p+=5;
		o=S();
		o=(-exp(-o)+exp(o))/2;
	}
	/* SQRT(x)|SQR(x) square root function 
	 * Domain: x -> R : x >= 0 
	 * Range:  y -> R : y >= 0 */ 
	else if ((cn=recog(2,p,"SQR(","SQRT("))) {
		p+=cn;
		o=S();
		/* domain error */
		if (o<0.0) perror_(__LINE__,123,l); 
		else o=sqrt(o);
	}
	/* SUM(x) summation of list terms */
	else if (!strncasecmp_(p,"SUM(",4)) {
		p+=4, o=0.0;
		while (*p!=')') {
			o+=S();
			if (*p==',') p++;
		}
	}
	/* VSUM(x) summation of a vector/matrix items */
	else if (!strncasecmp_(p,"VSUM(",5)) {
	       	int cn=0;
		p+=5, o=0.0;
		forceArray=TRUE;
		cn=findDEC(p);
		forceArray=FALSE;
		if (!cn && vsumLast) cn=vsumLast;
		else vsumLast=cn;
		/* invalid cases */
		if (cn && vn[cn].rettype==STR) perror_(__LINE__,215,l);
		else if (cn && !vn[cn].isArr) perror_(__LINE__,34,l);
		/* sum of array items */
		else if (cn) {
			int i,j,m=dim[cn].row, n=dim[cn].col;
			p+=_len;
			for (i=0;i<=m;i++) {
				for (j=0;j<=n;j++) {
					o+=getArrayVal(cn,dim[cn].type,i,j);
				}
			}
		}
		else o=0;
	}
	/* TAN(x) tangent function 
	 * Domain: x -> R - { pi/2+kpi } (angular measure)
	 * Range:  y -> R 
	 * Note: the case x=pi/2+kpi is treated */
	else if (!strncasecmp_(p,"TAN(",4)) {
		p+=4;
		o=S();
		if (!isRadians) o/=RATIO;
		/* set value to the range -pi/2 +pi/2 */
		while (o>HALFPI) o-=TBASPI;
		while (o<-HALFPI) o+=TBASPI;
		if (fabs(fabs(o)-HALFPI)<ZEROCALC) o=o<0?MINFF:INFF;
		else o=tan(o);
	}
	/* TANH(x) hyperbolic tangent function
	 * Domain: x -> R
	 * Range:  y -> R : -1 < y < 1 */
	else if (!strncasecmp_(p,"TANH(",5)) {
		p+=5;
		o=S();
		o=(exp(o)-exp(-o))/(exp(o)+exp(-o));
	}
	/* TEN(x) ten power function
	 * ~tenth power function (base 10)
	 * Domain: x -> R 
	 * Range:  y -> R : y >= 0 */
	else if (!strncasecmp_(p,"TEN(",4)) {
		p+=4;
		o=S();
		o=pow(10.0,o);
	}
	/* VAL(x$) function 
	 * ~return the value of the number or ground formula in the argument, 
	 * which must be a regular string */
	else if (!strncasecmp_(p,"VAL(",4)) {
		char op[COMPSTR];
		p+=4;
		o=0.0;
		strncpy_(op,SS(),COMPSTR);
		/* Note: the only incorrect forms of the argument I can think 
		 * of is if it's null; if so, return 0.0 with a warning */
		if (!op[0]) perror_(__LINE__,131,l);
		/* in case it holds a number or a ground formula, turn string 
		 * to float using tbas rules 
		 * Note: isValOn is a flag that forces interpretation of
		 * string as a ground formula (no variables) */
		else {
			char *pp, pe[COMPSTR];
			stripBlanks(op,pe);
			pp=p;
			isValOn=TRUE;
			p=pe;
			o=S();
			if (*p) perror_(__LINE__,130,l);
			isValOn=FALSE;
			p=pp;
		}
	}
	/* DIM-RELATED FUNCTIONS */
	/* LBOUND()/LDIM() function
	 * ~return the lowest index */
	else if ((cn=recog(2,p,"LBOUND(","LDIM("))) {
		p+=cn;
		o=getBound(FALSE);
	}
	/* UBOUND()/UDIM() function
	 * ~return the highest index */
	else if ((cn=recog(2,p,"UBOUND(","UDIM("))) {
		p+=cn;
		o=getBound(TRUE);
	}
	/* TIME-RELATED FUNCTIONS */
	/* CLK/CLK(d) variable or function with dummy argument
	 * ~return hour in the format hh.nnnn (hour plus minutes and seconds
	 * in a percentage as decimals)
	 * ~es. 15h 26' 34" = 15.4428 (= 44% of the 15th hour) */
	else if (!strncasecmp_(p,"CLK",3) && !isalnum(*(p+3))) {
		p+=3;
		if (*p=='(') {
			p++;
			o=S(); // dummy
		}
		else p--, paren=FALSE;
		o=(getMin()*60.0+getSec())/3600.0+getHour();
	}	
	/* HOUR/HOUR() function 
	 * ~if argument is absent or 0, return current hour
	 * ~if argument is in the decimal format hh.nnnn, extract hour  */
	else if (!strncasecmp_(p,"HOUR",4) && !isalnum(*(p+4))) {
		p+=4;
		o=0.0;
		if (*p=='(') {
			p++;
			o=S();
		}
		else p--, paren=FALSE;
		if (o==0.0) o=getHour();
		else o=(int)o;		// int
	}
	/* MINUTE/MINUTE() function 
	 * ~if argument is absent or 0, return current minute
	 * ~if argument is in the decimal format hh.nnnn, extract minutes  */
	else if (!strncasecmp_(p,"MINUTE",6) && !isalnum(*(p+6))) {
		p+=6;
		o=0.0;
		if (*p=='(') {
			p++;
			o=S();
		}
		else p--, paren=FALSE;
		if (o==0.0) o=getMin();
		else {
			o-=(int)o;	// frac
			o=(int)(60*o);	
		}
	}
	/* SECOND/SECOND() function 
	 * ~if argument is absent or 0, return current second
	 * ~if argument is in the decimal format hh.nnnn, extract seconds  */
	else if (!strncasecmp_(p,"SECOND",6) && !isalnum(*(p+6))) {
		p+=6;
		o=0.0;
		if (*p=='(') {
			p++;
			o=S();
		}
		else p--, paren=FALSE;
		if (o==0.0) o=getSec();
		else {
			int m;
			o-=(int)o; 	// frac
			m=(int)(60*o); 	// min
			o=(int)(3600.0*(o-m/60.0)+0.5);
		}
	}
	/* DATE/DATE(d) variable or function with dummy value
	 * ~return current date in the decimal format yy.mmdd */
	else if (!strncasecmp_(p,"DATE",4) && !isalnum(*(p+4))) {
		int y,m,d;
		y=getYear();
		p+=4;
		if (*p=='(') {
			p++;
			o=S(); // dummy
		}
		else p--, paren=FALSE;
		/* retain year's last two digits only */
		while (y>1000) y-=1000;
		while (y>100) y-=100;
		m=getMonth();
		d=getDay();
		o=y+m/100.0+d/10000.0;
	}	
	/* TIM/TIM(d) variable or function with dummy value 
	 * ~'begin' time is calculated in RUN() with the first program 
	 * (i.e. out of a CHAIN sequence) 
	 * Note: TIM returns the elapsed execution time in milliseconds up to 
	 * the TIM call, including all the programs (in the CHAIN) met so far */
        else if (!strncasecmp_(p,"TIM",3) && !isalnum(*(p+3))) { 
		double now;
		p+=3;
        	gettimeofday(&tv, NULL);
           	now = ((double)tv.tv_sec + tv.tv_usec / 1000000.0);
		if (*p=='(') {
			p++;
			o=S(); // dummy
		}
		else p--, paren=FALSE;
		o=((int)(1000*(now-beginS)+0.5)); // milliseconds
	}
	/* DAY/DAY() function 
	 * ~if argument is absent or 0, return current day
	 * ~if argument is in the format yy.mmdd or yyyy.mmdd, extract day */
	else if (!strncasecmp_(p,"DAY",3) && !isalnum(*(p+3))) {
		p+=3;
		o=0.0;
		if (*p=='(') {
			p++;
			o=S();
		}
		else p--, paren=FALSE;
		if (o==0.0) o=getDay();
		else {
			o*=100;
			o-=(int)o;
			o=(int)(o*100.0+0.5);
		}
	}
	/* MONTH/MONTH() function 
	 * ~if argument is absent or 0, return current month
	 * ~if argument is in the format yy.mmdd or yyyy.mmdd, extract month */
	else if (!strncasecmp_(p,"MONTH",5) && !isalnum(*(p+5))) {
		p+=5;
		o=0.0;
		if (*p=='(') {
			p++;
			o=S();
		}
		else p--, paren=FALSE;
		if (o==0.0) o=getMonth();
		else {
			o-=(int)o;
			o=(int)(o*100);
		}
	}
	/* YEAR/YEAR() function 
	 * ~if argument is absent or 0, return current year
	 * ~if argument is in the format yy.mmdd or yyyy.mmdd, extract year 
	 * for current century 
	 * ~year is returned in yyyy format */
	else if (!strncasecmp_(p,"YEAR",4) && !isalnum(*(p+4))) {
		p+=4;
		o=0.0;
		if (*p=='(') {
			p++;
			o=S();
		}
		else p--, paren=FALSE;
		o=(int)setYear(o);
	}
	/* STRING-RELATED FUNCTIONS */
	/* ASC(x)|ASCII(x)/ORD(x) function 
	 * ~return ASCII code of character argument; argument can be:
	 * ~a label identifying a standard ASCII name
	 * ~a character only in the range 0-127
	 * ~a string in any form (enclosed into double quotes, a variable,
	 * a string function): in this case the first character of the string 
	 * is returned.
	 * Note: the ASCII table represented by ASC() is in the range 0-255 */
	else if ((cn=recog(3,p,"ASC(","ASCII(","ORD("))) {
		p+=cn;
		/* void argument */
		if (*p==')') o=32; // ASC() = ASC( )
		else if (!strncasecmp_(p,"NUL",3)) o=0,p+=3;
		else if (!strncasecmp_(p,"SOH",3)) o=1,p+=3;
		else if (!strncasecmp_(p,"STX",3)) o=2,p+=3;
		else if (!strncasecmp_(p,"ETX",3)) o=3,p+=3;
		else if (!strncasecmp_(p,"EOT",3)) o=4,p+=3;
		else if (!strncasecmp_(p,"ENQ",3)) o=5,p+=3;
		else if (!strncasecmp_(p,"ACK",3)) o=6,p+=3;
		else if (!strncasecmp_(p,"BEL",3)) o=7,p+=3;
		else if (!strncasecmp_(p,"BS",2)) o=8,p+=2;
		else if (!strncasecmp_(p,"HT",2)) o=9,p+=2;
		else if (!strncasecmp_(p,"LF",2)) o=10,p+=2;
		else if (!strncasecmp_(p,"VT",2)) o=11,p+=2;
		else if (!strncasecmp_(p,"FF",2)) o=12,p+=2;
		else if (!strncasecmp_(p,"CR",2)) o=13,p+=2;
		else if (!strncasecmp_(p,"SO",2)) o=14,p+=2;
		else if (!strncasecmp_(p,"SI",2)) o=15,p+=2;
		else if (!strncasecmp_(p,"DLE",3)) o=16,p+=3;
		else if (!strncasecmp_(p,"DC1",3)) o=17,p+=3;
		else if (!strncasecmp_(p,"DC2",3)) o=18,p+=3;
		else if (!strncasecmp_(p,"DC3",3)) o=19,p+=3;
		else if (!strncasecmp_(p,"DC4",3)) o=20,p+=3;
		else if (!strncasecmp_(p,"NAK",3)) o=21,p+=3;
		else if (!strncasecmp_(p,"SYN",3)) o=22,p+=3;
		else if (!strncasecmp_(p,"ETB",3)) o=23,p+=3;
		else if (!strncasecmp_(p,"CAN",3)) o=24,p+=3;
		else if (!strncasecmp_(p,"EM",2)) o=25,p+=2;
		else if (!strncasecmp_(p,"SUB",3)) o=26,p+=3;
		else if (!strncasecmp_(p,"ESC",3)) o=27,p+=3;
		else if (!strncasecmp_(p,"FS",2)) o=28,p+=2;
		else if (!strncasecmp_(p,"GS",2)) o=29,p+=2;
		else if (!strncasecmp_(p,"RS",2)) o=30,p+=2;
		else if (!strncasecmp_(p,"US",2)) o=31,p+=2;
		else if (!strncasecmp_(p,"SP",2)) o=32,p+=2;
		else if (!strncasecmp_(p,"DEL",3)) o=127,p+=3;
		else if (*(p+1)==')') o=*p++; // any single char
		else if (isanystring(p)) {
			char str[COMPSTR];
			strncpy_(str,SS(),COMPSTR);
			o=*str;
			if (o<0) o+=256;
		}
		else if ((cn=findDEC(p)) && _rettype==STR && !_arr) {
			char str[COMPSTR];
			strncpy_(str,SS(),COMPSTR);
			o=*str;
			if (o<0) o+=256;
		}
		else perror_(__LINE__,32,l);
		if (*p!=')') perror_(__LINE__,65,l,makechar(p),")");
	}
	/* INSTR([n,]a$,b$) function 
	 * ~return position of string b$ into string a$ (starting at
	 * position n if specified, or else at first character of a$) */
	else if (!strncasecmp_(p,"INSTR(",6)) {
		int pos=0;
		char op1[COMPSTR], op2[COMPSTR];
		char *P1;
		p+=6;
		/* the starting position has been set */
		if (!isanystring(p)) {
			pos=(int)S()-1;
			if (*p==',') p++;
			else perror_(__LINE__,43,l);
		}
		/* first string is the base string */
		strncpy_(op1,SS(),COMPSTR);
		if (pos<0) pos=0;
		if (pos>(int)strlen(op1)) pos=(int)strlen(op1);
		/* set starting position */
		P1=op1+pos;
		if (*p==',') p++; else perror_(__LINE__,37,l);
		/* second string is the search string */
		strncpy_(op2,SS(),COMPSTR);
		o=(double)(strcasestr_(P1,op2)-op1)+1;
		if (o<0 || o>(int)strlen(op1)) o=0.0;
	}
	/* POS(a$,b$[,n]) function (Dartmouth and HP BASIC tradition)
	 * ~return position of string b$ into string a$ (starting at
	 * position n if specified, or else at first character of a$) */
	else if (!strncasecmp_(p,"POS(",4)) {
		int pos=0;
		char op1[COMPSTR], op2[COMPSTR];
		char *P1;
		p+=4;
		/* first string is the base string */
		strncpy_(op1,SS(),COMPSTR);
		if (*p==',') p++; else perror_(__LINE__,43,l);
		/* second string is the search string */
		strncpy_(op2,SS(),COMPSTR);
		/* the starting position has been set */
		if (*p==',') {
			p++;
			pos=(int)S()-1;
		}
		/* set starting position*/
		if (pos<0) pos=0;
		if (pos>(int)strlen(op1)) pos=(int)strlen(op1);
		P1=op1+pos;
		o=(double)(strcasestr_(P1,op2)-op1)+1;
		if (o<0 || o>(int)strlen(op1)) o=0.0;
	}
	/* FILE-RELATED FUNCTIONS */
	/* EXISTS() function
	 * checks if a file exists and what permissions it has
	 * Note: only read and write permissions are given here, because
	 * it's supposed that data file written or read by tbas are not
	 * executable but only text files or random access files */
	else if (!strncasecmp_(p,"EXISTS(",7)) {
		char op[COMPSTR];
		p+=7;
		strncpy_(op,SS(),COMPSTR);
		o=(double)doesExist(op);
	}
	/* EOF(#x)/EOF(:x)/EOF(x)/EOFx/EOF#x/EOF:x
	 * END(#x)/END(:x)/END(x)/ENDx/END#x/END:x
	 * ~return true if EOF of the file on channel x is reached */
	else if ((cn=recog(2,p,"EOF","END")) && !isalnum(*(p+cn))) {
		int ch, type=SEQ;
		p+=cn; paren=FALSE;
		if (*p=='#') p++, type=SEQ; // case 5
		else if (*p==':') p++, type=RANDOM; // case 6
		else if (*p=='(') {
			p++;
			if (*p=='#') p++, type=SEQ; // case 2
			else if (*p==':') p++, type=RANDOM; // case 3
			paren=TRUE;
		}
		ch=(int)S();
		if (!paren) p--;
		/* test for sequential file*/
		if (channel[ch].type==SEQ) {
			/* check syntax */
			if (type && type==RANDOM) perror_(__LINE__,101,l); 
			/* checking if channel is valid */
			if (ch<1 || ch>STREAMS) perror_(__LINE__,133,l);
			/* checking if file is open */
			if (!channel[ch].isopen) perror_(__LINE__,80,l);
			if (channel[ch].wayout!=VREAD) perror_(__LINE__,85,l);
		}
		/* test for random-access file */
		else if (channel[ch].type==RANDOM) {
			/* check syntax */
			if (type && type==SEQ) perror_(__LINE__,101,l); 
			/* checking if channel is valid */
			if (ch<1 || ch>STREAMS) perror_(__LINE__,133,l);
			/* checking if file is open */
			if (!channel[ch].isopen) perror_(__LINE__,80,l);
		}
		/* unrecognized type */
		else perror_(__LINE__,104,l);
		/* check EOF */
		o=(double)testEOF(ch); 
	}
	/* MORE(x)/MORE(#x)/MORE(:x)/MOREx/MORE#x/MORE:x
	 * ~return true if EOF of the file on channel x is not reached */
	else if (!strncasecmp_(p,"MORE",4) && !isalnum(*(p+4))) {
		int ch, type=SEQ;
		p+=4; paren=FALSE;
		if (*p=='#') p++, type=SEQ; // case 5
		else if (*p==':') p++, type=RANDOM; // case 6
		else if (*p=='(') {
			p++;
			if (*p=='#') p++, type=SEQ; // case 2
			else if (*p==':') p++, type=RANDOM; // case 3
			paren=TRUE;
		}
		ch=(int)S();
		if (!paren) p--;
		/* test for sequential file*/
		if (channel[ch].type==SEQ) {
			/* check syntax */
			if (type && type==RANDOM) perror_(__LINE__,101,l); 
			/* checks if channel is valid */
			if (ch<1 || ch>STREAMS) perror_(__LINE__,133,l);
			/* checks if file is open */
			if (!channel[ch].isopen) perror_(__LINE__,80,l);
			if (channel[ch].wayout!=VREAD) perror_(__LINE__,85,l);
		}
		/* test for random-access file */
		else if (channel[ch].type==RANDOM) {
			/* check syntax */
			if (type && type==SEQ) perror_(__LINE__,101,l); 
			/* checks if channel is valid */
			if (ch<1 || ch>STREAMS) perror_(__LINE__,133,l);
			/* checks if file is open */
			if (!channel[ch].isopen) perror_(__LINE__,80,l);
		}
		/* unrecognized type */
		else perror_(__LINE__,104,l);
		/* check EOF */
		o=-(double)(!testEOF(ch)); 
	}
	/* LOC([:]x) function - (LOCation)
	 * ~return the number of the current record in the random file
	 * on the specified channel */
	else if (!strncasecmp_(p,"LOC",3) && !isalnum(*(p+3))) {
		int ch;
		p+=3;
		if (*p==':') p++;
		else if (*p=='(') {
			p++;
			if (*p==':') p++;
			paren=TRUE;
		}
		ch=(int)S();
		if (!paren) p--;
		/* check if file is open and random */
		if (ch<1 || ch>STREAMS) perror_(__LINE__,133,l);
		if (!channel[ch].isopen) perror_(__LINE__,80,l);
		if (channel[ch].type!=RANDOM) perror_(__LINE__,101,l);
		o=channel[ch].currseek;
	}
	/* LOF([:]x) function - (Last Of File)
	 * ~return the number of the last record in the random file
	 * on the specified channel */
	else if (!strncasecmp_(p,"LOF",3) && !isalnum(*(p+3))) {
		int ch;
		p+=3;
		if (*p==':') p++;
		else if (*p=='(') {
			p++;
			if (*p==':') p++;
			paren=TRUE;
		}
		ch=(int)S();
		if (!paren) p--;
		/* check if file is open and random */
		if (ch<1 || ch>STREAMS) perror_(__LINE__,133,l);
		if (!channel[ch].isopen) perror_(__LINE__,80,l);
		if (channel[ch].type!=RANDOM) perror_(__LINE__,101,l);
		o=getEOF(ch);
	}
	/* ERROR-RELATED FUNCTIONS */
	/* ESM/ESM()/ERR/ERR() with (): dummy argument
	 * Error State Message / Error
	 * ~return error message code */
	else if ((cn=recog(2,p,"ESM","ERR")) && !isalnum(*(p+cn))) {
		p+=cn;
		if (*p=='(') {
			p++;
			o=S(); // dummy
		}
		else p--, paren=FALSE;
		o=errValue;
	}
	/* ASL/ASL() with (): dummy argument
	 * Attention State Line
	 * ~return line number where interrupt occurred */
	else if (!strncasecmp_(p,"ASL",3) && !isalnum(*(p+3))) {
		p+=3;
		if (*p=='(') {
			p++;
			o=S(); // dummy
		}
		else p--, paren=FALSE;
		o=arrLine;
	}
	/* ESL/ESL()/ERL/ERL() with (): dummy argument
	 * Error State Line / Error Line
	 * ~return line number where error occurred */
	else if ((cn=recog(2,p,"ESL","ERL")) && !isalnum(*(p+cn))) {
		p+=cn;
		if (*p=='(') {
			p++;
			o=S(); // dummy
		}
		else p--, paren=FALSE;
		o=errLine;
	}
	/* NXL() function - NeXt Line
	 * ~return next line number of line in argument (if the argument is
	 * the line where error occurred - ESL - it returns the line 
	 * immediately after, suitable for a JUMP */
	else if (!strncasecmp_(p,"NXL(",4)) {
		p+=4;
		o=(int)S()+1;
		if (o==0.0) perror_(__LINE__,147,l);
	}
	/* DECLARE variables */
	else if (!isValOn && (ns=findDEC(p))) {
		int var=ns, M=0, N=0, T=1;
		p+=_len;
		if (_rettype && _rettype==STR) perror_(__LINE__,217,l);
		/* return a number */
		else if (*(p)!='(' && !_arr) {
			if (vn[ns].isInt) o=(int)P[ns];
			else o=P[ns];
			paren=FALSE;
			p+=-1;
		}
		/* return an array */
		else if (*(p)=='(' && _arr) {
			p++;
			M=(int)S();
			if (*p==',') {
				p++;
				N=(int)S();
				T=2;
			}
			/* check dimensions */
			if (T>dim[var].type||M>dim[var].row||N>dim[var].col)
				perror_(__LINE__,76,l);
			else if (vn[ns].isInt) 
				o=(int) getArrayVal(ns,T,M,N);
			else 
				o=getArrayVal(ns,T,M,N);
		}
	}
	/* any other sequence of characters is a syntax error 
	 * Note: within VAL, the error message is dedicated. */
	else if (!isValOn) perror_(__LINE__,37,l);
	else perror_(__LINE__,115,l);

	/* check closing parenthesis for functions requiring an argument;
	 * p is updated into getVal() */
	if (paren) {
		if (!*p) perror_(__LINE__,65,l,"A FF,LF,VT, OR CR",")");
		else if (*p!=')') perror_(__LINE__,33,l,*p);
	}
	/* return sweated value! */
	return o;
}


/* optoken()
 * substitute operators names with establishes symbols,
 * according to the following list:
 * && for AND
 * || for OR
 * !! for XOR
 * -> for IMP
 * ## for EQV
 * '{' for NOR
 * '}' for NAND 
 * Also, get rid of inline REM comments
 * The substitution has the advantage to consider operators as established
 * operators already in use
 * Note: using option -r (debug), current line is printed after tokenization 
 * and blank-stripping, to see the 'real' string under evaluation
 * System function */
void optoken(char* q) {
	char *qq;
	while((qq=strcasestr_oq(q,"AND")) && qq>q && isante(qq) && ispost(qq+2))
		*qq++='&', *qq++='&',*qq=' ';
	while((qq=strcasestr_oq(q,"OR")) && qq>q && isante(qq) && ispost(qq+1))
		*qq++='|', *qq++='|',*qq=' ';
	while((qq=strcasestr_oq(q,"XOR")) && qq>q && isante(qq) && ispost(qq+2))
		*qq++='!', *qq++='!',*qq=' ';
	while((qq=strcasestr_oq(q,"IMP")) && qq>q && isante(qq) && ispost(qq+2))
		*qq++='-', *qq++='>',*qq=' ';
	while((qq=strcasestr_oq(q,"EQV")) && qq>q && isante(qq) && ispost(qq+2))
		*qq++='#', *qq++='#',*qq=' ';
	while((qq=strcasestr_oq(q,"NOR")) && qq>q && isante(qq) && ispost(qq+2))
		*qq++='{', *qq++=' ',*qq=' ';
	while((qq=strcasestr_oq(q,"NAND"))&& qq>q && isante(qq) && ispost(qq+3))
		*qq++='}', *qq++=' ',*qq++=' ', *qq=' ';
	if ((qq=strcasestr_oq(q,":"))) {
		char *yy=qq++;
		while (isblank(*qq)) qq++;
		if (!strncasecmp_(qq,"REM",3)) *yy='\0';
	}
}


/******************************************************************************
 * The following functions define general system environment subroutines: 
 * they are internal routines not directly accessible as BASIC statements.
 ******************************************************************************/

/*****************************************
 *            STATEMENTS AID             *
 *****************************************/

/* printFormatted()
 * This is the routine used by all the PRINT/WRITE USING statements and the 
 * FORMAT$ function; its purpose is to build a C-format string from the BASIC 
 * format string, to ensure a correct output according to the required features
 * ad leaving the result in _FORMAT[] (global variable).
 * The number-printing routine is delegated to printFormattedNumber() and the
 * string-printing routine is delegated to printFormattedString().
 * Note: the C-formatter is not a classic C-string formatter (es. "%05f") but
 * a special series of data that identifies the objects that are included in
 * the BASIC formatter, and which is used by the prin:tFormatted subroutines.
 * BForm: the BASIC formatter string
 * list: the sequence of what is to be printed (it follows the USING part)
 *       (FORMAT$() uses NIL here)
 * isWrite: the WRITE flag: if TRUE, prints a Line Number + tab (WRITE)
 * 	 (FORMAT$() uses FALSE here)
 * val: the value to be printed (given only by FORMAT$)
 * 	 (PRINT USING uses 0.0 here)
 * System function for the USING feature */
void printFormatted(char* BForm, char* list, int isWrite, double val) {
	char cForm[SCREENLINE+1], *cf=cForm;
	int ic=0,dc=0;	// ic=integers count, dc=decimals count
	int sL=0,nL=0; 	// sL=strings count, nL=numbers count in list
	int sF=0,nF=0; 	// sF=strings count, nF=numbers count in BForm
	int in,is;	// in and is are indexes for numbers/strings
	int inf,isf;	// inf and ins are field indexes
	double numb[MAXARGS];   	   // the stack for numbers in 'list'
	char stri[MAXARGS][SCREENLINE+1];  // the stack for strings in 'list'
	char fin[COMPSTR];
	char *pb=p;		// secure program counter backup
	char fl=' ', com=sep; 	// filler and separator
	TUsing dat[MAXARGS], datS[MAXARGS];// dat for numbers, datS for strings
	int isInt;	// isInt = is the number in Integer form?
	int isExp;	// isExp = is the number in exponential form?
	int isMin;	// isMin = is the minus sign to be written at the end?
	int isDol;	// isDol = is the dollar format enabled?
	int isCom;	// isCom = is the thousands separator format enabled?
	int isZer;	// isZer = is the zeroes format enabled?
	int isCR=TRUE;	// isCR  = is the final Carriage Return to be printed?
	int CNimage=0,VNlist=0;	// CNimage=num counter, VNlist=num counter
	int CSimage=0,VSlist=0;	// CSimage=str counter, VSlist=str counter
	int i,len,type,items;   // process tools

	/* erase the C-formatter and the destination string (global) */
	for (i=0; i<SCREENLINE; i++) cForm[i]='\0';
	for (i=0; i<COMPSTR; i++) _FORMAT[i]='\0';

	/* check length of BASIC formatter */
	if ((int)strlen(BForm)>MAXSTRLEN) perror_(__LINE__,112,l);
	
	/* step 1. build C images */
	while (*BForm) {
		isInt=TRUE, isExp=FALSE, isMin=FALSE;
		isDol=FALSE, isCom=FALSE, isZer=FALSE;
		len=0; type=LEFTJUST;

		ic=dc=0; isInt=TRUE;
		/* get numeric formatter */
		if ((*BForm=='#' && *(BForm+1)=='#')\
				||((*BForm=='.') && (*(BForm+1)=='#'))\
				||((*BForm=='#') && (*(BForm+1)=='.'))\
				||((*BForm=='*'&&*(BForm+1)=='*')&&((fl='*')))\
				||((*BForm=='0'&&*(BForm+1)=='0')&&((fl='0')))\
				||(*BForm=='$'&&*(BForm+1)=='$') ) {
			*cf++=NUMID;
			CNimage++;
			/* integer part */
			while (*BForm && strchr(",.*#$0",*BForm)) {
				if (*BForm==',') isCom=TRUE,BForm++,ic++;
				if (*BForm=='$') isDol=TRUE,BForm++,ic++;
				if (*BForm=='*') BForm++,ic++;
				if (*BForm=='0') BForm++,ic++;
				if (*BForm=='#') BForm++,ic++;
				/* decimal part */
				if (*BForm=='.') {
					BForm++;
					isInt=FALSE;
					while(*BForm&&strchr(",.*#$0",*BForm)){ 
						if (*BForm==',') 
						    isCom=TRUE,BForm++,dc++;
						else if (*BForm=='$') 
						    isDol=TRUE,BForm++,dc++;
						else if (*BForm=='*') 
						    BForm++,dc++;
						else if (*BForm=='0') 
						    BForm++,dc++;
						else if (*BForm=='#') 
						    BForm++,dc++;
						else break;
					}
					break;
				}
			}
			/* exponential in case of float */
			if (!isInt && !strncmp(BForm,"^^^^",4)) {
				BForm+=4;
				isExp=TRUE;
			}
			/* minus at the end */
			if (*BForm=='-') {
				BForm++;
				isMin=TRUE;
			}

			/* store current-image data */
			dat[nF].ic=ic;
			dat[nF].dc=dc;
			dat[nF].isInt=isInt;
			dat[nF].isExp=isExp;
			dat[nF].isMin=isMin;
			dat[nF].isCom=isCom;
			dat[nF].isDol=isDol;
			dat[nF].fill=fl;
			dat[nF].com=com;
			dat[nF].isZer=isZer;
			nF++;

		}
		/* get string formatter #1 <###, <###< and <###> */	
		else if (*BForm=='<' && *(BForm+1)=='#') {
			*cf++=STRID;
			CSimage++;
			len=1;
			BForm++;
			while (*BForm=='#') BForm++, len++;
			if (*BForm=='>') BForm++, len++, type=CENTERED;
			else if (*BForm=='<') BForm++, len++, type=EXTENDED;
			else type=LEFTJUST;
			datS[sF].len=len;
			datS[sF].type=type;
			sF++;
		}
		/* get string formatter #1 >### */	
		else if (*BForm=='>' && *(BForm+1)=='#') {
			*cf++=STRID;
			CSimage++;
			len=1;
			BForm++;
			while (*BForm=='#') BForm++, len++;
			type=RIGHTJUST;
			datS[sF].len=len;
			datS[sF].type=type;
			sF++;
		}
		/* get string formatter #2 (DEC-20 features)*/	
		else if (*BForm=='\'') {
			*cf++=STRID;
			CSimage++;
			len=1;
			BForm++;
			/* E Format Character */
			if (*BForm=='E') {
				type=EXTENDED;
				while (*BForm=='E') BForm++,len++;
			}
			/* C Format Character */
			else if (*BForm=='C') {
				type=CENTERED;
				while (*BForm=='C') BForm++,len++;
			}
			/* L Format Character */
			else if (*BForm=='L') {
				type=LEFTJUST;
				while (*BForm=='L') BForm++,len++;
			}
			/* R Format Character */
			else if (*BForm=='R') {
				type=RIGHTJUST;
				while (*BForm=='R') BForm++,len++;
			}
			/* store current-image data */
			datS[sF].len=len;
			datS[sF].type=type;
			sF++;
		}
		else {
			/* escape */
			if (*BForm=='\\') BForm++;
			*cf++=*BForm++;
		}
	}
	
	/* step 2. close the C-formatter */
	*cf='\0'; cf=cForm;

	/* step 3a. add val argument (called with list=NIL)*/
	if (!list) {
		numb[nL++]=val;
		VNlist++;
		isCR=FALSE;

	}
	/* step 3b. grab arguments to be printed in 'list' and add them */
	else while (*list) {
		/* string in any form */
		if (isanystring(p)) {
			p=list;
			strncpy_(stri[sL++],SS(),COMPSCR);
			list=p;
			VSlist++;
		}
		/* number in any form */
		else {
			double res=0;
			p=list;
			res=S();
			numb[nL++]=res;
			list=p;
			VNlist++;
		}
		/* is string ended? */
		if (!*list) isCR=TRUE;
		else if (*list==',' || *list==';') {
			list++;
			isCR=FALSE;
		}
		else if (*list) perror_(__LINE__,65,l,FFLFVT,",");
	}

	/* step 4. check errors and discontinuities */
	if (VSlist && !VNlist && !CSimage && CNimage) 
		perror_(__LINE__,68,l); // mixed
	if (!VSlist && VNlist && CSimage && !CNimage) 
		perror_(__LINE__,68,l); // mixed
	if (VSlist && !CSimage) 
		perror_(__LINE__,105,l); // no fields in str image
	if (VNlist && !CNimage) 
		perror_(__LINE__,105,l); // no fields in num image

	/* step 5. add elements in order */
	in=is=0;
	inf=isf=0;
	items=sL+nL;
	
	while (items>0) {
		/* restart formatter if there are still items to print */
		if (!*cf) {
			chrcat_(_FORMAT,'\n',COMPSTR);
			cf=cForm;
		}
		/* if the formatter is at the beginning, do the WRITE line
		 * prepending a five digits number followed by a tab */
		if (cf==cForm && isWrite) { 
			sprintw(fin,COMPSTR,"%05d\t",writeSeqStart);
			strncat_(_FORMAT,fin,COMPSTR);
			writeSeqStart+=writeSeqStep;
		}
		/* add a number in a number field */
		if (*cf==NUMID) {
			printFormattedNumber(numb[in],dat[inf],_FORMAT);
			in++;
			cf++;
			inf++; if (inf>=nF) inf=0;
			items--;
		}
		/* add a string */
		else if (*cf==STRID) {
			printFormattedString(stri[is],datS[isf],_FORMAT);
			is++;
			cf++;
			isf++; if (isf>=sF) isf=0;
			items--;
		}
		/* add the rest of the formatter string */
		while (*cf && *cf!=NUMID && *cf!=STRID && *cf!='\n') {
			if (*cf==',' && !*(cf+1)) break;
			chrcat_(_FORMAT,*cf++,COMPSTR);
		}
	}
	/* step 6. close line */
	if (isCR) chrcat_(_FORMAT,'\n',COMPSTR);

	/* the result is in _FORMAT: let the PRINTUSING family to print
	 * result after calling this printFormatted() and let FORMAT$() to
	 * add _FORMAT to mypad[] */

	/* final step. restore general pointer */
	p=pb;
}


/* printFormattedString() 
 * build a formatted string according to information contained in the 'us' 
 * formatter.
 * str: the string to print
 * us: a TUsing structure holding formatting data
 * dest: the destination string 
 * System function for the USING management */
void printFormattedString(char* str, TUsing us, char *dest) {
	int thislen=(int)strlen(str),i;
	char format[FORMATLINE*4+1];
	char fin[COMPSTR]="";

	/* set up left justification (with cut) */
	if (us.type==LEFTJUST) { // "%-s"
		if (thislen<us.len) 
			sprintw(format,FORMATLINE*4+1,"%%-%d.%ds",us.len,us.len);
		else if (thislen>us.len) 
			sprintw(format,FORMATLINE*4+1,"%%%d.%ds",us.len,us.len);
		else sprintw(format,FORMATLINE*4+1,"%%s");
	}
	/* set up right justification (with cut) */
	else if (us.type==RIGHTJUST) 
		sprintw(format,FORMATLINE*4+1,"%%%d.%ds",us.len,us.len);
	/* set up centered justification (with cut) */
	else if (us.type==CENTERED) {
		if (thislen<us.len) {
			for (i=0;i<(int)((us.len-thislen)/2);i++) {
				chrcat_(dest,' ',COMPSTR);
			}
			sprintw(format,FORMATLINE*4+1,"%%s");
		}
		else if (thislen>us.len) 
			sprintw(format,FORMATLINE*4+1,"%%%d.%ds",us.len,us.len);
		else sprintw(format,FORMATLINE*4+1,"%%s");
	}
	/* set up extended (with no cut) */
	else if (us.type==EXTENDED) 
		sprintw(format,FORMATLINE*4+1,"%%-%ds",us.len);
	/* append result */
	sprintw(fin,COMPSTR,format,str);
	strncat_(dest,fin,COMPSTR);
	/* complete centered justification appending spaces as needed*/
	if (us.type==CENTERED) 
		for(i=0;i<us.len-thislen-(int)((us.len-thislen)/2);i++) {
			chrcat_(dest,' ',COMPSTR);
		}
}


/* printFormattedNumber() 
 * build a float number according to the information contained in the 'us' 
 * formatter, leaving to outNum() the task to append the number digits. 
 * num: the floating point number to print
 * us: a TUsing structure holding formatting data
 * dest: the destination string 
 * System function for the USING management */
void printFormattedNumber(double num, TUsing us, char *dest) {
	int sign=1,nlen=0,b=0;
	char temp[MAXSTRLEN], *t=temp;

	/* make checks */
	if (us.isExp && (us.isDol || us.fill=='*'|| us.fill=='0')) 
		perror_(__LINE__,78,l) ;

	/* convert num to a positive number, saving sign */
	if (num<0) sign=-1, num*=-1;
	if (num>INFF) num=INFF;

	/* Trap for numbers in the range ]-1,1[ when an integer is to be
	 * printed (that is, no decimal was specified in the BASIC format):
	 * since num has already been turned into a positive number, 
 	 * restart to avoid seeing -0 or 0- in the output figure */
	if (num<1.0 && us.isInt && !us.isZer) {
		us.isZer=TRUE;
		printFormattedNumber(num,us,dest);
		return;
	}

	/* Treat minus identifier, which causes the possible negative minus 
	 * sign to be printed afterwards;
	 * Note: the normal negative sign printing is embedded in the 
	 * following printing routine */
	if (sign<0) {
		if (us.isMin) {
			printFormattedNumber(num,us,dest);
			chrcat_(dest,'-',COMPSTR);
			return;
		}
		else if ((us.isDol || us.fill=='*'|| us.fill=='0')) 
			perror_(__LINE__,67,l);
	}
	nlen=intLen(num);
	if (us.isCom) nlen+=(int)((nlen-1)/3); // add commas number
	b=nlen;

	/* reset number in case of the special case 1E-N as a difference number
	 * because it cannot be rendered exactly in memory due to the storing 
	 * technique and the fact there is only one significant digit.
	 * The algorithm is: print the number in the exponential format of C
	 * on the string temp; if the number of significant digits is only 1, 
	 * convert the number from this string, otherwise use original one */
	sprintw(temp, MAXSTRLEN,"%g",num);
	while (*t!='e') t++;
	if ((int)(t-temp)<2) num=strtod(temp,NULL);

	/* normal form (options: dollar, commas or asterisk-fill) */
	if (!us.isExp) {
		/* minus at the end */
		if (us.isMin) {
			if (nlen < us.ic) {
				while (b++ < us.ic-1) {
					chrcat_(dest,us.fill,COMPSTR);
				}
				if (us.isDol) chrcat_(dest,'$',COMPSTR); 
				else chrcat_(dest,us.fill,COMPSTR);
			}
			if (nlen > us.ic || (nlen==us.ic && us.isDol)) {
				chrcat_(dest,'&',COMPSTR);
				if (us.isDol) chrcat_(dest,'$',COMPSTR); 
			}
		}
		/* normal minus ahead of number */
		else {
			if (nlen < us.ic) {
				while (b++ < us.ic-1) {
					chrcat_(dest,us.fill,COMPSTR);
				}
				if (sign<0) chrcat_(dest,'-',COMPSTR);
				else if (us.isDol) chrcat_(dest,'$',COMPSTR);
				else chrcat_(dest,us.fill,COMPSTR);
			}
			if (nlen > us.ic || (nlen==us.ic && sign<0)) {
				chrcat_(dest,'&',COMPSTR);
				if (sign<0) chrcat_(dest,'-',COMPSTR);
				else if (us.isDol) chrcat_(dest,'$',COMPSTR);
			}
		}
		outNum(num,us,dest);
	}
	/* exponential form 
	 * Note: exponentials never overflow; they adapt their format
	 * to the length of the integer part (minimum two placeholders),
	 * so that at least the sign and a digit can be represented; 
	 * if no dot is in the number image, the ^^^^ symbol is taken "as-is";
	 * if the dot is present, and isn't followed by any digits placeholders 
	 * in the number image, no decimal is printed, but the dot is. */
	else {
		char esign, fin[COMPSTR]="";
		int i=0, E=0, cut, w;
		if (num>ZERO) {
			/* reduce to the 0.nnnnnnnn form */
			while (num < 0.1) num*=10.0, E--;
			while (num > 1.0) num/=10.0, E++;
			/* expand to the maximum ic (integer digits count)
			 * space; in case of postponed minus, the space is 
			 * the  whole ic length, otherwise reserve one space 
			 * for the sign position (a white space in case 
			 * of positive numbers) */
			cut=us.isMin?0:1;
			while (i<us.ic-cut) i++, num*=10.0, E--;
			/* append sign */
			if (E<0) esign='-', E*=-1;
			else esign='+';
			if (sign<0) chrcat_(dest,'-',COMPSTR);
			else if (!us.isMin) chrcat_(dest,us.fill,COMPSTR);
			/* append integer part, dot and float part */
			outNum(num,us,dest);
			/* append E.XX*/
			sprintw(fin,COMPSTR,"E%c%02d%n",esign,E,&w);
			strncat_(dest,fin,COMPSTR);
		}
		else {
			/* in case of ZERO, fill before 0.0 */
			while (i<us.ic-1) {
				i++;
				chrcat_(dest,us.fill,COMPSTR);
			}
			outNum(0.0,us,dest);
			strncat_(dest,"E+00",COMPSTR);
		}
	}
}


/* intLen()
 * find the length of the integer part of any number passed as double, up to 
 * INFF (suitable for outNum). If the number is negative, it is turned to 
 * positive (the minus sign is not counted because printFormattedNumber() 
 * arranges it).
 * num: the float to be printed
 * System function for the USING management */
int intLen(double num) {
	int len=0;
	double integ;
	if (num<0) num*=-1;
	if (num<10) return 1;
	modf(num,&integ);
	while (!(1.0 > integ)) integ/=10,len++;
	return len;
}


/* outNum()
 * build a positive number in float according to the BASIC format, appending
 * it to the destination string.
 * num: the number to be printed 
 * us: a TUsing structure holding formatter data
 * dest: the destination string
 * System function for the USING management */
void outNum(double num, TUsing us, char* dest) {
	char nums[FORMATLINE*4+1], *np=nums, format[FORMATLINE*4+1];
	int e=0, i=0, zero=FALSE, m=0;
	double temp, integ;

	/* if number must have no decimals and it's not an integer, it should 
	 * be rounded; so print it to a string, restore it and continue */
	if (!us.dc && !us.isInt) {
		sprintw(nums,FORMATLINE*4+1,"%.0f", num);
		num=strtod(nums,NULL);
	}

	/* cut out significant digits after fifteenth
	 * by building number string in bare format through C */
	if (num<0) num*=-1; // just to be sure...
	if (num<ZERO) zero=TRUE;
	else {
		while (num >= 1.0) num/=10, e++;
		while (num < 0.1) num*=10, e--;
	}
	sprintw(nums,FORMATLINE*4+1,"%.15f",num);

	/* now number is in the form 0.XXXXXXXXXXXXXXX (15 significant digits)
	 * and 'e' holds the exponential factor */

	/* BUILD INTEGER PART */
	np+=2; // skip 0 and dot
	if (!isAmericanFormat) sep='\'';
	if (e<=0 || zero) {
		chrcat_(dest,'0',COMPSTR);
	}
	else {
		/* add all significant digits */
		while (i<e-1) {
			if (!*np) break;
			chrcat_(dest,*np,COMPSTR);
			m=(e-i)%3;
			if (m==1 && us.isCom) {
				chrcat_(dest,sep,COMPSTR);
			}
			i++,np++;
		}
		if (i<e && *np) {
			chrcat_(dest,*np,COMPSTR);
			i++,np++;
		}
		/* if number is not finished, fill the rest with zeroes */
		while (i<e-1) {
			chrcat_(dest,'0',COMPSTR);
			m=(e-i)%3;
			if (m==1 && us.isCom) {
				chrcat_(dest,sep,COMPSTR);
			}
			i++;
		}
		if (i<e) {
			chrcat_(dest,'0',COMPSTR);
		}
	}

	/* end up in case of integer */
	if (us.isInt) return;

	/* APPEND THE DOT */
	if (isAmericanFormat) chrcat_(dest,'.',COMPSTR);
	else chrcat_(dest,',',COMPSTR);

	/* BUILD FRACTIONAL PART */
	i=0;

	/* end up in case there are no decimals */
	if (!us.dc) return;

	/* add all necessary zeros that follow the dot */
	while(e<0 && i<us.dc) chrcat_(dest,'0',COMPSTR), i++, e++;

	/* end up in case all digits were printed (num may not be over) */
	if (e<0 || i>=us.dc) return;

	/* add the rest of the decimal digits */
	*(np-1)='.';
	//temp=strtod(np-1,NULL);
	temp=modf(atof(np-1),&integ);
	sprintw(format,FORMATLINE*4+1,"%%0.%df",us.dc-i);
	sprintw(nums,FORMATLINE*4+1,format,temp); // autoformat && autorounding
	strncat_(dest,nums+2,COMPSTR);   // skip '0' and '.'
}


/* checkNames()
 * control that a file cannot be opened more than once, using FILES or FILE.
 * c: the channel id to check
 * fname: the file name to check 
 * return TRUE if a file already opened does exist, FALSE otherwise.
 * System function for the file management */
int checkNames(int ch, char* fname) {
	int i=1, res=FALSE;
	while (i<=STREAMS) {
		if (i==ch);
		else if (!strcasecmp(channel[i].ioname,fname)) {
			res=TRUE; 
			break;
		}
		i++;
	}
	return res;
}


/* doesExists()
 * this function determines if the file whose name is in the argument is 
 * accessible in read, write or both modes by current tbas session.
 * The sequence of the permissions follows these lines:
 * - if the user owns the file, the permission is the user file permission
 * - if the file is not owned by the user, but the groups are the same, the 
 *   permission is the group file permission
 * - if the file is not owned by the user, and the groups differ, the 
 *   permission is the world file permission
 * - else the file is not accessible.
 * Return values are as follows:
 * - For a file that exists but is not a file (a directory or a block file) -4 
 *   is returned; if the string is empty, return also -4 along with a warning
 * - For a file that does not exist, -8 is returned
 * - For a file that exists but is not accessible by tbas, 0 is returned
 * - For a file that exists and is accessible in read mode only, 1 is returned
 * - For a file that exists and is accessible in write mode only, 2 is returned
 * - For a file that exists and is accessible in read and write mode, 3 is 
 *   returned (that is 1+2) */
int doesExist(char *fn) {
	int read=0, write=0;
	int readu=0, readg=0, reado=0;
	int writeu=0, writeg=0, writeo=0;
	int fgroup=0, ugroup=0, fuser=0, uuser=0;

	/* if string is null, it is not a file */
	if (!*fn) {
		perror_(__LINE__,159,l);
		return NOTAFILE;
	}
	
	/* get file attributes and check existence */
	if (stat(fn,&statbuf)<0) return NOEXIST;
	/* is it a directory? */ 
	if (S_ISDIR(statbuf.st_mode)) return NOTAFILE;
	/* is it a block file? */
	if (S_ISBLK(statbuf.st_mode)) return NOTAFILE;
	
	/* find file and tbas process group ids */
	fgroup=(int)statbuf.st_gid;
	ugroup=(int)getegid();
	/* find file and tbas process user ids */
	fuser=(int)statbuf.st_uid;
	uuser=(int)geteuid();
	
	/* calculate READ permissions */
	readu=statbuf.st_mode & S_IRUSR;
	readg=statbuf.st_mode & S_IRGRP;
	reado=statbuf.st_mode & S_IROTH;
	/* check READ permissions */
	if (fuser==uuser) {
		if (readu) read=DOREAD;
	}
	else if (fuser!=uuser && fgroup==ugroup) {
		if (readg) read=DOREAD;
	}
	else if (fuser!=uuser && fgroup!=ugroup) {
		if (reado) read=DOREAD;
	}
	else read=NOACCESS;
	
	/* calculate WRITE permissions */
	writeu=statbuf.st_mode & S_IWUSR;
	writeg=statbuf.st_mode & S_IWGRP;
	writeo=statbuf.st_mode & S_IWOTH;
	/* check WRITE permissions */
	if (fuser==uuser) {
		if (writeu) write=DOWRITE;
	}
	else if (fuser!=uuser && fgroup==ugroup) {
		if (writeg) write=DOWRITE;
	}
	else if (fuser!=uuser && fgroup!=ugroup) {
		if (writeo) write=DOWRITE;
	}
	else write=NOACCESS;

	/* return the total permissions */
	return (read + write);
}


/* setupName()
 * check file name, returning correct file name with extension "bas", if not
 * specified for an unexisting file; deprive name of file specifiers $ or %.
 * If an improper % character followed by number was chosen, len is -1.
 * fn: the file name string
 * mode: return character specifier '$' or '%' or NIL (e.g. $ or %)
 * len: return len of string field if specified (e.g. $15)
 * Note: the file specifiers $, $<num> and % are applied to 'names' of files,
 * or to 'extensions'; thus, a file may be named NUMBER%.DTA or NUMBER.DTA%
 * with the very same effect: random number file called NUMBER.DTA
 * Note: this version does not attach .bas to files called as names only. This
 * characteristic belongs to the DEC-20 O.S. and family
 * System function for the file management */
void setupName(char *fn, int *mode, int *len) {
	int isExt=TRUE;
        char*q=fn, ext[COMPSTR], tempf[COMPFIL];
	FILE *fd;

	/* setup defaults */
	*mode=0;
	*len=0;
	ext[0]='\0';

	/* check if the input file is a directory */
	stat(fn, &statbuf);
	if (S_ISDIR(statbuf.st_mode)) {
		beginS=-1;
		perror_(__LINE__,3,-1,fn);
	}
	
	/* find extension start (and check eventual specifiers set after
	 * extension in place of after the name) */
	while (*q) q++;
	while (*fn=='.') fn++;
	while (*q !='.' && q > fn) {
		if (*q=='$' || *q=='%') {
			*mode=*q;
			if (*q=='$' && isdigit(*(q+1)))
				*len=strtol(q+1,NULL,10);
			*q='\0';
		}
		q--;
	}

	/* there is an extension */
	if (q > fn) {
		strncpy_(ext,q,COMPSTR);
		*q='\0';
		q--;
	}
	/* there's no extension */
	else {
		if ((int)strlen(q)>0) q=fn+(int)strlen(q)-1;
		isExt=FALSE;
	}

	/* there's a length (or at least I suppose there is) 
	 * Note: this function assumes that no file may be named with an ending 
	 * '$' or '%' followed by any digit, because it would be interpreted as 
	 * a random file; in any case, since it's bad practice in any OS naming 
	 * files using '$' or '%', I take it for granted. */
	if (isdigit(*q)) {
		while ((int)(q-fn)>0 && isdigit(*q)) q--;
		/* it's really a field length */
		if (*q=='$') *len=(int)strtol(q+1,NULL,10);
		/* it's a wrong assignment */
		else if (*q=='%') {
			*len=-1;
			sprintw(fn,COMPSCR,"%d",(int)strtol(q+1,NULL,10));
		}
		/* it's not a field length: number is part of the filename */
		else {
			q++;
			while (isdigit(*q)) q++;
		}
	}
	
	/* there's a specifier $ or % */
	if (*q=='$' || *q=='%') *mode=*q, *q--='\0';

	/* if a dot was printed, but no extension specified, remove the dot */
	if (!strcasecmp(ext,".")) ext[0]='\0', isExt=FALSE;

	/* if an extension was given as input, attach it */
	if (isExt) strncat_(fn, ext, COMPSCR);
	/* search for implicit extension or check if it's needed */
	else {
		/* first attempt: the name alone, extension not present */
		sprintw(tempf,COMPFIL,"%s",fn);
		if ((fd=fopen(tempf,"r"))) {
			fclose(fd);
			return;
		}
		/* second attempt: .BAS extension not specified */
		sprintw(tempf,COMPFIL,"%s%s",fn,EXTU);
		if ((fd=fopen(tempf,"r"))) {
			fclose(fd);
			strncat_(fn,EXTU,COMPSCR);
			return;
		}
		/* third attempt: .bas extension not specified */
		sprintw(tempf,COMPFIL,"%s%s",fn,EXTL);
		if ((fd=fopen(tempf,"r"))) {
			fclose(fd);
			strncat_(fn,EXTL,COMPSCR);
			return;
		}
	}
}


/* testEOF()
 * return TRUE if current file has passed the last character/item, or FALSE.
 * Note: it does NOT control if stream is open.
 * ch: the testing channel id 
 * System function for the file management */
int testEOF(int ch) {
	int res=FALSE, after, before=ftell(channel[ch].stream);
	fseek(channel[ch].stream,0,SEEK_END);
	after=ftell(channel[ch].stream);
	if (before-after>=-1) res=TRUE;
	fseek(channel[ch].stream,before,SEEK_SET);
	return res;
}


/* getEOF()
 * return location of last record for random access files
 * Note: it does NOT control if stream is open 
 * ch: the testing channel id 
 * System function for the file management */
int getEOF(int ch) {
	int back=ftell(channel[ch].stream);
	int res;
	/* calculate total offset */
	fseek(channel[ch].stream,0,SEEK_END);
	res=((int)(ftell(channel[ch].stream)-FIRSTLEN)/channel[ch].reclen);
	/* restore position */
	fseek(channel[ch].stream,back,SEEK_SET);
	return res;
}


/* printContainedRaw() 
 * print a text as-is
 * does not update channel fcounter
 * ch: the output channel id
 * string: the string to be printed (number or string) 
 * System function for the printing management */
void printContainedRaw(int ch, char* st) {
	char out[COMPSTR], *o=out;
	int i=0;
	if (!*st) return;
	while (*st) {
		/* build part of string to be printed */
		o=out;
		while (*st && i< COMPSTR) {
/* alarm beep */	     if (*st=='\\' && *(st+1)=='a') *o++='\a',st+=2;
/* backspace */		else if (*st=='\\' && *(st+1)=='b') *o++='\b',st+=2;
/* formfeed */		else if (*st=='\\' && *(st+1)=='f') *o++='\f',st+=2;
/* newline */		else if (*st=='\\' && *(st+1)=='n') *o++='\n',st+=2;
/* carriage return */	else if (*st=='\\' && *(st+1)=='r') *o++='\r',st+=2;
/* horizontal tab */	else if (*st=='\\' && *(st+1)=='t') *o++='\t',st+=2;
/* vertical tab */	else if (*st=='\\' && *(st+1)=='v') *o++='\v',st+=2;
/* backslash */		else if (*st=='\\' && *(st+1)=='\\') *o++='\\',st+=2;
/* single quote */	else if (*st=='\\' && *(st+1)=='\'') *o++='\'',st+=2;
             		else *o++=*st++; 
			channel[ch].fcounter++;
			i++;
		}
		*o='\0';
		if (isCap) turnToUpper(out);
		/* print string */	
		fprintf(channel[ch].stream,"%s",out);
	}
}



/* printContainedPlain() 
 * print a text into the MARGIN limit;
 * updates channel fcounter
 * ch: the output channel id
 * string: the string to be printed (number or string) 
 * margin: the channel margin (for speed)
 * System function for the printing management */
void printContainedPlain(int ch, char* string, int margin) {
	char out[margin], *o=out;
	if (!*string) return;
	while (*string) {
		/* build part of string to be printed */
		o=out;
		while (*string && channel[ch].fcounter<margin) {
			if (*string==_CR) channel[ch].fcounter=0;
			*o++=*string++; 
			channel[ch].fcounter++;
		}
		*o='\0';
		if (isCap) turnToUpper(out);
		/* print string */	
		fprintf(channel[ch].stream,"%s",out);
		if (channel[ch].fcounter>=margin && *string) {
			FCR(ch);
			channel[ch].fcounter=0;
		}
	}
}


/* printContained()
 * wrapper for directing the print of strings to the Plain (BASIC) and
 * the Raw (print as-is and filters escape characters))
 * System function for the printing management */
void printContained(int ch, char* string, int margin) {
	if (isRawPrint) printContainedRaw(ch,string);
	else printContainedPlain(ch,string,margin);
}


/* fillGap()
 * In a random file, fill the intermediate elements of a random access file 
 * with spaces, whenever the index is moved at least one element over the last;
 * used within FWRITECOLON();
 * No checks is done upon channel
 * ch: the channel id
 * actualIndex: the current value of the random file position
 * System function for the printing management */
void fillGap(int ch, int actualIndex) {
	int i;
	for (i=1; i<channel[ch].currseek-actualIndex; i++) 
		fputc(' ',channel[ch].stream);
}


/* matStrReadAssign(result-matrix,type,rows,columns)
 * iterate through matrix indexes to assign string DATA values
 * var: the matrix code
 * t: the matrix type
 * m, n: matrix rows and columns
 * System routine for matrix management, used into MATREAD() */
void matStrReadAssign(int var, int t, int m, int n){
	int i,j;
	/* vectors */
	if (1==t) {
		for(i=1;i<=m;i++) {
			if (dcs>dataLimitS) perror_(__LINE__,110,l);
			setArrayStr(var,1,i,0,DS[dcs++]);
		}
	}
	/* matrices */
	else {
		/* horizontal vectors */
		if (!m && n) {
			for (i=1;i<=n;i++) {
				if (dcs>dataLimitS) perror_(__LINE__,110,l);
				setArrayStr(var,2,0,i,DS[dcs++]);
			}
		}
		/* normal matrices */
		else for(i=1;i<=m;i++) 
			for (j=1;j<=n;j++) {
				/* if out of data, print warning */
				if (dcs>dataLimitS) perror_(__LINE__,110,l);
				setArrayStr(var,2,i,j,DS[dcs++]);
			}
	}
}


/* matReadAssign(result-matrix,type,rows,columns)
 * iterate through matrix indexes to assign numerical DATA values
 * var: the matrix code
 * t: the matrix type
 * m, n: matrix rows and columns numbers
 * System routine for matrix management, used into MATREAD() */
void matReadAssign(int var, int t, int m, int n){
	int i,j;
	/* vectors */
	if (1==t) {
		for(i=1;i<=m;i++) {
			if (dc>dataLimit) perror_(__LINE__,110,l);
			setArrayVal(var,1,i,0,D[dc++]);
		}
	}
	/* matrices */
	else {
		/* horizontal vectors */
		if (!m && n) {
			for (i=1;i<=n;i++) {
				if (dc>dataLimit) perror_(__LINE__,110,l);
				setArrayVal(var,2,0,i,D[dc++]);
			}
		}
		/* normal matrices */
		else for(i=1;i<=m;i++) 
			for (j=1;j<=n;j++) {
				/* if out of data, print warning */
				if (dc>dataLimit) perror_(__LINE__,110,l);
				setArrayVal(var,2,i,j,D[dc++]);
			}
	}
}


/* isfn(string-position) 
 * (IS a FN function implicit assignment) 
 * test if the current string **begins** with a FN + char
 * p is not touched
 * System function for the parsing management */
int isfn(char *p) {
	if (!strncasecmp_(p,"FN",2) && isalpha(*(p+2))) return TRUE;
	else return FALSE;
}


/* issaa(string-position)
 * (IS a String ArrAy)
 * System function for the DIM management */
int issaa(char *lp) {
	if (!isbeginnamechar(lp)) return FALSE;
	while (isnamechar(lp)) lp++;
	if (*lp!='$') return FALSE;
	else lp++;
	if (*lp!='(') return FALSE;
	return TRUE;
}	


/* isstrvar(string-position)
 * (IS a STRing VAR)
 * System function for the DIM management 
 * Note: the figure var() is trated as a var name (used in sub arguments) */
int isstrvar(char *lp) {
	if (!isbeginnamechar(lp)) return FALSE;
	while (isnamechar(lp)) lp++;
	if (*lp!='$') return FALSE;
	else lp++;
	if (*lp=='(' && *(lp+1)==')') return TRUE;
	if (*lp!='(') return TRUE;
	return FALSE;
}	


/* isnumvar(string-position)
 * (IS a NUMeric VAR)
 * System function for the DIM management 
 * Note: the figure var() is trated as a var name (used in sub arguments) */
int isnumvar(char *lp) {
	if (!isbeginnamechar(lp)) return FALSE;
	while (isnamechar(lp)) lp++;
	if (*lp=='$') return FALSE;
	if (*lp=='(' && *(lp+1)==')') return TRUE;
	return TRUE;
}	


/* dimArray(matrix-index,type,rows-number,columns-number)
 * dimension a numerical array allocating the necessary memory space.
 * var: the index of the variable holding the matrix name
 * T: matrix type, 1 for unidimensional vectors, 2 for matrices
 * M: the row number (=0 for one rows matrices)
 * N: the column number (=0 for vectors and one columns matrices)
 * Note: the one-dimensional vector is supposed to be vertical (one column).
 * Note: controls over M&N are supposed to happen elsewhere.
 * Note: all elements of declared array elements are set to 0.0 = zero. 
 * System function for the DIM management */
void dimArray(int var, int T, int M, int N) {
	int i,j;
	int size1=sizeof(double)*((M+1)*2+1);
	int size2=sizeof(double)*((M+1)*(N+1)+1);
	dim[var].brow  = M;
	dim[var].bcol  = N;
	dim[var].row   = M;
	dim[var].col   = N;
	dim[var].type  = T;

	/* reserve memory space for the array */
	if (dim[var].elem!=NIL) {
		free(dim[var].elem);
		dim[var].elem=NIL;
	}
	if (T==1) dim[var].elem=xmalloc((size_t)size1);  
	else if (T==2) dim[var].elem=xmalloc((size_t)size2);
	else perror_(__LINE__,35,l);
	if (!dim[var].elem) perror_(__LINE__,190,l);

	/* erase all array elements */
	if (1==T) for(i=0;i<M+1;i++) 
		setArrayValFast(var,T,i,0,0);
	else for(i=0;i<M+1;i++) for(j=0;j<N+1;j++) 
		setArrayValFast(var,T,i,j,0);
}	


/* setArrayValFast(matrix-index,type,row-position,column-position,value)
 * a subset of setArrayVal() for fast assignments 
 * System function for the dimArray function to override the BASE */
void setArrayValFast(int var, int t, int m, int n, double val){
	/* vectors */
	if (m>0 && !n && t==1) *(dim[var].elem+(dim[var].col+1)*m)=val;
	/* horizontal vectors */
	else if (!m && n>0 && t==1) *(dim[var].elem+n)=val;
	/* matrix identified by one column A(I) = A(0,I) */
	else if (!m && n>0 && t==2) *(dim[var].elem+n)=val;
	/* regular matrix */
	else if (m>0 && n>=0 && t==2) *(dim[var].elem+(dim[var].col+1)*m+n)=val;
}


/* setArrayStrFast(matrix-index,type,row-position,column-position,value)
 * a subset of setArrayStr() for fast assignments 
 * System function for the dimArray function to override the BASE */
void setArrayStrFast(int var, int t, int m, int n, char* str){
	/* vectors */
	if (m>0 && !n && t==1) strncpy_(dimS[var].elem+(dimS[var].col+1)*m*COMPSTR,str,COMPSTR);
	/* horizontal vectors */
	else if (!m && n>0 && t==1) strncpy_(dimS[var].elem+n*COMPSTR,str,COMPSTR);
	/* matrix identified by one column A(I) = A(0,I) */
	else if (!m && n>0 && t==2) strncpy_(dimS[var].elem+n*COMPSTR,str,COMPSTR);
	/* regular matrix */
	else if (m>0 && n>=0 && t==2) strncpy_(dimS[var].elem+(n+(dimS[var].col+1)*m)*COMPSTR,str,COMPSTR);
}


/* getArrayVal(matrix-index,type,row-position,column-position)
 * get value of an element of a numerical array, given its coordinates.
 * var: the index of the variable holding the matrix name
 * T: matrix type, 1 for unidimensional vectors, 2 for matrices
 * M: the row number (=0 for one rows matrices)
 * N: the column number (=0 for vectors and one columns matrices)
 * Note: a two dimensional array may be referenced using the first subscript;
 * Note: the second subscript defaults to zero for DIM and to one for MAT.
 * System function for the DIM management */
double getArrayVal(int var, int t, int m, int n) {
	/* check BASE validity */
	if (t==1) {
	       if (m<BASE) perror_(__LINE__,108,l);
	}
	else if (t==2 && !m && n) {
		if (n<BASE) perror_(__LINE__,108,l);
	}
	else if (m<BASE || n<BASE) perror_(__LINE__,108,l);
	/* trivial case of A(0,0) */
	if (!m && !n) {
		if (!dim[var].type) dimArray(var,2,10,10);
		return *(dim[var].elem);
	}
	/* vectors */
	else if (m>0 && !n && t==1) {
		if (m<=10 && 0==dim[var].type) dimArray(var,1,10,0);
		if (m>dim[var].row) perror_(__LINE__,76,l);
		return *(dim[var].elem+m);
	}
	/* horizontal vectors */
	else if (!m && n>0 && t==1) {
		if (n<=10 && 0==dim[var].type) dimArray(var,1,0,10);
		if (n>dim[var].col) perror_(__LINE__,76,l);
		return *(dim[var].elem+n);
	}
	/* matrix identified by one column A(I) = A(0,I) - first row */
	else if (!m && n>0 && t==2) {
		if (n<=10 && 0==dim[var].type) dimArray(var,2,10,10);
		if (n>dim[var].col) perror_(__LINE__,76,l);
		return *(dim[var].elem+n);
	}
	/* regular matrix */
	else if (m>0 && n>=0 && t==2) {
		if (m<=10 && 0==dim[var].row && n<=10 && dim[var].col==0)       
			dimArray(var,2,10,10);
		if (m>dim[var].row || n>dim[var].col) perror_(__LINE__,76,l);
		return *(dim[var].elem+(dim[var].col+1)*m+n); 
	}
	else perror_(__LINE__,76,l);
	return 0;
}


/* setArrayVal(matrix-index,type,row-position,column-position,value)
 * set an element of a numerical array to a specific value, given the value 
 * itself and the element coordinates.
 * var: the index of the variable holding the matrix name
 * t: matrix type, 1 for unidimensional vectors, 2 for matrices
 * m: the row number (=0 for one rows matrices)
 * n: the column number (=0 for vectors and one columns matrices)
 * val: the double value to be stored.
 * Note: a two dimensional array may be referenced using the first subscript;
 * Note: the second subscript defaults to zero  for DIM and to one for MAT.
 * System function for the DIM management */
void setArrayVal(int var, int t, int m, int n, double val){
	/* set integer value if the case */
	if (vn[var].isInt) val=(int)val;
	/* check BASE validity */
	if (t==1) {
	       if (m<BASE) perror_(__LINE__,108,l);
	}
	else if (t==2 && !m && n) {
		if (n<BASE) perror_(__LINE__,108,l);
	}
	else if (m<BASE || n<BASE) perror_(__LINE__,108,l);
	/* trivial case of A(0,0) */
	if (!m && !n) {
		if (!dim[var].type) dimArray(var,2,10,10);
		*(dim[var].elem)=val;
	}
	/* vectors */
	else if (m>0 && !n && t==1) {
		if (m<=10 && 0==dim[var].type) dimArray(var,1,10,0);
		if (m>dim[var].row) perror_(__LINE__,76,l);
		*(dim[var].elem+m)=val;
	}
	/* horizontal vectors */
	else if (!m && n>0 && t==1) {
		if (n<=10 && 0==dim[var].type) dimArray(var,1,0,10);
		if (n>dim[var].col) perror_(__LINE__,76,l);
		*(dim[var].elem+n)=val;
	}
	/* matrix identified by one column A(I) = A(0,I) - first row */
	else if (!m && n>0 && t==2) {
		if (n<=10 && 0==dim[var].type) dimArray(var,2,10,10);
		if (n>dim[var].col) perror_(__LINE__,76,l);
		*(dim[var].elem+n)=val;
	}
	/* regular matrix */
	else if (m>0 && n>=0 && t==2) {
		if (m<=10 && 0==dim[var].row && n<=10 && dim[var].col==0) 
			dimArray(var,2,10,10);
		if (m>dim[var].row || n>dim[var].col) perror_(__LINE__,76,l);
		*(dim[var].elem+(dim[var].col+1)*m+n)=val;
	}
	else perror_(__LINE__,76,l);
}


/* dimStrArray(matrix-index,type,rows-number,columns-number)
 * dimension a string array, allocating the necessary memory space; 
 * var: the index of the variable holding the matrix name
 * T: the matrix type, 1 for unidimensional vectors, 2 for matrices (currently
 *    not available)
 * M: the row number (=0 for one rows matrices)
 * N: the column number (=0 for vectors and one columns matrices)
 * Note: the one-dimensional vector is supposed to be vertical (one column).
 * Note: controls over M&N are supposed to happen elsewhere 
 * Note: all elements of declared array are set to the void string.
 * System function for the DIM management */
void dimStrArray(int var, int T, int M, int N) {
	int i,j;
	int size1=sizeof(char)*((M+1)*COMPSTR+1);
	int size2=sizeof(char)*((M+1)*(N+1)*COMPSTR+1);
	dimS[var].brow = M;
	dimS[var].row  = M;
	dimS[var].bcol = N;
	dimS[var].col  = N;
	dimS[var].type = T;
	
	/* reserve memory space for the array */
	if (dimS[var].elem!=NIL) {
		free(dimS[var].elem);
		dimS[var].elem=NIL;
	}
	if (T==1) dimS[var].elem=xmalloc((size_t)size1);  
	else if (T==2) dimS[var].elem=xmalloc((size_t)size2);
	else perror_(__LINE__,35,l);

	if (!dimS[var].elem) perror_(__LINE__,190,l);

	/* erase all array elements */
	if (1==T) for(i=0;i<M+1;i++) 
		setArrayStrFast(var,T,i,0,"");
	else for(i=0;i<M+1;i++) for(j=0;j<N+1;j++) 
		setArrayStrFast(var,T,i,j,"");

}		

/* getArrayStr(matrix-index,type,row-position,column-position)
 * get value of an element of a string array, given its coordinates.
 * var: the index of the variable holding the matrix name
 * t: matrix type, 1 for unidimensional vectors, 2 for matrices (currently
 *    not available)
 * m: the row number (=0 for one rows matrices)
 * n: the column number (always 0)
 * Note: I keep the form alive, to implement the bidimensional arrays in some
 * future.
 * System function for the DIM management */
char* getArrayStr(int var, int t, int m, int n) {
	/* check BASE validity */
	if (t==1) {
	       if (m<BASE) perror_(__LINE__,108,l);
	}
	else if (t==2 && !m && n) {
		if (n<BASE) perror_(__LINE__,108,l);
	}
	else if (m<BASE || n<BASE) perror_(__LINE__,108,l);
	/* trivial case of A$(0,0) */
	if (!m && !n) {
		if (!dimS[var].type) dimStrArray(var,2,10,10);
		return dimS[var].elem;
	}
	/* vectors */
	else if (m>0 && !n && t==1) {
		if (m<=10 && 0==dimS[var].type) dimStrArray(var,1,10,0);
		if (m>dimS[var].row) perror_(__LINE__,76,l);
		return dimS[var].elem+m*COMPSTR;
	}
	/* horizontal vectors */
	else if (!m && n>0 && t==1) {
		if (n<=10 && 0==dimS[var].type) dimStrArray(var,1,0,10);
		if (n>dimS[var].col) perror_(__LINE__,76,l);
		return dimS[var].elem+n*COMPSTR;
	}
	/* matrix identified by one column A(I) = A(0,I) - first row */
	else if (!m && n>0 && t==2) {
		if (n<=10 && 0==dimS[var].type) dimStrArray(var,2,10,10);
		if (n>dimS[var].col) perror_(__LINE__,76,l);
		return dimS[var].elem+n*COMPSTR;
	}
	/* regular matrix */
	else if (m>0 && n>=0 && t==2) {
		if (m<=10 && 0==dimS[var].row && n<=10 && dimS[var].col==0)
			dimStrArray(var,2,10,10);
		if (m>dimS[var].row || n>dimS[var].col) perror_(__LINE__,76,l);
		return dimS[var].elem+(n+(dimS[var].col+1)*m)*COMPSTR; 
	}
	else perror_(__LINE__,76,l);
	return NULLSTR;
}


/* setArrayStr(matrix-index,type,row-position,column-position,value)
 * set an element of a string array to a specific value, given the value itself 
 * and the element coordinates 
 * Parameter n is not used, since string matrices are not yet implemented;
 * var: the index of the variable holding the matrix name
 * t: matrix type, 1 for unidimensional vectors, 2 for matrices (currently
 *    not available)
 * m: the row number (=0 for one rows matrices)
 * n: the column number (always 0)
 * Note: I keep the form alive, to implement the bidimensional arrays in some
 * future.
 * System function for the DIM management */
void setArrayStr(int var, int t, int m, int n, char *str){
	/* check BASE validity */
	if (t==1) {
	       if (m<BASE) perror_(__LINE__,108,l);
	}
	else if (t==2 && !m && n) {
		if (n<BASE) perror_(__LINE__,108,l);
	}
	else if (m<BASE || n<BASE) perror_(__LINE__,108,l);
	/* trivial case of A$(0,0) */
	if (!m && !n) {
		if (!dimS[var].type) dimStrArray(var,2,10,10);
		strncpy_(dimS[var].elem,str,COMPSTR);
	}
	/* vectors */
	else if (m>0 && !n && t==1) {
		if (m<=10 && 0==dimS[var].type) dimStrArray(var,1,10,0);
		if (m>dimS[var].row) perror_(__LINE__,76,l);
		strncpy_(dimS[var].elem+m*COMPSTR,str,COMPSTR);
	}
	/* horizontal vectors */
	else if (!m && n>0 && t==1) {
		if (n<=10 && 0==dimS[var].type) dimStrArray(var,1,0,10);
		if (n>dimS[var].col) perror_(__LINE__,76,l);
		strncpy_(dimS[var].elem+n*COMPSTR,str,COMPSTR);
	}
	/* matrix identified by one column A(I) = A(0,I) - first row */
	else if (!m && n>0 && t==2) {
		if (n<=10 && 0==dimS[var].type) dimStrArray(var,2,10,10);
		if (n>dimS[var].col) perror_(__LINE__,76,l);
		strncpy_(dimS[var].elem+n*COMPSTR,str,COMPSTR);
	}
	/* regular matrix */
	else if (m>0 && n>=0 && t==2) {
		if (m<=10 && 0==dimS[var].row && n<=10 && dimS[var].col==0) 
			dimStrArray(var,2,10,10);
		if (m>dimS[var].row || n>dimS[var].col) perror_(__LINE__,76,l);
		strncpy_(dimS[var].elem+(n+(dimS[var].col+1)*m)*COMPSTR,str,COMPSTR);
	}
}


/* getBound()
 * return the lowest or the upper bound of array in argument.
 * The request is in the form L/UBOUND(V[,N]), where V is the array name and
 * N (optional) is the dimension. If omitted, dimension 1 is considered;
 * At present, N may be 1 or 2; if different, an error message is shown
 * if mode = TRUE search for UBOUND (upper bound)
 * if mode = FALSE, search for LBOUND (lower bound = BASE).
 * The array is identified by its letters, if () is reported, it is skipped
 * system function for the LBOUND and UBOUND functions */
int getBound(int mode) {
	int lb=BASE, ub=0,cs=0;
	int var=0, alt=1;
	int number=TRUE;	// if false, it is a string array

	/* get array name */
	forceArray=TRUE;
	cs=findDEC(p);
	if (cs && _rettype==NUM && _arr) 
		var=cs, p+=_len;
	else if (cs && _rettype==STR && _arr) 
		var=cs,p+=_len,number=FALSE;
	else perror_(__LINE__,42,l);
	forceArray=FALSE;
	if (*p=='(' && *(p+1)==')') p+=2;
	/* dimension index is explicit */
	if (*p==',') {
		p++;
		alt=(int)S();
		if (alt<1 || alt>2) perror_(__LINE__,148,l);
	}

	/* LBOUND returns always BASE */
	if (!mode) return lb;

	/* return number UBOUND */
	if (number) {
		if (alt==1) ub=dim[var].row;
		else ub=dim[var].col;
	}
	/* return string UBOUND */
	else  {
		if (alt==1) ub=dimS[var].row;
		else ub=dimS[var].col;
	}
	return ub;
}


/* getRandom()
 * return a pseudo-random number, called by the RND function.
 * The algorithm is the following:
 *  1. randSeed and randDivi at start are respectively 7139 and 32768
 *  2. divide randSeed by randDivi, and get the decimals part, from third digit
 *     to obtain a number greater than 0 (excluded) and lower than 1 (excluded) 
 *     as the generated random number (0.217864990 -> 0.7864990)
 *  3. the first five decimals of this number are taken as the new randSeed for 
 *     the next iteration (randSeed=21787)
 *  4. proceed from step 2.
 * Note: to reduce the risk of circular ranges I add two variables, varying
 * continuously: 'fix' that alters randSeed and div that alters randDivi; to
 * reduce also the risk of having 'exact' divisions (like 0.250000, which is 
 * 'unfair' as random number) randSeed is always odd, randDivi is even.
 * Note: This algorithm, after a variable length list of generated numbers, 
 * falls into a cycle, whose length is exactly 8532532; the start of the cycle
 * is variable, and depends on the initial values for randSeed and randDivi.
 * Note: the values of 27268 and 32767 for the starting values of randSeed and
 * randDivi were chosen to return as the first random number the same first
 * random number generated by the DEC-BASIC
 * System function for the pseudo-random number generation management */
double getRandom(void) {
	static double fix=0.3; // variable randSeed part
	static double div=0.1; // variable randDivi part
	double o=100.0*randSeed/randDivi;

	/* take fractional part of the division of randSeed by randDivi
	 * (take digits from 3rd to 8th position) */
	o-=(int)o;

	/* update randSeed */
	randSeed=(unsigned int)((o*1E5)+fix);
	if (randSeed<1E3) randSeed=27268;  
	if (!randSeed%2) randSeed++;

	/* update randDivi */
	randDivi=(unsigned int)((randDivi*randSeed)%0xFFFFF+div);
	if (randDivi<1E3) randDivi=32767; 
	if (randDivi%2) randDivi++;

	/* update variable parts */
	div+=3.0; if (div>499) div=-997;
	fix+=167.0; if (fix>4999) fix=0;

	/* ensure limits are respected (see previous note) */
	if (o<ZERO||o>=1.0) return getRandom(); // discard also neg. numbers
	return o;
}


/* isARing()
 * check if GOSUB is calling an address that is the GOSUB chain; if so,
 * the subroutine is calling itself
 * System function used into GOSUB() */
int isARing(int num) {
	int i=0;
	
	while (i<=gosubTOS) {
		if (gosubStack[i]==num) return TRUE;
		i++;
	}
	return FALSE;
}


/* isSingleDEF()
 * check if the DEF function starting at d is a single line DEF (return SINGLE)
 * a multiline DEF (return MULTI) or a function with bugs (return FALSE) 
 * d: points to the characters after DEF (which must be 'FN')  
 * System function for the DEFFN management */
int isSingleDEF(char* d) {
	/* FN */
	if (!strncasecmp_(d,"FN",2)) d+=2;
	/* XX - error */
	else return FALSE;
	/* FNA */
	if (isalpha(*d)) d++;
	/* FN_ - error */
	else return FALSE;
	/* FNA (multi line introduction) with no arguments */
	if (!*d) return MULTI;
	/* FNA( */
	else if (*d=='(') d++;
	/* FNA= (no arguments) */
	else if (*d=='=') return SINGLE;
	/* FNA_ - error */
	else return FALSE;
	/* FNA(.... */
	while (*d && *d!=')') d++;
	/* FNA(....) */
	if (*d==')') d++;
	/* FNA(... - error */
	else return FALSE;
	/* FNA(....)= */
	if (*d=='=') return SINGLE;
	/* FNA(....) */
	else return MULTI;
}


/* calcDEF() 
 * calculates the DEF FN result for a single/multi line DEF FNF; 
 * data relative to DEF FN functions must have been preprocessed by preParse(),
 * In order for calcDEF to run.
 * the double result of DEF calculation will be store in calcres
 * p points to the character after FN, which contains the function index.
 * Return the updated position of p
 * System function for the DEFFN management */
char *calcDEF(void) {
	char *pb;
	int defn=*p++, type, i;
	int fmb[MEMLIMIT], cnt, decCountBack=decCount;
	int isVarBack, isRunBack, isDefCalcBack, ic;
	static int stackLim;

	/* control self calling subroutine
	 * Note: the control is done this way:
	 * 1. store information about current DEF FN only if it differs;
	 *    this enables writing recursion functions; 
	 * 2. the recursion is limited to STACKLIMIT jumps;
	 * 3. search until previous position;
	 *    this avoids calling error in case of recursion, but detects
	 *    circular calls (FNA calling FNB calling FNA) */
	if (!FNXTOS || FNXlist[FNXTOS]!=defn) 
		FNXlist[++FNXTOS]=defn, stackLim=0;
	else stackLim++; 
	if (stackLim>STACKLIM) perror_(__LINE__,187,l);
	for (i=1;i<FNXTOS-1;i++) {
		if (FNXlist[i]==defn) perror_(__LINE__,126,l);
	}

	/* determine type (SINGLE/MULTI) */
	type=fn[defn].type;
	
	/* backup masking line state to enable execution of DEF */
	for (i=fn[defn].start;i<=fn[defn].end;i++) fmb[i]=fm[i],fm[i]=FALSE;
	
	/* if there are arguments store them */
	if (*p=='(')  {
		cnt=0;
		ic=1;
		++p; // skip '('
		/* cycle through SUB arguments and get values to be used; 
		 * cnt holds number of arguments of called proc, stored by 
		 * DEF() in fn[defn].vars[0] list (which must be respected) */
		while (cnt < fn[defn].vars[0]) {
			cnt++;
			/* DEC variable */
			if (fn[defn].vars[ic]>=DECADD) {
				int rettype=0, namelen=0;
				char name[COMPSTR], *np=name;
				ic++; // skip DECADD
				/* store name until -1 */
				while (fn[defn].vars[ic]>=0)
					*np++=fn[defn].vars[ic++];
				*np='\0'; ic++; // skip -1
				rettype=fn[defn].vars[ic]-DECADD;
				ic++;
				/* DEC numeric variable */
				if (rettype==NUM) {
					int var=decCount+1;
					if (isanystring(p)) 
						perror_(__LINE__,215,l);
					P[var]=S();
					strncpy_(PS[var],"",COMPSTR);
				}
				else perror_(__LINE__,200,l);
				/* instantiate DEC variable */
				decCount++;
				if (decCount>2*LIMIT) perror_(__LINE__,7,l);
				vn[decCount].isInt=FALSE;
				vn[decCount].isArr=FALSE;
				vn[decCount].isConst=FALSE;
				vn[decCount].rettype=rettype;
				vn[decCount].isCommon=FALSE;
				namelen=((int)strlen(name)+1)*sizeof(char);
				if (vn[decCount].name!=NIL) {
					free(vn[decCount].name);
					vn[decCount].name=NIL;
				}
				vn[decCount].name=xmalloc((size_t)namelen);
				if (!vn[decCount].name) 
					perror_(__LINE__,173,l);
				else strncpy_(vn[decCount].name,name,namelen);
				/* debug feature */
				if (isDebug) showDEC(decCount, FALSE);
			}
			else perror_(__LINE__,119,l);
			if (*p==',') p++;
		}
	}
	else cnt=0;
	
	/* wrong number of arguments */
	if (fn[defn].vars[0]!=cnt) perror_(__LINE__,37,l);

	/* calculate a single line DEF */
	if (type==SINGLE) {
		pb=p; 
		p=fn[defn].def;
		calcres=S(); 
		p=pb; 
	}
	/* calculate a multi line DEF */
	else {
		/* execute lines */
		pb=p;
		isRunBack=isDefRun; isDefRun=TRUE;
		isVarBack=isDefVar; isDefVar=defn;
		isDefCalcBack=isDefCalc;
		pushRUN();
		parse(m,fn[defn].start,fn[defn].end);
		popRUN();
		isDefVar=isVarBack;
		isDefRun=isRunBack;
		/* there's no error message for cases in which
		 * LET FNF= is not used within a multiline DEF */
		if (!isDefCalc) calcres=0.0;
		else calcres=fn[defn].res;
		isDefCalc=isDefCalcBack;
		p=pb;
	}

	/* restore masking state */
	for (i=fn[defn].start;i<=fn[defn].end;i++) fm[i]=fmb[i];

	/* update counter for circular calling check */
	FNXTOS--;
	decCount=decCountBack;

	return p;
}


/* findSUB() 
 * find SUB in current parsing line, among SUBs found by preParse()
 * p points to the SUB name in getFunc(), exec() or SS() 
 * return in p2 updated position of p (just after the proc name), unless
 *    p2 is NIL
 * store in type and rettype relative values for an immediate check 
 * System function for the SUB management */
int findSUB(char *p, char **p2, int *type, int *rettype) {
	int i=subCount,isStr;
	char name[COMPSTR], *n=name;

	/* get SUB name */
	isStr=FALSE;
	while (isnamechar(p)) *n++=*p++;
	if (*p=='$') *n++=*p++,isStr=TRUE;
	*n='\0';

	/* return immediately if it's last call */
	if (!strcmp(rememberName,name)) {
		if (p2) *p2=p;
		*type=sn[rememberCode].type;
		*rettype=sn[rememberCode].rettype;
		return rememberCode;
	}

	/*  find SUB */
	while (i>0) {
		if (sn[i].name!=NIL) {
			if (!strcmp(sn[i].name,name)) {
				*type=sn[i].type;
				*rettype=sn[i].rettype;
				if (p2) *p2=p;
				if (*type==FUNC && *rettype==NUM && isStr) {
					i++;
					continue;
				}
				strncpy_(rememberName,sn[i].name,COMPSTR);
				rememberCode=i;
				return i;
			}
		}
		i--;
	}
	/* restore p: SUB not found */
	if (p2) *p2=p;
	return FALSE;
}


/* execSUB() 
 * execute a subroutine; 
 * data relative to the SUB are preprocessed by preParse(), in order for 
 * execSUB to run.
 * subn: the sub number found by findSUB()
 * pc: the position in the incoming string where arguments (if any) are stored,
 *    pointing to the character after SUB name. pc is used local, instead of
 *    global p, because execSUB() is supposed to execute any type of string
 * numres: the double result pointer
 * strres; the string result pointer
 * return the updated position of pc in pc
 * System function for the SUB management */
char* execSUB(int subn, char*pc, double *numres, char* strres) {
	char *pb, subname[COMPSTR];
	int i,isSubVarBack,isSubCalcBack,isSubRunBack;
	int fmb[MEMLIMIT], cnt, ic, isInt=FALSE, isArr=FALSE;

	/* SUB does not exist */
	if (!sn[subn].number) perror_(__LINE__,212,l);
	subTOS=subn;
	XCount++;
	if (XCount>=RUNSTACK) perror_(__LINE__,195,l);
	saveSubCount[XCount]=decCount;

	/* set sub name */
	if (sn[subn].name[0]) strncpy_(subname,sn[subn].name,COMPSTR);
	else strncpy_(subname,"obj.",COMPSTR);

	/* backup masking lines state to enable execution of SUB */
	for (i=0;i<MEMLIMIT;i++) fmb[i]=fm[i],fm[i]=FALSE;
	
	/* if there are arguments store them (it works even in case of 
	 * the () figure */
	if (*pc=='(')  {
		cnt=0;
		ic=1;
		++pc; // skip '('
		/* cycle through SUB arguments and get values to be used; 
		 * cnt holds number of arguments of called proc, stored by 
		 * SUB() in sn[subn].vars[0] list (which must be respected) */
		while (cnt < sn[subn].vars[0]) {
			cnt++;
			/* DEC variable */
			if (sn[subn].vars[ic]>=DECADD) {
				int rettype=0, namelen=0;
				char name[COMPSTR], *np=name;
				ic++; // skip DECADD
				/* store name until -1 */
				while (sn[subn].vars[ic]>=0)
					*np++=sn[subn].vars[ic++];
				*np='\0'; ic++; // skip -1
				rettype=sn[subn].vars[ic]-DECADD;
				ic++; // skip DECADD + ret
				/* DEC numeric variable */
				if (rettype==NUMINT) {
					int var=decCount+1;
					if (isanystring(pc)) 
						perror_(__LINE__,215,l);
					p=pc;
					P[var]=(int)S();
					pc=p;
					strncpy_(PS[var],"",COMPSTR);
					isInt=TRUE;
					rettype=NUM;
				}
				else if (rettype==NUM) {
					int var=decCount+1;
					if (isanystring(pc)) 
						perror_(__LINE__,215,l);
					p=pc;
					P[var]=S();
					pc=p;
					strncpy_(PS[var],"",COMPSTR);
				}
				/* DEC string variable */
				else if (rettype==STR) {
					int var=decCount+1;
					char lop[COMPSTR];
					if (!isanystring(pc)) 
						perror_(__LINE__,214,l);
					p=pc;
					strncpy_(lop,SS(),COMPSTR);
					pc=p;
					strncpy_(PS[var],lop,COMPSTR);
					P[var]=0.0;
				}
				/* instantiate DEC variable */
				decCount++;
				if (decCount>2*LIMIT) 
					perror_(__LINE__,223,l);
				vn[decCount].isInt=isInt;
				vn[decCount].isArr=isArr;
				vn[decCount].rettype=rettype;
				vn[decCount].isCommon=FALSE;
				namelen=strlen(name)+1;
				if (vn[decCount].name!=NIL) {
					free(vn[decCount].name);
					vn[decCount].name=NIL;
				}
				vn[decCount].name=xmalloc((size_t)namelen);
				if (!vn[decCount].name) 
					perror_(__LINE__,173,l);
				else strncpy_(vn[decCount].name,name,namelen);
				/* debug feature */
				if (isDebug) showDEC(decCount, FALSE);
			}
			else perror_(__LINE__,119,l);
			if (*pc==',') pc++;
		}
	}
	else cnt=0;

	/* wrong number of arguments */
	if (sn[subn].vars[0]!=cnt) perror_(__LINE__,37,l);

	/* execute lines */
	pb=pc;
	isSubRunBack=isSubRun; isSubRun=TRUE;
	isSubVarBack=isSubVar; isSubVar=subn;
	isSubCalcBack=isSubCalc;
	pushRUN();
	parse(m,sn[subn].start,sn[subn].end);
	popRUN();
	isSubVar=isSubVarBack;
	isSubRun=isSubRunBack;
	/* there's no error message for cases in which
	 * RETURN is not called within a SUB stored as function */
	if (sn[subn].type==FUNC) {
	       if (!isSubCalc && sn[subn].rettype==NUM) *numres=0.0;
	       else if (!isSubCalc && sn[subn].rettype==STR) 
		       strncpy_(strres,"",COMPSTR);
	       else if (isSubCalc && sn[subn].rettype==NUM) 
		       *numres=sn[subn].numres;
	       else if (isSubCalc && sn[subn].rettype==STR) 
		       strncpy_(strres,sn[subn].strres,COMPSTR);
	}
	isSubCalc=isSubCalcBack;
	pc=pb;
	
	/* restore masking lines state and make SUB invisible again */
	for (i=0;i<MEMLIMIT;i++) fm[i]=fmb[i];
	
	/* closing parenthesis is automatic for getFunc() and SS() */
	if (sn[subn].type==PROC && *pc==')') pc++;
	/* restore and go */
	decCount=saveSubCount[XCount];
	XCount--;
	p=pc;
	return pc;
}


/* findDEC() 
 * find a DECLAREd variable/array in the vn[] structure;
 * lp points to the DEC name in getFunc(), LET() or SS() and get its value;
 * store in _len the length of the variable name
 * store in _rettype relative values for an immediate check
 * store in _arr the array flag
 * Instantiate the variable if not found and if not in EXPLICIT mode
 * System function for the DECLARE management */
int findDEC(char *lp) {
	int i=decCount, rr=NUM, ii=FALSE;
	char name[COMPSTR], *n=name;
	int lArray=FALSE;	// locally detection of arrays
	int found=FALSE;

	/* set anyway a valid value as a safety measure */	
	_len=0;
	_arr=0;
	_rettype=0;

	if (!*lp) return FALSE;
	if (!isbeginnamechar(lp)) return FALSE;

	/* get DEC name */
	while (isnamechar(lp)) *n++=*lp++;
	if (*lp=='%' && *(lp+1)!='%') *n++=*lp++, ii=TRUE;
	else if (*lp=='%' && *(lp+1)=='%' && *(lp+2)=='%') 
		*n++=*lp++, ii=TRUE;
	else if (*lp=='$') *n++=*lp++, rr=STR;
	*n='\0';
	_len=(int)strlen(name);
	if (!_len || !name[0]) return FALSE;
	if (*lp=='(') lArray=TRUE;

	/*  find DEC and attribute values */
	while (i>0) {	// search all variables
		if (vn[i].name!=NIL && !strcmp(vn[i].name,name)) {
			if (!vn[i].isArr && !forceArray && !lArray)
				found=TRUE;
			else if (vn[i].isArr && lArray) 
				found=TRUE;
			else if (vn[i].isArr && !lArray && forceArray)
				found=TRUE;
			if (found) {
				_rettype=vn[i].rettype;
				_arr=vn[i].isArr;
				return i;
			}
		}
		found=FALSE;
		i--;
	}

	/* DEC not found: try to instantiate the variable */
	if (!isExplicit && *lp!='(' && !forceArray) {
		int namelen=0;
		decCount++;
		if (decCount>2*LIMIT) perror_(__LINE__,223,l);
		vn[decCount].isInt=ii;
		vn[decCount].rettype=_rettype=rr;
		vn[decCount].isArr=_arr=FALSE;
		vn[decCount].isConst=FALSE;
		vn[decCount].isCommon=FALSE;
		_len=(int)strlen(name);
		namelen=(_len+1)*sizeof(char);
		if (vn[decCount].name!=NIL) {
			free(vn[decCount].name);
			vn[decCount].name=NIL;
		}
		vn[decCount].name=xmalloc((size_t)namelen);
		if (!vn[decCount].name) perror_(__LINE__,173,l);
		else strncpy_(vn[decCount].name,name,namelen);
		P[decCount]=0.0;
		strncpy_(PS[decCount],"",COMPSTR);
		/* debug feature */
		if (isDebug) showDEC(decCount, FALSE);
		return decCount;
	}
	/* DEC not found: try to instantiate the array 
	 * Note: if forceArray is TRUE, the name comes from the MAT 
	 * istructions, and thus, even if it's a variable name, it must be 
	 * considered a matrix name, and it's made 11x11 because it's the
	 * most general use. Vectors must be instantiated explicitly */
	else if (!isExplicit && (*lp=='(' || forceArray)) {
		int namelen=0, M=0, N=0, T=0;
		char *pp=p;
		decCount++;
		if (*lp=='(' && *(lp+1)!=')') {
			T=1;
			lp++;
			p=lp; M=(int)S();
			if (*p==',') {
				p++;
				N=(int)S(); 
				T=2;
			}
			lp=p;
		}
		T=2;
		/* implicit dimensioning if values are into the 0-10 range */
		if (rr==NUM && M<=10 && N<=10) {
			if (T==1) dimArray(decCount,1,10,0);
			else if (T==2) dimArray(decCount,2,10,10);
		}
		else if (rr==STR && M<=10 && N<=10) {
			if (T==1) dimStrArray(decCount,1,10,0);
			else if (T==2) dimStrArray(decCount,2,10,10);
		}
		else perror_(__LINE__,76,l);
		if (decCount>2*LIMIT) perror_(__LINE__,223,l);
		vn[decCount].isInt=ii;
		vn[decCount].rettype=_rettype=rr;
		vn[decCount].isArr=_arr=TRUE;
		vn[decCount].isConst=FALSE;
		vn[decCount].isCommon=FALSE;
		_len=(int)strlen(name);
		namelen=(_len+1)*sizeof(char);
		if (vn[decCount].name!=NIL) {
			free(vn[decCount].name);
			vn[decCount].name=NIL;
		}
		vn[decCount].name=xmalloc((size_t)namelen);
		if (!vn[decCount].name) perror_(__LINE__,173,l);
		else strncpy_(vn[decCount].name,name,namelen);
		/* debug feature */
		if (isDebug) showDEC(decCount, FALSE);
		p=pp;
		return decCount;
	}

	/* DEC irremediably not found */
	_len=0;
	return FALSE;
}


/*****************************************
 *             TIME & STACK              *
 *****************************************/

/* pushRUN() 
 * store current running state
 * used into calcDEF() */
void pushRUN(void) {
	lb[runTOS]=l;
	sb[runTOS]=startingLine;
	pb[runTOS]=channel[0].fprintn;
	cb[runTOS]=channel[0].fcounter;
	eCb[runTOS]=endCore;
	nowb=now;
	runTOS++;
}


/* popRUN() 
 * restore current running state
 * used into calcDEF() */
void popRUN(void) {
	runTOS--;
	l=lb[runTOS];
	startingLine=sb[runTOS];
	channel[0].fprintn=pb[runTOS];
	channel[0].fcounter=cb[runTOS];
	endCore=eCb[runTOS];
	now=nowb;
}


/* swapDEC()
 * swap two DECLARE variables, given their var-value as arguments 
 * Variables are swapped in their totality, included the values in the
 * proper value-structures P/PS/dim/dimS */
void swapDEC(int v1, int v2) {
	char name1[COMPSTR], name2[COMPSTR]; // names buffer
	int i1=v1, i2=v2, namelen=0;
	// vars buffers 
	int nex;
	double ex;
	char Sex[COMPSTR];
	Tarray ax;
	TSarray Sax;

	/* exchange names in temporary placeholders*/
	if (vn[i2].name) strncpy_(name1,vn[i2].name,COMPSTR);
	else name1[0]='\0';
	if (vn[i1].name) strncpy_(name2,vn[i1].name,COMPSTR);
	else name2[0]='\0';
	/* set first name */
	if (vn[i1].name!=NIL) {
		free(vn[i1].name);
		vn[i1].name=NIL;
	}
	namelen=((int)strlen(name1)+1)*sizeof(char);
	vn[i1].name=xmalloc((size_t)namelen);
	if (!vn[i1].name) perror_(__LINE__,173,l);
	else strncpy_(vn[i1].name,name1,namelen);
	/* set second name */
	if (vn[i2].name!=NIL) {
		free(vn[i2].name);
		vn[i2].name=NIL;
	}
	namelen=((int)strlen(name2)+1)*sizeof(char);
	vn[i2].name=xmalloc((size_t)namelen);
	if (!vn[i2].name) perror_(__LINE__,173,l);
	else strncpy_(vn[i2].name,name2,namelen);

	/* exchange flags */
	nex=vn[i1].isInt;
	vn[i1].isInt=vn[i2].isInt;
	vn[i2].isInt=nex;	

	nex=vn[i1].isArr;
	vn[i1].isArr=vn[i2].isArr;
	vn[i2].isArr=nex;

	nex=vn[i1].rettype;
	vn[i1].rettype=vn[i2].rettype;
	vn[i2].rettype=nex;	
	
	nex=vn[i1].isCommon;
	vn[i1].isCommon=vn[i2].isCommon;
	vn[i2].isCommon=nex;	
	
	/* exchange values */
	ex=P[v1]; 
	P[v1]=P[v2]; 
	P[v2]=ex;
	
	strncpy_(Sex,PS[v1],COMPSTR); 
	strncpy_(PS[v1],PS[v2],COMPSTR); 
	strncpy_(PS[v2],Sex,COMPSTR);
	
	ax=dim[v1]; 
	dim[v1]=dim[v2]; 
	dim[v2]=ax;
	
	Sax=dimS[v1]; 
	dimS[v1]=dimS[v2]; 
	dimS[v2]=Sax;
}


/* fprintb() - used with BIN$()
 * return an integer in binary format according to the following format:
 *   - bit 0 appears on the right (e.g. 13 => 1101);
 *   - negative numbers occupy all 32 bits;
 *   - if mode=FALSE return a number with leading zeroes
 * 'res' is the output string;
 * 'n' is the integer to be converted. */
int fprintb(char *res, int n, int mode) {
	int num[BITS+1], c=-1;
	char *pp=res;
	if (!n && mode) *pp++='0';
	while (n && (c < (BITS - 1))) num[++c] = n & 1, n >>= 1;
	if (!mode) {
		int y=BITS;
		while (--y>c) *pp++='0';
	}
	while (c >= 0) *pp++='0'+num[c--];
	*pp='\0';
	return (int)strlen(res);
}


/* modulo() - modulo and division calculation 
 * if flag=TRUE perform classic math, otherwise perform Integer Theory math
 * A = dividend
 * B = divisor
 * *div = result of division
 * *mod = result of modulo 
 * Note: the case B=0 is treated with the restitution of mod=A and div=A;
 * actually, if 1/0 is performed, the result is an infinite, but to build
 * the original number with a divisor, a dividend, a quotient and a rest,
 * my algorithm worlks on the following theory:
 * A normal integer division, such as 7/2 yields 7=3x2+1, where
 * 	dividend = 7
 * 	divisor  = 2
 * 	quotient = 3
 * 	rest     = 1 (modulo)
 * Suppose now we have to treat 1/0; I suppose the result is inf; so I can 
 * assume that inf x 0 = 1 (or any computable number, but 1 fits my scopes). 
 * Thus we have:
 * 	dividend = 1
 * 	divisor  = 0
 * 	quotient = inf
 * 	rest     = ?
 * I assume rest=0 because inf multiplied by zero returns the divisor; thus, 
 * the rest must yield zero, as in the following equation:
 *      1 = inf x 0 + 0 = 1 + 0 = 1 */
void modulo(int A, int B, int *div, int *mod, int flag) {
	/* special case of null divisor */
	if (!B) {
		perror_(__LINE__,77,l); // warning
		*div=A<0?(int)MINFF:(int)INFF;
		*mod=0;
	}
	/* classic */
	else if (flag) {
		*div=A/B;
		*mod=A-*div*B;
	}
	/* integer theory */
	else {
		int k=A<0?(B<0?-1:1):0;
		*div=A/B;
		*mod=A-*div*B;
		if (*mod<0) {
			*div=*div-k;
			*mod=A-*div*B;
		}
	}
}


/* sign()
 * if argument is positive, return 1
 * if argument is negative, return -1
 * if argument is exactly zero, return 0 */
int sign(double s) {
	if (s<0.0) return -1;
	else if (s>0.0) return 1;
	return 0;
}


/* getTime()
 * return current time structure */
struct tm *getTime(void) {
	time_t st_lt;
	st_lt=time(NULL);
	return localtime(&st_lt);
}


/* getMin()
 * return current min (0-59) */
int getMin(void) {
	struct tm* vt = getTime();
	return (int) vt->tm_min;
}


/* getSec()
 * return current second (0-59) */
int getSec(void) {
	struct tm* vt = getTime();
	return (int) vt->tm_sec;
}


/* getHour()
 * return current hour (0-23) */
int getHour(void) {
	struct tm* vt = getTime();
	return (int) vt->tm_hour;
}


/* getWDay()
 * return current week day (0-6) */
int getWDay(void) {
	struct tm* vt = getTime();
	return (int) vt->tm_wday;
}


/* getDay()
 * return current day (1-31) */
int getDay(void) {
	struct tm* vt = getTime();
	return (int) vt->tm_mday;
}


/* getMonth()
 * return current month (1-12) */
int getMonth(void) {
	struct tm* vt = getTime();
	return (int) (vt->tm_mon+1);
}


/* getYear()
 * return current year */
int getYear(void) {
	struct tm* vt = getTime();
	return (int) (vt->tm_year+1900);
}


/* getMonthString()
 * return current month string */
char *getMonthString(void) {
	return month[getMonth()];
}


/* gregDayNumber()
 * return gregorian day number for date in arguments */
int gregDayNumber(int year,int month, int day) {
	int y,m;
	y=year; // it must be greater than 0
	if (y<=0) y=1;
	m=month;
	if (month < 3) y=y-1, m=m+12;
	return floor(365.25*y)-floor(y/100)+(y/400)+floor(30.6*(m+1))+day-62;
}


/* setYear()
 * adjust year value from a yy to a yyyy (setting for current century) 
 * - if value is negative or it is in the form yyyy and lesser than YEARLIMIT, 
 *   set it to YEARLIMIT and print a warning
 * - if yy is in the range 0-99, if it goes past 15 years from now, return
 *   current century + yy
 * - if yy is in the range 0-99, if it goes back 85 year from now, return
 *   previous century + yy
 * - otherwise return the year itself in the form yyyy
 * Note: since the gregDayNumber return wrong days count for my
 * algorithm if the year is lower than YEARLIMIT, I limit it to YEARLIMIT */
int setYear(int year) {
	int offset=getYear()+TIMEOFFSET-currcentury;
	if (year<=0.0) year=getYear();
	else if (year>=100) {
		if (year<YEARLIMIT) {
			perror_(__LINE__,160,l,YEARLIMIT);
			year=YEARLIMIT;
		}
	}
	else if (year<offset) year+=currcentury;
	else year+=currcentury-100;
	return year;
}
 


/*****************************************
 *        INPUT/OUTPUT FORMATTING        *
 *****************************************/

/* Note: the FAC()/FCR() functions are strictly connected with the 
 * PAGE/NOPAGE/<PA> features for the simulated terminal because they all 
 * use/set the page size limits */ 


/* FAC()
 * Advance Carriage on a file stream*/
void FAC(int ch) {
	channel[ch].fcounter++;
	fputc(' ',channel[ch].stream);
	if (!isRawPrint && channel[ch].fcounter > channel[ch].margin) 
		FCR(ch);
}


/* FCR()
 * Carriage Return onto file stream */
void FCR(int ch) {
	channel[ch].fcounter=0;
	channel[ch].pagerow++;
	/* force printing footer if this is last line */
	if (channel[ch].pagerow==channel[ch].height && channel[ch].pagenumber) {
		printFooter(ch);
		channel[ch].pagerow++;
	}
	if (channel[ch].pagerow>channel[ch].height) {
		if (channel[ch].page) channel[ch].page++;
		channel[ch].pagerow=1;
	}
	fputc('\n',channel[ch].stream);
}


/* fillPage()
 * fill page on current channel with void lines in case of PAGE statement set */
void fillPage(int ch) {
	while (channel[ch].pagerow<channel[ch].height) {
		channel[ch].pagerow++;
		fputc('\n',channel[ch].stream);
	}
	if (channel[ch].pagerow==channel[ch].height && channel[ch].pagenumber)
		printFooter(ch);
	else fputc('\n',channel[ch].stream);
	channel[ch].pagerow=1;
	if (channel[ch].page) channel[ch].page++;
}


/* printFooter()
 * print the "- PAGE # -" string at the page footer */
void printFooter(int ch) {
	int i,k=(channel[ch].margin-12)/2;
	for (i=0;i<k;i++) fputc(' ',channel[ch].stream);
	fprintf(channel[ch].stream,"- PAGE %d -\n",channel[ch].page);
}


/* emitNUM()
 * output number (with sign) contained into formatted string num to stream ch
 * or to string dest; update file counter ; 
 * a space follows always the number to stdout/file;
 * if flag is TRUE, redirect printing to dest, otherwise to ch 
 * Note: dest width must be COMPSTR = MAXSTRLEN + 1 */
int emitNUM(int ch, char minus, char *num, int flag, char* dest) {
	int ret=(int)strlen(num);
	char build[COMPSTR];

	/* reset screen/file position if number length exceeds width */
	if (ret+2+channel[ch].fcounter>channel[ch].margin) {
		if (!isRawPrint) FCR(ch);
	}

	/* intercept the RAW PRINT mode with positive number*/
	if (isRawPrint && minus!='-') sprintw(build,COMPSCR,"%s",num);
	/* intercept the RAW PRINT mode with negative number*/
	else if (isRawPrint) sprintw(build,COMPSCR,"%c%s",minus,num);
	/* intercept the normal mode with any-sign number*/
	else sprintw(build,COMPSCR,"%c%s ",minus,num,' ');

	/* print into a string */
	if (flag) sprintw(dest,COMPSCR,"%s",build);
	/* print to screen/file */
	else fprintf(channel[ch].stream,"%s",build);

	/* return correct value counter/file counter */
	return channel[ch].fcounter+2+ret;
}


/* printNUM()
 * format number num to a printing output according to DEC BASIC rules 
 * (ref. LBMAA-A-D 6-4) onto stream identified by ch
 * return updated position of counter (as yielded by emitNum) 
 * if flag is true, the number (in emitNUM) will be printed on string dest 
 * rather than on stdout 
 * Note: dest width must be COMPSTR = MAXSTRLEN + 1 
 * Note: this function does not print anything, delegating emitNUM() */
int printNUM(int ch, double num, int flag, char* dest) {
	int isFloat=TRUE;
	char minus=' ';
	char nums[COMPSCR], *nump;
	double base=0.0;
	int exp=0;
	
	/* intercept the OPTION ZERO feature (only in printing processes)
	 * print number lesser than ZEROLIM as zero */
	if (isZeroLim) {
		if (fabs(num) < ZEROLIM) num=0;
	}
	
	/* intercept the floating depth feature
	 * print number with the specified precision, rather than with the 
	 * standard BASIC printer engine */
	if (PRECISION>=0) {
		char pnum[64] ,dnum[64];
		int isInt=FALSE;
		pnum[0]='\0';
		if (num<0) num=-num, minus='-';
		if (num-(int)num==0.0) isInt=TRUE;
		if (num<pow(10,-PRECISION) && !isInt)
			sprintw(pnum,64,"%%.%dE",PRECISION);
		else if (!isInt)
			sprintw(pnum,64,"%%.%df",PRECISION);
		else 
			sprintw(pnum,64,"%%d");
		if (isInt) snprintf(dnum,64,pnum,(int)num);
		else snprintf(dnum,64,pnum,num);
		return emitNUM(ch,minus,dnum,flag,dest);
	}
	/* intercept the OPTION RAW PRINT ON feature */
	if (isRawPrint) 

	/* if number is null, shorten the process! */
	if (num==0) return emitNUM(ch,' ',"0",flag,dest);
	
	/* operate only with positive numbers, storing the sign aside */
	if (num<0) num=-num , minus='-';

	/* catch UNDERFLOW for a number less than ZERO */
	if (num<ZERO) return emitNUM(ch,minus,"0",flag,dest); 

	/* catch OVERFLOW for a number > INFF in absolute value */
	else if (num > INFF) num=INFF;

	/* temporarily, use exponential or normal form 
	 * in representing the number in a raw form */
	if (num < 1E-10 || num > 1E10) sprintw(nums,COMPSCR,"%.5E", num);
	else sprintw(nums,COMPSCR,"%21.10F", num);

	/* rule 4: suppress final zeroes */
	if ((int)strlen(nums)>0) nump=nums+(int)strlen(nums)-1;
	else nump=nums;
	while((int)(nump-nums)>0 && '0'==*nump) *nump--='\0';

	/* rule 1 part 1: if there's a final dot in an integer, suppress it */
	if ('.'==*nump) {
		*nump--='\0';
		isFloat=FALSE;
	}

	/* suppress any possible trailing nulls */
	nump=nums;
	while((int)(nump-nums)<COMPSCR && (*nump=='0' || *nump==' ')) nump++;

	/* print real number */
	if (isFloat){
		/* exception to rule 3: print real number as-is if it fits */
		if (num <= 1 && (int)strlen(nump) < 8) {
			sprintw(nums,COMPSCR,"0%s",nump);
			return emitNUM(ch,minus,nums,flag,dest);
		}
		else if ((int)strlen(nump) < 8) {
			sprintw(nums,COMPSCR,"%s",nump);
			return emitNUM(ch,minus,nums,flag,dest);
		}

		/* rule 3: print real with exponent */
		else if (num < 0.1) {
			sprintw(nums,COMPSCR,"%.5E",num);
			doExp(nums,&base,&exp);
			if (exp<0) sprintw(nums,COMPSCR,"%.5fE%d",base,exp);
			else sprintw(nums,COMPSCR,"%.5fE+%d",base,exp);
			return emitNUM(ch,minus,nums,flag,dest);
		}
		/* rule 2: print real with 6 significant digits 
		 * retaining decimal dot (use of #)*/ 
		else {
			sprintw(nums,COMPSCR,"%#.6G",num);
			nump=strstr(nums,"E");
			/* depict number in exponential form */
			if (nump) {
				doExp(nums,&base,&exp);
				if (exp<0) 
				     sprintw(nums,COMPSCR,"%.5fE%d",base,exp);
				else sprintw(nums,COMPSCR,"%.5fE+%d",base,exp);
			}
			/* rule 4 reapplied: suppress trailing zeroes at 
			 * the end if not and exponential */
			if (!strstr(nums,"E")) {
				if ((int)strlen(nums)>0) 
					nump=nums+strlen(nums)-1;
				else nump=nums;
				while ((int)(nump-nums)>0 && *nump=='0') 
					*nump--='\0';
			}
			return emitNUM(ch,minus,nums,flag,dest);
		}
	}
	/* print integer number */
	else {
		/* rule 1 part 2: print integer till 9 significant digits */
		if ((int)strlen(nump) < 9) 
			return emitNUM(ch,minus,nump,flag,dest);

		/* rule 1 part 2 (reverse case): print integer with exponent
		 * and 6 significant digits (that is, 5 decimals) */
		else { 
			sprintw(nums,COMPSCR,"%.5E",num);
			nump=strstr(nums,"E");
			if (nump) {
				doExp(nums,&base,&exp);
				if (exp <0) 
				     sprintw(nums,COMPSCR,"%.5fE%d",base,exp);
				else sprintw(nums,COMPSCR,"%.5fE+%d",base,exp);
			}
			return emitNUM(ch,minus,nums,flag,dest);
		}
	}

	/* avoiding warnings.... */
	return 0;

}


/* getString()
 * get a formatted string from user keyboard, excluding user delay */
char * getString(void) {
	char *lp=line;
	int i;
	double extraS;

	/* get timer */
        gettimeofday(&tv, NULL);
       	extraS = ((double)tv.tv_sec + tv.tv_usec / 1000000.0);

	/* get string */
	for (i=0; i<SCREENLINE; i++) line[i]='\0';
	while (TRUE) {
		*lp=getchar();
		if (*lp==10 || *lp==13 || *lp=='\n') {
			*lp='\0';
			break;
		}
		else lp++;
	}

	/* get timer again to exclude user delay from timing 
	 * and update beginS */
        gettimeofday(&tv, NULL);
       	extraS = ((double)tv.tv_sec + tv.tv_usec / 1000000.0)-extraS;
	beginS=beginS+extraS;

	/* reset counter (since the user pressed ENTER) */
	channel[CONSOLE].fcounter=0;
	
	/* return string */
	return lp=line;
}


/*****************************************
 *             ASSIGNMENTS               *
 *****************************************/

/* assign()
 * The assign subroutine called by LET()
 * assign a number to a number-variable or assign a string to a string-variable
 * variable name is taken from input string
 * return TRUE if assignment went OK, false otherwise 
 * System function */
int assign(void) {
	char *e=p;
	int ns;
	/* return left assignment of LEFT$ */
	if (!strncasecmp_(p,"LEFT$(",6)) {
		p+=6;
		assignLeft();
	}
	/* return left assignment of RIGHT$ */
	else if (!strncasecmp_(p,"RIGHT$(",7)) {
		p+=7;
		assignRight();
	}
	/* return left assignment of MID$ */
	else if (!strncasecmp_(p,"MID$(",5)) {
		p+=5;
		assignMid();
	}
	/* assign FN value */
	else if (!strncasecmp_(e,"FN",2)) {
		if (isDefRun) {
			int ward;
			e+=2, ward=*e++;
			if (!isalpha(ward)) return FALSE;
			if (ward!=isDefVar) return FALSE;
			if (*e=='=') e++; else return FALSE;
			if (isanystring(e)) perror_(__LINE__,45,l); 
			p=e; fn[ward].res=S(); 
			return isDefCalc=TRUE;
		}
		else {
			int ward;
			e+=2, ward=*e++;
			perror_(__LINE__,60,l,ward);
		}
	}
	/* assign to variables */
	else if ((ns=findDEC(p))) {
		/* constants cannot be assigned */
		if (vn[ns].isConst) perror_(__LINE__,205,l,vn[ns].name);
		p+=_len;
		/* numerical variables */
		if (_rettype==NUM && !vn[ns].isArr) {
			if (*p=='=') p++; 
			else perror_(__LINE__,37,l);
			if (vn[ns].isInt) P[ns]=(int)S();
			else P[ns]=S();
		}
		/* string variables */
		else if (_rettype==STR && !vn[ns].isArr) {
			char op[COMPSTR];
			if (*p=='=') p++; 
			else perror_(__LINE__,37,l);
			strncpy_(op,SS(),COMPSTR);
			strncpy_(PS[ns],op,COMPSTR);
		}
		/* numeric and string arrays */
		else if (vn[ns].isArr) {
			int M=0, N=0, T=1, rb=_rettype;
			if (*p=='(') p++; 
			else perror_(__LINE__,37,l);
			M=(int)S();
			if (*p==',') {
				p++;
				N=(int)S();
				T=2;
			}
			if (*p==')') p++; 
			else perror_(__LINE__,37,l);
			if (*p=='=') p++; 
			else perror_(__LINE__,37,l);
			_rettype=rb;
			/* numeric arrays */
			if (_rettype==NUM) {
				double val=S();
				if (vn[ns].isInt) 
					setArrayVal(ns,T,M,N,(int)val);
				else
					setArrayVal(ns,T,M,N,val);
			}
			/* string arrays */
			else {
				char op[COMPSTR];
				strncpy_(op,SS(),COMPSTR);
				setArrayStr(ns,T,M,N,op);
			}
		}
	}
	else return FALSE;

	return TRUE;
}


/* assignLeft()
 * calculate the left-assignment with LEFT$()*/
char *assignLeft(void) {
	int index=0, num=0, curr=0, cs;
	char lefts[COMPSTR], op[COMPSTR];
	
	/* check that a variable is used and nothing else */
	if ((cs=findDEC(p)) && _rettype==STR && !_arr) {
		index=cs;
		p+=_len;
	}
	else perror_(__LINE__,42,l);
	
	/* get string argument */
	strncpy_(op,PS[index],COMPSTR);

	/* get length (implicitly set to 1 if not given) */
	if (*p==',') {
		p++;
		num=(int)S();
	}
	else num=1;
	curr=(int)strlen(op);
	/* check length */
	if (num<0) num=0;
	if (num>curr+1) num=curr+1;
	if (strncmp(p,")=",2)) perror_(__LINE__,35,l);
	else p+=2;
	/* get argument string (after =) and compose */
	strncpy_(lefts,SS(),COMPSTR);
	strncat_(lefts,op+num,COMPSTR);
	/* allocate space for the new string and abandon previous */
	num=(int)strlen(lefts)+1;
	strncpy_(PS[index],lefts,num);
	return p;
}


/* assignRight()
 * calculate the left-assignment with RIGHT$()*/
char *assignRight(void){
	int index=0, num=0, curr=0, cs;
	char rights[COMPSTR], op[COMPSTR];
	
	/* check that a variable is used and nothing else */
	if ((cs=findDEC(p)) && _rettype==STR && !_arr) {
		index=cs;
		p+=_len;
	}
	else perror_(__LINE__,42,l);

	/* get string argument */
	strncpy_(op,PS[index],COMPSTR);

	/* get length (implicitly set to 1 if not given) */
	if (*p==',') {
		p++;
		num=(int)S();
	}
	else num=1;
	curr=(int)strlen(op);
	/* check length */
	if (num<0) num=0;
	if (num>curr+1) num=curr+1;
	if (strncmp(p,")=",2)) perror_(__LINE__,35,l);
	else p+=2;
	/* get argument string (after =) and compose */
	strncpy_(rights,op,curr-num+1);
	strncat_(rights,SS(),COMPSTR);
	/* allocate space for the new string and abandon previous */
	num=(int)strlen(rights)+1;
	strncpy_(PS[index],rights,num);
	return p;
}


/* assignMid()
 * calculate the left-assignment with MID$()*/
char *assignMid(void){
	int index=0, start=0, len=0, curr=0, cs;
	char mids[COMPSTR], op[COMPSTR];

	/* check that a variable is used and nothing else */
	if ((cs=findDEC(p)) && _rettype==STR && !_arr) {
		index=cs;
		p+=_len;
	}
	else perror_(__LINE__,42,l);

	/* get string argument */
	strncpy_(op,PS[index],COMPSTR);

	/* get and check arguments */
	if (*p!=',') perror_(__LINE__,65,l,makechar(p),",");
	else p++;
	curr=(int)strlen(op);
	/* get start position */
	start=(int)S();
	if (start<=0) start=1;
	if (start>curr+1) start=curr+1;
	/* get length (implicitly set to 1 if not given) */
	if (*p==',') {
		p++;
		len=(int)S();
	}
	else len=1;
	/* check length */
	if (len<0) len=0;
	if (len>curr+1) len=curr+1;
	if (strncmp(p,")=",2)) perror_(__LINE__,35,l);
	else p+=2;
	/* get argument string (after =) and compose */
	strncpy_(mids,op,start);
	strncat_(mids,SS(),COMPSTR);
	strncat_(mids,op+start+len-1,COMPSTR);
	/* allocate space for the new string and abandon previous */
	len=(int)strlen(mids)+1;
	strncpy_(PS[index],mids,len);
	return p;
}


/*****************************************
 *             MATH & STRINGS            *
 *****************************************/

/* getLabelAddr()
 * return the physical address of a label, both in numeric and in alphanumeric
 * format;
 * return -1 in case no useful value has been found */
int getLabelAddr(char *lp, char **rp) {
	int val;
	char num[10], *n=num;
	/* numeric label */
	if (isdigit(*lp)) {
		int c=0;
		while (c<5 && isdigit(*lp)) *n++=*lp++, c++;
		*n='\0';
		val=strtol(num,NULL,10);
		*rp=lp;
	}
	/* alphanumeric label */
	else val=setAddr(lp,rp); // rp is updated in setAddr() 
	/* check value */
	if (norow(val)) return -1;
	return getAddr(val);
}	


/* setAddr()
 * calculate the numerical correspondence of any string label 
 * in chunks of 8 characters*/
int setAddr(char *lp, char **rp) {
	int val=CORELIM, c=0, i=0;
	while (isnamechar(lp) && i<NAMEDIM) {
		c=0;
		while (isnamechar(lp) && c<8) {
			if (isdigit(*lp))
				val+=pow(10,c++)*(toupper(*lp++)-'0'+1);
			else val+=pow(10,c++)*(toupper(*lp++)-'A'+1);
			i++;
		}
	}
	if (**rp) *rp=lp;
	return val;
}


/* getAddr()
 * convert the label into a physical address
 * return -1 in case no address was found */
int getAddr(int label) {
	int i;
	for (i=0;i<MEMLIMIT;i++) if (ln[i]==label) return i;
	return -1;
}


/* check(string1,string2)
 * check if sb syntax is correct against sc; if it is, put a space in the 
 * position found, so that a subsequent parsing by S() may stop without 
 * issuing an error message from getFunc() and return true; 
 * if syntax it's not correct, return false.
 * This is done in statements with some core-words, like IF <Test> THEN
 * where the parsing of <Test> must stop just before THEN; check() inserts 
 * a space in place of the T of THEN, causing the parser to stop there */
int check(char *sb, char *sc){
	char *qq=strcasestr_oq(sb,sc);
	/* put a space to stop S() parsing */
	if (qq) { *qq=' '; return TRUE; }
	return FALSE;
}


/* doExp(str, &base, &exp, work) 
 * find base and exponent of a number which must be printed in exponential 
 * form given the pointer to the string str containing the number in raw form.
 * Note: See the BCLM manual, for some output of numbers in exponential form.
 * str is the base string hosting the number
 * base and exp are char arrays to host base and exponent pictures
 * work is a space area for the exponent calculation
 * System function used into printNUM() */
void doExp(char *pos, double *base, int *exp) {
	char *e = strstr(pos, "E");
	if (!e) e = strstr(pos, "e");
	if (!e) return;
	*e=' ';
	*base=strtod(pos,NULL);
	*exp=strtol(e,NULL,10);
}


/* power(base,exponent)
 * safe powering function */
double power(double base, double expon) {
	/* if base is negative, take some cautions... */
	if (base<0) {
		/* integer exponent is always evaluable */
		if (expon==(double)((int)expon)) {
			if ((int)expon%2) 
				// odd exponent
				return -pow(-base,expon);
				// even exponent (regular case)
				else return pow(base,expon);
		}
		/* real exponent is not evaluable if base is < 0 */
		else perror_(__LINE__,66,l);
	}
	/* error in case of 0^(<0) */
	else if (expon<0 && base==0) perror_(__LINE__,132,l);
	/* 1 is the limit of function x^x if x approaches zero from right */
	else if (expon==0 && base==0) return 1.0;
	/* default case for base > 0 and any exponent (regular case) */
	return pow(base,expon);
}


/* stripBlanks()
 * strip blanks from src and copy stripped string to dest; 
 * stop stripping while into a string */
void stripBlanks(char *src, char *dest) {
	int j=1; // strip flag/counter - don't change!
	while (*src) {
		if (*src=='\"') j++;
		if (j&1) { 
			if (*src!=' ' && *src!='\t') 
				*dest++=*src; 
		}
		else *dest++=*src;
		src++;
	}
	*dest='\0';
}


/* makechar()
 * returns first character of string str as a string */
char *makechar(char* str) {
	static char onechar[2];

	onechar[0]=*str;
	onechar[1]='\0';
	return onechar;
}


/* spaces() 
 * return a pointer to a string of v spaces up to LIMIT+1 characters */
char *spaces(int v) {
	static char spacePad[LIMIT+1];
	char *sp;
	int n;

	/* set limits */
	sp=spacePad;
	if (v<0) n=0;
	else if (v>LIMIT) n=LIMIT;
	else n=v;
	
	/* build spaces string */
	while (n-->0) *sp++=32;
	*sp='\0';
	return spacePad;
}


/* turnToUpper()
 * turn a line to all upper characters */
void turnToUpper(char *str) {
	while (*str) *str=toupper(*str), str++;
}


/* checkCommand()
 * check current against liner not considering the inner blanks, so that
 * "END WHILE" is equal to "END    WHILE" or to "END\tWHILE"
 * The 'current' string may be longer after the matching substring, while
 * liner must be complete
 * System function for preParse() */
int checkCommand(char *current, char *liner, char **final) {
	char *c=current, *l=liner;
	while (isblank(*c)) c++; while (isblank(*l)) l++;
	while (*c && *l) {
		if (toupper(*c)!=toupper(*l)) break;
		c++; l++;
		while (isblank(*c)) c++; while (isblank(*l)) l++;
	}
	/* if second string got to the end, they match! */
	if (!*l) {
		if (*final) *final=c;
		return TRUE;
	}
	else return FALSE;
}


/* Note
 * The following strncpy_() and strncat_() are my own personal versions for 
 * strcpy/strncpy and strcat/strncat respectively.
 * I must use mine because strncpy/strncat standard functions in C (gcc) have 
 * erratic behaviors within different operating systems like Linux-Suse, 
 * OpenBSD and Mac OS X 10.4 (the one I use).
 * Since many reported strange things happening because of strncpy/strncat
 * (from error in compilation to completely wrong string functions output), I 
 * made up my mind and built these substitutes for both. */

/* strncpy_()
 * dst: the destination string (which must have its proper dimension)
 * src: the source to be copied onto dst
 * dw:  the destination width, the unity used to declare dst (for the sake
 * of precision, if dst is declared as 10 characters wide, it means that text
 * goes from pos. 0 to pos 8, and in pos. 9 is the last terminator; thus a width
 * equal to 10 means 9 characters at most for the string; dw must be put equal 
 * to the width, in this example 10; of course, any lower value is also good.)
 * The process will take care of never going past the limit of dst, and to fill 
 * the rest of the space (if any) with nulls; 
 * System function replacing strcpy/strncpy */
char *strncpy_(char *dst, char *src, int dw) {
	char *dp=dst;
	/* check length of copy and source */
	if (dw<=1) *dst='\0';
	else {
		/* erase destination */
		*dst='\0';
		/* copy src onto dst but not past the end of dst */
		while ((int)(dp-dst)<(dw-1) && *src) *dp++=*src++;
		*dp++='\0';
		/* fill the rest with nulls */
		while ((int)(dp-dst)<(dw)) *dp++='\0';
	}
	/* Note: return the same value returned by strcpy/strncpy */
	return dst;
}


/* strncat_()
 * dst: the destination string (which must have its proper dimension)
 * src: the source to be copied onto dst
 * dw:  the destination width, the unity used to declare dst (for the sake
 * of precision, if dst is declared as 10 characters wide, it means that text
 * goes from pos. 0 to pos 8, and in pos. 9 is the last terminator; thus width
 * equal to 10 means 9 characters at most for the string; dw must be put equal 
 * to the width, in this example 10; of course, any lower value is good.)
 * The process will take care of never going past the limit of dst, and to fill 
 * the rest of the space (if any) with nulls; if dst is found not initialized 
 * (that is nowhere the terminator is found) it is initialized as void.
 * System function replacing strcat/strncat */
char *strncat_(char *dst, char *src, int dw) {
	char *dp=dst;
	/* search for current terminator */
	while (*dp && (int)(dp-dst)<dw) dp++;
	/* terminator not found: initialize destination */
	if ((int)(dp-dst)>=dw) dp=dst, *dp='\0';
	/* copy src onto dst but not past the end of dst */
	while ((int)(dp-dst)<(dw-1) && *src) *dp++=*src++;
	*dp='\0';
	/* fill the rest with nulls */
	while ((int)(dp-dst)<(dw)) *dp++='\0';
	/* Note: return the same value returned by strcat/strncat */
	return dst;
}


/* chrcat_()
 * add character c to list dst, taking care of checks the limit
 * dst: the destination string (which must have its proper dimension)
 * c: the character to be appended to dst
 * dw:  the destination width, the unity used to declare dst (for the sake
 * of precision, if dst is declared as 10 characters wide, it means that text
 * goes from pos. 0 to pos 8, and in pos. 9 is the last terminator; thus width
 * equal to 10 means 9 characters at most for the string; dw must be put equal 
 * to the width, in this example 10; of course, any lower value is good.)
 * The process will take care of never going past the limit of dst, and to fill 
 * the rest of the space (if any) with nulls; if dst is found not initialized 
 * (that is no terminator is found) dst is initialized as void.
 * Note: this function is a complement of the strncat_ function
 * System function */
char *chrcat_(char *dst, int c, int dw) {
	char *dp=dst;
	/* search for current terminator */
	while (*dp && (int)(dp-dst)<dw) dp++;
	/* terminator not found: initialize destination */
	if ((int)(dp-dst)>=dw) dp=dst, *dp='\0';
	/* add c to dst but not past the end of dst */
	if (c && (int)(dp-dst)<(dw-1)) *dp++=c;
	*dp='\0';
	/* fill the rest with nulls */
	while ((int)(dp-dst)<(dw)) *dp++='\0';
	/* Note: return the same value returned by strncat_ */
	return dst;
}


/* Note
 * The following sprintw() is my own personal version wrapper for snprintf(), 
 * which appear to behave erratically on my system (is does not work with 
 * function return values, but works with real arrays) and probably the same 
 * (with inverse properties) on other systems.
 * This wrapper assures that only physical arrays are treated by sprintw() */

/* sprintw() 
 * the snprintf wrapper 
 * dest is the destination string,
 * len is the string width (including the terminator character)
 * format is the formatting string (with the same printf rules)
 * '...' is the printing string argument (it may be a function return value) 
 * return the number of characters written in the host string 
 * System function replacing sprintf/snprintf */
int sprintw(char *dest, int len, char* format, ...) {
	va_list ap;
	char base[COMPSTR];
	va_start(ap,format);
	vsnprintf(base,COMPSTR,format,ap);
	snprintf(dest,len,"%s",base);
	va_end(ap);
	return (int)strlen(base);
}


/* Note
 * The following strncasecmp_() is my own personal version of strncasecmp.
 * I must use mine because strncasecmp seems to cause some problems if tbas is
 * compiled under openSuse with gcc and or Linux/openBSD with gcc.
 * Since many reported errors in compiling when the standard strncasecmp is
 * used, I made up my mind and built this substitute. */

/* strncasecmp_()
 * the string compare function
 * dst and src are the strings to be compared;
 * N is the number of characters to be checked against;
 * return the flag for positive comparison result, 0 for match, -1 for failure.
 * Note: return values mimic the original return values, with the difference
 * that I don't check here for dst being greater or less than src, and in case
 * they differ, -1 is returned anyway; this is done for speed issues. */
int strncasecmp_(char *dst, char *src, int N) {
	if (N<=0) return -1;
	while (N && *dst && *src && toupper(*dst)==toupper(*src)) 
		src++, dst++, N--;
	if (!N) return 0;
	return -1;
}


/* Note
 * The following memcpy_() is my own personal version of memcpy and memmove.
 * I must use mine because memcpy seems to be sometimes rejected by openBSD.
 * Since many reported errors in compiling if the standard memcpy_/memmove is
 * used, I made up my mind and built this substitute. */

/* memcpy_()
 * copy a portion of memory from src to dst, using memmove (that takes care
 * of overlapping), for a number of characters as in size, checks before
 * if any of arrays are null, or size is zero */
void memcpy_(void *dst, void *src, int size) {
	if (src==NIL || dst==NIL || !size) return;
	memmove(dst,src,size);
}


/* Note
 * The following strcasestr_() is my own personal version of strcasestr().
 * I must use mine because strcasestr() in C has someway an erratic behavior
 * with Linux.
 * Since many reported errors in compiling if the standard strcasestr() is
 * used, I made up my mind and built this substitute. */

/* strcasestr_()
 * find the first occurrence of the substring 'little' in string 'big', 
 * ignoring the case of both arguments.
 * The terminating null bytes (i.e. '\0') are not compared. 
 * If 'little' is null, return NIL; if the substring is found, return the 
 * updated address in 'big', otherwise return NIL. */
char *strcasestr_(char *big, char *little) {
	int l=(int)strlen(little);

	if (!l) return NIL;
	while (*big) {
		if (!strncasecmp_(big,little,l)) return big;
		big++;
	}
	return NIL;
}


/* strcasestr_oq()
 * find the first occurrence of the substring string 'pr' into string 'pl'
 * this function has the same usage of strcasestr_(), but it works only if 
 * outside of a string, in BASIC normally surrounded by double quotes 
 * (strcasestr_oq = strcasestr_ Out of Quotes) */
char *strcasestr_oq(char *pl, char* pr) {
	int valid=1, l=(int)strlen(pr);

	if (!l) return pl;
	while (*pl) {
		if (*pl=='\"') valid++;
		if (!strncasecmp_(pl,pr,l) && (valid&1)) return pl;
		pl++;
	}
	return NIL;
}


/* storeSingleLineToCore()
 * saves into core memory the line in argument, which must be constituted
 * by a line number lesser than 100000 (not required) and a BASIC statement;
 * return TRUE on successful storing;
 * line holds the textual statements part
 * lnum holds the physical line number
 * a void line is transparent for this command
 * System function */ 
int storeSingleLineToCore(char* line, int lnum) {
	char *lp=line, *z, *sp=line;
	char linenum[10], label[COMPSCR], *lab=label;
	int row=0, lk=0;

	/* check NIL (empty line...) */
	if (!*line) return TRUE;

	if ((int)strlen(line)>0) lp=line+(int)strlen(line)-1;
	else lp=line;
	/* get rid of final blanks (to intercept alphalabels) */
	while ((int)(lp-line)>0 && isblank(*lp)) *lp--='\0';
	/* advance till first useful command (for colon and IMAGE) */
	while (isdigit(*sp)) sp++;
	while (isblank(*sp)) sp++;

	/* search for the tick comment and get rid of it if out of strings,
	 * not in first position and if not argument of the ASC() function */
	if (line[0]=='\'') line[0]='\0';
	if (*sp!=':' && strncasecmp_(sp,"IMAGE",5)) {
		if ((z=strcasestr_oq(line,"\'")) && (int)(z-line)>0 &&\
			       *(z-1)!='(' && *(z+1)!=')') 
			*z='\0';
	}

	/* get line number 
	 * numbers must be in the range 0-99999 (any up-to-five digits) */
	lp=line;
	if (isdigit(*lp)) {
		row=0;
		/* get and check line number */
		while (row<5 && isdigit(*lp)) linenum[row++]=*lp++;
		linenum[row]='\0';
		row = strtol(linenum, NULL, 10);
		/* check label validity */
		if (row<0 || row > DATLIM) perror_(__LINE__,71,lnum);
		/* set reference */
		ln[lnum]=row; // store label for debugging purposes
	}
	/* get label code number 
	 * labels must stand on their own line, terminated by colon and not
	 * be followed by a statement; labels are constituted by numbers, 
	 * letters and the underscore, but they must begin with a letter 
	 * or an underscore character */
	else if ((int)strlen(line)>0) {
		if (line[(int)strlen(line)-1]==':' && isbeginnamechar(lp)) {
			while (isnamechar(lp) && *lp) *lab++=*lp++;
			*lab='\0';
			row=setAddr(label,&lp);
			if (norow(row)) perror_(__LINE__,57,lnum);
			ln[lnum]=row;
		}
	}
	
	/* remove trailing spaces ahead (after the line number) and behind */
	while ((int)(lp-line)<MAXSTRLEN && isblank(*lp)) lp++;
	if ((int)strlen(lp)>0) lk=(int)strlen(lp)-1;
	else lk=0;
	if ((int)(lp-line)+lk<MAXSTRLEN) {
		z=lp+lk;
		while ((int)(z-line)>0 && isblank(*z)) *z--='\0';
	}

	/* store line; if lp is void line with only spaces and/or tabs, it's 
	 * stored anyway, but it won't be executed (see up) */
	if (*lp && (int)strlen(lp)>0) {
		lk=((int)strlen(lp)+1)*sizeof(char);
		m[lnum]=xmalloc((size_t)lk);
		if (!m[lnum]) perror_(__LINE__,194,-1,lnum);
		strncpy_(m[lnum],lp,lk);
		optoken(m[lnum]);
	}
	/* if line is void (i.e. with line number/label but without any
	 * statements), fill it with some spaces; this is done because some 
	 * programmers may GO TO lines with line number but void; a whim that I
	 * have got to circumvent! */
	else {
		lk=(5+1)*sizeof(char);
		m[lnum]=xmalloc((size_t)lk);
		if (!m[lnum]) perror_(__LINE__,194,-1,lnum);
		strncpy_(m[lnum],"     ",lk); // five spaces
	}

	/* normal behavior */
	return TRUE;
}


/* printHeader()
 * print Header in the DEC style */
void printHeader(void) {
	char hour[24], date[24];
	/* cut extension in ANY CASE */
	strncpy_(longname,filename,COMPSCR);
	turnToUpper(longname);
	if ((int)strlen(longname)>3) {
		if (!strcasecmp(longname+(int)strlen(longname)-4, EXTU))
			longname[(int)strlen(longname)-4]='\0';
	}
	sprintw(hour,24,"%02d:%02d", getHour(), getMin());
	sprintw(date,24,"%02d-%s-%02d",getDay(),getMonthString(),
			getYear()-2000);
	fprintf(stdout,"\n");
	fprintf(stdout,THISHEADER, basename(longname), hour, date);
	fprintf(stdout,"\n");
	fprintf(stdout,"\n\n\n");
}


/******************************************
 *                 PARSING                *
 ******************************************/


/* doVintage()
 * set Vintage properties */
void doVintage(void) {
	isRadians=TRUE;
	BASE=0;
	isCap=TRUE;
	isCaseSensitive=FALSE;
	isComparison=RELATIVE;
	isDebug=FALSE;
	ZEROEQUAL=1.0E-6;
	isEcho=TRUE;
	isExplicit=FALSE;
	isAmericanFormat=TRUE;
	PRECISION=-1;
	isPrompt=TRUE;
	isRawPrint=FALSE;
	isToReset=TRUE;
	isRoute=FALSE;
	isSpacing=FALSE;
	ZONE=ZONEBASE;
	isTiming=TRUE;
	isZeroLim=FALSE;
	ZEROLIM=0.0;
}



/* preParse()
 * A fundamental subroutine to be run BEFORE each process.
 * Scan source in core memory and executes DATA and DEF recordings, sets DIM
 * (DIMENSION) arrays spaces, opens FILES files, and in general activates all 
 * commands that must be preprocessed before program execution.
 * Check also FOR-NEXT crossed references while GOTO, GOSUB and THEN addresses 
 * are checked into their relative functions.
 * Note for subsequent addons: This command is executed BEFORE stripblanks()
 * into parse(), so assure to do stripblanks() before processing the line,
 * or use isblank() to skip all blanks!!! 
 * m: the array of the string pointers containing the program lines to be
 * executed.
 * System basic function */
void preParse(char **m, int start, int end) {
	int l=start-1, thisisHeader=FALSE;
	char *q,*qq;
	dc=dataLimit=0;

	/* preset DEFFN values */
	isDefRun=FALSE;
	isDefCalc=FALSE;
	isDefVar=0;
	isDefFor=FALSE;

	/* scan source */
	do {
		l++;
		if (!m[l]) continue;
		q=m[l];
		while (isblank(*q)) q++;
		/* image lines (for USING) are not checked */
		if (!strncasecmp_(q,"IMAGE",5)) continue;
		if (*q==':') continue;
		/* REM statements are not checked */
		else if (!strncasecmp_(q,"REM",3)) continue;
		/* PRINT and WRITE may contain strings that shouldn't be 
		 * checked, but in any case, PRINT() and WRITE() and their
		 * coworkers can themselves detect all errors;
		 * so avoid checks here */
		else if (!strncasecmp_(q,"PRINT",5)) continue;
		else if (!strncasecmp_(q,"WRITE",5)) continue;

		/* DATA Statement - saving elements 
		 * Note: this is the actual execution of the DATA instruction, 
		 * which in runtime is effectively a no-op */
		if (!strncasecmp_(q,"DATA",4)) {
			/* debug info */
			if (isDebug) {
				printf(INVCOLOR);
				printf("Storing DATA line %d\n",ln[l]);
				printf(RESETCOLOR);
			}
			/* check line length */
			if ((int)strlen(q)>SCREENLINE-7) perror_(__LINE__,75,l);
			q+=4;
			while (*q) {
				while (isblank(*q)) q++;
				/* store naked strings (comma separated) */
				if (isalpha(*q)) {
					char op[COMPSTR], *P1=op;
					int lk;
					/* check if DATA space is full */
					if (dcs>=DATLIM-1) 
						perror_(__LINE__,97,l);
					while (*q&&*q!=','&&*q!='\n') 
						*P1++=*q++;
					*P1='\0';
					++dcs;
					lk=((int)strlen(op)+1)*sizeof(char);
					if (DS[dcs]!=NIL) {
						free(DS[dcs]);
						DS[dcs]=NIL;
					}
					DS[dcs]=xmalloc((size_t)lk);
					if (!DS[dcs]) perror_(__LINE__,193,l);
					*DS[dcs]='\0';
					strncpy_(DS[dcs],op,lk);
					DSN[dcs]=l;
				}
				/* store quoted strings */
				else if (*q=='\"') {
					char op[COMPSTR], *P1=op;
					int lk;
					/* check if DATA space is full */
					if (dcs>=DATLIM-1) 
						perror_(__LINE__,97,l);
					q++;
					while (*q && *q!='\"') *P1++=*q++;
					*P1='\0';
					++dcs;
					lk=((int)strlen(op)+1)*sizeof(char);
					if (DS[dcs]!=NIL) {
						free(DS[dcs]);
						DS[dcs]=NIL;
					}
					DS[dcs]=xmalloc((size_t)lk);
					if (!DS[dcs]) perror_(__LINE__,193,l);
					*DS[dcs]='\0';
					strncpy_(DS[dcs],op,lk);
					DSN[dcs]=l;
					if (*q=='\"') q++;
				}
				/* store numbers; Note: in DATA statements 
				 * formulas are accepted (they must begin
				 * with a number in order to be recognized as
				 * "numbers" and not as "unquoted strings") */
				else if (*q) {
					/* check if space is over */
					if (dc>=DATLIM-1)perror_(__LINE__,97,l);
					while (isblank(*q)) q++;
					++dc;
					p=q, D[dc]=S(), q=p, DN[dc]=l;
				}
				while (isblank(*q)) q++;
				/* discard apostrophe comment */
				if (*q=='\'') while (*q) q++;
				/* check if datum is followed by
				 * something else than a comma */
				if (*q && *q==',') q++;
				else if (*q) perror_(__LINE__,144,l);
			}
			continue;
		}
		/* DECLARE SUB statements - delegate SUB() */
		else if (checkCommand(q,"DECLARE SUB",&q)) {
			char *x=J;
			p=q;
			stripBlanks(p,x); p=J;
			SUB(l); 
			continue;
		}
		/* SUB statements - delegate SUB() */
		else if (!strncasecmp_(q,"SUB",3)) {
			char *x=J;
			p=q+3;
			stripBlanks(p,x); p=J;
			SUB(l); 
			continue;
		}
		/* DEF statements - delegate DEF() */
		else if (!strncasecmp_(q,"DEF",3)) {
			char *x=J;
			p=q+3;
			stripBlanks(p,x); p=J;
			DEF(l); 
			continue;
		}
		/* OPTION statements */
		else if (strcasestr_oq(q,"OPTION")) {
			q+=6;
			/* OPTION HEADER 
			 * option header must be detected before execution */
			if ((qq=strcasestr_oq(q,"HEADER"))) {
				thisisHeader=TRUE;
			}
			/* OPTION VINTAGE
			 * whimsical option for those who remember the dear old
			 * days...
			 * OPTION VINTAGE rends the behaviour and aspect of 
			 * programs in monitors of first Seventies; it sets the 
			 * following parameters:
			 * ANGLE RADIANS, BASE 0, CAPS ON, CASE OFF, DEBUG OFF,
			 * COMPARISON RELATIVE, HEADER ON, PRECISION OFF, 
			 * PROMPT ON, RESET TRUE, TIMING ON, ZERO OFF, 
			 * RAW PRINT OFF, SPACING OFF, ROUTING OFF,  
			 * EXPLICIT OFF, FORMAT AMERICAN
			 * Note: since the header should be written before any 
			 * output, this option is intercepted here, and may so 
			 * be written anywhere, but any other instance of 
			 * OPTION may alter the parameters, so you should avoid 
			 * them (unless this is what you want, of course...). 
			 * Note ECHO is executed directly by OPTION ECHO and 
			 * OPTION PAGING is executed directly by OPTION PAGING.
			 */
			if ((qq=strcasestr_oq(q,"VINTAGE"))) {
				doVintage();
				thisisHeader=TRUE;
			}
		}
		/* FILES statements - delegate FFILE() 
		 * Note: this is the actual execution of the FILES instruction, 
		 * which in runtime is effectively a no-op*/	
		else if (!strncasecmp_(q,"FILES",5)) {
			char *x=J;
			p=q+5;
			stripBlanks(p,x); p=J;
			FFILE(TRUE,l);
			continue;
		}
		/* FOR must correspond to a NEXT statement */
		else if (!strncasecmp_(q,"FOR",3)) {
			int bl=l+1, wcount=0;
			while (bl<=end) {
				if (m[bl]==NIL) { bl++; continue; }
				q=m[bl];
				if (!strncasecmp_(q,"FOR",3)) 
					wcount++;
				else if (wcount && !strncasecmp_(q,"NEXT",4)) {
					wcount--;
					/* count commas to know how many
					 * variables are in the same NEXT */
					while (*q && *q!='\'') 
						if (*q++==',') wcount--;
					if (wcount<0) wcount=0, bl--;
				}
				else if (!wcount && !strncasecmp_(q,"NEXT",4)){
					cAddr[l]=forAddr[l]=bl+1;
					cAddr[bl]=forAddr[bl]=l;
					break;
				}
				bl++;
			}
			if (bl>end) perror_(__LINE__,29,l);
			continue;
		}
		/* WHEN ERROR IN must correspond to a USE statement 
		 * Note: controls the missing END WHEN and delegate other
		 * error-catching to the WHEN ERROR functions */
		else if (checkCommand(q,"WHEN ERROR IN",&q)) {
			int cuse=0, bl=l+1;
			/* search for USE */
			while (bl<=end) {
				if (m[bl]!=NIL && !strncasecmp_(m[bl],"USE",3)){
					cAddr[l]=bl;
					whenAddr[l]=bl;
					cuse=bl;
					break;
				}
				bl++;
			}
		        if (!cuse) perror_(__LINE__,164,l);	
			/* search for END WHEN */
			bl++;
			while (bl<=end) {
				q=m[bl];
				if (m[bl]!=NIL&&checkCommand(q,"END WHEN",&q)){
					cAddr[cuse]=bl;
					whenAddr[cuse]=bl;
					cAddr[bl]=l;
					whenAddr[bl]=l;
					break;
				}
				bl++;
			}
		        if (bl>end) perror_(__LINE__,166,l);	
			continue;
		}
		/* WHEN ERROR USE must correspond to a END WHEN statement 
		 * Note: controls the missing END WHEN and delegate other
		 * error-catching to the WHEN ERROR functions */
		else if (checkCommand(q,"WHEN ERROR USE",&q)) {
			int bl=l+1;
			while (bl<=end) {
				q=m[bl];
				if (m[bl]!=NIL&&checkCommand(q,"END WHEN",&q)){
					cAddr[l]=bl;
					whenAddr[l]=bl;
					cAddr[bl]=l;
					whenAddr[bl]=l;
					break;
				}
				bl++;
			} 
		        if (bl>end) perror_(__LINE__,166,l);	
			continue;
		}
		/* HANDLER must correspond to an END HANDLER statement;
		 * also, fill the handler data 
		 * Note: controls the missing END HANDLER and delegate other
		 * error-catching to the HANDLER function */
		else if (!strncasecmp_(q,"HANDLER",7)) {
			int bl=l+1, hnlen;
			char hname[COMPSTR], *h=hname;
			q+=7; hnlen=0;
			/* get handler name */
			while (isblank(*q)) q++;
			while (*q) *h++=*q++;
			*h='\0';
			handlerTOS++;
			while (bl<=end) {
				if (m[bl]==NIL) {
					bl++;
					continue;
				}
				else if (!strncasecmp_(m[bl],"END",3)&&\
				(int)strlen(m[bl])>6 &&\
				!strcasecmp(m[bl]+(int)strlen(m[bl])-7,\
						"HANDLER")) {
					cAddr[l]=handlerAddr[l]=bl;
					cAddr[bl]=handlerAddr[bl]=l;
					break;
				}
				bl++;
			}
		        if (bl>end) perror_(__LINE__,165,l);	
			handler[handlerTOS].pos=l;
			handler[handlerTOS].endwhenpos=bl;
			hnlen=((int)strlen(hname)+1)*sizeof(char);
			if (handler[handlerTOS].hname!=NIL) {
				free(handler[handlerTOS].hname);
				handler[handlerTOS].hname=NIL;
			}
			handler[handlerTOS].hname=xmalloc((size_t)hnlen);
			if (!handler[handlerTOS].hname) perror_(__LINE__,174,l);
			else strncpy_(handler[handlerTOS].hname,hname,hnlen);
			continue;
		}
		/* WHILE must correspond to an END WHILE statement
		 * Note: controls the missing END WHILE and delegate other 
		 * error-catching to the WHILE functions */
		else if (!strncasecmp_(q,"WHILE",5)) {
			int bl=l+1, wcount=0;
			while (bl<=end) {
				if (m[bl]==NIL) { bl++; continue; }
				q=m[bl];
				if (!strncasecmp_(q,"WHILE",5)) 
					wcount++;
				else if (wcount &&\
					       checkCommand(q,"END WHILE",&q))
					wcount--;
				else if (wcount &&\
					       !strncasecmp_(q,"WEND",4))
					wcount--;
				else if (!wcount &&\
					       checkCommand(q,"END WHILE",&q)){
					cAddr[l]=whileAddr[l]=bl;
					cAddr[bl]=whileAddr[bl]=l;
					break;
				}
				else if (!wcount &&\
					       !strncasecmp_(q,"WEND",4)) {
					cAddr[l]=whileAddr[l]=bl;
					cAddr[bl]=whileAddr[bl]=l;
					break;
				}
				bl++;
			}
			if (bl>end) perror_(__LINE__,28,l);
			continue;
		}
		/* DO WHILE/UNTIL must correspond to a LOOP statement
		 * Note: controls the missing UNTIL and  delegate other
		 * error-catching to the DO WHILE/UNTIL functions */
		else if (!strncasecmp_(q,"DO",2)) {
			int bl=l+1, wcount=0;
			while (bl<=end) {
				if (m[bl]==NIL) { bl++; continue; }
				if (!strncasecmp_(m[bl],"DO",2))
					wcount++;
				else if (wcount &&\
						!strncasecmp_(m[bl],"LOOP",4))
					wcount--;
				else if (!wcount &&\
					        !strncasecmp_(m[bl],"LOOP",4)){
					cAddr[l]=doAddr[l]=bl;
					cAddr[bl]=doAddr[bl]=l;
					break;
				}
				bl++;
			}
			if (bl>end) perror_(__LINE__,28,l);
			continue;
		}
		/* SELECT and CASE must correspond to an END SELECT statement
		 * Note: controls the missing END SELECT and  delegate other 
		 * error-catching to the SELECT/CASE functions */
		else if (!strncasecmp_(q,"SELECT",6)) {
			int bl=l+1, wcount=0, endpos=0;
			/* look for end*/
			while (bl<=end) {
				if (m[bl]==NIL) {
					bl++;
					continue;
				}
				q=m[bl];
				if (!strncasecmp_(q,"SELECT",2))
					wcount++;
				else if (wcount &&\
						checkCommand(q,"END SELECT",&q))
					wcount--;
				else if (!wcount &&\
					       checkCommand(q,"END SELECT",&q)){
					cAddr[l]=selectAddr[l]=bl;
					cAddr[bl]=selectAddr[bl]=l;
					break;
				}
				bl++;
			}
			endpos=bl;
			if (bl>end) perror_(__LINE__,27,l);
			bl=l+1; wcount=0;
			/* now bl is the END SELECT pos: set up CASE */
			while (bl<=end) {
				if (m[bl]==NIL) {
					bl++;
					continue;
				}
				q=m[bl];
				if (!strncasecmp_(q,"SELECT",6))
					wcount++;
				else if (wcount &&\
					       checkCommand(q,"END SELECT",&q))
					wcount--;
				else if (!wcount &&\
					       !strncasecmp_(q,"CASE",4))
					cAddr[bl]=selectAddr[bl]=endpos;
				else if (!wcount &&\
	 				       checkCommand(q,"END SELECT",&q)){
					break;
				}
				bl++;
			}
			continue;
		}
		/* IF THEN must correspond to an END IF statement
		 * Note: controls the missing END IF and delegate other 
		 * error-catching to the IF THEN functions */
		else if (!strncasecmp_(q,"IF",2) && (int)strlen(q)>3 &&\
				!strcasecmp(q+(int)strlen(q)-4,"THEN")) {
			/* only for clauses IF..THEN followed by spaces/EOL */
			int bl=l+1, wcount=0, endpos=0, elsepos=0;
			/* look for closing END IF */
			while (bl<=end) {
				if (m[bl]==NIL) {
					bl++;
					continue;
				}
				if (!strncasecmp_(m[bl],"IF",2)&&\
				(int)strlen(m[bl])>3 &&\
				!strcasecmp(m[bl]+(int)strlen(m[bl])-4,"THEN"))
					wcount++;
				else if (wcount&&!strncasecmp_(m[bl],"END",3)&&\
				(int)strlen(m[bl])>1 &&\
				!strcasecmp(m[bl]+(int)strlen(m[bl])-2,"IF"))
					wcount--;
				else if(!wcount&&!strncasecmp_(m[bl],"END",3)&&\
						(int)strlen(m[bl])>1 &&\
						!strcasecmp(m[bl]+\
						(int)strlen(m[bl])-2,"IF")) {
					cAddr[bl]=ifAddr[bl]=l;
					break;
				}
				bl++;
			}
			endpos=bl;
			if (bl>end) perror_(__LINE__,26,l);
			bl=endpos-1; wcount=0; elsepos=endpos;
			cAddr[endpos]=ifAddr[endpos]=l;
			/* now bl is the END IF pos: set up every ELSE 
			 * to point to next ELSE/END */
			while (bl>=l) {
				if (m[bl]==NIL) {
					bl--;
					continue;
				}
				if (!strncasecmp_(m[bl],"END",3)&&\
				(int)strlen(m[bl])>1 &&\
				!strcasecmp(m[bl]+(int)strlen(m[bl])-2,"IF"))
					wcount++;
				if (wcount&&!strncasecmp_(m[bl],"IF",2)&&\
				(int)strlen(m[bl])>3 &&\
				!strcasecmp(m[bl]+(int)strlen(m[bl])-4,"THEN"))
					wcount--;
				else if(!wcount&&!strncasecmp_(m[bl],"ELSE",4)){
					cAddr[bl]=ifAddr[bl]=elsepos; 
					elsepos=bl;
				}
				else if (!wcount&&!strncasecmp_(m[bl],"IF",2)&&\
				(int)strlen(m[bl])>3 &&\
				!strcasecmp(m[bl]+(int)strlen(m[bl])-4,"THEN")){
					cAddr[bl]=ifAddr[bl]=elsepos;
				}
				bl--;
			}
			continue;
		}
	} while (l<=end+1);

	/* debug feature */
	if (isDebug) {
		int i,some=0;
		printf(INVCOLOR);
		printf("\n");
		printf("debug: crossed references:");
		for (i=start;i<=end;i++) {
			if (cAddr[i]) {
				some++;
				if (!strncasecmp_(m[i],"FOR",3))
					printf("\nref.#%d = %s ==> #%d (%s)\n",i,m[i],cAddr[i]-1,m[cAddr[i]-1]);
				else 
					printf("\nref.#%d = %s ==> #%d (%s)\n",i,m[i],cAddr[i],m[cAddr[i]]);
			}
		}
		if (some) printf("\n\n");
		else printf(" none.\n\n");
		printf(RESETCOLOR);
	}
	
	/* debug feature for handlers */
	if (isDebug && handlerTOS>0) {
		int i;
		printf(INVCOLOR);
		for (i=1;i<=handlerTOS;i++) 
		printf("handler #%d has pos %d, end pos %d, and name %s.\n",
				i, handler[i].pos, handler[i].endwhenpos,
				handler[i].hname!=NIL?handler[i].hname:" ");
		printf("\n");
		printf(RESETCOLOR);
	}

	/* set up starting parameters for READ */
	dataLimit=dc,   dc=1;
	dataLimitS=dcs, dcs=1;

	/* print header in after all option cases */
	if (!isHeader && thisisHeader) printHeader();
	isHeader=thisisHeader;
}


/* recog() 
 * checks a command against a list of commands
 * num: the number of the list of arguments (after the source)
 * the first argument in the list must be a char* which is interpreted as
 * the source argument (variable length)
 * the following arguments are the list of commands 
 * return the number of characters in the selected command, or zero if
 * not found 
 * System command for the getFunc() function */
int recog(int num, ...) {
	va_list ap;
	int i=0, ret=0, l=0;
	char *c, *s;

	va_start(ap, num);
	s=va_arg(ap,char*); // source
	while (i<num) {
		c=va_arg(ap, char*); // next item
		l=(int)strlen(c);
		if (!strncasecmp_(s,c,l)) ret=l;
		i++;
	}
	va_end(ap);
	return ret;
}


/* parse()
 * The parser 
 * It's a general purpose routine that executes lines contained into the m 
 * line-memory buffer, from the 'start' line to the 'end' line 
 * System basic function */
void parse(char **m, int start, int end) {
	char BASESTR[COMPSCR]; // don't make it global, it must be local

	/* check/set starting parameters */
	if (!end) return;
	l=start;
	
	/* execute line by line */
	while(l<=end){
		/* skip empty or masked lines in memory */
		if (m[l] && !fm[l] && m[l][0]) {
			strncpy_(BASESTR,m[l],COMPSCR);
			parseLine(BASESTR,l);
		}
		l++;
	}
}


/* parseLine()
 * The single line parser 
 * It's a general purpose routine that executes one line contained into s 
 * (a line-memory buffer), and using sline as current execution physical line 
 * number.
 * (Note for me: the stripBlanks (s,B) returns in B the stripped copy of s, 
 * which is the pad created in parse;
 * the strncpy_(s,B) returns in s the copy of B (which is a global pad); 
 * now, never change this, because tbas cannot work on B, since other uses it!)
 * System basic function */
int parseLine(char *s, int sline) {
	int tlen=0;
	/* skip formatting lines */
	while (isblank(*s)) s++;
	if (*s==':') return TRUE;
	
	/* skip void lines */
	if (!*s && !ln[sline]) return TRUE;

	/* strip blanks if out of string (checked into stripBlanks)
	 * buffer B will hold purged execution line */
	tlen=strlen(s)+1;
	stripBlanks(s,B);	// don't change!
	strncpy_(s,B,tlen); 	// don't change!


	/* DEBUGGING PURPOSES - print pure/stripped line */
	if (isDebug && (m[sline]!=NIL && m[sline][0])) {
		printf(INVCOLOR);
		if (!debugRaw) {
			if (ln[sline]) 
			   printf("\n[%-5d %s] ",ln[sline],m[sline]);
			else printf("\n[      %s] ",m[sline]);
		}
		else {
			if (ln[sline]) 
			   printf("\n[%-5d %s] ",ln[sline],B);
			else printf("\n[      %s] ",B);
		}
		printf(RESETCOLOR);
		printf("\n");
	}
	
	p=s;
	/* maybe it's a statement (or a SUB call): process it through exec() */ 
	if (exec(p));
	/* otherwise it's an (implicit) assignment: call multiLET() */
	else if (multiLET());
	/* otherwise it's an illegal instruction */
	else if (!isExplicit) perror_(__LINE__,44,l);
	else perror_(__LINE__,40,l);
	/* OK. Sure result */
	return TRUE;
}


/* exec()
 * the statement execution subroutine (the core of the engine, which calls
 * all the statements references in order to execute the commands)
 * lpc is the address of the execution line (the first character after the
 * line number)
 * return TRUE if execution goes well, FALSE otherwise
 * Note: all statements executed in preParse, are here simple no-op, as well
 * as comment statements REM and ' (tick)
 * Note: in future versions, I'm thinking about a tokenization
 * of all instructions, that should make tbas faster 
 * System basic function */
int exec(char* lpc) {
	int ns, type, rettype;
	char *lpr;
	if (*lpc=='\0'); // nop for void lines (and tick comments)
	/* check if statement is a SUB */
	else if ((ns=findSUB(lpc,&lpr,&type,&rettype))) {
		if (type!=PROC) perror_(__LINE__,211,l);
		execSUB(ns,lpr,&numRes,NIL);
	}
	else if (!strncasecmp_(lpc,"ABORT",5)) p=lpc+5, ABORT();
	else if (!strncasecmp_(lpc,"ACCEPT#",7)) p=lpc+7,FREADSHARP(1,1);
	else if (!strncasecmp_(lpc,"ACCEPT",6)) p=lpc+6, INPUT(0);
	else if (!strncasecmp_(lpc,"APPEND",6)) p=lpc+6, FAPPEND();
	else if (!strncasecmp_(lpc,"BREAK",5)) p=lpc+5, BREAK();
	else if (!strncasecmp_(lpc,"CALL",4)) { p=lpc+4; return exec(p); }
	else if (!strcasecmp(lpc,"CASEDEFAULT")) CASEDEFAULT();
	else if (!strcasecmp(lpc,"CASEELSE")) CASEDEFAULT();
	else if (!strncasecmp_(lpc,"CASE",4)) p=lpc+4, CASE();
	else if (!strncasecmp_(lpc,"CAUSEERROR",10)) p=lpc+10, CAUSEERROR();
	else if (!strncasecmp_(lpc,"CHAIN",5)) {p=lpc+5; CHAIN(); return TRUE; }
	else if (!strncasecmp_(lpc,"CHANGE",6)) p=lpc+6, CHANGE(); 
	else if (!strncasecmp_(lpc,"CLEAR",5)) p=lpc+5,TCLEAR();
	else if (!strncasecmp_(lpc,"CLOSE",5)) p=lpc+5,FCLOSE();
	else if (!strncasecmp_(lpc,"COMMAND",7)) p=lpc+7, COMMAND(); 
	else if (!strncasecmp_(lpc,"COMMON",6)) p=lpc+6, TCOMMON(); 
	else if (!strncasecmp_(lpc,"CONSTANT",8)) p=lpc+8,DECLARE(TRUE);
	else if (!strncasecmp_(lpc,"CONST",5)) p=lpc+5,DECLARE(TRUE);
	else if (!strncasecmp_(lpc,"CONTINUE",8)) p=lpc+8, DOCONTINUE();
	else if (!strncasecmp_(lpc,"DATA",4)); //no-op
	else if (!strncasecmp_(lpc,"DECLARESUB",10)); // no-op
	else if (!strncasecmp_(lpc,"DECLARE",7)) p=lpc+7, DECLARE(FALSE);
	else if (!strncasecmp_(lpc,"DEF",3)) {
		/* Multiple line DEFs cannot be nested */
		if (isDefRun) perror_(__LINE__,46,l);
	}
	else if (!strncasecmp_(lpc,"DIM",3)) {
		p=lpc=lpc+3;
		if (!strncasecmp_(lpc,"ENSION",6)) p=lpc=lpc+6;
		DECLARE(FALSE);
	}
	else if (!strncasecmp_(lpc,"DOUNTIL",7)) p=lpc+7, DOUNTIL();
	else if (!strncasecmp_(lpc,"DOWHILE",7)) p=lpc+7, DOWHILE();
	else if (!strcasecmp(lpc,"DO")) ; // no-op;
	else if (!strncasecmp_(lpc,"ELSEIF",6)) p=lpc+6, ELSEIF();
	else if (!strncasecmp_(lpc,"ELSE",4)) ELSE();
	else if (!strcasecmp(lpc,"ENDHANDLER")) ENDHANDLER();
	else if (!strcasecmp(lpc,"ENDIF")) ENDIF();
	else if (!strcasecmp(lpc,"ENDSELECT")) ENDSELECT();
	else if (!strcasecmp(lpc,"ENDSUB")); // no-op
	else if (!strcasecmp(lpc,"ENDWHEN")) ENDWHEN();
	else if (!strcasecmp(lpc,"ENDWHILE")) ENDWHILE();
	else if (!strcasecmp(lpc,"END")) END(EXIT_SUCCESS);
	else if (!strncasecmp_(lpc,"ENTER",5)) p=lpc+5, INPUT(1);
	else if (!strncasecmp_(lpc,"ERASE",5)) p=lpc+5,ERASE();
	else if (!strncasecmp_(lpc,"EXITSUB",7)) p=lpc+7, EXITSUB();
	else if (!strcasecmp(lpc,"EXITHANDLER")) EXITHANDLER();
	else if (!strcasecmp(lpc,"EXITWHEN")) EXITWHEN();
	else if (!strncasecmp_(lpc,"EXIT",4)) p=lpc+4, EXIT(0);
	else if (!strncasecmp_(lpc,"EXPORT",6)) p=lpc+6,EXPORT();
	else if (!strncasecmp_(lpc,"FILES",5)); //no-op
	else if (!strncasecmp_(lpc,"FILE",4)) p=lpc+4,FFILE(FALSE,l); 
	else if (!strncasecmp_(lpc,"FILLPAGE",8)) p=lpc+8,FFILLPAGE(); 
	else if (!strcasecmp(lpc,"FNEND")); // no-op
	else if (!strncasecmp_(lpc,"FOR",3)) p=lpc+3,FOR();
	else if (!strncasecmp_(lpc,"FREE",4)) p=lpc+4, FSCRATCH();
	else if (!strncasecmp_(lpc,"GET",3)) p=lpc+3, DOGET();
	else if (!strncasecmp_(lpc,"GOSUB",5)) p=lpc+5,GOSUB();
	else if (!strncasecmp_(lpc,"GOTO",4)) p=lpc+4,GOTO();
	else if (!strncasecmp_(lpc,"HANDLER",7)) p=lpc+7,HANDLER();
	else if (!strncasecmp_(lpc,"IF",2)) p=lpc+2, IFTHEN();
	else if (!strncasecmp_(lpc,"IMAGE",5)); // no-op
	else if (!strncasecmp_(lpc,"INPUT:",6)) p=lpc+6, FREADCOLON();
	else if (!strncasecmp_(lpc,"INPUT#",6)) p=lpc+6, FREADSHARP(1,0);
	else if (!strncasecmp_(lpc,"INPUT",5)) p=lpc+5, INPUT(1);
	else if (!strncasecmp_(lpc,"JUMP",4)) p=lpc+4, JUMP();
	else if (!strncasecmp_(p=lpc,"LET",3) && multiLET());
	else if (!strncasecmp_(lpc,"LIBRARY",7)) p=lpc+7, OPENASLIBRARY();
	else if (!strncasecmp_(lpc,"LINEINPUT#",10)) p=lpc+10,FREADSHARP(1,1);
	else if (!strncasecmp_(lpc,"LINEINPUT",9)) p=lpc+9, INPUT(0);
	else if (!strncasecmp_(lpc,"LINPUT#",7)) p=lpc+7,FREADSHARP(1,1);
	else if (!strncasecmp_(lpc,"LINPUT",6)) p=lpc+6, INPUT(0);
	else if (!strncasecmp_(lpc,"LINEPRINT",9)) p=lpc+9, LPRINT();
	else if (!strncasecmp_(lpc,"LPRINT",6)) p=lpc+6, LPRINT();
	else if (!strncasecmp_(lpc,"LOOPWHILE",9)) p=lpc+9,LOOPWHILE();
	else if (!strncasecmp_(lpc,"LOOPUNTIL",9)) p=lpc+9,LOOPUNTIL();
	else if (!strcasecmp(lpc,"LOOP")) p=lpc+4,LOOP();
	else if (!strncasecmp_(lpc,"LINEREAD#",9)) p=lpc+9,FREADSHARP(0,1);
	else if (!strncasecmp_(lpc,"LREAD#",6)) p=lpc+6,FREADSHARP(0,1);
	else if (!strncasecmp_(lpc,"MARGIN",6)) p=lpc+6, FMARGIN();
	else if (!strncasecmp_(lpc,"MATINPUT#",9)) p=lpc+9, MATFINPUT(FALSE);
	else if (!strncasecmp_(lpc,"MATINPUT",8)) p=lpc+8, MATINPUT(0,FALSE);
	else if (!strncasecmp_(lpc,"MATPRINT#",9)) p=lpc+9, MATFPRINT(FALSE);
	else if (!strncasecmp_(lpc,"MATPRINT",8)) p=lpc+8, MATPRINT(0,FALSE);
	else if (!strncasecmp_(lpc,"MATWRITE#",9)) p=lpc+9, MATFPRINT(TRUE);
	else if (!strncasecmp_(lpc,"MATWRITE",8)) p=lpc+8, MATPRINT(0,TRUE);
	else if (!strncasecmp_(lpc,"MATREAD#",8)) p=lpc+8, MATFINPUT(TRUE);
	else if (!strncasecmp_(lpc,"MATREAD",7)) p=lpc+7, MATREAD();
	else if (!strncasecmp_(lpc,"MAT",3)) p=lpc+3, MATtype();
	else if (!strncasecmp_(lpc,"NEXT",4)) p=lpc+4, NEXT();
	else if (!strncasecmp_(lpc,"NODATA",6)) p=lpc+6, NODATA();
	else if (!strncasecmp_(lpc,"NOMARGIN",8)) p=lpc+8, FNOMARGIN();
	else if (!strncasecmp_(lpc,"NOPAGENUM",9)) p=lpc+9, FPAGENUM(FALSE);
	else if (!strncasecmp_(lpc,"NOPAGE",6)) p=lpc+6, FNOPAGE();
	else if (!strncasecmp_(lpc,"NOQUOTE",7)) p=lpc+7,FQUOTE(NOQUOTE);
	else if (!strncasecmp_(lpc,"ONATTENTION",11)) p=lpc+11, ONATTENTION();
	else if (!strncasecmp_(lpc,"ONERROR",7)) p=lpc+7, ONERROR();
	else if (!strncasecmp_(lpc,"ON",2)) p=lpc+2, ONGOTO();
	else if (!strncasecmp_(lpc,"OPENASLIBRARY",13))p=lpc+13,OPENASLIBRARY();
	else if (!strncasecmp_(lpc,"OPEN",4)) p=lpc+4, FOPEN();
	else if (!strncasecmp_(lpc,"OPTION",6)) p=lpc+6, OPTION();
	else if (!strncasecmp_(lpc,"PAGENUM",7)) p=lpc+7, FPAGENUM(TRUE);
	else if (!strncasecmp_(lpc,"PAGE",4)) p=lpc+4, FPAGE();
	else if (!strncasecmp_(lpc,"PREPEND",7)) p=lpc+7, FPREPEND();
	else if (!strncasecmp_(lpc,"PRINTUSING#",11)) 
		p=lpc+11, FWRITEUSINGSHARP(TRUE,TRUE,0);
	else if (!strncasecmp_(lpc,"PRINTUSING",10)) 
		p=lpc+10, FWRITEUSINGSHARP(TRUE,FALSE,0);
	else if (!strncasecmp_(lpc,"PRINT",5)) p=lpc+5, PRINT(TRUE);
	else if (!strncasecmp_(lpc,"PUT",3)) p=lpc+3, DOPUT();
	else if (!strncasecmp_(lpc,"QUEUE",5)) p=lpc+5, QUEUE();
	else if (!strncasecmp_(lpc,"QUOTE",5)) p=lpc+5, FQUOTE(QUOTE);
	else if (!strncasecmp_(lpc,"RANDOMIZE",9)) p=lpc+9, RANDOMIZE();
	else if (!strncasecmp_(lpc,"READ:",5)) p=lpc+5, FREADCOLON();
	else if (!strncasecmp_(lpc,"READ#",5)) p=lpc+5, FREADSHARP(0,0);
	else if (!strncasecmp_(lpc,"READ",4)) p=lpc+4, READ();
	else if (!strncasecmp_(lpc,"REDIM",5)) p=lpc+5,REDIM();
	else if (!strncasecmp_(lpc,"REM",3)); // no-op
	else if (!strncasecmp_(lpc,"RESET",5)) p=lpc+5, RESTORE();
	else if (!strncasecmp_(lpc,"RESTORE#",8)) p=lpc+7, FRESTORE();
	else if (!strncasecmp_(lpc,"RESTORE:",8)) p=lpc+7, FRESTORE();
	else if (!strncasecmp_(lpc,"RESTORE",7)) p=lpc+7, RESTORE();
	else if (!strncasecmp_(lpc,"RESUME",6)) p=lpc+6, RESUME();
	else if (!strncasecmp_(lpc,"RETRY",5)) p=lpc+5, RETRY();
	else if (!strcasecmp(lpc,"RETURN")) RETURN();
	else if (!strncasecmp_(lpc,"RETURN",6)) p=lpc+6,RETURNTOSUB();
	else if (!strncasecmp_(lpc,"ROUTE",5)) p=lpc+5,ROUTE();
	else if (!strncasecmp_(lpc,"SCRATCH",7)) p=lpc+7, FSCRATCH();
	else if (!strncasecmp_(lpc,"SELECT",6)) p=lpc+6, SELECT();
	else if (!strncasecmp_(lpc,"SETDIGITS",9)) p=lpc+9, SETDIGITS();
	else if (!strncasecmp_(lpc,"SET",3)) p=lpc+3, FSET();
	else if (!strncasecmp_(lpc,"STOP",4)) END(EXIT_STOP);
	else if (!strncasecmp_(lpc,"SUB",3)); // no-op
	else if (!strncasecmp_(lpc,"SYSTEM",6)) END(EXIT_SYSTEM);
	else if (!strncasecmp_(lpc,"UPDATE",6)) p=lpc+6, FUPDATE();
	else if (!strncasecmp_(lpc,"USE",3)) p=lpc+3, USE();
	else if (!strncasecmp_(lpc,"WAIT",4)) p=lpc+4, DOWAIT();
	else if (!strcasecmp(lpc,"WEND")) ENDWHILE();
	else if (!strncasecmp_(lpc,"WHENERROR",9)) p=lpc+9, WHENERROR();
	else if (!strncasecmp_(lpc,"WHILE",5)) p=lpc+5, WHILE();
	else if (!strncasecmp_(lpc,"WRITEUSING#",11)) 
		p=lpc+11, FWRITEUSINGSHARP(FALSE,TRUE,0);
	else if (!strncasecmp_(lpc,"WRITEUSING",10)) 
		p=lpc+10, FWRITEUSINGSHARP(TRUE,FALSE,0);
	else if (!strncasecmp_(lpc,"WRITE",5)) p=lpc+5, PRINT(FALSE);
	else return FALSE;
	return TRUE;
}



/*****************************************
 *      SIGNALS & ERROR TREATMENT        *
 *****************************************/

/* perror_()
 * print a string message based upon the following information data:
 * - errNum is the error code number defined in errors.h and must always be
 *   given as first argument 
 * - ll is the line number in which the error is occurred; if null, the 
 *   line reference is not printed (general errors); this must always be given 
 *   as the second argument
 * - all the other parameters should be the ones needed in the error strings,
 *   to fill the data in the % specifiers, according to their type.
 * If an error-catching technique is in use, set codes and exit.
 * The file name is added only in a CHAINed process, and only from the second
 * file on; the first file run has no file name reference*/
void perror_(int line,int errNum, int ll, ...) {
	va_list ap;
	int diff=0, delay=0;

	/* if USE or HANDLER are in action, their error messages must be
	 * printed and the program stopped,  because an error in the handler
	 * must not be re-catched. So redirect message to syserr_() */
	if ((isUseActive || isHandlerActive) && error[errNum][0]!='%') {
		va_start(ap, ll);
		syserr_(line,errNum, ll, ap);
		va_end(ap);
	}

	/* intercept the WHEN ERROR USE feature first; this is conceived as the
	 * inner WHEN ERROR error catching technique  */
	if (isWhenErrorUsePos && error[errNum][0]!='%') {
		errValue=errNum;
		errLine=l;
		l=handler[isHandlerNum].pos-1;
		return;
	}
	/* intercept the WHEN ERROR IN feature then; this is conceived as the
	 * outer WHEN ERROR error catching technique */
	else if (isWhenErrorInPos && error[errNum][0]!='%') {
		errValue=errNum;
		errLine=l;
		l=isUsePos-1;
		return;
	}
	/* intercept the NO DATA error catching technique */
	else if (isNoDataPos && error[errNum][0]!='%') {
		errValue=errNum;
		errLine=l;
		l=isNoDataPos-1;
		return;
	}
	/* intercept the ON ERROR <code> error catching technique; this
	 * responds only if a specific error has to be trapped  */
	else if (isOnErrorPos && isOnErrorTrap && error[errNum][0]!='%') {
		if (errNum==isOnErrorTrap) {
			errValue=errNum;
			errLine=l;
			l=isOnErrorPos-1;
			return;
		}
	}	
	/* intercept the ON ERROR error catching technique; this is conceived
	 * as the most general error catching technique, which should be
	 * set as the outer of all possible error catching features */
	else if (isOnErrorPos && error[errNum][0]!='%') {
		errValue=errNum;
		errLine=l;
		l=isOnErrorPos-1;
		return;
	}

	/* print error message */
	if (channel[0].fcounter || isRawPrint) fprintf(stdout,"\n");
	va_start(ap, ll);
	/* real error: print regular error message with arguments */
	if (ll!=ERRORCODE) vprintf(error[errNum], ap);
	else {
		/* print string as-is; if error message string is void, print 
		 * default string at position 0; this is the case of CAUSE 
		 * ERROR called with a reference to a not-used code error */
		if (!(int)strlen(error[errNum])) fprintf(stdout,"%s",error[0]);
		else fprintf(stdout,"%s",error[errNum]);
	}
	/* add other data (line number, label, things like that) */
	if (ll==ERRORCODE) ll=l;
	if (ll>=0) {
		if (!isSubRun) fprintf(stdout," in line %d",ll);
		else fprintf(stdout," in sub %s at line %d",\
				sn[isSubVar].name!=NIL?sn[isSubVar].name:"\b",\
				ll);
		if (ln[ll]) fprintf(stdout," (labelled as %d)",ln[ll]);
	}
	if (isChain) fprintf(stdout," into chained file %s", filename);
	va_end(ap);
	fprintf(stdout,".");
	if (isPrintSourceCode) printf(" #%d#\n",line);
	else fprintf(stdout,"\n");

	/* print line and indicator for errors (not for warnings) */
	if (isRunOn && error[errNum][0]!='%') {
		char *qq;
		int diff=0;
		if (p!=NIL) {
			qq=p;
			while (*qq) qq++; // find end of p
			diff=(int)(qq-p);
		}
		fprintf(stdout,"%s\n",m[ll]);
		if (m[ll] && (int)strlen(m[ll])>diff-1) 
			delay=(int)((int)strlen(m[ll])-diff);
		else delay=0;
		/* 'here' improvement */
		if (delay>0) while (delay>0) fputc(' ',stdout), delay--;
		fprintf(stdout,"^\n");
	}

	/* reset color */
	printf(RESETCOLOR);

	/* rebuild cursor position in case of a warning */
	if (error[errNum][0]=='%') while (diff++ < channel[0].fcounter) 
		putchar(' ');
	/* otherwise, stop execution */
	else END(EXIT_FAILURE);
}


/* syserr_()
 * print a system string message based upon the following information data:
 * - errNum is the error code number defined in errors.h and must always be
 *   given as first argument 
 * - all the other parameters should be the ones needed in the error strings,
 *   to fill the data in the % specifiers, according to their type 
 * No interception is done, here, so the printing of the error force 
 * program end. 
 * If ll=-1 then print Usage */
void syserr_(int line ,int errNum, int ll, ...) {
	va_list ap;

	/* print error message */
	if (channel[0].fcounter) fprintf(stdout,"\n");
	va_start(ap, ll);
	vprintf(error[errNum], ap);
	if (ll>=0) {
		if (!isSubRun) fprintf(stdout," in line %d",ll);
		else fprintf(stdout," in sub %s at line %d",\
				sn[isSubVar].name!=NIL?sn[isSubVar].name:"\b",\
				ll);
		if (ln[ll]) fprintf(stdout," (labelled as %d)",ln[ll]);
	}
	if (isChain) fprintf(stdout," into chained file %s", filename);
	va_end(ap);
	fprintf(stdout,".");
	if (isPrintSourceCode) printf(" #%d#\n",line);
	else fprintf(stdout,"\n");

	/* reset color */
	fprintf(stdout,RESETCOLOR);

	/* quit printing usage and emitting failure exit code */
	if (ll==-1) printUsage();
	END(EXIT_FAILURE);
}


/* sigtrap()
 * trap procedure for ON ATTENTION when an interrupt has been trapped */
void sigtrap(int sig) {
	/* reassign user destination SIGINT */
	if (isOnAttentionPos) {
		signal(SIGINT, sigint);
		signal(SIGHUP, sigint);
		arrLine=l;
		arrValue=ERRLIM;
		FCR(CONSOLE);
		parse(m,isOnAttentionPos,endCore);
	}
	else sigint(sig);
}


/* sigquit()
 * end procedure for stopping when a SIGBUS or a SIGSEGV signal 
 * has been trapped */
void sigquit(int sig) {
	/* reassign the normal behavior to SIGINT */
	signal(SIGBUS, SIG_DFL);
	signal(SIGSEGV, SIG_DFL);
	signal(SIGABRT, SIG_DFL);
	closeFiles();
	syserr_(__LINE__,6,-1);
}


/* sigint()
 * end procedure if a SIGHUP or a SIGINT signal was trapped */
void sigint(int sig) {
	/* reassign the normal behavior to signals */
	signal(SIGINT, SIG_DFL);
	signal(SIGHUP, SIG_DFL);
	isInterrupt=TRUE;
	END(EXIT_STOP);
}


/* xmalloc()
 * create a memory space for strings
 * size is the size of the string (included the terminator) 
 * Note: this function was created to be portable among all UNIX systems
 * with gcc installed. */
void *xmalloc(size_t size) {
	void *value = malloc(size);
	if (value == NIL) perror_(__LINE__,5,l);
	return value;
}


/***************************************
 *      DOCUMENTATION AND LICENSE      *
 ***************************************/

/* printHelp()
 * help message */
void printHelp(void) {
	printExistence();
	fprintf(stdout,"Usage: tbas [options] [filename]\n\n");
	fprintf(stdout,"Options:\n");
	fprintf(stdout,"      --conditions  Print GPL3 redistribution conditions and exit\n");
	fprintf(stdout,"  -d                Normal debug mode (print full lines)\n");
	fprintf(stdout,"      --errors-list Print a complete errors list and exit\n");
	fprintf(stdout,"  -h, --help        Display this information and exit\n");
	fprintf(stdout,"  -H                Set Header and Timing display (DEC-20 mode)\n");
	fprintf(stdout,"  -r                Raw debug mode (print preprocessed lines)\n");
	fprintf(stdout,"  -v, --version     Display version and exit\n");
	fprintf(stdout,"      --warranty    Print warranty conditions from GPL3 and exit\n");
	fprintf(stdout,"\nYou should have received a copy of the GNU General Public License\n");
	fprintf(stdout,"along with this program. If not, see <http:/www.gnu.org/licenses/>.\n\n");
	fprintf(stdout,"Consult the man or the info page to know about tbas more in detail.\n\n");
	fprintf(stdout,"Email bugs reports to <tbin at libero dot it>.\n\n"); 
}

/* printHiddenHelp()
 * help on hiddend options */
void printHiddenHelp(void) {
	printExistence();
	fprintf(stdout,"Usage: tbas [options] [filename]\n\n");
	fprintf(stdout,"Options (included hidden ones):\n");
	fprintf(stdout,"      --conditions  Print GPL3 redistribution conditions and exit\n");
	fprintf(stdout,"  -d                Normal debug mode (print full lines)\n");
	fprintf(stdout,"      --errors-list Print a complete errors list and exit\n");
	fprintf(stdout,"  -h, --help        Display the standard help and exit\n");
	fprintf(stdout,"      --hidden      Display this extended help and exit\n");
	fprintf(stdout,"  -H                Set Header and Timing display (DEC-20 mode)\n");
	fprintf(stdout,"  -r                Raw debug mode (print preprocessed lines)\n");
	fprintf(stdout,"  -R                set raw print mode (testing purposes)\n");
	fprintf(stdout,"  -S                print error source position (debug purposes)\n");
	fprintf(stdout,"  -t                set OPTION TIMING (speed evaluation purposes)\n");
	fprintf(stdout,"  -v, --version     Display version and exit\n");
	fprintf(stdout,"  -V                print version and exit (versioning purposes)\n");
	fprintf(stdout,"      --warranty    Print warranty conditions from GPL3 and exit\n");
	fprintf(stdout,"  -X                exit after preprocessing (debug purposes)\n");
	fprintf(stdout,"  -z                set OPTION ZERO 6 (testing purposes)\n");
	fprintf(stdout,"\nYou should have received a copy of the GNU General Public License\n");
	fprintf(stdout,"along with this program. If not, see <http:/www.gnu.org/licenses/>.\n\n");
	fprintf(stdout,"Consult the man or the info page to know about tbas more in detail.\n\n");
	fprintf(stdout,"Email bugs reports to <tbin at libero dot it>.\n\n"); 
}

/* printUsage()
 * usage message */
void printUsage(void) {
	fprintf(stdout,"Usage: tbas [options] [filename]\n");
	fprintf(stdout,"Type tbas -h for help\n");
	s_exit(EXIT_FAILURE);
}

/* printExistence()
 * existence message */
void printExistence(void) {
	fprintf(stdout,"tbas BASIC interpreter\n");
}

/* printVersion()
 * help banner */
void printVersion(void) {
	printExistence();
	fprintf(stdout,"version %s (built on %s, at %s)", 
		VERSION, __DATE__ ,__TIME__);
	fprintf(stdout," - Copyleft (C) %s\n%s <%s>. \n", CY, AU, EM);
}

/* printHelpBanner()
 * version message */
void printHelpBanner(void) {
	fprintf(stdout,"This free software comes with ABSOLUTELY NO WARRANTY; for details about this,\ntype 'tbas --warranty'.  You are welcome to redistribute it under certain\nconditions; type 'tbas --conditions' for details.\nThis software is released under the terms of the GNU General Public License 3.\n\n");
}

/* printWarranty()
 * warranty message (required by GPL3) */
void printWarranty(void) {
	fprintf(stdout,"THERE IS NO WARRANTY FOR THE PROGRAM, TO THE EXTENT PERMITTED BY\nAPPLICABLE LAW. EXCEPT WHEN OTHERWISE STATED IN WRITING THE COPYRIGHT\nHOLDERS AND/OR OTHER PARTIES PROVIDE THE PROGRAM \"AS IS\" WITHOUT\nWARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING, BUT NOT\nLIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A\nPARTICULAR PURPOSE. THE ENTIRE RISK AS TO THE QUALITY AND PERFORMANCE OF\nTHE PROGRAM IS WITH YOU. SHOULD THE PROGRAM PROVE DEFECTIVE, YOU ASSUME\nTHE COST OF ALL NECESSARY SERVICING, REPAIR OR CORRECTION.\n\n");
}

/* printConditions()
 * conditions message (required by GPL3) */
void printConditions(void) {
	fprintf(stdout,"You may make, run and propagate covered works that you do not convey,\nwithout conditions so long as your license otherwise remains in force.\nYou may convey covered works to others for the sole purpose of having\nthem make modifications exclusively for you, or provide you with\nfacilities for running those works, provided that you comply with the\nterms of this License in conveying all material for which you do not\ncontrol copyright. Those thus making or running the covered works for\nyou must do so exclusively on your behalf, under your direction and\ncontrol, on terms that prohibit them from making any copies of your\ncopyrighted material outside their relationship with you.\n\n");
}


/******************************************************************************
 * MAIN SECTION
 ******************************************************************************/

/* main() 
 * parse options, get file name, setup environment and perform, at request, 
 * the following actions:
 * OLD to create the file in memory
 * RUN to execute the file
 * Here is where it all begins... and where it all ends */
int main(int argc, char **argv) {
	int c, ac=0;

	/* parse options; double and single dash options may be alternated,
	 * options requiring an argument may or not be attached (but must
	 * be the last options of that group);
	 * ~Antonio Maschio - March 2014 for the decb project */
	while (++ac < argc) {
		if (*argv[ac]=='-') {
			/* parse double-dash options */
			if (!strcasecmp(argv[ac], "--help")) { 
			        putchar('\n');	
				printHelp(); 
				exit(EXIT_SUCCESS); 
			}
			else if (!strcasecmp(argv[ac], "--hidden")) { 
			        putchar('\n');	
				printHiddenHelp(); 
				exit(EXIT_SUCCESS); 
			}
			else if (!strcasecmp(argv[ac], "--errors-list") ||\
				!strcasecmp(argv[ac], "--errorslist") ||\
				!strcasecmp(argv[ac], "--errors")) { 
				int i;
				char *y;
				fprintf(stdout,"\nErrors list for tbas ");
				fprintf(stdout,"version %s\n\n",VERSION);
				for (i=0;i<ERRLIM;i++) {
				    char n='\0';
				    y=error[i];
				    if (*y) {
					while (*y=='\n') y++;
				    	if (!strrchr(y,'\n')) n='\n';
				        if (!strncasecmp_(y,"%%%%",2))	
				    	    fprintf(stdout,"%3d %s%c",i,y+1,n);
					else
					    fprintf(stdout,"%3d %s%c",i,y,n);
				    }
				}
				fprintf(stdout,"\n");
				exit(EXIT_SUCCESS); 
			}
			else if (!strcasecmp(argv[ac], "--version")) {
			        putchar('\n');	
				printVersion();
				printHelpBanner();
				exit(EXIT_SUCCESS); 
			}
			else if (!strcasecmp(argv[ac], "--warranty")) {
			        putchar('\n');	
				printWarranty();
				exit(EXIT_SUCCESS);
			}
			else if (!strcasecmp(argv[ac], "--conditions")) {
			        putchar('\n');	
				printConditions();
				exit(EXIT_SUCCESS);
			}
			/* this feature is intentionally undocumented 
			 * I use it to generate the README file */
			else if (!strcasecmp(argv[ac], "--greetings")) {
				fprintf(stdout,"tbas README page\n");
				fprintf(stdout,WELCOME);
				printVersion();
				printHelpBanner();
				printHelp();
				fprintf(stdout,GREET);
				printWarranty();
				printConditions();
				fprintf(stdout,"It's GPL. Enjoy!\n\n");
				exit(EXIT_SUCCESS);
			}
			/* Ignore unrecognized options */
			else if (!strncmp(argv[ac],"--",2)) {
				fprintf(stdout,"(%s%s)\n",argv[ac],error[8]);
				continue;
			}
			
			/* parse single-dash options */
			while (*++argv[ac]) {
				c=*argv[ac];
				/* option -h
				 * print help*/
				if (c=='h') {
			        	putchar('\n');	
					printHelp();
					exit(EXIT_SUCCESS); 
				}
				/* option -v 
				 * print version */
				else if (c=='v') {
			        	putchar('\n');	
					printVersion();
					printHelpBanner(); 
					exit(EXIT_SUCCESS); 
				}
				/* option -d
				 * set normal Debug mode */
				else if (c=='d') {
					isDebug=TRUE;
					debugRaw=FALSE;
				}
				/* option -r
				 * set raw Debug mode */
				else if (c=='r') {
					isDebug=debugRaw=TRUE;
				}
				/* option -t (UNDOCUMENTED)
				 * set timing (speed evaluation purposes) */
				else if (c=='t') {
					isTiming=TRUE;
				}
				/* option -S  (UNDOCUMENTED)
				 * print source code line number in error message */
				else if (c=='S') {
					isPrintSourceCode=TRUE;
				}
				/* option -H
				 * set DEC-20 screen mode */
				else if (c=='H') {
					isHeader=isTiming=TRUE;
				}
				/* option -R (UNDOCUMENTED)
				 * set raw print mode (testing purposes) */
				else if (c=='R') {
					isRawPrint=TRUE;
				}
				/* option -X (UNDOCUMENTED)
				 * don't execute; exit after preprocessing
				 * in debug mode */
				else if (c=='X') {
					isNoExecute=TRUE;
					isDebug=TRUE;
				}
				/* option -z (UNDOCUMENTED)
				 * set ZEROLIM to 10^-6 for rough execution */
				else if (c=='z') {
					ZEROLIM=1.0E-6;
					isZeroLim=TRUE;
				}
				/*  (UNDOCUMENTED)
				 * option -V print only version id */
				else if (c=='V') {
					fprintf(stdout,"%s\n",VERSION);
					exit(EXIT_SUCCESS); 
				}
				/* Ignore unrecognized options */
				else fprintf(stdout,"(-%c%s)\n",c,error[8]);
			}
		}
		else break;
	}

	/* OK; now, if after any option no file name was given
	 * (or if tbas was launched with no parameters), issue an error */
	if ((ac==argc)||(1==argc)) perror_(__LINE__,4,-1); 

	/* get input file name (discard following arguments) 
	 * and setup name, attaching extension if not present */ 
	strncpy_(filename, argv[ac], COMPSCR);
	setupName(filename,&c ,&c);
	

	/* set features that depend on starting time */
	now=time(NULL);

	/* set up console channels data (which is the terminal, in tbas) 
	 * Note: only meaningful data for sequential output are set 
	 * Note: this is done here because every tbas functions must find
	 * CONSOLE set up and running for error messages printing */
	channel[CONSOLE].stream=stdout;
	channel[CONSOLE].isopen=TRUE;
	channel[CONSOLE].type=SEQ;
	channel[CONSOLE].fcounter=0;
	channel[CONSOLE].wayout=VWRITE;
	channel[CONSOLE].quote=NOQUOTE;
	channel[CONSOLE].page=1;
	channel[CONSOLE].height=PAGEHEIGHT;
	channel[CONSOLE].pagerow=1;
	channel[CONSOLE].pagenumber=FALSE;
	channel[CONSOLE].margin=SCRLIM;
	channel[CONSOLE].fprintn=1;
	channel[CONSOLE].isReadonly=FALSE;
	channel[CONSOLE].isPrepend=FALSE;
	for (c=1; c<=STREAMS; c++) 
		channel[c].isopen=FALSE, channel[c].stream=NIL;

	/* search for existing routefile */
	while (TRUE) {
		FILE *f;
		char fn[COMPSTR];
		sprintw(fn,COMPSTR,"%s%d.txt",TEMPROUTE,queueFileCount);
		f=fopen(fn,"r");
		if (f) {
			queueFileCount++;
			fclose(f);
		}
		else break;
	}
	
	/* Trap overflow error */
	signal(SIGABRT, sigquit);
	signal(SIGSEGV, sigquit);
	signal(SIGBUS, sigquit);
	/* Trap interrupts */
	signal(SIGHUP, sigint);
	signal(SIGINT, sigint);

#ifndef NOUNIX
	/* Store tcgetattr values */
	tcgetattr(0, &modes);
	savemodes = modes;
#endif

	/* OLD (implicit!)
	 * store program in memory */
	OLD();

	/* RUN (finally!)
	 * Note: RUN records also starting timer */
	if (isHeader) printHeader();
	RUN();
    	
	/* END (implicit!) */
	END(EXIT_SUCCESS);

	/* this is here only for C! */
	return EXIT_SUCCESS; 


} /* end main */

/* end of tbas.c */

